#include <syslib/types.h>

/*
  lsdev -- dump informations about devices available in the system
  Copyright (C) 2006 Jozef Behran

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
  Usage: lsdev 
*/

#include <syslib/traps.h>
#include <syslib/env.h>
#include <syslib/pdbase.h>
#include <syslib/pddev.h>

#include <libhs/strings.h>
#include <libhs/process.h>
#include <libhs/devices.h>
#include <libhs/errno.h>
#include <libhs/dirname.h>
#include <libhs/output.h>

/* The name this program was run with. */
char *ProgramName;

/*
  These calls display various portions of the help and versions.
*/

void Hint(void)
{
  PrintErr("Try '");
  PrintErr(ProgramName);
  PrintErr(" --help' for more information.\n");
}

void DumpErrorMessage(const char *Msg)
{
  PrintErr(ProgramName);
  PrintErr(": ");
  PrintErr(Msg);
  PrintErr("\n");
}

void DumpError(const char *Msg)
{
  DumpErrorMessage(Msg);
  Hint();
  Halt(1);
}

void ReportDeviceFetchError(void)
{
  DumpErrorMessage("Error obtaining device info(s)");
  PrintErr("System said: ");
  PrintErr(ComplainLastError());
  PrintErr("\n");
  Halt(1);
}

void Usage(void)
{
  Print ("\nUsage:\t");
  Print(ProgramName);
  Print("\n\t");
  Print(ProgramName);
  Print(
    " OPTION\n\n"
    "List all devices and their IDs known to the system.\n"
    "\n"
    "--help\t\tdisplay this help and exit\n"
    "--version\toutput version information and exit\n"
    "-l\t\tList valid types, then exit (may be specified only once)\n"
    "-t<nums>\tList only devices with type in <nums>, which is one or more\n"
    "\t\thex digit pairs denoting the device types. When used multiple\n"
    "\t\ttimes, the lists are unified into one big list. Cannot be used\n"
    "\t\twith -l\n"
    "\n"
    "Report bugs to <"
  );
  Print("jozef.behran@krs.sk");
  Print(">.\n");
  Halt(0);
}

void Version(void)
{
  Print(
    "lsdev (OSHS coreutils) 0.0.1\n"
    "written by The Holy Spirit through Jozef Behran\n"
    "\n"
    "Copyright (C) 2003 Free Software Foundation, Inc.\n"
    "This is free software; see the source for copying conditions.\n"
    "There is NO WARRANTY; not even for MERCHANTABILITY or FITNESS FOR A\n"
    "PARTICULAR PURPOSE."
  );
  Halt(0);
}

/*
  Output specified count of spaces.
*/
static void Spaces(int16 Count)
{
  while (Count>0) {
    Print(" ");
    Count--;
  }
}

/*
  Processing the command arguments
*/

bool LFlag;
char Allowed[256];

/*
  Setup codes 
*/
bool SetupCodes(const char *Arg)
{
  int16 i;
  char Ch;
  int16 Digit;
  int16 Num;
  char Buf[3];
  bool Invalid;
  bool Result;

  Buf[2]='\0';
  i=0;
  Invalid=false;
  Result=true;
  do {
    Ch=*(Arg++);
    Buf[i%2]=Ch;
    if (Ch=='\0') break;
    Digit=Ch;
    Digit-='0';
    if (Ch>'9') Digit-=7;
    if (Ch>'Z') Digit-=32;
    if (Digit<0 || Digit>15) Invalid=true;
    if (Invalid) {
      if (i%2==1) {
        Print("Invalid type code: ");
        PrintLn(Buf);
        Result=false;
        Invalid=false;
      }
    } else {
      if (i%2==0) {
        Num=Digit*16;
      } else {
        Num+=Digit;
        if (Allowed[Num]=='Y') {
          Print("Type code '");
          Print(Buf);
          PrintLn("' specified more than once");
          Result=false;
        } else {
          Allowed[Num]='Y';
        }
      }
    }
    i++;
  } while (true);
  if (i%2==1) {
    Print("Incomplete type code: ");
    PrintLn(Buf);
    Result=false;
  }
  return(Result);
}

/*
  Process commandline arguments
*/
static void ProcessArgs(void)
{
  uint16 argc;
  char *Arg;
  int16 i;
  bool Invalid;
  bool Terminate;
  bool SomeTypes;

  MemSet(Allowed,sizeof(Allowed),'N');
  Terminate=false;
  LFlag=false;
  SomeTypes=false;
  argc=GetArgCnt();
  for (i=0;i<argc;i++) {
    Invalid=false;
    Arg = GetArg(i);
    if (!StrCmp(Arg,"--help")) Usage();
    if (!StrCmp(Arg,"--version")) Version();
    ActivateErr();
    if (Arg[0]=='-') {
      switch (Arg[1]) {
        case 'l': {
          if (Arg[2]=='\0') {
            if (LFlag) {
              PrintLn("'-l' specified more than once");
              Terminate=true;
            } else {
              LFlag=true;
            }
          } else {
            Invalid=true;
          }
          break;
        }
        case 't': {
          if (Arg[2]=='\0') {
            PrintLn("'-t' must contain at least one pair of hex digits");
          } else {
            if (!SetupCodes(Arg+2)) Terminate=true;
            SomeTypes=true;
          }
          break;
        }
        default: {
          Invalid=true;
          break;
        }
      }
    } else {
      Invalid=true;
    }
    if (Invalid) {
      Print("Invalid argument: ");
      PrintLn(Arg);
      Terminate=true;
    }
    DeactivateErr();
  }
  if (Terminate) {
    Hint();
    Halt(1);
  }
  if (!SomeTypes) {
    MemSet(Allowed,sizeof(Allowed),'Y');
  }
}

#define MAX_DESC_SIZE 13

datablock(Descriptions)
  streami8u dtCharacter   endstream    streamstr "Char device"   endstream
  streami8u dtBlock       endstream    streamstr "Block device"  endstream
  streami8u dtScreen      endstream    streamstr "Screen"        endstream
  streami8u dtKeyboard    endstream    streamstr "Keyboard"      endstream
  streami8u dtMouse       endstream    streamstr "Mouse"         endstream
  streami8u dtNetworkCard endstream    streamstr "Network iface" endstream
  streami8u dtInternet    endstream    streamstr "Internet"      endstream
  streami8u dtUnknown     endstream    streamstr "unknown"       endstream
  streami8u dtInvalid     endstream
enddatablock

/*
  Perform sanity check of the table, fire a bug detection trap, if problems
  found.
*/
static void TestStrings(void)
{
  const upint8 *Data;
  TDeviceType Current;
  const char *Desc;
  uint16 Len;
  bool Invalid;

  Invalid=false;
  Data=Descriptions();
  do {
    Current=*(Data++);
    if (Current==dtInvalid) break;
    Desc=(const char *)Data;
    Len=StrLen(Desc);
    Data+=Len+1;
    if (Len>MAX_DESC_SIZE) {
      Print("Description '");
      Print(Desc);
      PrintLn("' too long");
      Invalid=true;
    }
  } while (true);
  if (Invalid) ReportBug("The description table failed sanity check");
}

/*
  Give a string for the specified device type
*/
static const char *TypeToString(TDeviceType Type)
{
  const upint8 *Data;
  const char *Result;
  TDeviceType Current;
  uint16 Len;

  Data=Descriptions();
  do {
    Current=*(Data++);
    if (Current==dtInvalid) return("???");
    Result=(const char *)Data;
    Len=StrLen(Result);
    Data+=Len+1;
  } while (Current!=Type);
  return(Result);
}

/*
  List all devices in the system
*/
static void ListDevices(void)
{
  TDevInfoReply R;
  bool DumpTableHeader;
  const char *Type;

  if (__DEBUG__) TestStrings();
  DumpTableHeader=true;
  if (!FindFirstDevice(&R)) ReportDeviceFetchError();
  while (R.Type!=dtInvalid) {
    if (Allowed[(int16)R.Type]=='Y') {
      if (DumpTableHeader) {
        DumpTableHeader=false;
        Print("ID");
	Spaces(8);
	Print("TYPE");
        Spaces(MAX_DESC_SIZE+3);
        PrintLn("DESC");
      }
      PrintHEX(R.ID,8);
      Print("  ");
      PrintHEX((int16)R.Type,2);
      Print(" (");
      Type=TypeToString(R.Type);
      Print(Type);
      Print(")");
      Spaces(MAX_DESC_SIZE+2-StrLen(Type));
      Print((const char *)R.InfoString);
      NewLine();
    }
    if (!FindNextDevice(&R)) ReportDeviceFetchError();
  }
}

/*
  Dump all types as hexadecimal numbers and as human readable strings.
*/
static void DumpDeviceTypes(void)
{
  const upint8 *Data;
  TDeviceType Current;
  const char *Desc;
  uint16 Len;

  Data=Descriptions();
  do {
    Current=*(Data++);
    if (Current==dtInvalid) break;
    Desc=(const char *)Data;
    Len=StrLen(Desc);
    Data+=Len+1;
    PrintHex((int16)Current,2);
    Print("  ");
    PrintLn(Desc);
  } while (true);
}

/*
  The main program itself
*/
int16 Main(void)
{
  ProgramName = StripBaseName(GetSelf());

  ProcessArgs();

  if (LFlag) {
    DumpDeviceTypes();
  } else {
    ListDevices();
  }

  return(0);
}
