/*
  The bootstrap program for OSHS

  Usage (after compilation): bstrap [-B<builddir>] <program> <program> ...

  where the <program> has the form <name>:<filespec> and denotes the program
  to be built. <name> is the name of the resulting binary and <filespec> is
  the relative specification of the file containing the Main() function of
  the program to be built (without the extension).
  
  The -B option determines the directory in which the programs are built. It
  defaults to /tmp/OSHSTMP.

  The purpose of this program is to create a makefile that builds the
  specified programs in the specified temporary directory. It is used to
  bootstrap OSHS on Linux

*/

/*
  Import various code from the OSHS core. Note that source code files are
  #included instead of the headers. This is because this program needs to be
  complete; there is no usable automatic build system available yet when
  compiling this ('make' does not count, since there are no 'makefiles' in
  the OSHS source code).

  Note also that the deprecated keyword detection is turned off here; it
  collides with the compilation of the source code
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslib/types.h>
#undef int
#undef long
#undef short
#undef unsigned
#undef signed
#undef main
#define GetMem KMEM__malloc
#define FreeMem KMEM__free
#include <syslib/conio.h>
#include <syslib/errno.h>
#include <syslib/memory.h>
#include <syslib/strings.h>
#include <syslib/filein.h>
#include "libhs/dictrb.c"

/*
  This procedure is called by the memory handling procedures. Its
  functionality is not needed during the bootstrap so we will supply an
  empty one instead.
*/
void CheckHeap(void)
{
}

/*
  Dependency file description

  The "FileName" contains the name of the dependency file (with the
  extension moved to the front). Only ".c" and ".h" files are supported by
  this simple bootstrapper (no autogenerated modules etc). The former are
  prepended with "C" and the latter with "H".
*/
typedef struct tagFile {
  const char *FileName;
  struct tagFile *Next;
  struct tagFile *NextLinked;
  uint16 ID;
  bool Unlinked;
} TFile;

static char *TempDir;
static TRBDict *Files;
static TFile *StackTop;
static TFile *Current;
static TFile *FirstLinked;
static uint16 CurID;
static bool GenMsDos;
static bool GenTurboC;
static bool UseStatic;
static int16 CondDepth;

static void Error(const char *Msg)
{
  ReportFileReadingError(Msg);
}

static TFile *CreateFileDesc(const char *FileName,char MayBeDupped)
{
  TFile *F;
  const char *FName;
  int16 Result;

  FName=StrDup(FileName);
  if (FName==NULL) Error("No memory for the file name");
  F=GetMem(sizeof(*F));
  if (F==NULL) Error("No memory for file descriptor");
  F->FileName=FName;
  Result=AddItemWithKey(Files,FName,F);
  switch (Result) {
    default: break;
    case rbNOMEM: Error("No memory adding file into dictionary");
    case rbDUPLICATE: {
      FreeMem(FName);
      FreeMem(F);
      if (MayBeDupped) return(NULL);
      Error("Attempt to register file more than once");
    }
  }
  F->Next=NULL;
  F->ID=65535;
  F->NextLinked=NULL;
  F->Unlinked=true;
  return(F);
}

static void FileError(const char *File,const char *Msg)
{
  fprintf(stderr,"%s: %s\n",File,Msg);
  exit(255);
}

static void ScheduleForLinking(TFile *F)
{
  if (F->Unlinked) {
    F->NextLinked=FirstLinked;
    FirstLinked=F;
    F->Unlinked=false;
  }
}

static void NewLine()
{
  printf("\n");
}

static void GetEnvVar(const char *VarName)
{
  printf(GenMsDos?"%%%s%%":"$%s",VarName);
}

static void SetEnvVar(const char *Name,const char *Value)
{
  if (GenMsDos) printf("set ");
  printf("%s=%s\n",Name,Value);
}

static void AssignIDForTurboC(TFile *Current)
{
  const char *FileName;
  const char *Ptr;
  char *WPtr;
  char FilePath[128];
  char SourceFile[129];
  char DestFile[129];
  char Ch;
  int Len;

  /*
    Split the filename into the path and the name of the file. In the path
    we need to convert "/" to "\\".
  */
  Ptr=Current->FileName+1;
  WPtr=FilePath;
  do {
    Ch=*(Ptr++);
    if (Ch=='\0') break;
    if (Ch=='/') {
      Ch='\\';
      FileName=Ptr;
    }
    *(WPtr++)=Ch;
  } while (true);
  FilePath[FileName-(Current->FileName+1)-1]='\0';
  /*
    Preprocess the file and store the result into the temporary directory.
    This is required for the stupid Turbo C which does not understand
    '#include "file.h"' in these source code files.
  */
  if (Current->FileName[0]!='C') {
    fprintf(stderr,"Compilation of header requested");
    exit(1);
  }
  Len=strlen(Current->FileName+1);
  if (Len>126) {
    fprintf(
      stderr,
      "Source file specification '%s.c' too long",
      Current->FileName+1
    );
    exit(1);
  }
  Ptr=Current->FileName+1;
  WPtr=SourceFile;
  do {
    Ch=*(Ptr++);
    if (Ch=='/') Ch='\\';
    *(WPtr++)=Ch;
  } while (Ch!='\0');
  strcpy(WPtr-1,".c");
  Len=strlen(TempDir);
  if (TempDir[Len-1]!='\\') Len++;
  if (Len>122) {
    fprintf(
      stderr,
      "Temporary directory specification '%s' is too long\n",
      TempDir
    );
    exit(1);
  }
  strcpy(DestFile,TempDir);
  DestFile[Len-1]='\\';
  sprintf(DestFile+Len,"%04X.c",CurID);
  if (!PreprocessFile(SourceFile,DestFile)) {
    fprintf(
      stderr,
      "Error preprocessing file '%s.c'\n",
      Current->FileName+1
    );
    exit(1);
  }
  /*
    Now generate the piece of the code which compiles the given file.
  */
  printf("C:\\TC\\tcc -ml -IC:\\TC -I");
  GetEnvVar("CURWD");
  printf(" -c %04X.c\n",CurID);
}

static void AssignID(TFile *Current)
{
  printf("echo \"Compiling %s.c\"",Current->FileName+1);
  NewLine();
  if (GenTurboC) {
    AssignIDForTurboC(Current);
  } else {
    printf("gcc -I. -I./syslib -c -g -O0 -o ");
    GetEnvVar("TMP");
    printf("/%04X.o %s.c\n",CurID,Current->FileName+1);
  }
  Current->ID=CurID++;
  ScheduleForLinking(Current);
}

static void AddAndPush(const char *FileName)
{
  TFile *F;

  F=CreateFileDesc(FileName,true);
  if (F==NULL) {
    if (FileName[0]=='C') {
      F=SearchKey(Files,FileName);
      if (F->ID!=65535) ScheduleForLinking(F);
    }
    return;
  }
  F->Next=StackTop;
  StackTop=F;
}

static void ParseInclude(char *Arg,const char *BaseDir)
{
  bool Local;
  char Buffer[178];
  uint16 Len;
  char EndMark;
  char *Ptr;

  /*
    If the #include is in a conditional block, ignore it
  */
  if (CondDepth>0) return;
  /*
    Check if we are #includeing a local header or a system one.
  */
  Local=(Arg[0]=='"');
  /*
    Remove the marks so only the filename is left in the "Arg" variable.
  */
  EndMark=Local?'"':'>';
  Len=StrLen(Arg);
  while (Arg[Len]!=EndMark) {
    if (Len==0) Error("Malformed #include directive detected");
    Len--;
  }
  /*
    Remove the extension. Change it to the extension prefix used for the
    filename.
  */
  Len-=2;
  if (Arg[Len]!='.' || (Arg[Len+1]!='h' && Arg[Len+1]!='c')) return;
  Arg[Len]='\0';
  /*
    If the header is a system-wide one, we need to check if it belongs to
    OSHS. The check is simple: if the base directory is "syslib/", the header
    isn't an OSHS header. Only OSHS headers are interesting for us, so if
    the header isn't an OSHS one, ignore it.
  */
  if (!Local && StrCmp(BaseDir,"syslib/")==0) return;
  /*
    If the header belongs to SYSLIB, ignore it. SYSLIB headers don't need
    scan as we can link against SYSLIB library directly.
  */
  MemCpy(Buffer,Arg+1,7);
  Buffer[7]='\0';
  if (StrCmp(Buffer,"syslib/")==0) return;
  /*
    Build the specification of the file referenced by the #include
  */
  Ptr=Buffer+1;
  Len=178;
  if (Local) {
    Ptr=StrNCpy(Ptr,BaseDir,178);
    Len=Buffer-Ptr;
  }
  StrNCpy(Ptr,Arg+1,Len);
  /*
    Add the ".c" and ".h" files into the file dictionary
  */
  Buffer[0]='H';
  AddAndPush(Buffer);
  Buffer[0]='C';
  AddAndPush(Buffer);
}

/*
  Parse one file. This parse is somewhat similar to the one 'makegen' does
  but is much simpler. There are no package description files; the specified
  programs refer only to packages that are in the OSHS core distribution.
*/
static void ParseSingleFile(const char *FileName,bool IsMainFile)
{
  char BaseDir[178];
  char LineBuf[181];
  char *Ptr;
  const char *Ext;
  int16 Len;
  bool IsSource;
  char Ch;

  /*
    Determine the directory in which the file resides.
  */
  Len=StrLen(FileName)-1;
  if (Len>178) FileError(FileName,"File spec too long");
  while ((Len>0) && (FileName[Len]!='/')) Len--;
  MemCpy(BaseDir,FileName+1,Len);
  BaseDir[Len]='\0';
  /*
    Build the file specification
  */
  Ptr=StrNCpy(LineBuf,FileName+1,178);
  IsSource=FileName[0]=='C';
  Ext=IsSource?".c":".h";
  StrCpy(Ptr,Ext);
  /*
    Open the file for reading. The main file and the headers must do exist;
    the other csource codes may be missing.
  */
  if (!OpenTheFile(LineBuf)) {
    if (!IsMainFile && IsSource) return;
    FileError(LineBuf,"Error opening the file");
  }
  /*
    If the file is a source code, assign an identifier to it.
  */
  if (IsSource) AssignID(Current);
  /*
    Initialize the conditional depth counter
  */
  CondDepth=IsSource?0:-1;
  /*
    Read the lines of the file and parse them
  */
  do {
    if (!ReadLine(LineBuf,181)) break;
    if (IsEOF()) break;
    Ptr=LineBuf;
    do {
      Ch=*(Ptr++);
    } while (Ch!=' ' && Ch!='\0' && Ch!='\t');
    if (Ch!='\0') *(Ptr-1)='\0';
    if (!StrCmp(LineBuf,"#include")) ParseInclude(Ptr,BaseDir);
    if (!StrCmp(LineBuf,"#endif")) CondDepth--;
    if (
      !StrCmp(LineBuf,"#if") ||
      !StrCmp(LineBuf,"#ifdef") ||
      !StrCmp(LineBuf,"#ifndef")
    ) CondDepth++;
  } while (true);
  /*
    Close the file after processing. Otherwise we could run out of handles.
  */
  CloseTheFile();
}

static void LinkProgramUsingTurboC(const char *ProgName,const TFile *Main)
{
  TFile *F;
  char Buffer[80];
  char *Ptr;

  Ptr=Buffer;
  F=FirstLinked;
  do {
    if (Ptr==Buffer) {
      Ptr=StrCpy(Ptr,"C:\\TC\\tlib /C ");
      Ptr=StrCpy(Ptr,ProgName);
      Ptr=StrCpy(Ptr,".lib");
    }
    sprintf(Ptr," +%04X",F->ID);
    Ptr+=6;
    F=F->NextLinked;
    if (Ptr-Buffer>72 || F==NULL) {
      printf("%s\n",Buffer);
      Ptr=Buffer;
    }
  } while (F!=NULL);
  printf(
    "C:\\TC\\tlink /c /i %04X.obj+%s.lib+c:\\tc\\c0l.obj+c:\\tc\\cl.lib,"
    "makegen\n",
    Main->ID,ProgName
  );
  printf("cd ");
  GetEnvVar("CURWD");
  printf("\n");
}

static void LinkProgramUsingGCC(const char *ProgName,const TFile *Main)
{
  TFile *F;

  printf("gcc -o %s %04X.o",ProgName,Main->ID);
  F=FirstLinked;
  do {
    printf(" %04X.o",F->ID);
    F=F->NextLinked;
  } while (F!=NULL);
  printf(" ");
  GetEnvVar("LIB");
  printf("/syslib%s\ncd ",UseStatic?".a $GPS_LIB":".so");
  GetEnvVar("CURWD");
  NewLine();
}

static void ParseMainFile(const char *ProgName,const char *MainFile)
{
  char SpecBuf[180];
  bool IsMainFile;
  TFile *Main;

  IsMainFile=true;
  StackTop=NULL;
  SpecBuf[0]='C';
  SpecBuf[179]='\0';
  StrNCpy(SpecBuf+1,MainFile,179);
  if (SpecBuf[179]!='\0') Error("File name too long");
  AddAndPush(SpecBuf);
  MainFile=SpecBuf;
  do {
    Current=StackTop;
    StackTop=StackTop->Next;
    ParseSingleFile(Current->FileName,IsMainFile);
    if (IsMainFile) {
      Main=FirstLinked;
      FirstLinked=Main->NextLinked;
    }
    IsMainFile=false;
  } while (StackTop!=NULL);
  /*
    Dump the chain of the linked files
  */
  if (!GenTurboC) {
    printf("cd ");
    GetEnvVar("TMP");
    NewLine();
  }
  printf("echo \"Linking %s\"",ProgName);
  NewLine();
  if (GenTurboC) {
    LinkProgramUsingTurboC(ProgName,Main);
  } else {
    LinkProgramUsingGCC(ProgName,Main);
  }
}

static void SetupCWD(void)
{
  char *Cwd;
  char Buffer[129];
  bool Bad;

  Cwd="`pwd`";
  if (GenTurboC) {
    do {
      Bad=true;
      if (!OpenTheFile("cwd.txt")) break;
      if (!ReadLine(Buffer,129)) break;
      Bad=false;
    } while (false);
    if (Bad) {
      Error("Unable to read current directory from \"cwd.txt\"");
    }
    Cwd=Buffer;
  }
  SetEnvVar("CURWD",Cwd);
}

static void BeginMsDos(void)
{
  if (GenMsDos) Error("Too many msdos format switches on commandline");
  GenMsDos=true;
  printf("@echo off\n");
}

int16 main(int16 argc,char **argv)
{
  int16 i;
  char *Arg;
  char *ProgName;
  char Ch;
  char *TmpDir;
  bool DefaultTemp;

  /*
    Create all the necessary structures.
  */
  Files=CreateDict();
  if (Files==NULL) Error("No memory for the file dictionary");
  /*
    Process the arguments and analyze the programs.
  */
  DefaultTemp=true;
  GenMsDos=false;
  GenTurboC=false;
  UseStatic=false;
  CurID=0;
  FirstLinked=NULL;
  TmpDir=argv[0];
  for (i=1;i<argc;i++) {
    Arg=argv[i];
    if (Arg[0]=='-') {
      if (TmpDir==NULL) {
        Error("Options allowed only at the start of the command line");
      }
      switch (Arg[1]) {
        case 'T': {
          TmpDir=Arg+2;
          DefaultTemp=false;
          break;
        }
        case 'D': {
          BeginMsDos();
          break;
        }
        case 't': {
          GenTurboC=true;
          BeginMsDos();
          break;
        }
        case 's': {
          UseStatic=true;
          break;
        }
      }
      continue;
    }
    if (DefaultTemp) {
      if (GenMsDos) {
        TmpDir="\\OSHSTMP";
      } else {
        TmpDir="/tmp/OSHSTMP";
      }
    }
    TempDir=TmpDir;
    ProgName=Arg;
    do {
      Ch=*(Arg++);
    } while ((Ch!=':') && (Ch!='\0'));
    if (Ch=='\0') {
      fprintf(stderr,"Invalid program specifier: %s\n",ProgName);
      return(255);
    }
    *(Arg-1)='\0';
    if (TmpDir!=NULL) {
      SetEnvVar("TMP",TmpDir);
      SetupCWD();
      TmpDir=NULL;
      if (GenTurboC) {
        printf("cd ");
        GetEnvVar("TMP");
        NewLine();
      }
    }
    ParseMainFile(ProgName,Arg);
  }
  /*
    Return success
  */
  return(0);
}
