#include <syslib/types.h>
#define __INTERFACE__
#include "datetime.h"

#include "strings.h"

static uint8 DaysInMonth(int16 Month)
/*
  Returns the number of days in a month. The months are numbered from 0.
*/
{
  upint8 *Days;
  Days=(upint8 *)"\x1F\x1C\x1F\x1E\x1F\x1E\x1F\x1F\x1E\x1F\x1E\x1F";
  return(Days[Month]);
}

void PackTimeStamp(TDateTime *T)
/*
  Pack the timestamp.
*/
{
  uint32 TimeStamp;
  uint16 Year;
  int16 LeapYear;
  int16 Month;
  int16 DayCnt;

  /*
    Setup the failure result.
  */
  T->TimeStamp=0xFFFFFFFF;
  /*
    Check the time.
  */
  if (T->Second>59) return;
  if (T->Minute>59) return;
  if (T->Hour>23) return;
  /*
    Convert the time.
  */
  TimeStamp=T->Second+T->Minute*60+T->Hour*3600;
  /*
    Convert and add the year.
  */
  if (T->Year<1970) return;
  Year=T->Year-1970;
  TimeStamp+=Year*365*86400+((Year+1)/4)*86400;
  /*
    Determine if the current year is the leap one.
  */
  LeapYear=(Year % 4)==2;
  /*
    Check and convert the day and month.
  */
  Month=T->Month;
  if (Month==0 || Month>12) return;
  if (T->Day==0 || (T->Day>DaysInMonth(Month-1)+LeapYear)) return;
  TimeStamp+=(T->Day-1)*86400;
  while (--Month) {
    DayCnt=DaysInMonth(Month-1);
    if (LeapYear && (Month==2)) DayCnt++;
    TimeStamp+=DayCnt*86400;
  }
  /*
    At the very end put the result to the apropiate place.
  */
  T->TimeStamp=TimeStamp;
}

static int16 AnalyzeNumber(const char *Number,uint8 Places,int16 *Buffer)
/*
  Analyze the number. The [Number] points to data with the number to be
  analyzed, [Places] is the count of places to be analyzed and [Buffer]
  points to buffer where the analyzed number is to be written. If the number
  was successfully analyzed, the call returns true and [Buffer] is filled.
  If the analysis fails, the call returns false and [Buffer] is left intact.
  [Places] should be less than 5.
*/
{
  int16 ResNum;
  int16 i;
  char Ch;
  
  ResNum=0;
  for (i=0;i<Places;i++) {
    Ch=*(Number++);
    if (Ch<'0' || Ch>'9') return(false);
    ResNum*=10;
    ResNum+=Ch-'0';
  }
  *Buffer=ResNum;
  return(true);
}

uint32 AnalyzeTimeStamp(const char *Format,const char *Buffer)
/*
  TYPE: Public Code

  PURPOSE: Formatted timestamp analysis

  This function analyses the given timestamp information according to the
  format string specified. The function returns the timestamp in the canonical
  format, if the timestamp is recognized or 0xFFFFFFFF if not. The timestamp
  is not recognized if it does not match the format string, the format string
  is invalid or if the timestamp interpretation is ambigous (for example for
  format ""
*/
{
  const char *F;
  const char *D;
  TDateTime R;
  int16 Value;
  int16 Good;
  char Ch;

  /*
    Initialize the result.
  */
  R.Year=0;
  R.Month=0;
  R.Day=0;
  R.Hour=100;
  R.Minute=100;
  R.Second=100;
  /*
    Parse the input.
  */
  F=Format;
  D=Buffer;
  do {
    Ch=*(F++);
    Value=0;
    switch (Ch) {
      case 'Y': {  /* year number (4-digit) */
        Good=AnalyzeNumber(D,4,&Value);
        if (R.Year==0) R.Year=Value;
        Good=Good && (R.Year==Value);
        D+=4;
        break;
      }
      case 'y': {  /* year number (2-digit) */
        Good=AnalyzeNumber(D,2,&Value);
        Value+=1900;
        if (Value<1970) Value+=100;
        if (R.Year==0) R.Year=Value;
        Good=Good && (R.Year==Value);
        D+=2;
        break;
      }
      case 'm': {  /* month number (2-digit) */
        Good=AnalyzeNumber(D,2,&Value);
        if (R.Month==0) R.Month=Value;
        Good=Good && (R.Month==Value);
        D+=2;
        break;
      }
      case 'N': {  /* month number ((1 or 2)-digit) */
        Good=false;
        break;
      }
      case 'F': {  /* month name (English) */
        Good=false;
        break;
      }
      case 'M': {  /* month name shortcut (English) */
        Good=false;
        break;
      }
      case 'd': {  /* day-in-month number (2-digit) */
        Good=AnalyzeNumber(D,2,&Value);
        if (R.Day==0) R.Day=Value;
        Good=Good && (R.Day==Value);
        D+=2;
        break;
      }
      case 'j': {  /* day-in-month number ((1 or 2)-digit) */
        Good=false;
        break;
      }
      case 'S': {  /* tail of the number (e.g. 'th') */
        Good=false;
        break;
      }
      case 'w': {  /* day-in-week number (1-digit, 0=Sunday) */
        Good=false;
        break;
      }
      case 'D': {  /* day-in-week name shortcut (English) */
        Good=false;
        break;
      }
      case 'l': {  /* day-in-week name (English) */
        Good=false;
        break;
      }
      case 'z': {  /* day-in-year number ((1 to 3)-digit) */
        Good=false;
        break;
      }
      case 'H': {  /* hour (00 to 23) */
        Good=AnalyzeNumber(D,2,&Value);
        if (R.Hour==100) R.Hour=Value;
        Good=Good && (R.Hour==Value);
        D+=2;
        break;
      }
      case 'G': {  /* hour (0 to 23) */
        Good=false;
        break;
      }
      case 'h': {  /* hour (01 to 12) */
        Good=false;
        break;
      }
      case 'g': {  /* hour (1 to 12) */
        Good=false;
        break;
      }
      case 'i': {  /* minute (00 to 59) */
        Good=AnalyzeNumber(D,2,&Value);
        if (R.Minute==100) R.Minute=Value;
        Good=Good && (R.Minute==Value);
        D+=2;
        break;
      }
      case 's': {  /* second (00 to 59) */
        Good=AnalyzeNumber(D,2,&Value);
        if (R.Second==100) R.Second=Value;
        Good=Good && (R.Second==Value);
        D+=2;
        break;
      }
      case 'A': {  /* AM/PM identifier */
        Good=false;
        break;
      }
      case 'a': {  /* am/pm identifier */
        Good=false;
        break;
      }
      default: {
        Good=(Ch==(*D));
	D++;
        break;
      }
    }
  } while (Good && Ch);
  /*
    If parsing failed, return failure and don't waste time with attempts to
    construct the result.
  */
  if (!Good) return(0xFFFFFFFF);
  /*
    Supply default values of the timestamp.
  */
  if (R.Year==0) R.Year=1970;
  if (R.Month==0) R.Month=1;
  if (R.Day==0) R.Day=1;
  if (R.Hour==100) R.Hour=0;
  if (R.Minute==100) R.Minute=0;
  if (R.Second==100) R.Second=0;
  /*
    Pack and return the timestamp. If the timestamp read was an invalid one,
    it will be detectected in the timestamp packing call.
  */
  PackTimeStamp(&R);
  return(R.TimeStamp);
}

static void ConvertNumber(uint32 Number,uint8 Places,char *Buffer)
/*
  TYPE: Local Code

  PURPOSE: Number conversion

  This procedure converts the number to it's string representation. If the
  number's representation is shorter than the number of characters requested,
  the number is padded right with zeroes. If the number's representation is
  longer than the number of characters requested, only the rightmost digits
  are stored.
*/
{
  Buffer+=Places;
  for (;Places>0;Places--) {
    Buffer--;
    (*Buffer)=(char)(Number%10+'0');
    Number/=10;
  }
}

static char *CopyBufWithoutZeroes(const char *Src,char *Dst,uint32 Len)
/*
  TYPE: Local Code

  PURPOSE: Buffer copying with zeroes removal

  Does the same as CopyBuf() but removes any leading zeroes first. If the
  source string is all zeroes, one zero will be written to output.
*/
{
  while ((Len>1) && (*Src=='0')) {
    Src++;
    Len--;
  }
  return(MemCpy(Dst,Src,Len));
}

static const char *GetStr(const char *List,int16 Num)
{
  while (Num) {
    List++;
    if ((*List)<'a') Num--;
  }
  return(List);
}

static char *CopyName(char *Dst,const char *Src)
{
  do {
    *(Dst++)=*(Src++);
  } while ((*Src)>='a');
  return(Dst);
}

void UnpackTimeStamp(TDateTime *T)
{
  uint32 TimePortion;
  uint32 DatePortion;
  uint32 DaysForYear;
  bool Continue;
  uint8 DayCnt;

  DatePortion=T->TimeStamp / 86400;
  T->DayInWeek=(DatePortion+4)%7;

  /* Determine the year */
  T->Year=(DatePortion / (365*4+1))*4+1970;
  DatePortion%=365*4+1;
  do {
    DaysForYear=(T->Year%4?365:366);
    Continue=(DatePortion>=DaysForYear);
    if (Continue) {
      T->Year++;
      DatePortion-=DaysForYear;
    }
  } while (Continue);

  /* Determine the month and day */
  T->DayInYear=DatePortion;
  T->Month=0;
  do {
    DayCnt=DaysInMonth(T->Month++);
    if ((T->Month==2) && !(T->Year%4)) DayCnt++;
    Continue=(DatePortion>=DayCnt);
    if (Continue) {
      DatePortion-=DayCnt;
    }
  } while (Continue);
  T->Day=DatePortion+1;

  /*
    Determine the time (hour, minute and second).
  */
  TimePortion=T->TimeStamp % 86400;
  T->Hour=TimePortion / 3600;
  TimePortion%=3600;
  T->Minute=TimePortion / 60;
  T->Second=TimePortion % 60;
}

void FormatTimeStamp(const char *Format,uint32 Stamp,char *Buffer)
/*
  TYPE: Public Code

  PURPOSE: Timestamp formatting

  This code takes the timestamp in the UNIX format (measured as seconds since
  the start of the time (commonly known as the start of Epoch) when the leap
  seconds are ignored and all years divisible by 4 are leap years) and
  converts it into a human readable string according to the format string also
  specified by the caller.
*/
{
  TDateTime T;
  const char *R;
  char *W;
  char Ch;
  char Day[2];
  const char *MonthName;
  char Month[2];
  char Year[4];
  char Hour[2];
  char Hour12[2];
  char Min[2];
  char Sec[2];
  const char *DayName;
  char DayInYear[3];
  char *ampmIndi;
  char *AMPMIndi;
  int16 TailNum;
  const char *TailStr;

  /*
    Unpack the timestamp.
  */
  T.TimeStamp=Stamp;
  UnpackTimeStamp(&T);

  /*
    Determine the names 
  */
  MonthName=GetStr(
    "January" "February" "March" "April" "May" "June" "July"
    "August" "September" "October" "November" "December",
    T.Month-1
  );
  DayName=GetStr(
    "Sunday" "Monday" "Tuesday" "Wednesday" "Thursday" "Friday" "Saturday",
    T.DayInWeek
  );

  /* determine the tail for the day number */
  TailNum=((T.Day % 10)+9)%10;
  if (TailNum>3 || (T.Day/10)==1) TailNum=3;
  TailStr="st" "nd" "rd" "th";
  TailStr+=TailNum*2;

  /*
    Convert to strings
  */
  ConvertNumber(T.Day,2,Day);
  ConvertNumber(T.Month,2,Month);
  ConvertNumber(T.Year,4,Year);
  ConvertNumber(T.Hour,2,Hour);
  ConvertNumber(((T.Hour+11)%12)+1,2,Hour12);
  ConvertNumber(T.Minute,2,Min);
  ConvertNumber(T.Second,2,Sec);
  ConvertNumber(T.DayInYear,3,DayInYear);
  ampmIndi=(T.Hour>11)?"pm":"am";
  AMPMIndi=(T.Hour>11)?"PM":"AM";

  /*
    Convert the string and put the date into it.
  */

  R=Format;
  W=Buffer;
  do {
    Ch=*(R++);
    switch (Ch) {
      case 'Y': {  /* year number (4-digit) */
        W=MemCpy(W,Year,4);
        break;
      }
      case 'y': {  /* year number (2-digit) */
        W=MemCpy(W,Year+2,2);
        break;
      }
      case 'm': {  /* month number (2-digit) */
        W=MemCpy(W,Month,2);
        break;
      }
      case 'N': {  /* month number ((1 or 2)-digit) */
        W=CopyBufWithoutZeroes(Month,W,2);
        break;
      }
      case 'F': {  /* month name (English) */
        W=CopyName(W,MonthName);
        break;
      }
      case 'M': {  /* month name shortcut (English) */
        W=MemCpy(W,MonthName,3);
        break;
      }
      case 'd': {  /* day-in-month number (2-digit) */
        W=MemCpy(W,Day,2);
        break;
      }
      case 'j': {  /* day-in-month number ((1 or 2)-digit) */
        W=CopyBufWithoutZeroes(Day,W,2);
        break;
      }
      case 'S': {  /* tail of the number (e.g. 'th') */
        W=MemCpy(W,TailStr,2);
        break;
      }
      case 'w': {  /* day-in-week number (1-digit, 0=Sunday) */
        *(W++)=(char)(T.DayInWeek+'0');
        break;
      }
      case 'D': {  /* day-in-week name shortcut (English) */
        W=MemCpy(W,DayName,3);
        break;
      }
      case 'l': {  /* day-in-week name (English) */
        W=CopyName(W,DayName);
        break;
      }
      case 'z': {  /* day-in-year number ((1 to 3)-digit) */
        W=MemCpy(W,DayInYear,3);
        break;
      }
      case 'H': {  /* hour (00 to 23) */
        W=MemCpy(W,Hour,2);
        break;
      }
      case 'G': {  /* hour (0 to 23) */
        W=CopyBufWithoutZeroes(Hour,W,2);
        break;
      }
      case 'h': {  /* hour (01 to 12) */
        W=MemCpy(W,Hour12,2);
        break;
      }
      case 'g': {  /* hour (1 to 12) */
        W=CopyBufWithoutZeroes(Hour12,W,2);
        break;
      }
      case 'i': {  /* minute (00 to 59) */
        W=MemCpy(W,Min,2);
        break;
      }
      case 's': {  /* second (00 to 59) */
        W=MemCpy(W,Sec,2);
        break;
      }
      case 'A': {  /* AM/PM identifier */
        W=MemCpy(W,AMPMIndi,2);
        break;
      }
      case 'a': {  /* am/pm identifier */
        W=MemCpy(W,ampmIndi,2);
        break;
      }
      default: {
        *(W++)=Ch;
        break;
      }
    }
  } while (Ch);
}
