#include <syslib/types.h>
#define __INTERFACE__
#include "dictrb.h"

#include <syslib/traps.h>

#include "strings.h"
#include "memory.h"

#define rbRED 1
#define ndTREE 2
#define ndOWNINGKEY 4
#define ndPOINTERDICT 8
#define ndITERATING 16

#define stFRESH 0
#define stGOING_LEFT 32
#define stGOING_RIGHT 64
#define stPROCESSED (32+64)
#define ndVISIT_FLAGS (stFRESH|stGOING_LEFT|stGOING_RIGHT|stPROCESSED)

bool IsRed(TRBDict *X)
{
  return((X->Flags & rbRED)>0);
}

bool IsBlack(TRBDict *X)
{
  return(!(X->Flags & rbRED));
}

void SetRed(TRBDict *X)
{
  X->Flags|=rbRED;
}

void SetBlack(TRBDict *X)
{
  X->Flags&=~rbRED;
}

void SetColor(TRBDict *X,TRBDict *Y)
{
  X->Flags=(X->Flags&~rbRED)+(Y->Flags&rbRED);
}

/*
  The NIL sentinel. Left and Right fields are directed back to NIL, Parent
  is used as scratch area 
*/
static TRBDict NIL;

/*
  The tree handle uses the filds differently than the tree node:
  
  Field		Tree node usage			Tree handle usage
  
  Left		The root of the left subtree	The root of the entire tree
  
  Right		The root of the right subtree	Pointer to the NIL sentinel
						during the tree color fixup.
						This is required by the tree
						insertion call, where it is
						stepping toward the root to
						correctly determine the
						root as being the left son of
						the tree handle node. When
						creating and inserting the
						new node into the tree, the
						node pointer is passed around
						also in this field.

  Parent 	The parent of the node		After successfull search for
						a key points to the node with
						the key. After unsuccessfull
						search for a key points to the
						node where the search ended.
						This field is used by the
						tree insertion procedure as
						the starting node for the new
						node. This may be the
						dictionary handle itself if
						the dictionary is still empty.

  Data		User supplied pointer		After unsuccessfull search
						for a key is NULL if the
						key of the T->Parent node is
						less than the key searched
						for, otherwise it is non-NULL
						(currently the NIL sentinel
						pointer is placed here in
						that case).
*/

TRBDict *CreateDict(void)
{
  TRBDict *T;

  if (!NIL.Left) {
    NIL.Left=&NIL;
    NIL.Right=&NIL;
  }
  T=GetMem(sizeof(*T));
  if (T==NULL) return(T);
  T->Left=&NIL;
  T->Right=&NIL;
  T->Flags=ndTREE;
  return(T);
}

TRBDict *CreatePointerDict(void)
{
  TRBDict *T;

  T=CreateDict();
  if (T!=NULL) T->Flags|=ndPOINTERDICT;
  return(T);
}

static void CancelIteration(TRBDict *T)
{
  if ((T->Flags & ndITERATING)>0) {
    T->Right=&NIL;
    T->Flags&=~ndITERATING;
  }
}

static void RotateLeft(TRBDict *X)
/*
                           P                           P
                           |                           |
                           X                           Y
                          / \                         / \
                         A   Y       -------->       X   C
                            / \                     / \
                           B   C                   A   B


 *X->Parent                P                         Y
  X->Left                  A                         A
 *X->Right                 Y                         B
 *Y->Parent                X                         P
 *Y->Left                  B                         X
  Y->Right                 C                         C
 *B->Parent                Y                         X
 *P->[Left|Right]          X                         Y

  Affected nodes: X,Y,P,B

  We assume X has right son.
*/
{
  TRBDict *Y;
  TRBDict *B;
  TRBDict *P;

  /*
    As said above, we assume that X has right son.
  */
  assert(X->Right!=&NIL);
  /*
    Some important nodes of the tree during the operation: XP = current
    parent of X, Y = the right son (must be non-NIL), YL = the left subtree
    of Y.
  */
  P=X->Parent;
  Y=X->Right;
  B=Y->Left;
  /*
    Left subtree of Y now become right subtree of X.
  */
  X->Right=B;
  B->Parent=X;
  /*
    The parent of Y is now XP (the old parent of X).
  */
  Y->Parent=P;
  /*
    Update the son of the XP (was X, now is Y).
  */
  if (X==P->Left) {
    P->Left=Y;
  } else {
    P->Right=Y;
  }
  /*
    X becomes left son of Y.
  */
  Y->Left=X;
  X->Parent=Y;
}

static void RotateRight(TRBDict *X)
/*
                             P                       P
                             |                       |
                             X                       Y
                            / \                     / \
                           Y   C     -------->     A   X
                          / \                         / \
                         A   B                       B   C


 *X->Parent                P                         Y
 *X->Left                  Y                         B
  X->Right                 C                         C
 *Y->Parent                X                         P
  Y->Left                  A                         A
 *Y->Right                 B                         X
 *B->Parent                Y                         X
 *P->[Left|Right]          X                         Y

  Affected nodes: X,Y,P,B

  We assume X has left son.
*/
{
  TRBDict *Y;
  TRBDict *B;
  TRBDict *P;

  /*
    As said above, we assume that X has left son.
  */
  assert(X->Left!=&NIL);
  /*
    Fetch affected nodes.
  */
  P=X->Parent;
  Y=X->Left;
  B=Y->Right;
  /*
    Left subtree of Y now become right subtree of X.
  */
  X->Left=B;
  B->Parent=X;
  /*
    The parent of Y is now P (the old parent of X).
  */
  Y->Parent=P;
  /*
    Update the son of P (was X, now is Y).
  */
  if (X==P->Left) {
    P->Left=Y;
  } else {
    P->Right=Y;
  }
  /*
    X becomes left son of Y.
  */
  Y->Right=X;
  X->Parent=Y;
}

static void *SearchKeyCore(TRBDict *T,const char *Key)
{
  TRBDict *X;
  TRBDict *Y;
  int16 Res;

  /*
    Sanity checks. The passed dictionary must be a valid tree handle.
  */
  assert((T->Flags & ndTREE)>0);
  /*
    Initialize. Note that T->LeftRoot is in fact T->Root.
  */
  Y=T;
  X=T->Left;
  Res=-1;
  /*
    Search the tree for the key.
  */
  while (X!=&NIL) {
    Y=X;
    if ((T->Flags & ndPOINTERDICT)>0) {
      if (Key<X->Key) {
        Res=-1;
      } else if (Key==X->Key) {
        Res=0;
      } else {
        Res=1;
      }
    } else {
      Res=StrCmp(Key,X->Key);
    }
    if (!Res) {
      T->Parent=X;
      return(X->Data);
    }
    if (Res<0) {
      X=X->Left;
    } else {
      X=X->Right;
    }
  }
  /*
    Export return values if not found.
  */
  T->Parent=Y;
  T->Data=NULL;
  if (Res<0) T->Data=&NIL;
  return(NULL);
}

void *SearchKey(TRBDict *T,const char *Key)
{
  assert((T->Flags & ndPOINTERDICT)==0);
  return(SearchKeyCore(T,Key));
}

void *SearchPointerKey(TRBDict *T,const void *Key)
{
  assert((T->Flags & ndPOINTERDICT)!=0);
  return(SearchKeyCore(T,Key));
}

static void AddTheNewNode(TRBDict *T)
{
  TRBDict *X;
  TRBDict *Y;
  TRBDict *P;
  TRBDict *PP;

  /*
    Get the new node being added.
  */
  X=T->Right;
  assert(X!=NULL && X!=&NIL);
  X->Left=&NIL;
  X->Right=&NIL;
  T->Right=&NIL;
  /*
    The node which should be parent to the key is placed into T->Parent.
    If the key should become left son, the T->Data will be set to non-NULL
    otherwise it will be set to NULL.

    However if the key should become root of the tree, the returned parent
    would be NIL. I'm using an optimization trick here. Since the root
    pointer is stored in T->Left, I return T->Parent==T and T->Data==non-NULL
    to fool the code placing the new node to T->Root (which is in fact
    T->Left).

    Note also that the pointer being updated must always be NIL.
  */
  Y=T->Parent;
  X->Parent=Y;
  if (T->Data) {
    assert(Y->Left==&NIL);
    Y->Left=X;
  } else {
    assert(Y->Right==&NIL);
    Y->Right=X;
  }
  /*
    Color the node we just added to RED.
  */
  X->Flags=rbRED;
  /*
    Restore the RB tree properties.
  */
  do {
    P=X->Parent;
    if (X==T->Left || IsBlack(P)) break;
    PP=P->Parent;
    if (P==PP->Left) {
      Y=PP->Right;
      if (IsRed(Y)) {
        SetBlack(P);
        SetBlack(Y);
        SetRed(PP);
	X=PP;
      } else {
        if (X==P->Right) {
	  X=P;
	  RotateLeft(X);
	  P=X->Parent;
	}
        SetBlack(P);
        SetRed(PP);
	RotateRight(PP);
      }
    } else {
      Y=PP->Left;
      if (IsRed(Y)) {
        SetBlack(P);
        SetBlack(Y);
        SetRed(PP);
	X=PP;
      } else {
        if (X==P->Left) {
	  X=P;
	  RotateRight(X);
	  P=X->Parent;
	}
        SetBlack(P);
        SetRed(PP);
	RotateLeft(PP);
      }
    }
  } while (true);
  SetBlack(T->Left);
}

static int16 BuildNode(const char *Key,void *Data,TRBDict *T)
{
  TRBDict *X;

  assert(Data!=NULL);
  CancelIteration(T);
  /*
    Search for the key to be inserted into the tree. If the key is found,
    return failure (I don't want duplicates in the dictionary).
  */
  if (SearchKeyCore(T,Key)) return(rbDUPLICATE);
  /*
    Attempt to create "container" for the data.
  */
  X=GetMem(sizeof(*X));
  if (X==NULL) return(rbNOMEM);
  X->Key=Key;
  X->Data=Data;
  T->Right=X;
  return(rbSUCCESS);
}

int16 AddItem(TRBDict *T,const char *Key,void *Data)
{
  TRBDict *X;
  int16 Result;

  /*
    This call is not allowed on dictionaries with pointers as keys.
  */
  assert((T->Flags & ndPOINTERDICT)==0);
  /*
    Build the node for the RB dictionary
  */
  Result=BuildNode(Key,Data,T);
  if (Result!=rbSUCCESS) return(Result);
  /*
    Create an instance of the dictionary key string
  */
  X=T->Right;
  X->Key=StrDup(Key);
  if (X->Key==NULL) {
    FreeMem(X);
    T->Right=&NIL;
    return(rbNOMEM);
  }
  /*
    Add the node into the dictionary
  */
  AddTheNewNode(T);
  /*
    Set its key as owned
  */
  X->Flags|=ndOWNINGKEY;
  /*
    Return success.
  */
  return(rbSUCCESS);
}

static int16 AddItemWithKeyCore(TRBDict *T,const void *Key,void *Data)
{
  int16 Result;

  /*
    Build the node for the RB dictionary
  */
  Result=BuildNode(Key,Data,T);
  if (Result!=rbSUCCESS) return(Result);
  /*
    Add the node into the dictionary
  */
  AddTheNewNode(T);
  /*
    Return success.
  */
  return(rbSUCCESS);
}

int16 AddItemWithKey(TRBDict *T,const char *Key,void *Data)
{
  assert((T->Flags & ndPOINTERDICT)==0);
  return(AddItemWithKeyCore(T,Key,Data));
}

int16 AddPointerItem(TRBDict *T,const void *Key,void *Data)
{
  assert((T->Flags & ndPOINTERDICT)!=0);
  return(AddItemWithKeyCore(T,Key,Data));
}

static void FreeAssociates(TRBDict *Z)
{
  if ((Z->Flags&ndOWNINGKEY)>0) FreeMem(Z->Key);
}

static bool DeleteItemCore(TRBDict *T, const char *Key)
{
  TRBDict *X;
  TRBDict *Y;
  TRBDict *Z;
  TRBDict *W;
  TRBDict *P;

  CancelIteration(T);
  /*
    Find the item to be deleted.
  */
  X=SearchKeyCore(T,Key);
  if (X==NULL) return(false);
  Z=T->Parent;
  /*
    Destroy the key of the node. I do this here because later I don't want
    to waste time with keeping that pointer around.
  */
  FreeAssociates(Z);
  /*
    Find the node to be deleted. Things are complicated a bit when the node
    we found has both subtrees nonempty. In that case I must find the
    successor (which is a suitable node) and swap it with the node being
    deleted to make the deletion feasible. So I do this scan and mark the
    node to be deleted as Y.
  */
  Y=Z;
  if ((Z->Left!=&NIL) && (Z->Right!=&NIL)) {
    X=Z->Right;
    while (X!=&NIL) {
      Y=X;
      X=X->Left;
    }
    Z->Data=Y->Data;
    Z->Key=Y->Key;
  }
  /*
    Disconnect node Y from the tree.
  */
  X=Y->Left;
  if (X==&NIL) X=Y->Right;
  P=Y->Parent;
  X->Parent=P;
  if (Y==P->Left) {
    P->Left=X;
  } else {
    P->Right=X;
  }
  /*
    Discard the deleted node. However I need to check it's color and do
    some fixup. So if this fixup is not needed, the code will return here
    with the success flag.
  */
  if (IsRed(Y)) X=NULL;
  FreeMem(Y);
  if (!X) return(true);
  /*
    The node is away but our tree is not RB-tree. The problem is with the
    black color that disappeared with the deleted node Y.
  */
  while ((X!=T->Left) && IsBlack(X)) {
    P=X->Parent;
    if (X==P->Left) {
      W=P->Right;
      if (IsRed(W)) {
        SetBlack(W);
        SetRed(P);
        RotateLeft(P);
        W=P->Right;
      }
      if (IsBlack(W->Left) && IsBlack(W->Right)) {
        SetRed(W);
        X=P;
      } else {
        if (IsBlack(W->Right)) {
          SetBlack(W->Left);
          SetRed(W);
          RotateRight(W);
          W=P->Right;
        }
        SetColor(W,P);
        SetBlack(P);
        SetBlack(W->Right);
        RotateLeft(P);
        X=T->Left;
      }
    } else {
      W=P->Left;
      if (IsRed(W)) {
        SetBlack(W);
        SetRed(P);
        RotateRight(P);
        W=P->Left;
      }
      if (IsBlack(W->Left) && IsBlack(W->Right)) {
        SetRed(W);
        X=P;
      } else {
        if (IsBlack(W->Left)) {
          SetBlack(W->Right);
          SetRed(W);
          RotateLeft(W);
          W=P->Left;
        }
        SetColor(W,P);
        SetBlack(P);
        SetBlack(W->Left);
        RotateRight(P);
        X=T->Left;
      }
    }
  }
  SetBlack(X);
  /*
    Everything went OK, so we are ready to exit.
  */
  return(true);
}

bool DeleteItem(TRBDict *T, const char *Key)
{
  assert((T->Flags & ndPOINTERDICT)==0);
  return(DeleteItemCore(T,Key));
}

bool DeletePointerItem(TRBDict *T, const void *Key)
{
  assert((T->Flags & ndPOINTERDICT)!=0);
  return(DeleteItemCore(T,Key));
}

void DestroyDict(TRBDict *T)
{
  TRBDict *X;
  TRBDict *Y;

  X=T->Left;
  if (X!=&NIL) do {
    Y=X->Left;
    if (Y!=&NIL) {
      X->Left=&NIL;
    } else {
      Y=X->Right;
      if (Y!=&NIL) {
        X->Right=&NIL;
      } else {
        Y=X->Parent;
        FreeAssociates(X);
        FreeMem(X);
      }
    }
    X=Y;
  } while (X!=T);
  FreeMem(T);
}

/*
  These calls allow the caller to iterate the dictionaries. The
  FindFirstKey() places the first key and its corresponding value to
  the 'Key' and 'Value' fields of the dictionary record. The FindNextKey()
  places the next key into the 'Key' and 'Value' fields into the dictionary.
  These keys are found in the sorted order and both calls return true if a
  key is found or false otherwise (in that case also the 'Key' field is set
  to NULL). The dictionary must not be changed during the iteration process,
  after any change the iteration is cancelled and must be restarted with
  FindFirstKey() or an assertion failure occurs.
*/

static void FetchData(TRBDict *T)
{
  T->Key=T->Right->Key;
  T->Data=T->Right->Data;
}

void *FindFirstKey(TRBDict *T)
{
  if (T->Left==&NIL) return(false);
  T->Right=T->Left;
  T->Right->Flags&=~ndVISIT_FLAGS;
  T->Flags|=ndITERATING;
  return(FindNextKey(T));
}

void *FindNextKey(TRBDict *T)
{
  int16 Status;
  TRBDict *New;
  bool Clear;
  void *Result;

  assert((T->Flags&ndITERATING)>0);
  T->Key=NULL;
  do {
    Status=T->Right->Flags & ndVISIT_FLAGS;
    T->Right->Flags |= ndVISIT_FLAGS;
    NIL.Parent=T->Right;
    Clear=false;
    switch (Status) {
      case stFRESH: {
        Status=stGOING_LEFT;
        New=T->Right->Left;
        Clear=true;
        break;
      }
      case stGOING_LEFT: {
        FetchData(T);
        Status=stGOING_RIGHT;
        New=T->Right->Right;
        Clear=true;
        break;
      }
      case stGOING_RIGHT: {
        Status=stFRESH;
        New=T->Right->Parent;
        break;
      }
      default: {
        ReportBug("Node with an invalid status found");
        return(false);
      }
    }
    T->Right->Flags &= ~ndVISIT_FLAGS;
    T->Right->Flags |= Status;
    if (New!=&NIL) {
      T->Right=New;
      if (Clear) T->Right->Flags&=~ndVISIT_FLAGS;
    }
  } while (T->Right!=T && T->Key==NULL);
  if (T->Key!=NULL) {
    Result=T->Data;
  } else {
    CancelIteration(T);
    Result=NULL;
  }
  return(Result);
}
