#include <syslib/hhead.h>
#ifndef __h__LIBHS_DICTRB__
#define __h__LIBHS_DICTRB__

/*T-Main------------------------ RB-tree dictionaries
  The RB tree dictionaries are used to hold string-indexed or pointer-indexed
  data collections. These dictionaries trade speed for memory usage; a
  typical dictionary operation lasts O(<EM>n</EM>*log(<EM>n</EM>)), where
  <EM>n</EM> is count of items in the dictionary. However the memory used to
  hold the dictionary is always linear to count of items in the dictionary.

  The library uses RB-trees to implement the dictionary operations.
*/
/*S-TRBTree--------------------- RB-tree node
  This structure is used to represent a RB-tree. The user of this module
  should treat this as an opaque handle.

  <BLOCKQUOTE>
  <langC>
*/
typedef struct tagRBDict {
  /*
    Various flags used to maintain the tree. One of the most important flags
    is the color of the node.
  */
  int8 Flags;
  /*
    Pointer to the left son of the node.
  */
  struct tagRBDict *Left;
  /*
    Pointer to the right son of the node.
  */
  struct tagRBDict *Right;
  /*
    Pointer to the parent of the node.
  */
  struct tagRBDict *Parent;
  /*
    Pointer to the string used as the key.
  */
  const char *Key;
  /*
    Pointer to the data stored along with the key.
  */
  void *Data;
} TRBDict;
/*
  </langC>
  </BLOCKQUOTE>

  <STRONG>DESCRIPTION</STRONG>

  This structure represents both things: the tree itself and the tree nodes.
  The description of this structure mentioned above is for the tree nodes.
  The tree itself holds in the <CODE>Left</CODE> field pointer to the
  structure itself, in the <CODE>Right</CODE> field pointer to NIL sentinel
  (to avoid confusion that would happen when this field contained garbage
  that happen to be equal to the content of <CODE>Left</CODE> field)
  and <CODE>Flags</CODE> holds information that it is the tree descriptor.
  The other fields are used as temporary workspace during the dictionary
  operations execution and contain garbage.
*/
/*------------------------------
*/
#define rbSUCCESS 0
#define rbDUPLICATE 1
#define rbNOMEM 2
/*------------------------------
*/
void *SearchKey(TRBDict *T,const char *Key);
TRBDict *CreateDict(void);
int16 AddItem(TRBDict *T,const char *Key,void *Data);
int16 AddItemWithKey(TRBDict *T,const char *Key,void *Data);
bool DeleteItem(TRBDict *T, const char *Key);
void DestroyDict(TRBDict *T);

TRBDict *CreatePointerDict(void);
void *SearchPointerKey(TRBDict *T,const void *Key);
int16 AddPointerItem(TRBDict *T,const void *Key,void *Data);
bool DeletePointerItem(TRBDict *T, const void *Key);

void *FindFirstKey(TRBDict *T);
void *FindNextKey(TRBDict *T);
/*------------------------------
*/
#endif
#include <syslib/hfoot.h>
