#include <syslib/hhead.h>
#ifndef __h__LIBHS_DISKIO__
#define __h__LIBHS_DISKIO__

#include <syslib/kernel.h>
#include <syslib/pdbase.h>
#include <syslib/pdfsys.h>
#include <syslib/pdfileio.h>

int16 Open(const char *FileSpec,uint8 Mode);
uint16 Read(int16 fd,void *Buf,uint16 Size);
uint16 Write(int16 fd,const void *Buf,uint16 Size);
int32 HugeRead(int16 fd,void *Buf,int32 Size);
int32 HugeWrite(int16 fd,const void *Buf,int32 Size);
bool Seek(int16 fd,uint32 Where);
bool SeekRel(int16 fd,int32 Where);
uint32 Tell(int16 fd);
bool Truncate(int16 fd);
bool FChMod(int16 fd,uint16 NewAttr);
bool FChOwn(int16 fd,uint16 NewOwner);
bool FChGrp(int16 fd,uint16 NewGroup);
bool FStat(int16 fd,TFileInfo *Info);
bool Close(int16 fd);

/*T-DirectoryScanningDiffs------ About differences of directory scanning calls
  The directory scanning calls are designed so that applications can achieve
  better performance of their scanning routinnes than with traditional POSIX
  calls.

  With POSIX calls you open a directory, read it's content and close it.
  This gives you the basic informations about the files: file names and
  something called "inode numbers".

  Sometimes the basic informations about the files in a particular directory
  is not enough for your program. The program may want informations such as
  file permission bits or file size. In this case the program is required for
  each such file (typically all the files read from the directory) to form
  the complete file specification and call <EM>stat()</EM> on it. Or it can
  go to the directory being read and then pass only the file names read from
  it to <EM>stat()</EM>. There is no POSIX call that takes an inode number
  and returns the result of <EM>stat()</EM> for file with that particular
  inode number.

  This design is somewhat cumbersome. It means that the underlying operating
  system in such a scenario must search for these files multiple times: not
  only when the filename is read from the directory, but also when the
  supplementary informations are being served. Thus serious performance
  penalty is involved. Also when the application is getting the supplementary
  file data using absolute file specifications, the system for each of the
  files  has to search the directory tree for that file to obtain the inode
  number and then use that inode number to get the data the program wants.
  There is no way to get rid of this performance penalty, since there is no
  "inode stat" function defined in POSIX.

  Modern systems are not suffered with this bad design very badly, because
  they cache the directory trees in the memory. But this means that they need
  lots of memory if they host large trees of files.

  In <EM>SYSLIB</EM> I wanted to get out this stupid performance, because
  OSHS will tend to be a simple system with the case being an optional
  item. Under plain OSHS I would simply define another system call, which
  takes inode and returns the supplementary data without time-consuming
  directory tree searching, but I still have to cope with the rest of the
  world, which doesn't use OSHS and doesn't have such an useful system
  call available.

  The need to fill this gap in POSIX prompted for modification of the
  irectory scanning process. You don't get inode numbers from the directory
  via the ~ReadItem() call. Instead <EM>file identifiers</EM> are provided by
  it. After the content of the directory read these identifiers can be used
  to obtain the supplementary informations about the files, where (among
  other things) the actual inode number will be returned. The validity of
  these file identifiers is restricted to the time when the directory they
  belong to is closed.

  This small change allows the underlying layer to be optimized so that
  reading the complete informations about files in a particular directory
  takes less time than with traditional POSIX calls. And this optimization
  can be done without affecting all the software walking around us and using
  our layer.

  In <EM>OSHS</EM> the numerical identificators obtained from the
  directory are simply the same as the inode numbers and ~ReadFileInfo() uses
  these inode numbers to get the supplementary informations about the files.
  This can be done, because the design of <EM>SYSLIB</EM> directory scanning
  functions we just described is compatible with that feature of OSHS.
  Thus the performance gain will be significant, especially on slow machines
  or long directories (or both, long directories on slow machines).
  
  But why the restriction saying that closing a directory will invalidate all
  file identifiers obtained from it? The reason is the other platforms.
  Since they (usually) have no such feature natively, <EM>SYSLIB</EM> must
  emulate it on them. Without the file identifier validity flags we need
  a dictionary which maps the inode numbers to the names of the files.
  However there can be from tens of thousands or even millions of files in
  typical filesystem. This means that the required dictionary would be
  extremely large, making the emulation of the OSHS-like directory reads
  almost impossible. With the validity restriction the size of the dictionary
  is dramatically reduced. Now we don't need to record all possible file
  identifiers, only the identifiers belonging to the open directories need
  to be maintained. This is much more reasonable, so the emulation is
  possible only with standard POSIX system calls without any special
  platform dependent help.

  This restriction has one small disadvantage. Since under OSHS actually
  the identifier isn't invalidated when the directory from which it was
  obtained, the behavior of the ~ReadFileInfo() call is platform-dependent
  when an invalid identifier is passed to it. One of the possible solutions
  is to force the OSHS version of the call to return eINVAL in this
  situation. But since this would decrease the performance of OSHS (which
  I want to keep as high as possible), I chose another solution: passing
  an invalid file identifier to ~ReadFileInfo() is declared being <EM>an
  undefined situation</EM>.
*/

/*P-OpenDir--------------------- Open a directory for reading
  Opens a directory for searching.
  
  <STRONG>SYNOPSIS</STRONG>
  
  <BLOCKQUOTE>
  <langC>
  #include <syslib/odskio.h>

*/
int16 OpenDir(const char *DirSpec);
/*
  </langC>
  </BLOCKQUOTE>
  
  <STRONG>PARAMETERS</STRONG>
  
  <ITEMS>
  <ITEM>DirSpec</ITEM>
    Specifies the directory to be opened
  </ITEMS>
  
  <STRONG>RETURN VALUE</STRONG>
  
  If successfull, the return value is a directory descriptor. If failed, the
  return value is -1.
  
  <STRONG>DESCRIPTION</STRONG>

  The <EM>OpenDir()</EM> system call is used to open a directory and
  prepare it for searching. It allows the ~ReadItem() to obtain it's entries
  and ~ReadFileInfo() to obtain informations about a particular file. The
  directory is "converted" to a directory descriptor (a small non-negative
  integer) that can be used to obtain data from the opened directory.

  <STRONG>ERRORS</STRONG>

  <ITEMS>
  <ITEM>eACCESS</ITEM>
    Permission denied.
  <ITEM>eMFILE</ITEM>
    Too many descriptors in use by process.
  <ITEM>eNFILE</ITEM>
    Too many files are currently open in the system.
  <ITEM>eNOMEM</ITEM>
    Not enough memory to complete the operation
  <ITEM>eNOTDIR</ITEM>
    <EM>DirSpec</EM> is not a directory
  <ITEM>eIN</ITEM>
    There was an error attempting to access the directory
  </ITEMS>
  
  <STRONG>NOTES</STRONG>

  <NOTES>
    The directory descriptors are not the same thing as file descriptors.
    Attempt to use a directory descriptor in ~Read(), ~Write(), etc. will
    fail. Also attempt to ~ReadItem() from a file descriptor will fail.
  <NOTE>
    Directories are "read only" from the view of <EM>OpenDir()</EM>. To
    modify a content of the directory use ~Open() (with onCREAT flag) and
    subsequent calls; ~Remove() and ~Rename().
  </NOTES>

  <STRONG>CONFORMING TO</STRONG>
  
  OSHS
  
  <STRONG>SEE ALSO</STRONG>
  
  ~ReadItem() ~CloseDir() ~ReadFileInfo()
*/

/*P-ReadItem-------------------- Read one item from opened directory
  Reads an item from the directory.

  <STRONG>SYNOPSIS</STRONG>
  
  <BLOCKQUOTE>
  <langC>
  #include <syslib/odskio.h>

*/
bool ReadItem(int16 dd,TDirItem *Buf);
/*
  </langC>
  </BLOCKQUOTE>

  <STRONG>PARAMETERS</STRONG>
  
  <ITEMS>
  <ITEM>dd</ITEM>
    Directory descriptor obtained from a call to ~OpenDir()
  <ITEM>Buf</ITEM>
    Pointer to place to which read the directory item.
  </ITEMS>
  
  <STRONG>RETURN VALUE</STRONG>
  
  On success the return value is true, on failure, the return value is false.
  
  <STRONG>DESCRIPTION</STRONG>
  
  This call reads items from the directory. It read only one item at a time.
  The item read is stored in the buffer pointed to <EM>Buf</EM>.

  If there is no directory item to be read, the call will set the name field
  of the buffer to empty string and the ID field to -1. All valid files have
  their names nonempty and their ID numbers non-negative.

  <STRONG>NOTES</STRONG>
  
  <NOTES>
    This call reads only the name of the file and an ID number. Use the
    ~ReadFileInfo() with the ID number to obtain the remaining informations
    about the file. This architecture of the call is due to performance
    reasons. More details on this are `DirectoryScanningDiffs here`.
  <NOTE>
    Directories are "read only" from the view of <EM>OpenDir()</EM>; there is
    no WriteItem() call. To modify a content of the directory use ~Open()
    (with onCREAT flag) and  subsequent calls; ~MkDir, ~RmDir, ~Remove() and
    ~Rename().
  <NOTE>
    The filename is returned with all-uppercase. On systems that have case
    sensitive filenames the files that have some uppercase characters in
    their names are skipped and ignored by the library, because ~Open() uses
    lowercase filenames on these platforms and thus it would fail on them
    with eNOENT.
  </NOTES>
*/

/*P-CloseDir-------------------- Close an opened directory
  Closes an opened directory.

  <STRONG>SYNOPSIS</STRONG>
  
  <BLOCKQUOTE>
  <langC>
  #include <syslib/odskio.h>

*/
bool CloseDir(int16 dd);
/*
  </langC>
  </BLOCKQUOTE>

  <STRONG>PARAMETERS</STRONG>
  
  <ITEMS>
  <ITEM>dd</ITEM>
    Directory descriptor obtained from a call to ~OpenDir()
  <ITEM>Buf</ITEM>
    Pointer to place to which read the directory item.
  </ITEMS>
  
  <STRONG>RETURN VALUE</STRONG>
  
  On success the return value is true, on failure, the return value is false.
  
  <STRONG>DESCRIPTION</STRONG>
  
  Closes the directory previously opened by ~OpenDir(). Use this call to
  free the descriptor when it is no longer needed.
*/

/*P-ReadFileInfo---------------- Read rest of informations about a file
  Reads informations about the file.
  
  <STRONG>SYNOPSIS</STRONG>
  
  <BLOCKQUOTE>
  <langC>
  #include <syslib/odskio.h>

*/
bool ReadFileInfo(uint32 Number,TFileInfo *Info);
/*
  </langC>
  </BLOCKQUOTE>

  <STRONG>PARAMETERS</STRONG>
  
  <ITEMS>
  <ITEM>Number</ITEM>
    The number of the file read from the directory content
  <ITEM>Info</ITEM>
    Buffer for the file information structure (~TFileInfo).
  </ITEMS>

  <STRONG>RETURN VALUE</STRONG>
  
  On success the return value is true, on failure, the return value is false.
  
  <STRONG>DESCRIPTION</STRONG>

  This call is used to obtain remaining data about a file such as file
  creation date, file size or permission bits.

  The number passed to this call must refer to a directory item that was read
  from an opened directory. Generally it is not an inode number read from the
  directory content, but a numeric identifier that is used to access the rest
  of the data associated with a file. However sometimes the inode number is
  used as an identifier, but the caller cannot depend on this.

  In fact this call is the same as ~FStat(), except file number is passed to
  it rather than file descriptor. The information returned is the same as the
  one returned from ~FStat().
  
  The rationale of separating the directory reads into two calls is that
  these file information data are often stored on a different location of the
  disk (for example in ext2 and ext3 filesystems have them stored in an area
  called "inode array") than the actual directory entries. Thus reading these
  data along with the directory content would enforce the disk to seek at
  each call to ~ReadItem(), significantly decreasing performance.

  <STRONG>UNDEFINED SITUATIONS</STRONG>

  <NOTES>
    The file numbers are valid only after they were read from the directory
    and before the directory from which they were read is closed. Attempt to
    use an invalid file number is considered undefined situation.
  </NOTES>

  <STRONG>NOTES</STRONG>

  <NOTES>
    The file number is valid only when the directory it was obtained from is
    opened. After the directory was closed, the file numbers (not to be
    confused with "inode numbers" used by some filesystems) become invalid
    and should not be used. This separation of file numbers and inode numbers
    is because Linux does not allow inode reads by inode number (you must
    provide a filename to it).
  <NOTE>
    The file numbers may change across scans. This means that if a directory
    is opened, read and closed, then opened and read again, the files may
    have different numbers.
  <NOTE>
    On some filesystems also the inode numbers may be changed unpredictably.
    This may be the case of for example <EM>vfat</EM> filesystem, where if a
    file is moved into another directory, it changes it's "inode number" (in
    fact, <EM>vfat</EM> does not use inode numbers at all).
  <NOTE>
    In OSHS this call is much more effective than constructing the file
    specification from the directory specification and the filename returned
    by ~ReadItem(), opening the file and then calling ~FStat() with it,
    because this call bypasses the time consuming filename lookup. However on
    other platform the performance gain may be not as significant as for
    OSHS.
  </NOTES>
*/

/*------------------------------
*/

bool MkDir(const char *DirSpec);
bool RmDir(const char *DirSpec);

bool Remove(const char *FileSpec);
bool Rename(const char *OldName,const char *NewName);

bool GetFirstDisk(TDriveInfo *Info);
bool GetNextDisk(TDriveInfo *Info);
char GetCurrentDrive(void);
bool SetCurrentDrive(char DriveLetter);

#endif
#include <syslib/hfoot.h>
