#include <syslib/types.h>
#define __INTERFACE__
#include "expbuf.h"

/*
  Expandable buffer support.

  The expandable buffers are somewhat similar to the expandable arrays. The
  only difference is that they contain bytes instead of void pointers and
  that multiple bytes can be placed into the array at once.
*/

#include <syslib/traps.h>

#include "strings.h"
#include "memory.h"

#define BUFFER_INCREMENT 4096

void InitBuffer(TBuffer *Buffer)
{
  Buffer->Size=0;
  Buffer->Ptr=NULL;
}

TBufferHandle *OpenBuffer(TBuffer *Buffer)
{
  TBufferHandle *Handle;

  Handle=GetMem(sizeof(*Handle));
  if (Handle==NULL) return(NULL);
  Handle->FreeIt=true;
  Handle->Buffer=Buffer;
  Handle->Space=Buffer->Size;
  return(Handle);
}

TBufferHandle *CreateBuffer(TBuffer *Buffer)
{
  InitBuffer(Buffer);
  return(OpenBuffer(Buffer));
}

void OpenBufferEx(TBuffer *Buffer, TBufferHandle *Handle)
{
  Handle->FreeIt=false;
  Handle->Buffer=Buffer;
  Handle->Space=Buffer->Size;
}

void CreateBufferEx(TBuffer *Buffer, TBufferHandle *Handle)
{
  InitBuffer(Buffer);
  OpenBufferEx(Buffer,Handle);
}

bool ExpandBuffer(TBufferHandle *Handle,uint16 DesiredSpace)
{
  int16 NewSize;
  void *New;

  if (Handle->Space-Handle->Buffer->Size<DesiredSpace) {
    NewSize=Handle->Space;
    NewSize+=BUFFER_INCREMENT*((DesiredSpace/BUFFER_INCREMENT)+1);
    New=ReallocMem(Handle->Buffer->Ptr,NewSize);
    if (New==NULL) return(false);
    Handle->Space=NewSize;
    Handle->Buffer->Ptr=New;
  }
  return(true);
}

bool PutData(TBufferHandle *Handle, const void *Data, uint16 Size)
{
  bool Result;
  upint8 *DestPtr;

  Result=ExpandBuffer(Handle,Size);
  if (Result) {
    DestPtr=Handle->Buffer->Ptr+Handle->Buffer->Size;
    MemCpy(DestPtr,Data,Size);
    Handle->Buffer->Size+=Size;
  }
  return(Result);
}

void DeleteData(TBufferHandle *Handle, uint16 Size)
{
  assert(Handle->Buffer->Size>=Size);
  Handle->Buffer->Size-=Size;
}

void CloseBuffer(TBufferHandle *Handle)
{
  Handle->Buffer->Ptr=ReallocMem(Handle->Buffer->Ptr,Handle->Buffer->Size);
  if (Handle->Buffer->Size>0) {
    if (Handle->Buffer->Ptr==NULL) ReportBug("Memory shrinking failed");
  }
  if (Handle->FreeIt) FreeMem(Handle);
}

void ClearBuffer(TBuffer *Buffer)
{
  DestroyBuffer(Buffer);
  InitBuffer(Buffer);
}

void DestroyBuffer(TBuffer *Buffer)
{
  if (Buffer->Ptr!=NULL) FreeMem(Buffer->Ptr);
}
