#include <syslib/hhead.h>
#ifndef __h__LIBHS_EXPBUF__
#define __h__LIBHS_EXPBUF__

/*
  Expandable buffers are arrays of bytes that can be expanded and shrinked
  as needed. However the buffer must be opened first before it can be
  expanded.

  The module works by maintaining a piece of free space after the last valid
  byte in the array. After the buffer is closed, the free space is stripped
  away.
*/

/*** Informations about the expandable buffer ***/
/*
  This structure holds informations about an expandable buffer.
*/
typedef struct tagTBuffer {
  /*
    Count of bytes in the array.
  */
  uint16 Size;
  /*
    Pointer to the array of the items.
  */
  upint8 *Ptr;
} TBuffer;
/*
  This structure is used to hold basic informations about the array. The
  array itself is allocated on the heap.
  
  The sparse names of the fields of this structure are intentional; normally
  the identifier which references the array itself says what is the
  particular array for.
*/

/*** The buffer handle internals ***/
/*
  This structure holds informations about an open expandable buffer.
*/
typedef struct tagTBufferHandle {
  /*
    Pointer to the asociated buffer.
  */
  TBuffer *Buffer;
  /*
    The size of the memory block for the buffer. This includes the size of
    the data in the buffer, which is stored in the buffer itself.
  */
  uint16 Space;
  /*
    If True, ~CloseBuffer() will free the handle after the array is closed.
  */
  bool FreeIt;
} TBufferHandle;
/*
  This 'opaque' structure acts as an identifier for an open expandable
  buffer. Although exposed to public, the content of this structure is
  internal to the module and the caller should not care about it nor inspect
  it or worse modify it. The content of this structure becomes meaningless
  when the buffer is closed.
*/

/*** Calls: Calls about the expandable arrays ***/

void InitBuffer(TBuffer *Buffer);
TBufferHandle *CreateBuffer(TBuffer *Buffer);
TBufferHandle *OpenBuffer(TBuffer *Buffer);
void CreateBufferEx(TBuffer *Buffer, TBufferHandle *Handle);
void OpenBufferEx(TBuffer *Buffer, TBufferHandle *Handle);
bool ExpandBuffer(TBufferHandle *Handle,uint16 DesiredSpace);
bool PutData(TBufferHandle *Handle, const void *Data, uint16 Size);
void DeleteData(TBufferHandle *Handle, uint16 Size);
void CloseBuffer(TBufferHandle *Handle);
void ClearBuffer(TBuffer *Buffer);
void DestroyBuffer(TBuffer *Buffer);

#endif
#include <syslib/hfoot.h>
