#include <syslib/types.h>
#define __INTERFACE__
#include "filebuf.h"

/*
  Name: filebuf.c

  Purpose: Simple buffered file access library.
*/

#include <syslib/kernel.h>
#include <syslib/pdbase.h>
#include <syslib/pdfsys.h>
#include <syslib/pdfileio.h>

#include "strings.h"
#include "memory.h"
#include "diskio.h"
#include "errno.h"

/********************************************/
/*** File I/O using the status structures ***/
/********************************************/
/*
  All calls in this sections take as one of their parameters the file status
  structure on which the operation should be performed.
*/

#define F_WRITING       256
#define F_OWNED         512

#define F_DEFAULT_BUFFER_SIZE	8192

/*
  Read bytes from handle. If the read operation returns fewer bytes than
  requested, return eREOF.
*/
static bool ReadAll(int16 fd,void *Buf,uint16 Size)
{
  bool Result;

  Result=Read(fd,Buf,Size)==Size;
  if (!Result && GetErrNo()==0) SetErrNo(lheREOF);
  return(Result);
}

/*
  Write bytes into handle. If the write operation writes fewer bytes than
  requested, return eIO.
*/
static bool WriteAll(int16 fd,const void *Buf,uint16 Size)
{
  bool Result;

  if (Size==0) return(true);
  Result=Write(fd,Buf,Size)==Size;
  if (!Result && GetErrNo()==0) SetErrNo(lheIO);
  return(Result);
}

/*
  Set the error flag in the provided file status structure according to the
  error informations provided by the external libraries.
*/
static void SetErrorFlag(PBufferedFile F)
{
  F->IOError=GetErrNo();
  if (F->IOError>0) F->IOError=0;
}

/*
  "Raw" seek of the file. This will move the file pointer of the underlying
  file to the location specified. The rest of the library maintains two file
  pointers (reading and writing) and uses this call when it needs to switch
  between them. Also the error flag in the file status structure is set if
  the seek operation fails.
*/
static bool RawSeek(PBufferedFile F,uint32 Temp)
{
  bool Result;

  Result=Seek(F->F,Temp);
  SetErrorFlag(F);
  return(Result);
}

/*
  Switch the file positions stored in the file status structure. The active
  one will become inactive and the other one will become active. One of the
  positions is the reading position and the other one is writing. This call
  is used when the file status structure is being switched between reading
  and writing operation.
*/
static bool SwitchFilePositions(PBufferedFile F)
{
  uint32 Temp;

  if (!FlushFile(F)) return(false);
  Temp=F->StoredPosition;
  F->StoredPosition=F->Position;
  F->Position=Temp;
  return(RawSeek(F,Temp));
}

/*
  Switch the given file status buffer into reading mode. This call is made
  before any reading operation. It checks the buffer if it is in reading mode
  and switches it if not. When the file is in reading mode, the active
  position is the reading one and the inactive position is the writing one.
  A flag is adjusted to show that the buffer is in reading mode.
*/
static bool SwitchRead(PBufferedFile F)
{
  bool Result;

  Result=true;
  if (F->Flags & F_WRITING) {
    if (!F->Flags & F_READ) {
      F->IOError=lheACCESS;
      return(false);
    }
    Result=SwitchFilePositions(F);
    F->Flags &= (~F_WRITING);
  }
  return(Result);
}

/*
  The same as SwitchRead() but for writing mode this time. This call is made
  before any writing operation. It checks the buffer if it is in writing
  mode and switches it if not. In writing mode the active file position is
  the writing one and the other file postion is the reading one. Also here
  is a flag adjusted to show the buffer is in writing mode as in the
  SwitchRead() call.
*/
static bool SwitchWrite(PBufferedFile F)
{
  bool Result;

  Result=true;
  if (!(F->Flags & F_WRITING)) {
    if (!F->Flags & F_WRITE) {
      F->IOError=lheACCESS;
      return(false);
    }
    Result=SwitchFilePositions(F);
    F->Flags |= F_WRITING;
  }
  return(Result);
}

/*
  Allocate a buffer for the file. Returns true if successfull and false
  otherwise.
*/
static bool AllocBufferForFile(PBufferedFile F,uint16 BufferSize)
{
  F->Buffer=GetMem(BufferSize);
  if (F->Buffer) {
    F->BufferSize=BufferSize;
    F->DataSize=0;
    F->DataPos=0;
    return(true);
  }
  return(false);
}

/*
  Open the file and associate it with file status structure. The mode
  argument tells the library the mode in which the file should be open.
  Some of the mode flags are stored in the file status structure to be
  able to do some sanity checks to catch errors like attempts to read from
  write-only files.
*/
bool OpenFile(PBufferedFile F,const char *Spec,int16 Mode)
{
  bool Success;

  /*
    Check if the mode is valid.
  */
  if (Mode & ~(F_READ+F_WRITE+F_CLEAR+F_EXCL)) {
    F->IOError=lheINVAL;
    return(false);
  }
  /*
    Initialize the file status structure.
  */
  F->IOError=0;
  F->Position=0;
  F->StoredPosition=0;
  F->Flags=Mode & (F_READ+F_WRITE);
  /*
    Add actions for exstence and nonexistence to the file.
  */
  if (F->Flags==0) {
    F->IOError=lheINVAL;
    return(false);
  } else {
    Mode+=(Mode&F_CLEAR)?((Mode&F_EXCL)?0:oeTRUNC)+onCREAT:oeOPEN;
  }
  /*
    Allocate a buffer for the file.
  */
  if (!AllocBufferForFile(F,F_DEFAULT_BUFFER_SIZE)) return(false);
  /*
    Open the file. Remember to get rid of the buffer if the file open fails.
  */
  F->F=Open(Spec,Mode);
  SetErrorFlag(F);
  Success=(F->F!=-1);
  if (!Success) FreeMem(F->Buffer);
  return(Success);
}

/*
  Flush file buffers and send their content to the file. It is used to send
  any pending data in the output buffer to the file. This will flush only
  the internal file buffers maintained in the file status structure and not
  any disk caches.
*/
bool FlushFile(PBufferedFile F)
{
  if (F->Flags & F_WRITING) {
    if (!WriteAll(F->F,F->Buffer,F->DataSize)) return(false);
  }
  F->DataSize=0;
  F->DataPos=0;
  return(true);
}

/*
  Close the file associated with the file status buffer. After this operation
  is done, the file status structure is invalid and should not be used for
  any file I/O operation until reopened with another OpenFile() call. However
  the status informations (such as flags) is preserved and can be used to
  check the file status immediately before close.
*/
bool CloseFile(PBufferedFile F)
{
  bool Result;

  FlushFile(F);
  FreeMem(F->Buffer);
  Result=Close(F->F);
  SetErrorFlag(F);
  return(Result);
}

/*
  Obtain the file I/O error flag and then reset the flag back to zero. If the
  file I/O error flag is nonzero, any file operation will be ignored until
  it is cleared.
*/
int16 FileIOResult(PBufferedFile F)
{
  int16 Temp;

  Temp=F->IOError;
  F->IOError=0;
  return(Temp);
}

static bool MoveBufferPtr(PBufferedFile F,uint32 NewPos)
{
  uint32 Pos;
  uint32 Dist;
  bool Result;

  Result=false;
  Pos=F->Position;
  if (NewPos>Pos) {
    Dist=NewPos-Pos;
    if (Dist>F->DataPos) {
      FlushFile(F);
    } else {
      F->DataPos-=Dist;
      Result=true;
    }
  } else {
    Dist=Pos-NewPos;
    if (Dist>F->DataSize-F->DataPos) {
      FlushFile(F);
    } else {
      F->DataPos+=Dist;
      Result=true;
    }
  }
  return(Result);
}

/*
  Move the reading position to the specified location. Due to the performance
  reasons the file is not switched into reading mode if it is in writing mode
  when this call is performed on it. However in each case the call will move
  the correct position in the file.
*/
bool SeekRead(PBufferedFile F,uint32 NewPos)
{
  bool Result;

  if (F->IOError) return(false);
  if (!(F->Flags & F_WRITING)) {
    Result=MoveBufferPtr(F,NewPos);
    F->Position=NewPos;
    if (!Result) Result=RawSeek(F,NewPos);
  } else {
    F->StoredPosition=NewPos;
    Result=true;
  }
  return(Result);
}

/*
  Obtain the current reading position in the file. Like the previous function
  the file is not switched into reading mode by this call but everytime the
  correct position is returned.
*/
uint32 TellRead(PBufferedFile F)
{
  if (F->IOError) return(0xFFFFFFFF);
  if (!(F->Flags & F_WRITING)) {
    return(F->Position);
  } else {
    return(F->StoredPosition);
  }
}

/*
  Move the writing position in the file. Like SeekRead() this call will not
  switch the file into write mode, but it will always move the correct
  position in it's status structure. This call is ignored if the file is
  open in APPEND mode.
*/
bool SeekWrite(PBufferedFile F,uint32 NewPos)
{
  bool Result;

  if (F->IOError) return(false);
  if (F->Flags & F_WRITING) {
    Result=MoveBufferPtr(F,NewPos);
    F->Position=NewPos;
    if (!Result) Result=RawSeek(F,NewPos);
  } else {
    F->StoredPosition=NewPos;
    Result=true;
  }
  return(Result);
}

/*
  Obtain the current writing position. Like TellRead() this will not switch
  the file into writing mode but always will return the correct file
  position. If the file is in APPEND mode, the return value is the current
  file size.
*/
uint32 TellWrite(PBufferedFile F)
{
  if (F->IOError) return(0xFFFFFFFF);
  if (F->Flags & F_WRITING) {
    return(F->Position);
  } else {
    return(F->StoredPosition);
  }
}

/*
  Obtain the current size of the file. This call will not disturb the file's
  mode nor the internal file's positions. If this operation fails, the file
  status is undefined.
*/
uint32 TellFileSize(PBufferedFile F)
{
  int16 Failure;
  TFileInfo I;
  uint32 Size;

  if (F->Flags & F_WRITING) FlushFile(F);
  Size=0xFFFFFFFF;
  if (!F->IOError) {
    Failure=!FStat(F->F,&I);
    SetErrorFlag(F);
    if (!Failure) Size=I.Size;
  }
  return(Size);
}

/*
  Fetch new data into the buffer. This call assumes that the file is in read
  mode (it does not switch it if it is not).
*/
static bool FetchIntoBuffer(PBufferedFile F)
{
  F->DataSize=Read(F->F,F->Buffer,F->BufferSize);
  F->DataPos=0;
  SetErrorFlag(F);
  return(F->IOError==0);
}

/*
  Read block of data from the file to user's memory. The file is switched
  into reading mode before the read is performed. The position of the data to
  be read is determined by the reading position of the file. Returns true if
  successfull and false if error occured. The attempt to read past end of
  file is treated as error.
*/
bool FileReadBlock(PBufferedFile F,void *Buf,uint32 Count)
{
  uint32 CountToRead;
  uint32 TotalRead;
  uint32 Rem;
  bool Result;

  /*
    Check if no pending errors are on the file.
  */
  if (F->IOError) return(false);
  /*
    Switch to the reading mode. If the switch fails, return failure return
    status.
  */
  if (!SwitchRead(F)) return(false);
  /*
    Initialize read counters. The size of the block must be copied to be able
    to detect an error at the end.
  */
  TotalRead=0;
  Rem=Count;
  /*
    There can be prefetched and unread data in the buffer. If so, move them
    to the output buffer and skip them. Don't try to reread them as the
    actual file position is located after these prefetched bytes.
  */
  CountToRead=(Rem>F->DataSize-F->DataPos)?F->DataSize-F->DataPos:Rem;
  if (CountToRead>0) {
    TotalRead=CountToRead;
    Buf=MemCpy(Buf,F->Buffer+F->DataPos,CountToRead);
    F->DataPos+=CountToRead;
    Rem-=CountToRead;
  }
  /*
    If there are no more bytes to be read now, we have nothing to do and this
    portion of code must be skipped. There may still be prefetched but unread
    data in the file's buffer and attempt to execute this code when there are
    no bytes to be read would cause them to be destroyed by FetchIntoBuffer()
    call.

    If there are still some bytes to be read, we have a choice. If there are
    too many of them (more than the size of the buffer), we will read these
    data directly from the file and don't waste time with buffer processing.
    Otherwise we will fetch the data into the buffer and then take out the
    required bytes.
  */
  if (Rem>0) {
    if (Rem>F->BufferSize) {
      Result=ReadAll(F->F,Buf,Rem);
      SetErrorFlag(F);
      if (Result) TotalRead+=Rem;
    } else {
      if (!FetchIntoBuffer(F)) return(false);
      CountToRead=(Rem>F->DataSize)?F->DataSize:Rem;
      if (CountToRead>0) {
        TotalRead+=CountToRead;
        MemCpy(Buf,F->Buffer,CountToRead);
        F->DataPos=CountToRead;
      }
    }
  }
  /*
    Advance the file pointer indicator and the counter of transferred bytes.
  */
  F->Position+=TotalRead;
  F->IOBytes=TotalRead;
  /*
    Return the success flag.
  */
  return(TotalRead==Count);
}

/*
  Write block of data from user's memory to the file. The file is switched
  into write mode before the write is performed. The position in the file to
  which the data are being written is determined by the file's writing
  position. Returns true if successfull and false if error occured. The disk
  full condition is treated as error.
*/
bool FileWriteBlock(PBufferedFile F,const void *Buffer,uint32 Count)
{
  uint32 CountToWrite;
  uint32 TotalWrite;
  uint32 Rem;
  bool Result;
  const char *Buf;

  /*
    Make it possible to advance the pointer to the buffer by using +=
    operator. This can be achieved by assigning it to a "const char *"
    pointer variable.
  */
  Buf=Buffer;
  /*
    Check if no pending errors are on the file.
  */
  if (F->IOError) return(false);
  /*
    Switch to the writing mode. If the switch fails, return failure return
    status.
  */
  if (!SwitchWrite(F)) return(false);
  /*
    Initialize read counters. The size of the block must be copied to be able
    to detect an error at the end.
  */
  TotalWrite=0;
  Rem=Count;
  /*
    Move as many bytes as possible into the buffer. Be aware that the
    F->DataPos here is significant and may be not equal to F->DataSize.
  */
  CountToWrite=(Rem>F->BufferSize-F->DataPos)?F->BufferSize-F->DataPos:Rem;
  if (CountToWrite>0) {
    TotalWrite=CountToWrite;
    MemCpy(F->Buffer+F->DataPos,Buf,CountToWrite);
    Buf+=CountToWrite;
    F->DataPos+=CountToWrite;
    if (F->DataSize<F->DataPos) F->DataSize=F->DataPos;
    Rem-=CountToWrite;
  }
  /*
    If there are still bytes to be transferred, then the buffer is full and
    needs flush. After the buffer is flushed we need to choose if the rest of
    the data should be written into the buffer or directly into the file. If
    there are more bytes than the buffer size, the data should be written
    directly into the file, because that is faster. Otherwise put the data
    into the buffer.
  */
  if (Rem>0) {
    FlushFile(F);
    if (Rem>F->BufferSize) {
      Result=WriteAll(F->F,Buf,Rem);
      SetErrorFlag(F);
      if (Result) TotalWrite+=Rem;
    } else {
      MemCpy(F->Buffer,Buf,Rem);
      TotalWrite+=Rem;
      F->DataPos=Rem;
    }
  }
  /*
    Ensure F->DataSize no less than F->DataPos.
  */
  if (F->DataSize<F->DataPos) F->DataSize=F->DataPos;
  /*
    Advance the file pointer indicator and the counter of transferred bytes.
  */
  F->Position+=TotalWrite;
  F->IOBytes+=TotalWrite;
  /*
    Return the success flag.
  */
  return(TotalWrite==Count);
}

/******************************************/
/*** Static variables for default files ***/
/******************************************/
/*
  These variables are where the informations about "current input", "current
  output" and "current input/output" files are stored and maintained. These
  "current" files are used for conevient access to files because there is no
  need for the caller to specify the file status structure each time it is
  about to do a file I/O operation.
*/

/*
  Variable for the input/output error occured during any file I/O operation.
  If this is nonzero, all file I/O operations are blocked. To pick the error
  code up for processing and clear this variable use the IOResult() call.
*/
static int16 IOError;

/*
  Status variables for the input, output and input/output files maintained
  by the library. The file status structure pointers are stored here too, so
  program is not required to supply them for each file I/O operation to be
  performed on these files.
*/
static PBufferedFile InOutF;
static PBufferedFile InF;
static PBufferedFile OutF;

/*
  The library will put the error number when a file operation fails into an
  internal variable. No file operations will then be then performed until the
  error is read and cleared by this function.
*/
int16 IOResult(void)
{
  int16 Temp;

  Temp=IOError;
  IOError=0;
  return(Temp);
}

/*
  Complain the last I/O error
*/
const char *ComplainLastIOError(void)
{
  return(ComplainError(IOResult()));
}

/***************************/
/*** Input file handling ***/
/***************************/
/*
  These calls are used to setup a "current input file" and to obtain the data
  from it. They are useful for programs which operate on one input file at a
  time, because the program don't need to specify the file status structure
  each time it needs to read that file.
*/

/*
  Setup specified file information structure as input file. The file can be
  either closed and released using CloseAndReleaseInputFile() or simply
  released using ReleaseInputFile().
*/
bool SetupInputFile(PBufferedFile F)
{
  if (IOError) return(false);
  if (InF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  InF=F;
  return(true);
}

/*
  Allocate the memoty for the file status structure and open a file in it.
  This call is used for AllocateAndXXXFile() file opening functions for
  conevient obtaining the file status structure. Returns pointer to the open
  file status structure if successfull or NULL if failed. If failure occurs,
  the error code is provided and file I/O operations are blocked. If this
  call is made when the file I/O operations are blocked, results are
  undefined.
*/
static PBufferedFile AllocAndOpen(const char *FileSpec,int16 Mode)
{
  PBufferedFile F;
  bool Success;

  F=GetMem(sizeof(*F));
  if (F==NULL) {
    IOError=lheNOMEM;
    return(NULL);
  }
  Success=OpenFile(F,FileSpec,Mode);
  if (Success) {
    F->Flags |= F_OWNED;
  } else {
    IOError=FileIOResult(F);
    FreeMem(F);
    F=NULL;
  }
  return(F);
}

/*
  Opens a file for reading with specification passed in "FileSpec". Returns
  nonzero if successfull, zero if failed. After the failure the IOResult()
  call must be used to obtain error code, because the file I/O operations
  are blocked by it.
*/
bool AllocateAndOpenInputFile(const char *FileSpec)
{
  if (IOError) return(false);
  if (InF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  InF=AllocAndOpen(FileSpec,F_READ);
  return(InF!=NULL);
}

/*
  Open the specified file in specified file status structure and setup the
  global error code on any errors. Calling this procedure when file I/O
  operations are blocked will produce undefined results.
*/
static bool OpenTheFile(PBufferedFile F,const char *FileSpec,int16 Mode)
{
  bool Success;

  Success=OpenFile(F,FileSpec,Mode);
  if (Success) {
    F->Flags&=~F_OWNED;
  } else {
    IOError=FileIOResult(F);
  }
  return(Success);
}

/*
  Opens a file for reading with specification passed in "FileSpec". Returns
  nonzero if successfull, zero if failed.
*/
bool SetupAndOpenInputFile(PBufferedFile F,const char *FileSpec)
{
  bool Success;

  if (IOError) return(false);
  Success=OpenTheFile(F,FileSpec,F_READ);
  if (Success) InF=F;
  return(Success);
}

/*
  Sets the new input file position. Next reads will occur at the position
  specified by this call. Returns nonzero on success, zero on failure.
  If the call fails, IOResult() must be called to obtain the error code,
  because all file I/O is blocked with it.
*/
bool SeekInputFile(uint32 NewPosition)
{
  bool Success;

  if (IOError) return(false);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (!(InF->Flags & F_READ)) {
    IOError=lheACCESS;
    return(false);
  }
  Success=SeekRead(InF,NewPosition);
  if (!Success) IOError=FileIOResult(InF);
  return(Success);
}

/*
  Determine the size of the input file. Returns the size of the file on
  success or 0xFFFFFFFF on failure. If the call fails, IOResult() must be
  called to obtain the error code and unblock the file I/O operations.
*/
uint32 GetInputFileSize(void)
{
  uint32 Size;

  if (IOError) return(0xFFFFFFFF);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  Size=TellFileSize(InF);
  if (Size==0xFFFFFFFF) IOError=FileIOResult(InF);
  return(Size);
}

/*
  Derermine the current reading position of the file. Returns the current
  reading position on success or 0xFFFFFFFF on failure. In the case of
  failure the IOResult() call must be made to obtain the error code and
  unblock the file I/O operations.
*/
uint32 GetInputFilePos(void)
{
  uint32 Result;

  if (IOError) return(0xFFFFFFFF);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  if (!(InF->Flags & F_READ)) {
    IOError=lheACCESS;
    return(0xFFFFFFFF);
  }
  Result=TellRead(InF);
  if (Result==0xFFFFFFFF) IOError=FileIOResult(InF);
  return(Result);
}

/*
  Read one byte (8 bit entity) from the input file. The byte is read from the
  current reading position and the reading position is moved to the next
  byte. If there is no byte to read or file I/O operations are blocked with
  a pending error code, this call returns zero. If a read failure occurs,
  an error code is set and the read operations are blocked until the error
  code is picked out by IOResult() call.
*/
uint16 ReadByte(void)
{
  int16 Success;
  upint8 Ch;

  if (IOError) return(0);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0);
  }
  if (!(InF->Flags & F_READ)) {
    IOError=lheACCESS;
    return(0);
  }
  if (!SwitchRead(InF)) return(0);
  if (InF->DataPos<InF->DataSize) {
    InF->Position++;
    return(InF->Buffer[InF->DataPos++]);
  }
  Success=FileReadBlock(InF,&Ch,1);
  if (!Success) {
    IOError=FileIOResult(InF);
    Ch=0;
  }
  return(Ch);
}

/*
  Reads an word (16 bit entity) from the input file. This call will read two
  adjacent bytes from the file, move the reading position after them and
  return them as unsigned short integer. The interpretation of the two bytes
  read depends on the current format setting. Default interpretation is
  "big-endian". If read failure occurs or the file I/O is blocked by an
  unpicked error code, the call returns 0. In the case of the failure an
  error code is provided and the read operations are blocked until the error
  code is picked out using IOResult().
*/
static uint16 ReadWord(bool BigEndian)
{
  uint16 Low;
  uint16 Hi;

  if (IOError) return(0);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0);
  }
  if (BigEndian) {
    Hi=ReadByte();
    Low=ReadByte();
  } else {
    Low=ReadByte();
    Hi=ReadByte();
  }
  return(Low+256*Hi);
}

uint16 Read16LE(void)
{
  return(ReadWord(false));
}

uint16 Read16BE(void)
{
  return(ReadWord(true));
}

/*
  Read a dword (32 bit entity) from the input file. This call will read four
  bytes, move the reading position after them and return them as unsigned
  long integer. The interpretation of the bytes depends on the file format
  setting. Default interpretation is "big-endian". If read failure occurs or
  the file I/O is blocked by an unpicked error code, the call returns 0. In
  the case of the failure an error code is provided and the read operations
  are blocked until the error code is picked out using IOResult().
*/
static uint32 ReadDWord(bool BigEndian)
{
  uint16 Low;
  uint16 Hi;

  if (IOError) return(0);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0);
  }
  if (BigEndian) {
    Hi=Read16BE();
    Low=Read16BE();
  } else {
    Low=Read16LE();
    Hi=Read16LE();
  }
  return(Low+65536UL*Hi);
}

uint32 Read32LE(void)
{
  return(ReadDWord(false));
}

uint32 Read32BE(void)
{
  return(ReadDWord(true));
}

/*
  Read block of bytes into user's memory. The bytes are read from the current
  file's pointer. Returns count of bytes successfully read. Attempt to read
  past end of the file is treated as error. If a read failure occurs, an
  error code is provided and the reading operations are blocked until the
  error code is picked up using IOResult().
*/
bool ReadBlock(void *Buffer,uint32 Size)
{
  bool Success;

  if (IOError) return(false);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (!(InF->Flags & F_READ)) {
    IOError=lheACCESS;
    return(false);
  }
  Success=FileReadBlock(InF,Buffer,Size);
  if (!Success) IOError=FileIOResult(InF);
  return(Success);
}

/*
  Skip the rest of a line and place the reading pointer to the beginning of
  the next line. The lines are recognized by character '\n' on all platforms.

  This call is faster than reading the bytes until '\n' or EOF is
  encountered.
*/
bool SkipRestOfLine(void)
{
  char Ch;

  Ch='\0';
  if (IOError) return(false);
  do {
    while (InF->DataSize<InF->DataPos) {
      Ch=InF->Buffer[InF->DataPos++];
      if (Ch=='\n') break;
    }
    if (Ch!='\n') Ch=ReadByte();
  } while ((Ch!='\n') && (InF->DataSize));
  return(!IOError);
}

/*
  Get count of transferred bytes after last ReadBlock() call. If this call is
  not made after ReadBlock() call, it's result is undefined.
*/
uint32 GetReadByteCnt(void)
{
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  return(InF->IOBytes);
}

/*
  Release the input file. The input file structure is released but not closed.
  This call can't be made on input files opened with AllocAndOpenInputFile(),
  because these file status structures are internally created and must be
  deleted using CloseAndFreeInputFile().
*/
PBufferedFile ReleaseInputFile(void)
{
  PBufferedFile F;

  if (IOError) return(NULL);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(NULL);
  }
  if (InF==InOutF) {
    IOError=lheACCESS;
    return(NULL);
  }
  F=InF;
  InF=NULL;
  return(F);
}

/*
  Close the file status structure and free it. This call is used by the
  CloseAndFreeXXXFile() functions to coneviently get rid of the file status
  structure and report any errors. Returns nonzero on success and zero on
  failure. If this call fails, error code is provided and file I/O operations
  are blocked. If this call is made when the file I/O operations are blocked,
  the result is undefined.
*/
static bool CloseAndFreeTheFile(PBufferedFile F)
{
  bool Success;

  if (!(F->Flags & F_OWNED)) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseFile(F);
  if (Success) {
    FreeMem(F);
  } else {
    IOError=FileIOResult(F);
  }
  return(Success);
}

/*
  Closes a file previously opened with AllocateAndOpenInputFile().
  Returns nonzero if successfull, zero if failed. This function will fail if
  no input nor inout file is mapped. If the inout file is mapped to the
  output too, it will be unmapped during the close operation.
*/
bool CloseAndFreeInputFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (InF==InOutF) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseAndFreeTheFile(InF);
  if (Success) InF=NULL;
  return(Success);
}

/*
  Close the file status structure but do not free it. This call is used by the
  CloseAndReleaseXXXFile() functions to coneviently close the file status
  structure and report any errors. Returns nonzero on success and zero on
  failure. If this call fails, error code is provided and file I/O operations
  are blocked. If this call is made when the file I/O operations are blocked,
  the result is undefined.
*/
static bool CloseTheFile(PBufferedFile F)
{
  bool Success;

  if (F->Flags & F_OWNED) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseFile(F);
  if (!Success) IOError=FileIOResult(F);
  return(Success);
}

/*
  Closes a file previously opened with SetupAndOpenInputFile().
  Returns nonzero if successfull, zero if failed. This function will fail if
  no input nor inout file is mapped. If the inout file is mapped to the
  output too, it will be unmapped during the close operation.
*/
bool CloseAndReleaseInputFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (InF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (InF==InOutF) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseTheFile(InF);
  if (Success) InF=NULL;
  return(Success);
}

/****************************/
/*** Output file handling ***/
/****************************/
/*
  These calls are used to setup a file as current writing file and then to
  send data to it. They are useful for programs which produce one file at a
  time, because the program don't need to specify the file status structure
  each time it needs to write into the file.
*/

uint32 OutputPos;
uint32 WrittenBytes;

/*
  Setup specified file information structure as output file. The format
  specifier used for this file will not be modified. The file can be either
  closed and released using CloseAndReleaseOutputFile() or simply released
  using ReleaseOutputFile().
*/
bool SetupOutputFile(PBufferedFile F)
{
  if (IOError) return(false);
  if (OutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  OutF=F;
  return(true);
}

/*
  Creates a file for writing with specification passed in "FileSpec". If the
  file existed before this procedure was called, the previous file's content
  will be discarded. Returns nonzero if successfull, zero if failed. If the
  call fails, all file I/O is blocked until the error code provided by it is
  picked up and cleared by an IOResult() call.
*/
bool AllocateAndCreateOutputFile(const char *FileSpec)
{
  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  OutF=AllocAndOpen(FileSpec,F_WRITE|F_CLEAR);
  return(OutF!=NULL);
}

/*
  This is the same as AllocateAndCreateFile(), except the file is opened into
  an user-supplied file status structure instead of automatic allocation of
  one.
*/
bool SetupAndCreateOutputFile(PBufferedFile F,const char *FileSpec)
{
  bool Success;

  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  Success=OpenTheFile(F,FileSpec,F_WRITE|F_CLEAR);
  if (Success) OutF=F;
  return(Success);
}

/*
  This is the same as AllocateAndCreateOutputFile() but it fails when the
  file already exists.
*/
bool AllocateAndCreateOutputFileIfNonex(const char *FileSpec)
{
  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  OutF=AllocAndOpen(FileSpec,F_WRITE|F_CLEAR|F_EXCL);
  return(OutF!=NULL);
}

/*
  This is the same as SetupAndCreateOutputFile() but it fails when the file
  already exists.
*/
bool SetupAndCreateOutputFileIfNonex(PBufferedFile F,const char *FileSpec)
{
  bool Success;

  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  Success=OpenTheFile(F,FileSpec,F_WRITE|F_CLEAR|F_EXCL);
  if (Success) OutF=F;
  return(Success);
}

/*
  Opens a file for writing with specification passed in "FileSpec". The
  previous file's content will be preserved, if the file existed before this
  call was made. Returns nonzero if successfull, zero if failed. In the
  case of failure the file I/O operations are disabled until the error code
  is picked up and cleared using IOResult() call.
*/
bool AllocateAndOpenOutputFile(const char *FileSpec)
{
  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  OutF=AllocAndOpen(FileSpec,F_READ|F_WRITE);
  return(OutF!=NULL);
}

/*
  This is the same as AllocateAndOpenOutputFile() but the file is opened into
  an user-supplied file status structure instead of allocating one.
*/
bool SetupAndOpenOutputFile(PBufferedFile F,const char *FileSpec)
{
  bool Success;

  if (OutF!=NULL) IOError=lheBUSY;
  if (IOError) return(false);
  OutputPos=0;
  WrittenBytes=0;
  Success=OpenTheFile(F,FileSpec,F_READ|F_WRITE);
  if (Success) OutF=F;
  return(Success);
}

/*
  Sets the new output file position. All subsequent file operations will
  occur at the new file position. If the file position is moved past end of
  file, the file will be expanded to the position at next write. Returns
  nonzero on success or zero on failure. In the case of failure the file I/O
  operations are blocked until the error code produced by the failure is
  picked up and cleared by an IOResult() call.
*/
bool SeekOutputFile(uint32 NewPosition)
{
  bool Success;

  OutputPos=NewPosition;
  WrittenBytes=0;
  if (IOError) return(false);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (!(OutF->Flags & F_WRITE)) {
    IOError=lheACCESS;
    return(false);
  }
  Success=SeekWrite(OutF,NewPosition);
  if (!Success) IOError=FileIOResult(OutF);
  return(Success);
}

/*
  Determine the output file size. Returns the file size on success or
  0xFFFFFFFF on failure or if the file I/O operations are blocked. The file
  size will take into account any file write operations that extended the
  file. If the operation fails, it will produce an error code and the file
  I/O operations will be blcked until the error code is picked up and cleared
  by an IOResult() call.
*/
uint32 GetOutputFileSize(void)
{
  uint32 Size;

  if (IOError) return(0xFFFFFFFF);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  Size=TellFileSize(OutF);
  if (Size==0xFFFFFFFF) IOError=FileIOResult(OutF);
  return(Size);
}

/*
  Get current writing position of the output file. Returns the current write
  position on success or 0xFFFFFFFF on failure or if the file I/O operations
  are blocked. If the call fails, the file I/O operations are blocked until
  the error code produced by the failure is picked up and cleared by an
  IOResult() call.
*/
uint32 GetOutputFilePos(void)
{
  uint32 Pos;

  if (IOError) return(-1);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  if (!(OutF->Flags & F_WRITE)) {
    IOError=lheACCESS;
    return(0xFFFFFFFF);
  }
  Pos=TellWrite(OutF);
  if (Pos==0xFFFFFFFF) IOError=FileIOResult(OutF);
  return(Pos);
}

/*
  Write a byte (8 bit entity) into the output file. The write is buffered and
  data are sent to the file if enough data is collected, the write position
  is moved far enough to get out of the buffer or FlushOutputFile() call is
  made. If the write operation fails, the error code is produced and any file
  I/O operations are blocked until the error code is picked up and cleared by
  an IOResult() call. If the file I/O operations are blocked before this
  call, no action is performed.
*/
void WriteByte(uint16 What)
{
  upint8 Ch;
  int16 Success;

  OutputPos++;
  WrittenBytes++;
  if (IOError) return;
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  if (!(OutF->Flags & F_WRITE)) {
    IOError=lheACCESS;
    return;
  }
  Ch=What;
  Success=FileWriteBlock(OutF,&Ch,1);
  if (!Success) IOError=FileIOResult(OutF);
  return;
}

/*
  Write a word (16 bit entity) into the output file. The order of the bytes
  forming the word written into the file is determined by the file format
  setting. The default file format setting for the order is "big-endian".
  The write is buffered and the data are sent to the file if enough data are
  collected, the writing file pointer is moved far enough or FlushOutputFile()
  call is made. If the write operation fails, the error code is produced and
  any file I/O operations are blocked until the error code is picked up and
  cleared by an IOResult() call. If the file I/O operations are blocked before
  this call, no action is performed.
*/
static void WriteWord(uint16 What,bool BigEndian)
{
  uint16 Low;
  uint16 Hi;

  if (IOError) return;
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  Low=What % 256;
  Hi=What / 256;
  if (BigEndian) {
    WriteByte(Hi);
    WriteByte(Low);
  } else {
    WriteByte(Low);
    WriteByte(Hi);
  }
}

void Write16LE(uint16 What)
{
  WriteWord(What,false);
}

void Write16BE(uint16 What)
{
  WriteWord(What,true);
}

/*
  Write a dword (32 bit entity) into the output file. The order of the bytes
  written is determined by the output file format setting. The default format
  setting is "big-endian". The write is buffered and the data are sent to the
  file if enough data are collected, the writing file pointer is moved far
  enough or FlushOutputFile() call is made. If the write operation fails, the
  error code is produced and any file I/O operations are blocked until the
  error code is picked up and cleared by an IOResult() call. If the file I/O
  operations are blocked before this call, no action is performed.
*/
static void WriteDWord(uint32 What,bool BigEndian)
{
  uint16 Low;
  uint16 Hi;

  if (IOError) return;
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  Low=What % 65536UL;
  Hi=What / 65536UL;
  if (BigEndian) {
    Write16BE(Hi);
    Write16BE(Low);
  } else {
    Write16LE(Low);
    Write16LE(Hi);
  }
}

void Write32LE(uint32 What)
{
  WriteDWord(What,false);
}

void Write32BE(uint32 What)
{
  WriteDWord(What,true);
}

/*
  Write block of bytes into the output file. If the file I/O operations are
  blocked, no action is performed. Writes of small data blocks are buffered
  and the data are sent to the file if enough data are collected, the writing
  file pointer is moved far enough or FlushOutputFile() call is made. If the
  data block is large enough (comparable with the buffer size), it's write
  will result in buffer flush and no data will pend in the buffer after the
  operation is completed. Returns count of bytes successfully prepared for
  write. If the write operation fails, the error code is provided and file
  I/O operations are blocked until the error code is picked up and cleared by
  the IOResult() procedure call.
*/
bool WriteBlock(const void *Buffer,uint32 Size)
{
  bool Success;

  OutputPos+=Size;
  WrittenBytes+=Size;
  if (IOError) return(false);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (!(OutF->Flags & F_WRITE)) {
    IOError=lheACCESS;
    return(false);
  }
  Success=FileWriteBlock(OutF,Buffer,Size);
  if (!Success) IOError=FileIOResult(OutF);
  return(Success);
}

static bool WriteStringCore(const char *Str,uint8 Add)
{
  uint16 Size;

  Size=StrLen(Str)+Add;
  return(WriteBlock(Str,Size));
}

/*
  Write chacters into the file
*/
bool WriteText(const char *Str)
{
  return(WriteStringCore(Str,0));
}

/*
  Write a NULL-terminated string into the file
*/
bool WriteString(const char *Str)
{
  return(WriteStringCore(Str,1));
}

/*
  Get count of transferred bytes after last ReadBlock() call. If this call is
  not made after ReadBlock() call, it's result is undefined.
*/
uint32 GetWriteByteCnt(void)
{
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(0xFFFFFFFF);
  }
  return(OutF->IOBytes);
}

/*
  Send any pending write data from the buffer to the file. This will empty
  only the internal write buffer, not any file buffers provided by the
  operating system itself. Returns nonzero on success and zero if failed. If
  file I/O operations are blocked, no action is performed and zero is
  returned. If the flush fails, the error code is provided and file I/O
  operations are blocked until the error code is piced up and cleared by
  the IOResult() call.
*/
void FlushOutputFile(void)
{
  bool Success;

  if (IOError) return;
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  Success=FlushFile(OutF);
  if (!Success) IOError=FileIOResult(OutF);
  return;
}

/*
  Release the input file. The input file structure is released but not closed.
  This call can't be made on input files opened with AllocAndOpenInputFile(),
  because these file status structures are internally created and must be
  deleted using CloseAndFreeInputFile().
*/
PBufferedFile ReleaseOutputFile(void)
{
  PBufferedFile F;

  if (IOError) return(NULL);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(NULL);
  }
  if (OutF==InOutF) {
    IOError=lheACCESS;
    return(NULL);
  }
  F=OutF;
  OutF=NULL;
  return(F);
}

/*
  Close the output file. The output file's buffer is flushed before the file
  is closed. Returns nonzero if successfull or zero if failed. If the file
  I/O operations are blocked, the file is not closed and zero is returned.
  If the call fails, error code is provided and the file I/O operations are
  blocked until the error code is picked up and cleared by the IOResult()
  call. This call should be used for files opened with AllocateAndXXXFile().
*/
bool CloseAndFreeOutputFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (OutF==InOutF) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseAndFreeTheFile(OutF);
  if (Success) OutF=NULL;
  return(Success);
}

/*
  The same as CloseAndFreeOutputFile() but this call should be used for files
  opened with SetupAndXXXFile() calls.
*/
bool CloseAndReleaseOutputFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (OutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  if (OutF==InOutF) {
    IOError=lheACCESS;
    return(false);
  }
  Success=CloseTheFile(OutF);
  if (Success) OutF=NULL;
  return(Success);
}

/***************************/
/*** InOut file handling ***/
/***************************/
/*
  These call maintain a file opened for both input and output. This file
  cannot be accessed directly. Instead it must be mapped for read if read
  operations are required or mapped for write if output operations are
  required. This feature is useful when program uses data from a file and
  modifies them at the same time even when it uses some other files.
*/

/*
  Setup specified file information structure as input/output file. The format
  specifier used for this file will not be modified. The file can be either
  closed and released using CloseAndReleaseInOutFile() or simply released
  using ReleaseInOutFile().
*/
bool SetupInOutFile(PBufferedFile F)
{
  if (IOError) return(false);
  if (InOutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  InOutF=F;
  return(true);
}

/*
  Opens a file for reading and writing with specification passed in
  "FileSpec". Returns nonzero if successfull, zero if failed. Altlough the
  file opened, it must be mapped first to be used. Use MapInOutFileRead() to
  map the file as input file UnmapInOutFileRead() to unmap it. Similarly use
  MapInOutFileWrite() to map the file as output file and
  UnMapInOutFileWrite() to unmap it.
*/
bool AllocateAndOpenInOutFile(char *FileSpec)
{
  if (IOError) return(false);
  if (InOutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  InOutF=AllocAndOpen(FileSpec,F_READ|F_WRITE);
  return(InOutF!=NULL);
}

/*
  The same as AllocateAndOpenInOutFile() but the file is open using an user
  supplied file status structure instead of allocating one.
*/
bool SetupAndOpenInOutFile(PBufferedFile F,char *FileSpec)
{
  bool Success;

  if (IOError) return(false);
  if (InOutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  Success=OpenTheFile(F,FileSpec,F_READ|F_WRITE);
  if (Success) OutF=F;
  return(Success);
}

/*
  Create a file and setup it as input/output file. Returns nonzero if
  successfull or zero if failed. If successfull and the file existed before,
  it's content is discarded. The use of the file opened with this call is
  similar to the one opened with OpenInOutFile() call.
*/
bool AllocateAndCreateInOutFile(char *FileSpec)
{
  if (IOError) return(false);
  if (InOutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  InOutF=AllocAndOpen(FileSpec,F_READ|F_WRITE|F_CLEAR);
  return(InOutF!=NULL);
}

/*
  The same as AllocateAndCreateInOutFile() but the file is opened in an user
  supplied file status structure instead of allocaring one in memory.
*/
bool SetupAndCreateInOutFile(PBufferedFile F,char *FileSpec)
{
  bool Success;

  if (IOError) return(false);
  if (InOutF!=NULL) {
    IOError=lheBUSY;
    return(false);
  }
  Success=OpenTheFile(F,FileSpec,F_READ|F_WRITE|F_CLEAR);
  if (Success) OutF=F;
  return(Success);
}

/*
  Map inout file for reading. This call is required before any read
  operations on the file can be performed. There must be no input file open
  before this call is made. After the input/output file is successfully
  mapped, it can be used as any other input file.
*/
void MapInOutFileRead()
{
  if (IOError) return;
  if (InF!=NULL) {
    IOError=lheBUSY;
    return;
  }
  if (InOutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  InF=InOutF;
}

/*
  Unmap the input/output file mapped for reading. Use this call when no more
  file read operations are requested or when another input file is to be
  opened. When the file I/O operations are blocked, no action is performed
  and the file is not mapped.
*/
void UnmapInOutFileRead()
{
  if (IOError) return;
  if (InF!=InOutF) {
    IOError=lheACCESS;
    return;
  }
  InF=NULL;
}

/*
  Map input/output file for writing. This call is similar to the
  MapInOutFileRead() except that the file is prepared for write operations
  instead of read operations. The call will fail if there is another output
  file opened. After the file is successfully mapped, it can be used as any
  other output file. When the file I/O operations are blocked, no action is
  performed and the file is not mapped.
*/
void MapInOutFileWrite()
{
  if (IOError) return;
  if (OutF!=NULL) {
    IOError=lheBUSY;
    return;
  }
  if (InOutF==NULL) {
    IOError=lheNOFILE;
    return;
  }
  OutF=InOutF;
}

/*
  Unmap the input/output file mapped for writing. This call is used when there
  are no more write operations for the input/output file or another output
  file is to be opened. When the file I/O operations are blocked, no action is
  performed and the file is still mapped after the call.
*/
void UnmapInOutFileWrite()
{
  if (IOError) return;
  if (OutF!=InOutF) {
    IOError=lheACCESS;
    return;
  }
  OutF=NULL;
}

/*
  Unmap the input/output file. Returns nonzero on success or zero on failure.
*/
static void DestroyInOutFilePointer(void)
{
  if (InF==InOutF) UnmapInOutFileRead();
  if (OutF==InOutF) UnmapInOutFileWrite();
  InOutF=NULL;
}

/*
  Release the input file. The input file structure is released but not closed.
  This call can't be made on input files opened with AllocAndOpenInputFile(),
  because these file status structures are internally created and must be
  deleted using CloseAndFreeInputFile().
*/
PBufferedFile ReleaseInOutFile(void)
{
  PBufferedFile F;

  if (IOError) return(NULL);
  if (InOutF==NULL) {
    IOError=lheNOFILE;
    return(NULL);
  }
  F=InOutF;
  DestroyInOutFilePointer();
  return(F);
}

/*
  Close the input/output file and free it's status structure. This call will
  unmap the file if it was mapped for reading or for writing before this call
  so there is no need to unmap the file manually. When I/O operations are
  blocked, no action is performed and the file is still opened.
*/
bool CloseAndFreeInOutFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (InOutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  Success=CloseAndFreeTheFile(InOutF);
  if (Success) DestroyInOutFilePointer();
  return(Success);
}

/*
  Close the input output file. This call will unmap the file if it was mapped
  for reading or for writing before this call so there is no need to unmap the
  file manually. When I/O operations are blocked, no action is performed and
  the file is still opened.
*/
bool CloseAndReleaseInOutFile(void)
{
  bool Success;

  if (IOError) return(false);
  if (InOutF==NULL) {
    IOError=lheNOFILE;
    return(false);
  }
  Success=CloseTheFile(InOutF);
  if (Success) DestroyInOutFilePointer();
  return(Success);
}
