#include <syslib/types.h>
#define __INTERFACE__
#include "gfxcon.h"

#include <syslib/traps.h>

#include "thread.h"
#include "errclass.h"
#include "misc.h"
#include "diskio.h"
#include "servers.h"
#include "output.h"
#include "errno.h"
#include "process.h"

static bool KeyboardOpen;
static int16 KeyboardHandle;
static int16 ScrLock;
static int16 TransferLock;
static uint32 RemainingPixels;
static bool ScreenModeSet;

static void Fatal(const char *Msg)
{
  Print(Msg);
  NewLine();
  Print("Driver said: ");
  Print(ComplainLastError());
  NewLine();
  Halt(1);
}

static uint16 GetNumber(uint8 Number)
{
  switch (Number) {
    case num100: return(100);
    case num160: return(160);
    case num200: return(200);
    case num320: return(320);
    case num350: return(350);
    case num400: return(400);
    case num480: return(480);
    case num600: return(600);
    case num640: return(640);
    case num768: return(768);
    case num800: return(800);
    case num960: return(960);
    case num1024: return(1024);
    case num1280: return(1280);
    default: TrapUndefined("Invalid mode width/height specifier");
  }
  return(0);
}

void UnpackModeNumber(uint16 Mode,TSetModeCommand *C)
{
  C->Width=GetNumber((Mode >> lWidth) & 63);
  C->Height=GetNumber((Mode >> lHeight) & 63);
  C->PixelSize=((Mode >> lPixelSize)&3);
  C->Flags=(Mode&flgAllFlags) >> 8;
}

static bool AllocateScreenLock(void)
{
  RemainingPixels=0;
  if (ScreenModeSet) return(true);
  ScrLock=CreateLock();
  TransferLock=CreateLock();
  if (TransferLock==-1) {
    if (ScrLock!=-1) DestroyLock(ScrLock);
  }
  return(ScrLock!=-1 && TransferLock!=-1);
}

int16 SetScreenMode(uint16 Mode)
{
  TSetModeCommand C;
  TServerInfo I;

  /*
    Attempt to allocate the lock for the screen
  */
  if (!AllocateScreenLock()) return(-1);
  /*
    Decode the mode and fill the command packet
  */
  C.Command=scrcmdSetMode;
  UnpackModeNumber(Mode,&C);
  /*
    Send the packet into the screen device
  */
  I.Server=hndCONSOLE;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=NULL;
  I.RepSize=0;
  if (!SendCmd(&I)) return(-1);
  /*
    Return the handle from the reply
  */
  CurHeight=C.Height;
  CurWidth=C.Width;
  CurBitsPerPixel=(C.PixelSize+1)*8;
  ScreenModeSet=true;
  return(hndCONSOLE);
}

int16 SetScreenModeByRes(uint16 Width,uint16 Height,uint8 BPP,uint8 Flags)
{
  TSetModeCommand C;
  TServerInfo I;

  if (!AllocateScreenLock()) return(-1);
  C.Command=scrcmdSetMode;
  C.Width=Width;
  C.Height=Height;
  C.PixelSize=BPP;
  C.Flags=Flags;
  I.Server=hndCONSOLE;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=NULL;
  I.RepSize=0;
  if (!SendCmd(&I)) return(-1);
  CurHeight=C.Height;
  CurWidth=C.Width;
  CurBitsPerPixel=(C.PixelSize+1)*8;
  return(hndCONSOLE);
}

void OpenScreenRegion(uint16 X,uint16 Y,uint16 Width,uint16 Height)
{
  TOpenScreenRegionCommand C;
  TServerInfo I;
  bool Result;

  Lock(TransferLock);
  Lock(ScrLock);
  assert(RemainingPixels==0);
  C.Command=scrcmdOpenScreenRegion;
  C.X=X;
  C.Y=Y;
  C.Width=Width;
  C.Height=Height;
  I.Server=hndCONSOLE;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=NULL;
  I.RepSize=0;
  Result=SendCmd(&I);
  RemainingPixels=Width*Height;
  if (!Result) Fatal("Unable to open screen region");
  Unlock(ScrLock);
}

void ScreenWrite(void *Buffer,uint32 Size)
{
  bool IOSize;

  if (Size==0) return;
  Lock(ScrLock);
  assert(RemainingPixels>0);
  IOSize=Write(hndCONSOLE,Buffer,Size);
  assert(IOSize==Size);
  RemainingPixels-=IOSize;
  Unlock(ScrLock);
  if (RemainingPixels==0) Unlock(TransferLock);
}

void UpdateScreen(void)
{
  TNoParamCmd C;
  TServerInfo I;

  Lock(ScrLock);
  C.Command=scrcmdUpdateScreen;
  I.Server=hndCONSOLE;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=NULL;
  I.RepSize=0;
  SendCmd(&I);
  Unlock(ScrLock);
}

bool EnsureKeyboardIsOpen(void)
{
  TOpenKeyboardCommand C;
  TOpenKeyboardReply R;
  TServerInfo I;

  if (!KeyboardOpen) {
    C.Command=scrcmdOpenKeyboard;
    I.Server=hndCONSOLE;
    I.Cmd=&C;
    I.CmdSize=sizeof(C);
    I.Reply=&R;
    I.RepSize=sizeof(R);
    if (!SendCmd(&I)) return(false);
    KeyboardHandle=R.Handle;
    KeyboardOpen=true;
  }
  return(true);
}

uint16 RawReadKey(void)
{
  upint16 Key;

  if (!EnsureKeyboardIsOpen()) return(65535);
  if (Read(KeyboardHandle,&Key,2)!=2) return(65535);
  return(Key);
}

void CloseKeyboard(void)
{
  Close(KeyboardHandle);
}
