#include <syslib/types.h>
#define __INTERFACE__
#include "memory.h"

#include <syslib/memory.h>
#include <syslib/kernel.h>

#include "process.h"
#include "output.h"

#define MAX_MEMORY_TRANSACTION_SIZE 128

typedef struct tagTMemoryBlocks {
  void *Blocks[MAX_MEMORY_TRANSACTION_SIZE];
  uint16 Cnt;
} TMemoryBlocks;

static TMemoryBlocks FreeOnCommit;
static TMemoryBlocks FreeOnRollback;
static bool TransactionInProgress;

static void RegisterBlock(TMemoryBlocks *B,void *Block)
{
  assert(B->Cnt<MAX_MEMORY_TRANSACTION_SIZE);
  if (TransactionInProgress) B->Blocks[B->Cnt++]=Block;
}

static void AbortMem(void)
{
  PrintErr("Aborting due to unexpected memory (re)allocation failure\n");
  Halt(255);
}

void *GetMem(uint16 MemSize)
{
  void *Ptr;

  Ptr=KMEM__malloc(MemSize);
  if (Ptr!=NULL) RegisterBlock(&FreeOnRollback,Ptr);
  return(Ptr);
}

void *GetHugeMem(uint32 MemSize)
{
  void *Ptr;

  Ptr=KMEM__hugemalloc(MemSize);
  if (Ptr!=NULL) RegisterBlock(&FreeOnRollback,Ptr);
  return(Ptr);
}

void *GetClearMem(uint16 MemSize)
{
  void *Ptr;

  Ptr=KMEM__calloc(MemSize);
  if (Ptr!=NULL) RegisterBlock(&FreeOnRollback,Ptr);
  return(Ptr);
}

void *GetHugeClearMem(uint32 MemSize)
{
  void *Ptr;

  Ptr=KMEM__hugecalloc(MemSize);
  if (Ptr!=NULL) RegisterBlock(&FreeOnRollback,Ptr);
  return(Ptr);
}

void *XGetMem(uint16 MemSize)
{
  void *p;

  p=GetMem(MemSize);
  if (!p) AbortMem();
  return(p);
}

void *XGetHugeMem(uint32 MemSize)
{
  void *p;

  p=GetHugeMem(MemSize);
  if (!p) AbortMem();
  return(p);
}

void *XGetClearMem(uint16 MemSize)
{
  void *p;
  
  p=GetClearMem(MemSize);
  if (!p) AbortMem();
  return(p);
}

void *XGetHugeClearMem(uint32 MemSize)
{
  void *p;
  
  p=GetHugeClearMem(MemSize);
  if (!p) AbortMem();
  return(p);
}

void *ReallocMem(void *Ptr,uint16 NewSize)
{
  assert(!TransactionInProgress);
  return(KMEM__realloc(Ptr,NewSize));
}

void *ReallocHugeMem(void *Ptr,uint32 NewSize)
{
  assert(!TransactionInProgress);
  return(KMEM__hugerealloc(Ptr,NewSize));
}

void *XReallocMem(void *Ptr,uint16 MemSize)
{
  Ptr=ReallocMem(Ptr,MemSize);
  if (!Ptr) AbortMem();
  return(Ptr);
}

void *XReallocHugeMem(void *Ptr,uint32 MemSize)
{
  Ptr=ReallocHugeMem(Ptr,MemSize);
  if (!Ptr) AbortMem();
  return(Ptr);
}

void FreeMem(const void *Ptr)
{
  if (TransactionInProgress) {
    RegisterBlock(&FreeOnCommit,(void *)Ptr);
  } else {
    KMEM__free((void *)Ptr);
  }
}

void StartMemoryTransaction(void)
{
  assert(!TransactionInProgress);
  TransactionInProgress=true;
  FreeOnCommit.Cnt=0;
  FreeOnRollback.Cnt=0;
}

static void FreeBlocks(TMemoryBlocks *B)
{
  uint16 i;

  for (i=0;i<B->Cnt;i++) {
    FreeMem(B->Blocks[i]);
  }
}

void CommitMemoryTransaction(void)
{
  TransactionInProgress=false;
  FreeBlocks(&FreeOnCommit);
}

void RollbackMemoryTransaction(void)
{
  TransactionInProgress=false;
  FreeBlocks(&FreeOnRollback);
}
