#include <syslib/types.h>
#define __INTERFACE__
#include "dop.h"

/*
  The package description parser main module

  This module is responsible for parsing the package description file and for
  constructing the basic root file.
*/

#include <syslib/kernel.h>
#include <syslib/pdfileio.h>
#include <syslib/pdfsys.h>

#include <libhs/strings.h>
#include <libhs/diskio.h>
#include <libhs/filebuf.h>

#include "config.h"
#include "args.h"
#include "errors.h"
#include "fileread.h"
#include "dopscan.h"
#include "pkgdata.h"

/*** Check if the given file is a package description ***/
static int16 IsDescriptionFile(TDirItem *Item)
{
  uint16 Len;

  /*
    Determine the length of the name of the file.
  */
  Len=StrLen(Item->Name);
  /*
    If the file name is shorter than 5 characters, it cannot be an .DOP file.
    Files with empty filenames such as ".DOP" are never treated as DOP files.
  */
  if (Len<5) return(0);
  /*
    The file must have the ".dop" suffix.
  */
  return(StrCmp(Item->Name+Len-4,".dop")==0);
}

/*** get the filename without extension and place it to the name buffer ***/
static bool ParseFileName(const char *Arg, char *PkgName)
{
  int16 i;
  int16 Stop;
  char Ch;
  int16 Len;

  /*
    Initialize the name parsing.
  */
  i=StrLen(Arg);
  Stop=i;
  /*
    Find the string between '/' (or '\') and ".". This string is the name.
  */
  while (i>0) {
    i--;
    Ch=Arg[i];
    if (Ch=='.') Stop=i;
    if ((Ch=='/') || (Ch=='\\')) {
      i++;
      break;
    }
  }
  /*
    Determine the length of the name.
  */
  Len=Stop-i;
  /*
    If the length is too large, return failure and cause the caller to reject
    the file.
  */
  if (Len>MAX_FILENAME_LENGTH) return(false);
  /*
    Copy the string to the destination place and lowercaseize it.
  */
  StrNToLwr(PkgName,Arg+i,Len);
  PkgName[Len]='\0';
  /*
    Return success.
  */
  return(true);
}

static bool OpenTheDescription(const char *FileSpec)
{
  bool Success;

  Success=AllocateAndOpenInputFile(FileSpec);
  if (!Success) SysError("Unable to open the package description file");
  SetupFileName(FileSpec);
  return(Success);
}

/* Find and open the description file */
static bool FindAndOpenTheDescription(char *PackageName)
{
  int16 Dir;
  TDirItem Item;
  bool Result;
  bool FoundSomething;

  /*
    Open the current directory. If it can't be opened, complain and return.
  */
  Dir=OpenDir(".");
  if (Dir==-1) {
    SysError("Unable to scan the current directory");
    return(0);
  }
  /*
    Read items and search for an *.dop file. This version of the program does
    not support filenames with multiple extensions, so if this is the case,
    the "dop" extension must be the last one in the file.
    
    The "Result" variable is set to -1 first. When we encounter an *.dop
    file, we will open it and keep reading. When we will encounter another
    *.dop file, we will close the one we already opened and complain about
    that situation (we need the user to help us and specify, which *.dop file
    to use). Otherwise we will return just the one we opened.

    FEATURE:
    If more than one .dop file is there, read all of them and construct a
    'makefile' to make each program described in them. This will be done
    in one of the next version, now we can handle only one .dop file.
  */
  Result=false;
  FoundSomething=false;
  do {
    /*
      Read the directory item. If an error occurs, report it and prepare to
      return. Remember that the directory must be closed.
    */
    if (!ReadItem(Dir,&Item)) {
      SysError("Error reading the directory");
      Result=false;
      break;
    }
    /*
      If this was the last file, break out of the loop.
    */
    if (Item.ID==-1) break;
    /*
      Check if the item is a DOP file. If yes, install it's name somewhere.
      Also open it.
    */
    if (IsDescriptionFile(&Item)) {
      if (!ParseFileName(Item.Name,PackageName)) {
        PlainError("Package name too long, ignoring");
        continue;
      }
      /*
        Attempt to open the file.
      */
      if (FoundSomething) {
        CloseInputFile();
        PlainError("More than one program description found");
	Result=false;
	break;
      }
      Result=AllocateAndOpenInputFile(Item.Name);
      if (!Result) {
        PlainError("Could not open\001");
        PutArg(Item.Name);
      }
      SetupFileName(Item.Name);
      FoundSomething=true;
    }
  } while (1);
  if (!FoundSomething) PlainError("No program description found");
  /*
    Close the directory. If an error occurs here, abort the program even if
    we have the requested information. Errors during closing procedures
    usually mean system in ill state, so if they occur, get out as soon as
    possible.
  */
  if (!CloseDir(Dir)) {
    SysError("Error closing the directory");
  }
  return(Result);
}

TPackage *FindPackageDescription(void)
{
  bool Result;
  TPackage *P;
  char PackageName[16];

  if (DescFile==NULL) {
    Result=FindAndOpenTheDescription(PackageName);
  } else {
    ParseFileName(DescFile,PackageName);
    Result=OpenTheDescription(DescFile);
  }
  P=NULL;
  if (Result) {
    P=CreatePackage(PackageName);
  }
  return(P);
}

bool ParsePackageDescription(TPackage *P)
{
  bool Result;
  Result=ParsePackageDescriptionCore(P);
  CloseInputFile();
  return(Result);
}
