#include <syslib/types.h>
#define __INTERFACE__
#include "dopscan.h"

/*
  Parsing the package description content

  This module does the actual package description parsing.
*/

#include <libhs/strings.h>
#include <libhs/keyword.h>

#include "dop.h"
#include "config.h"
#include "errors.h"
#include "fileread.h"
#include "pkgdata.h"

typedef enum {
  kUnknown,
  kPROGRAM,
  kTRANSLATOR,
  kMODULEGEN,
  kLIBRARY,
  kSYSTEM,
  kAUTHOR,
  kLICENSE,
  kCREDIT,
  kTITLE,
  kCOPYRIGHT,
  kBRIEF,
  kDESC,
  kSOURCE,
  kHEADER,
  kSHARED,
  kTEST,
  kEOF
} TKeyWord;

static TKeyWord ReadKeyword(char *KeyWordStr,int16 BufSize)
{
  do {
    ReadWordIntoBuf(KeyWordStr,BufSize);
    if (KeyWordStr[0]=='\n') GetChar();
    if (KeyWordStr[0]=='#') {
      ReadToNL();
      KeyWordStr[0]='\n';
    }
  } while (KeyWordStr[0]=='\n');
  if (KeyWordStr[0]=='\0') return(kEOF);

  return(RecognizeWord(sensINSENSITIVE,KeyWordStr,
    "Program" "Translator" "Modulegen" "Library"
    "System"
    "Author" "License" "Credit" "Title"
    "Copyright"
    "Brief" "Desc" "Source" "Header" "Shared" "Test"
  ));
}

static TPackageItem *CreatePackageItemWithFile(char *Ext,TPackageItemType T)
{
  TPackageItem *R;
  TFileName Buf;

  Buf[MAX_FILENAME_LENGTH]='\0';
  ReadWordIntoBuf(Buf,MAX_FILENAME_LENGTH+1);
  if (Buf[MAX_FILENAME_LENGTH]!='\0') {
    TokError("Package item name too long");
    return(NULL);
  }
  R=CreatePackageItem(Buf);
  if (R!=NULL) {
    R->Type=T;
    R->Output=RegisterFileSplitted(Buf,Ext);
    if (R->Output==NULL) {
      DestroyPackageItem(R);
      R=NULL;
    }
  }
  return(R);
}

static TPackageItem *CreateProgram(void)
{
  return(CreatePackageItemWithFile(MachineCodeExt->Ext,ptProgram));
}

static TPackageItem *CreateLibrary(void)
{
  return(CreatePackageItemWithFile(MachineCodeExt->Ext,ptLibrary));
}

static bool ReadExtension(char *Ext)
{
  Ext[3]='\0';
  ReadWordIntoBuf(Ext,4);
  if (Ext[3]!='\0') {
    TokError("File extension too long");
    return(false);
  }
  return(true);
}

static TPackageItem *CreateTranslator(void)
{
  TPackageItem *I;
  char IExt[4];
  char OExt[4];
  TFileExtension *IExtPtr;
  TFileExtension *OExtPtr;
  bool Valid;

  I=CreatePackageItemWithFile(MachineCodeExt->Ext,ptTranslator);
  if (I!=NULL) {
    Valid=ReadExtension(IExt);
    if (!ReadExtension(OExt)) Valid=false;
    if (Valid) {
      IExtPtr=RegisterExtension(IExt);
      if (IExtPtr==NULL) Valid=false;
      OExtPtr=RegisterExtension(OExt);
      if (OExtPtr==NULL) Valid=false;
    }
    if (Valid) {
      I->ISrc=IExtPtr;
      I->OHdr=OExtPtr;
      OExtPtr->Translator=I;
      OExtPtr->Type=ftHeader;
    } else {
      DestroyPackageItem(I);
      I=NULL;
    }
  }
  return(I);
}

static TPackageItem *CreateModuleGenerator(void)
{
  TPackageItem *I;
  char SExt[4];
  TFileExtension *SExtPtr;
  bool Valid;

  I=CreateTranslator();
  I->Type=ptModuleGenerator;
  Valid=I!=NULL;
  if (Valid) {
    Valid=ReadExtension(SExt);
    if (Valid) {
      SExtPtr=RegisterExtension(SExt);
      I->OSrc=SExtPtr;
      Valid=(SExtPtr!=NULL);
    }
  }
  if (!Valid) {
    DestroyPackageItem(I);
    I=NULL;
  }
  return(I);
}

typedef TPackageItem *TCreationHandler(void);

static TPackageItem *I;
static TArrayHandle Sources;
static TArrayHandle Headers;

static void OpenArraysInItem(void)
{
  assert(I!=NULL);
  OpenArrayEx(&I->Sources,&Sources);
  OpenArrayEx(&I->Headers,&Headers);
}

static void CloseArraysInItem(void)
{
  if (I!=NULL) {
    CloseArray(&Sources);
    CloseArray(&Headers);
  }
}

static void RegisterFileOwner(TFileInModule *F,TFileInModule *Owner)
{
  TFileInModule *CurrentOwner;

  if (F->Owner==NULL) {
    F->Owner=Owner;
  } else { 
    CurrentOwner=F->Owner;
    if (CurrentOwner!=Owner) {
      TokError("File\1illegally shared between\2and\3");
      PutFileArg(F);
      PutArg(CurrentOwner->Module->Name);
    } else {
      TokError("File\1already owned by\2");
      PutFileArg(F);
    }
    PutArg(Owner->Module->Name);
  }
}

static void AddRootFiles(
  TFileType Type,
  TArrayHandle *H,
  TPackageItem *Owner,
  bool Shared
)
{
  TFileSpec Buffer;
  TFileInModule *F;
  TFileInModule *SF;

  if (I==NULL) {
    TokError("Directive not allowed at package level");
    return;
  }
  do {
    Buffer[MAX_FILESPEC_LENGTH]='\0';
    ReadWordIntoBuf(Buffer,MAX_FILESPEC_LENGTH+1);
    if (Buffer[0]=='\0' || Buffer[0]=='\n') break;
    if (Buffer[MAX_FILESPEC_LENGTH]!='\0') {
      TokError("File specification is too long");
      continue;
    }
    F=RegisterFile(Buffer);
    if (F!=NULL) {
      if (CheckFile(F,Type)) {
        RegisterFileOwner(F,Owner->Output);
        if (H!=NULL) PutItem(H,F);
        if (Shared) F->Shared=true;
        if (F->Ext->Type==ftHeader) {
          SF=RegisterFileSplitted(F->Module->Name,F->Ext->Assoc1->Ext);
          if (SF!=NULL) {
            RegisterFileOwner(SF,Owner->Output);
            if (Shared) SF->Shared=true;
          }
        }
      }
    }
  } while (true);
}

bool ParsePackageDescriptionCore(TPackage *P)
{
  TCreationHandler *CreationHandler;
  TArrayHandle H;
  TKeyWord Key;
  char KeyWordStr[18];
  bool Valid;

  assert(P!=NULL);
  Valid=true;
  InitFileReader(true);
  I=NULL;
  OpenArrayEx(&P->Items,&H);
  do {
    Key=ReadKeyword(KeyWordStr,18);
    CreationHandler=NULL;
    switch (Key) {
      case kEOF: break;
      case kPROGRAM: CreationHandler=CreateProgram; break;
      case kLIBRARY: CreationHandler=CreateLibrary; break;
      case kTRANSLATOR: CreationHandler=CreateTranslator; break;
      case kMODULEGEN: CreationHandler=CreateModuleGenerator; break;
      case kSOURCE: AddRootFiles(ftSource,&Sources,I,false); break;
      case kHEADER: AddRootFiles(ftHeader,&Headers,I,true); break;
      case kSHARED: AddRootFiles(ftHeader,NULL,I,true); break;
      case kSYSTEM: {
        TokError("System packages are not buildable with this");
        ReadToNL();
        break;
      }
      case kUnknown: {
        TokError("Unrecognized keyword\1");
        PutArg(KeyWordStr);
	/* fall through */
      }
      case kAUTHOR:
      case kLICENSE:
      case kCREDIT:
      case kTITLE:
      case kDESC:
      case kCOPYRIGHT:
      default: ReadToNL(); break;
    }
    if (CreationHandler!=NULL) {
      CloseArraysInItem();
      I=CreationHandler();
      if (I!=NULL) {
        if (!PutItem(&H,I)) {
          DestroyPackageItem(I);
          ReportError(CurLine,0,"Unable to bind the item into its package");
          Valid=false;
        }
        OpenArraysInItem();
      } else {
        Valid=false;
      }
    }
  } while (Key!=kEOF);
  CloseArray(&H);
  CloseArraysInItem();
  return(Valid);
}
