#include <syslib/types.h>
#define __INTERFACE__
#include "errors.h"

#include <libhs/strings.h>
#include <libhs/process.h>
#include <libhs/errno.h>
#include <libhs/filebuf.h>
#include <libhs/memory.h>
#include <libhs/output.h>
#include <libhs/dictrb.h>

#include "fileread.h"

void Fatal(const char *Msg)
{
  Print(Msg);
  NewLine();
  Halt(2);
}

static TMessage *Cur;
static TMessage *Last;
static TMessage MemErrMsg;

static uint8 ArgCnt;
static uint8 CurArg;

static TRBDict *FileNames;
static char *LastFileName;
static char *CurrentFileName;

#define MAX_NAMES_IN_STACK 4

static char *FileNamesStack[MAX_NAMES_IN_STACK];
static int16 StackPtr;

static const char *InfoMsg;
static const char *InfoArgs[3];

void InitErrors(void)
{
  assert(MemErrMsg.Column==0);
  MemErrMsg.Column=8192;
  MemErrMsg.Text="Out of memory";
  FileNames=CreateDict();
  if (FileNames==NULL) Fatal("Unable to create dictionary of file names");
  LastFileName=NULL;
  CurrentFileName=NULL;
  ErrorCount=0;
}

void ReportOutOfMemory(void)
{
  bool Discard;

  if (OutOfMmemory) return;
  assert(MemErrMsg.Column>0);
  Discard=false;
  if (MsgHook) Discard=MsgHook(&MemErrMsg);
  OutOfMmemory=true;
  if (!Discard) Last->Next=&MemErrMsg;
  ErrorCount+=1;
}

void SetupFileName(const char *FileName)
{
  char *NameCopy;

  if (FileName!=NULL) {
    NameCopy=SearchKey(FileNames,FileName);
    if (NameCopy==NULL) {
      NameCopy=StrDup(FileName);
      if (NameCopy!=NULL) {
        if (AddItemWithKey(FileNames,NameCopy,NameCopy)!=rbSUCCESS) {
          FreeMem(NameCopy);
          NameCopy=NULL;
        }
      }
      if (NameCopy==NULL) {
        ReportOutOfMemory();
      }
    }
    CurrentFileName=NameCopy;
  } else {
    CurrentFileName=NULL;
  }
}

void PushFileName(void)
{
  assert(StackPtr<MAX_NAMES_IN_STACK);
  FileNamesStack[StackPtr++]=CurrentFileName;
}

void PopFileName(void)
{
  assert(StackPtr>0);
  StackPtr--;
  CurrentFileName=FileNamesStack[StackPtr];
}

static void FreeMessage(TMessage *Ptr)
{
  int16 i;
  const char *Arg;

  for (i=0;i<3;i++) {
    Arg=Ptr->Args[i];
    if (Arg) FreeMem(Arg);
  }
  FreeMem(Ptr);
}

static void EndMsgCore(void)
{
  bool Discard;

  if (OutOfMmemory) return;
  assert(MemErrMsg.Column>0);
  assert(Cur!=NULL);
  Discard=false;
  if (MsgHook) Discard=MsgHook(Cur);
  if (Discard) {
    FreeMessage(Cur);
  } else {
    if (First==NULL) {
      First=Cur;
    } else {
      Last->Next=Cur;
    }
    Last=Cur;
  }
  Cur=NULL;
}

static void EndMsg(void);

static void PutMsg(uint16 Line,uint16 Column,char Class,const char *Text)
{
  const char *Ptr;
  char Ch;
  int16 i;

  ErrorCount+=1;
  if (OutOfMmemory) return;
  if (CurrentFileName!=LastFileName) {
    LastFileName=CurrentFileName;
    PutMsg(0,0,'f',CurrentFileName);
  }
  assert(MemErrMsg.Column>0);
  assert(Cur==NULL);
  Cur=GetMem(sizeof(*Cur));
  if (Cur==NULL) {
    ReportOutOfMemory();
    return;
  }
  Cur->Class=Class;
  Cur->Line=Line;
  Cur->Column=Column;
  Cur->Text=Text;
  Cur->Next=NULL;
  for (i=0;i<3;i++) Cur->Args[i]=NULL;
  /*
    Determine the count of arguments. The argument placement is identified
    by the \1, \2 or \3 characters in the text of the message.
  */
  assert(ArgCnt==0);
  assert(CurArg==0);
  if (Class!='f') {
    assert(Text!=NULL);
    Ptr=Text;
    do {
      Ch=*(Ptr++)&0xF7;
      if (Ch>='\1' && Ch<='\3') {
        if (ArgCnt<Ch) ArgCnt=Ch;
      }
    } while (Ch);
  }
  /*
    If there are no arguments, terminate the message
  */
  if (!ArgCnt) EndMsg();
}

void ReportError(uint16 Line,uint16 Column,const char *Text)
{
  PutMsg(Line,Column,'E',Text);
}

void ReportWarning(uint16 Line,uint16 Column,const char *Text)
{
  PutMsg(Line,Column,'W',Text);
}

void PutArg(const char *Arg)
{
  char *Ptr;

  /*
    Preconditions
  */
  assert(Cur!=NULL);
  assert(CurArg<ArgCnt);
  /*
  */
  Ptr=GetMem(StrLen(Arg)+1);
  if (Ptr==NULL) {
    FreeMessage(Cur);
    ReportOutOfMemory();
    return;
  }
  StrCpy(Ptr,Arg);
  /*
    Store the argument into the error message record.
  */
  Cur->Args[CurArg++]=Ptr;
  /*
    If this was the last argument in the field, terminate the message and
    cleanup the argument pointers.
  */
  if (CurArg==ArgCnt) {
    CurArg=0;
    ArgCnt=0;
    EndMsg();
  }
}

const char *Number2Text(uint16 Number)
{
  static char Buf[6];
  char *Ptr;

  Ptr=Buf+5;
  *Ptr='\0';
  do {
    *(--Ptr)='0'+(Number % 10);
    Number/=10;
  } while (Number);
  return(Ptr);
}

void PutNumArg(uint16 Number)
{
  PutArg(Number2Text(Number));
}

static void FlushInfoMsg(void)
{
  const char *Msg;

  if (InfoMsg!=NULL) {
    Msg=InfoMsg;
    InfoMsg=NULL;
    PutMsg(0,0,'I',Msg);
    while (CurArg<ArgCnt) PutArg(InfoArgs[CurArg]);
  }
}

static void SendInfoMsg(void)
{
  if (ArgCnt==0) FlushInfoMsg();
}

static void EndMsg(void)
{
  EndMsgCore();
  FlushInfoMsg();
}

void ClearMessages(void)
{
  TMessage *Ptr;

  assert(MemErrMsg.Column>0);
  /*
    If the "out of memory" message is there, disconnect it from the list
    to prevent it from being freed. This message must be permanently
    available for the library because it would be impossible to report the
    out of error memory condition cleanly.
  */
  if (OutOfMmemory) {
    Last->Next=NULL;
    OutOfMmemory=false;
  }
  /*
    Free the messages in the list.
  */
  while (First) {
    Ptr=First;
    First=(void*)Ptr->Next;
    FreeMessage(Ptr);
  }
}

#define LINE_WIDTH (80-16)

static char Buf[81];
static char Word[81];
static char Spaces[14];
static char *BPtr;
static char *WPtr;

void FlushMessageLine(void)
{
  Print(Buf);
  NewLine();
  Print(Spaces);
  BPtr=Buf;
}

void FlushMessageWord(void)
{
  assert(WPtr!=Word);
  if ((BPtr-Buf)+(WPtr-Word)+1>LINE_WIDTH) FlushMessageLine();
  *(BPtr++)=' ';
  *WPtr='\0';
  WPtr=Word;
  BPtr=StrCpy(BPtr,Word);
}

bool IsFileName(const TMessage *Msg)
{
  return(Msg->Class=='f');
}

void DumpFileName(const TMessage *Msg)
{
  if (Msg->Text==NULL) {
    PrintLn("Not in a file");
  } else {
    Print("In file '");
    Print(Msg->Text);
    PrintLn("'");
  }
}

bool DumpMsg(const TMessage *Msg)
{
  const char *Ptr;
  char Ch;
  const char *Arg;
  uint16 Len;

  /*
    If the error message is the filename reporter, report the filename and
    terminate.
  */
  if (IsFileName(Msg)) {
    DumpFileName(Msg);
    return(true);
  }
  /*
    Initialize the buffer used to output spaces at the bginning of the next
    lines (to align them with the first header).
  */
  MemSet(Spaces,13,' ');
  Spaces[13]='\0';
  /*
    Print the error message header.
  */
  Buf[0]=Msg->Class;
  Buf[1]=0;
  Print(Buf);
  if (Msg->Line) {
    PrintNum(Msg->Line,6);
    PrintNum(Msg->Column,6);
  } else {
    Spaces[12]='\0';
    Print(Spaces);
    Spaces[12]=' ';
  }
  /*
    Print the error message. Split it into several rows if it is too long.
  */
  Ptr=Msg->Text;
  BPtr=Buf;
  WPtr=Word;
  do {
    Ch=*(Ptr++);
    if ((Ch>='\1' && Ch<='\3')) {
      FlushMessageWord();
      Arg=Msg->Args[Ch-1];
      assert(Arg!=NULL);
      Len=StrLen(Arg);
      assert(Len<=78);
      Word[0]='"';
      WPtr=StrCpy(Word+1,Msg->Args[Ch-1]);
      *(WPtr++)='"';
      Ch=*Ptr;
      if (Ch!='.' && Ch!=',' && Ch!=':' && Ch!=';') FlushMessageWord();
    } else if ((Ch>='\11' && Ch<='\13')) {
      WPtr=StrCpy(WPtr,Msg->Args[Ch-'\11']);
    } else {
      if ((Ch==' ') || (Ch=='\0')) {
        FlushMessageWord();
      } else {
        assert(WPtr-Word<LINE_WIDTH);
        *(WPtr++)=Ch;
      }
    }
  } while (Ch);
  Print(Buf);
  NewLine();
  /*
    Return true so this dumper can be used as a message hook.
  */
  return(true);
}

void PlainError(const char *Msg)
{
  ReportError(0,0,Msg);
}

void FileError(const char *Msg)
{
  ReportError(CurLine,CurColumn,Msg);
}

void TokError(const char *Msg)
{
  ReportError(TokLine,TokColumn,Msg);
}

void SendErrorInfoMsg(int16 ErrorCode)
{
  InfoMsg="System said:\001";
  InfoArgs[0]=ComplainError(ErrorCode);
  SendInfoMsg();
}

void SysError(const char *Msg)
{
  PlainError(Msg);
  SendErrorInfoMsg(GetErrNo());
}

bool CheckIOResultAndReport(const char *Msg)
{
  int16 Result;

  Result=IOResult();
  if (Result<0) {
    if (Msg!=NULL) PlainError(Msg);
    SendErrorInfoMsg(Result);
  }
  return(Result>=0);
}
