#include <syslib/types.h>
#define __INTERFACE__
#include "fileext.h"

/*
  File extension relationship management for OBS

  This module contains the file extension relationship manager. The
  relationships managed by this module are the relationships between file
  extensions and the properties of the file (which is the second portion of
  the module, which translator to be used to process the file etc).
*/

#include <libhs/strings.h>
#include <libhs/memory.h>
#include <libhs/dictrb.h>
#include <libhs/libman.h>

#include "errors.h"

static TRBDict *Extensions;

static TFileExtension *RegisterBuiltinExtension(const char *Ext)
{
  TFileExtension *E;

  E=RegisterExtension(Ext);
  if (E==NULL) Fatal("Unable to register a builtin extension");
  return(E);
}

void InitExtensions(void)
/*
  DESCRIPTION

  Creates a dictionary of extensions and places the builtin extensions. These
  builtin extensions are defined by the core ESPM language and every ESPM
  compiler must know about them: The "c" and "h" are for sources written in C
  compatibility mode of P language, the "p" and "ph" are for sources written
  in the full P language, "rmc" are for the compiled modules ("Relocatable
  Module Code") and "bcf" are for binaries ("Binary Code File"), which are
  either the programs, the libraries or "binary mixins" (a file that acts as
  both, the program and the binary.
*/
{
  char BinExtBuf[4];
  TFileExtension *E_C;
  TFileExtension *E_H;
  TFileExtension *E_P;
  TFileExtension *E_PH;
  TFileExtension *E_RMC;
  TFileExtension *E_BCF;

  assert(Extensions==NULL);
  Extensions=CreateDict();
  if (Extensions==NULL) {
    Fatal("Unable to create main dictionary of file extensions");
  }
  E_RMC=RegisterBuiltinExtension("rmc");
  E_RMC->Type=ftCompiledModule;
  E_C=RegisterBuiltinExtension("c");
  E_C->Type=ftSource;
  E_C->Assoc1=E_RMC;
  E_H=RegisterBuiltinExtension("h");
  E_H->Type=ftHeader;
  E_H->Assoc1=E_C;
  E_C->Assoc2=E_H;
  E_P=RegisterBuiltinExtension("plb");
  E_P->Type=ftSource;
  E_P->Assoc1=E_RMC;
  E_PH=RegisterBuiltinExtension("plh");
  E_PH->Type=ftHeader;
  E_PH->Assoc1=E_P;
  E_P->Assoc2=E_PH;
  PutBinaryExtension(BinExtBuf,false);
  E_BCF=RegisterBuiltinExtension(BinExtBuf);
  E_BCF->Type=ftBinary;
  E_RMC->Assoc1=E_BCF;
  MachineCodeExt=E_BCF;
}

void DestroyExtensions(void)
{
  TFileExtension *E;

  assert(Extensions!=NULL);
  E=FindFirstKey(Extensions);
  while (E!=NULL) {
    FreeMem(E->Ext);
    FreeMem(E);
    E=FindNextKey(Extensions);
  }
  DestroyDict(Extensions);
  Extensions=NULL;
}

TFileExtension *SearchExtension(const char *Ext)
/*
  DESCRIPTION

  This function searches the database of file extensions for the extension
  specified by a string. It returns pointer to the extension info (declared
  in the header) or NULL if the extension has not yet been registered.
*/
{
  return(SearchKey(Extensions,Ext));
}

static void ReportMemoryFailure(void)
/*
  DESCRIPTION

  This local procedure reports a memory failure and rollbacks the memory
  transaction in progress. Used from ~RegisterExtension() on a memory
  allocation failure to cleanup after the memory transaction before returning
  an error result.
*/
{
  RollbackMemoryTransaction();
  ReportError(0,0,"Not memory for file extension information structures");
}

TFileExtension *RegisterExtension(const char *Ext)
/*
  DESCRIPTION

  Registers the specified extension, i.e. creates an extension info record
  and inserts it into the file extension database. The extension record is
  initialized to a generic extension; the caller has to adjust the info
  according to the data it encounters (for example that the extension belongs
  to a simple source file when parsing a translator declaration in the .DOP
  file). It returns NULL if there was insufficient memory for the necessary
  data structures. It creates and commits a memory transaction so it cannot
  be used when the caller has opened its own memory transaction.
*/
{
  TFileExtension *R;

  R=SearchExtension(Ext);
  if (R==NULL) {
    StartMemoryTransaction();
    R=GetMem(sizeof(*R));
    if (R==NULL) {
      ReportMemoryFailure();
      return(NULL);
    }
    R->Ext=StrDup(Ext);
    if (R->Ext==NULL) {
      ReportMemoryFailure();
      return(NULL);
    }
    if (AddItemWithKey(Extensions,R->Ext,R)!=rbSUCCESS) {
      ReportMemoryFailure();
      return(NULL);
    }
    R->Type=ftUnknown;
    R->Assoc1=NULL;
    R->Translator=NULL;
    CommitMemoryTransaction();
  }
  return(R);
}
