#include <syslib/types.h>
#define __INTERFACE__
#include "fileread.h"

/*
  Simple file reading module

  This module allows the program to read one file at a time. It tracks the
  line and column position in the file (including peculiarities like tabs)
  and allows the caller to go one character back in the input stream.
*/

#include <libhs/filebuf.h>

#include "errors.h"

static uint8 LastChar;
static int16 LastLine;
static int16 LastColumn;

static uint8 StoredChar;
static int16 StoredLine;
static int16 StoredColumn;

static uint8 InternalStoredChar;

static bool NewlinesAreWhitespace;

static bool DoCCommentFiltering;
static bool InCComment;

void InitFileReader(bool NewlinesSignificant)
{
  CurLine=1;
  CurColumn=0;
  StoredChar=0;
  NewlinesAreWhitespace=!NewlinesSignificant;
  DoCCommentFiltering=false;
  InCComment=false;
  InternalStoredChar=0;
}

void SetupCCommentFiltering(void)
{
  DoCCommentFiltering=true;
}

void CloseInputFile(void)
{
  CloseAndFreeInputFile();
  SetupFileName(NULL);
}

uint8 GetChar(void)
{
  if (StoredChar) {
    CurChar=StoredChar;
    CurLine=StoredLine;
    CurColumn=StoredColumn;
    StoredChar=0;
  } else {
    LastLine=CurLine;
    LastColumn=CurColumn;
    LastChar=CurChar;
    do {
      if (InternalStoredChar!=0) {
        CurChar=InternalStoredChar;
        InternalStoredChar=0;
      } else {
        CurChar=ReadByte();
      }
    } while (CurChar=='\r');
    switch (CurChar) {
      case '\n': {
        CurLine++;
        CurColumn=0;
        break;
      }
      case '\t': {
        CurColumn=((CurColumn+7)&7)+1;
        break;
      }
      default: {
        CurColumn++;
        if (DoCCommentFiltering) {
          if (InCComment) {
            if (CurChar=='*') {
              InternalStoredChar=ReadByte();
              if (InternalStoredChar=='/') {
                InCComment=false;
                InternalStoredChar=' ';
              }
            }
            CurChar=' ';
          } else {
            if (CurChar=='/') {
              InternalStoredChar=ReadByte();
              if (InternalStoredChar=='*') {
                InCComment=true;
                CurChar=' ';
              }
            }
          }
        }
        break;
      }
    }
  }
  return(CurChar);
}

void MarkToken(void)
{
  TokLine=CurLine;
  TokColumn=CurColumn;
}

void GoBack(void)
{
  assert(StoredChar==0);
  StoredChar=CurChar;
  StoredLine=CurLine;
  StoredColumn=CurColumn;
  CurChar=LastChar;
  CurLine=LastLine;
  CurColumn=LastColumn;
}

bool IsWhitespace(char Ch)
{
  return((NewlinesAreWhitespace && (Ch=='\n')) || (Ch==' ') || (Ch=='\t'));
}

bool IsDigit(void)
{
  return((CurChar>='0') && (CurChar<='9'));
}

bool IsLowerCase(void)
{
  return((CurChar>='a') && (CurChar<='z'));
}

bool IsStartingIDChar(void)
{
  return(
    ((CurChar>='A') && (CurChar<='Z')) ||
    IsLowerCase() ||
    (CurChar=='_')
  );
}

void ReadToNL(void)
{
  if (StoredChar==10) {
    GetChar();
  } else {
    StoredChar=0;
    SkipRestOfLine();
    CurLine++;
    CurColumn=0;
  }
}

void GoNonwhitespaceBack(void)
{
  if (!IsWhitespace(CurChar)) GoBack();
}

void SkipWhitespace(void)
{
  char Ch;
  
  do {
    Ch=GetChar();
  } while (IsWhitespace(Ch));
  GoBack();
}

void ReadWordIntoBuf(char *Buf,uint16 BufSize)
{
  char Ch;
  char *Ptr;

  Ptr=Buf;
  SkipWhitespace();
  MarkToken();
  do {
    Ch=GetChar();
    if (Ch==' ' || Ch=='\t' || Ch=='\n' || Ch=='\0') break;
    if (BufSize>0) {
      if ((BufSize--)==1) {
        FileError("Word too long");
	continue;
      }
      *(Buf++)=Ch;
    }
  } while (1);
  if (Ch=='\n') {
    if (Buf==Ptr) *(Buf++)='\n';
    GoBack();
  }
  *Buf='\0';
}
