#include <syslib/types.h>
#define __INTERFACE__
#include "files.h"

/*
  Generic file managing module for OBS

  Files are stored in dictionaries which have two levels. In the first level
  the modules are recorded, which give the names of the files and within them
  the file extensions are recorded.
*/

#include <syslib/kernel.h>
#include <syslib/pdfsys.h>
#include <syslib/traps.h>

#include <libhs/strings.h>
#include <libhs/libman.h>
#include <libhs/diskio.h>
#include <libhs/memory.h>
#include <libhs/filebuf.h>
#include <libhs/output.h>

#include "config.h"
#include "errors.h"
#include "filespec.h"
#include "fileext.h"
#include "filestk.h"
#include "filetype.h"

static TRBDict *Modules;

/*** Initialize the file handling machinery ***/
void InitFiles(void)
{
  assert(Modules==NULL);
  Modules=CreateDict();
  if (Modules==NULL) Fatal("Unable to create main dictionary of modules");
}

void DestroyFileDict(TRBDict *T)
{
  TFileInModule *F;

  F=FindFirstKey(T);
  while (F!=NULL) {
    DestroyArray(&F->References);
    DestroyArray(&F->Libraries);
    FreeMem(F);
    F=FindNextKey(T);
  }
  DestroyDict(T);
}

void DestroyFiles(void)
{
  TModule *M;

  assert(Modules!=NULL);
  /*
    Destroy the names and the 
  */
  M=FindFirstKey(Modules);
  while (M!=NULL) {
    FreeMem(M->Name);
    DestroyFileDict(M->Files);
    FreeMem(M);
    M=FindNextKey(Modules);
  }
  /*
    Destroy the dictionary of modules
  */
  DestroyDict(Modules);
  Modules=NULL;
}

static void ReportMemoryFailure(void)
{
  RollbackMemoryTransaction();
  ReportError(0,0,"Not enough memory for file information structures");
}

static char *CheckAndAllocateName(const char *Name)
{
  TFileSpec NameBuf;
  char *Result;

  if (CheckAndCanonizeModuleFileName(NameBuf,Name)==NULL) {
    RollbackMemoryTransaction();
    return(NULL);
  }
  Result=StrDup(NameBuf);
  if (Result==NULL) ReportMemoryFailure();
  return(Result);
}

static TFileInModule *RegisterFileSplittedCore(
  TLocation Loc,
  const char *Name,
  const char *Ext
)
{
  char ModKey[MAX_FILESPEC_LENGTH+2];
  char *ModName;
  TModule *M;
  opaqueTFileInModule *F;
  TFileExtension *E;
  TStackID i;

  /*
    Attempt to register the file extension. If this attempt fails, the file
    is not registrable due to lack of memory - bail out. Note that there is
    no error message displayed here - the error message is displayed by the
    RegisterExtension() call.
  */
  E=RegisterExtension(Ext);
  if (E==NULL) return(NULL);
  /*
    Allocate the memory for the names and put the names into them.
  */
  StartMemoryTransaction();
  ModName=CheckAndAllocateName(Name);
  if (ModName==NULL) return(NULL);
  ModKey[0]=Loc;
  StrCpy(ModKey+1,ModName);
  /*
    Create the module for the file if it does not exist yet.
    However if the module exist, check if the file is already in it and if
    so, rollback the memory transaction and exit with success (it is ok to
    register a file more than once; each succeeding registration must
    return the same file record than the first one).
  */
  M=SearchKey(Modules,ModKey);
  if (M==NULL) {
    M=GetMem(sizeof(*M));
    if (M==NULL) {
      ReportMemoryFailure();
      return(NULL);
    }
    if (AddItem(Modules,ModKey,M)!=rbSUCCESS) {
      ReportMemoryFailure();
      return(NULL);
    }
    M->Location=Loc;
    M->Name=ModName;
    M->Files=CreatePointerDict();
    if (M->Files==NULL) {
      ReportMemoryFailure();
      return(NULL);
    }
  } else {
    F=SearchPointerKey(M->Files,E);
    if (F!=NULL) {
      RollbackMemoryTransaction();
      return((void *)F);
    }
  }
  /*
    Create the structure for the file and add it into the dictionary of the
    module.
  */
  F=GetMem(sizeof(*F));
  if (F==NULL) {
    ReportMemoryFailure();
    return(NULL);
  }
  if (AddPointerItem(M->Files,E,F)!=rbSUCCESS) {
    ReportMemoryFailure();
    return(NULL);
  }
  F->P.Module=M;
  F->P.Ext=E;
  F->P.OldTimeStamp=0;
  F->P.MaxTimeStamp=0;
  for (i=0;i<MAX_STACKS;i++) F->StackNext[i]=NULL;
  F->P.Owner=NULL;
  F->P.Shared=false;
  F->P.Listed=false;
  F->P.Status=statMayBeOutdated;
  InitArray(&F->P.References);
  InitArray(&F->P.Libraries);
  F->P.Assoc1=NULL;
  F->NextPending=NULL;
  F->P.ID=65535;
  /*
    Everything went OK.
  */
  CommitMemoryTransaction();
  return((void *)F);
}

static uint32 GetTimeStampOfFileSpec(const char *FileSpec)
/*
  DESCRIPTION

  Given a file name returns the timestamp recorded in the database for the
  specified file. The procedure looks into the filesystem via FStat() call.
  It returns the timestamp or 0 if the file is not found in the filesystem.
  The file database is not updated.
*/
{
  TFileInfo Info;
  int16 FileID;
  int16 Result;

  FileID=Open(FileSpec,oeOPEN+oRDONLY);
  Result=FileID;
  if (FileID>0) {
    Result=FStat(FileID,&Info);
    Close(FileID);
  }
  if (Result>=0) {
    return(Info.ModifiedTime);
  } else {
    return(0);
  }
}

static uint32 GetTimeStamp(TFileInModule *F)
{
  TFileSpec Buffer;

  ExtractNameFromFile(F,Buffer);
  return(GetTimeStampOfFileSpec(Buffer));
}

void UpdateTimeStamp(TFileInModule *F)
/*
  DESCRIPTION

  Synchronize the tiomestamp in the file record passed to this call with the
  filesystem.
*/
{
  F->TimeStamp=GetTimeStamp(F);
}

TFileInModule *RegisterFileSplittedLoc(
  TLocation Loc,
  const char *Name,
  const char *Ext
)
/*
  DESCRIPTION

  Register the file into the database. The timestamp of the file is updated
  from the filesystem. This call accepts the file specification in three
  distinct portions: the location, the name (along with the path) and the
  extension of the file.
*/
{
  TFileInModule *F;

  F=RegisterFileSplittedCore(Loc,Name,Ext);
  if (F!=NULL) UpdateTimeStamp(F);
  return(F);
}

TFileInModule *RegisterFileSplitted(const char *Name, const char *Ext)
/*
  DESCRIPTION

  This is a convenient procedure for registering files in the working
  directory. This accepts the filespec splitted into the filename (along with
  the path) and the extension.
*/
{
  return(RegisterFileSplittedLoc(locWorkDir,Name,Ext));
}

/*** Register the specified file in the file handling machinery ***/
TFileInModule *RegisterFileLoc(TLocation Loc,const char *FileName)
/*
  DESCRIPTION

  Similar to RegisterFileSplittedLoc() but accepts the filename as a single
  string. It parses the string to the name of the file and the extension by
  itself.
*/
{
  uint16 Len;
  uint16 ExtLen;
  const char *Extension;
  char Ch;
  TFileSpec Buffer;
  TFileInModule *F;

  /*
    Split the filename into two pieces: the module name (the directory + the
    name of the file) and the extension. To do so, find the last dot in the
    file specification and split the name of the file around it. The file
    specification must be relative.
  */
  Len=StrLen(FileName);
  Extension=FileName+Len;
  ExtLen=0;
  do {
    ExtLen++;
    Ch=*(--Extension);
  } while (ExtLen<Len && (Ch!='.'));
  if (Ch=='.') {
    Extension++;
    Len-=ExtLen;
    if (Len>MAX_FILESPEC_LENGTH) {
      TokError("Filename is awfully long");
      return(NULL);
    }
    MemCpy(Buffer,FileName,Len);
    Buffer[Len]='\0';
    FileName=Buffer;
  } else {
    Extension=FileName+Len;
  }
  /*
    Create the file information record and fill the timestamp in the record.
  */
  F=RegisterFileSplittedLoc(Loc,FileName,Extension);
  return(F);
}

/*** Register a file in the working directory ***/
TFileInModule *RegisterFile(const char *FileName)
/*
  DESCRIPTION

  A convenience procedure similar to RegisterFileSplitted(). This one accepts
  the file specification relative to the working directory as a single string
  and registers the file in the working directory.
*/
{
  return(RegisterFileLoc(locWorkDir,FileName));
}

/*** Get the module to which the specified file belongs ***/
/*
  WARNING: Changes the content of the input buffer
*/
TModule *GetModuleOfFile(char *FileName)
/*
  DESCRIPTION

  Returns the module record of the module that contains the specified file.
  The file is expected in the working directory. The call modifies the input
  filename by splitting the extension out of it. This changing sideeffect is
  due to performance reasons (in most cases the procedure gets a temporary
  buffer which is discarded after use).
*/
{
  uint16 Len;
  char *Ptr;
  char Ch;

  /*
    Strip the extension of the file out. The rest is the name of the module.
  */
  Len=StrLen(FileName);
  Ptr=FileName+Len;
  do {
    Ch=*(--Ptr);
  } while (Ptr!=FileName && (Ch!='.'));
  if (Ch=='.') *Ptr='\0';
  /*
    Search for the module and return the result.
  */
  return(SearchKey(Modules,FileName));
}

static char *ConstructFileName(TModule *Module,const char *Ext,char *Buffer)
/*
  DESCRIPTION

  Construct a filename from the given module and the extension. This
  generates either a filename relative to the working directory or an
  absolute filename. The location of the file is taken from the module
  record.
*/
{
  char *Ptr;
  char *End;
  uint16 Remaining;

  Remaining=MAX_FILESPEC_LENGTH+1;
  switch (Module->Location) {
    case locWorkDir: End=Buffer; break;
    case locBinaries: End=GetBinaryDirectory(Buffer,Remaining);break;
    case locNonexecLibraries: {
      End=GetNonexecLibraryDirectory(Buffer,Remaining);
      break;
    }
    case locHeaders: End=GetHeaderDirectory(Buffer,Remaining);break;
    default: {
      ReportBug("Unsupported location");
      End=Buffer;
    }
  }
  Remaining-=End-Buffer;
  End=StrNCpy(End,Module->Name,Remaining);
  Remaining-=End-Buffer;
  if (Ext[0]!='\0') {
    if (Remaining>0) {
      *(End++)='.';
      Remaining--;
    }
    Ptr=End;
    End=StrNCpy(Ptr,Ext,Remaining);
    Remaining-=End-Ptr;
  }
  if (Remaining==0) {
    PlainError("File specification too long");
    return(NULL);
  }
  return(End);
}

char *ExtractNameFromFile(TFileInModule *F,char *Buffer)
/*
  DESCRIPTION

  Extract the filename from the given file info record. The buffer must be
  a pointer to the memory block of at least MAX_FILESPEC_LENGTH+1 bytes.
*/
{
  char *Result;

  Result=ConstructFileName(F->Module,F->Ext->Ext,Buffer);
  if (Result!=NULL) Buffer[MAX_FILESPEC_LENGTH]='\0';
  return(Result);
}

static bool OpenSourceFile(TModule *Module,const char *Ext,uint16 Flags)
{
  TFileSpec Buffer;
  bool Result;
  TBufferedFile F;

  if (ConstructFileName(Module,Ext,Buffer)) {
    if ((Flags&OF_CHECKONLY) != 0) {
      Result=OpenFile(&F,Buffer,F_READ);
      if (Result) CloseFile(&F);
    } else {
      if ((Flags&OF_WRITABLE) != 0) {
        Result=AllocateAndOpenInOutFile(Buffer);
        if (Result) {
          MapInOutFileRead();
          MapInOutFileWrite();
        }
      } else if ((Flags&OF_OUTPUT) != 0) {
        Result=AllocateAndCreateOutputFileIfNonex(Buffer);
      } else {
        Result=AllocateAndOpenInputFile(Buffer);
      }
      if (!Result) IOResult();
      SetupFileName(Buffer);
    }
    return(Result);
  } else {
    return(false);
  }
}

static uint16 LastOpenFileFlags;

bool OpenFileInModule(TFileInModule *F,uint16 Flags)
{
  LastOpenFileFlags=Flags;
  return(OpenSourceFile(F->Module,F->Ext->Ext,Flags));
}

void CloseFileInModule(void)
{
  if ((LastOpenFileFlags&OF_WRITABLE) != 0) {
    CloseAndFreeInOutFile();
  } else {
    CloseAndFreeInputFile();
  }
}

bool DeleteFileInModule(TFileInModule *F)
{
  TFileSpec Buffer;

  if (!ExtractNameFromFile(F,Buffer)) return(false);
  return(Remove(Buffer));
}

const char *TypeToStr(TFileType Type)
{
  switch (Type) {
    default: {
      case ftUnknown:return("unknown");
      case ftSimpleSource:return("simple source");
      case ftHeader:return("header");
      case ftSource:return("source");
      case ftCompiledModule:return("compiled module");
      case ftBinary:return("binary");
    }
  }
}

static void SearchForAssociatedFile(TFileInModule *F)
{
  TFileInModule *Other;

  if (F->Ext->Assoc1!=NULL) {
    if (OpenSourceFile(F->Module,F->Ext->Assoc1->Ext,true)) {
      Other=RegisterFileSplitted(F->Module->Name,F->Ext->Assoc1->Ext);
      if (Other==NULL) TokError("No memory for associated file");
    }
  }
}

void PrintFileName(TFileInModule *F)
{
  Print(F->Module->Name);
  if (F->Ext->Ext[0]!='\0') {
    Print(".");
    Print(F->Ext->Ext);
  }
}

void PutFileArg(TFileInModule *F)
{
  TFileSpec Buffer;
  ExtractNameFromFile(F,Buffer);
  PutArg(Buffer);
}

bool CheckFile(TFileInModule *F,TFileType Type)
{
  if (F!=NULL) {
    if (F->Ext->Type==ftUnknown) {
      TokError("File with unknown extension\1");
      PutArg(F->Ext->Ext);
    } else if (F->Ext->Type!=Type) {
      TokError("File is\1and expected is\2");
      PutArg(TypeToStr(F->Ext->Type));
      PutArg(TypeToStr(Type));
    } else if (!OpenFileInModule(F,OF_CHECKONLY)) {
      TokError("File\1not found");
      PutFileArg(F);
    } else {
      SearchForAssociatedFile(F);
      return(true);
    }
  }
  return(false);
}

/*** List of pending files ***/
/*
  See description of PrevPending and NextPending in filetype.h
*/

static opaqueTFileInModule *FirstPending;

#define NO_PENDING ((void *)(&FirstPending))

void InitPendingList(void)
{
  FirstPending=NO_PENDING;
}

void AddPendingFile(TFileInModule *File)
/*
  Adds a file into the list of pending files.
*/
{
  opaqueTFileInModule *oFile=(void *)File;
  bool Result;

  assert(oFile!=NULL);
  assert(oFile->P.Ext->Type==ftHeader);
  assert(FirstPending!=NULL);
  Result=(oFile->NextPending==NULL);
  if (Result) {
    oFile->NextPending=FirstPending;
    oFile->PrevPending=NO_PENDING;
    FirstPending->PrevPending=oFile;
    FirstPending=oFile;
  }
}

static void RemovePendingFile(TFileInModule *File)
{
  opaqueTFileInModule *oFile=(void *)File;

  if (oFile->NextPending==NULL) return;
  if (oFile->NextPending!=NO_PENDING) {
    oFile->NextPending->PrevPending=oFile->PrevPending;
  }
  if (oFile->PrevPending!=NO_PENDING) {
    oFile->PrevPending->NextPending=oFile->NextPending;
  }
  oFile->PrevPending=NULL;
  oFile->NextPending=NULL;
}

void PushPendingFiles(void)
{
  opaqueTFileInModule *F;
  opaqueTFileInModule *OldF;

  F=FirstPending;
  while (F!=NO_PENDING) {
    assert(F->P.Ext->Type==ftHeader);
    OldF=F;
    F=F->NextPending;
    PushFileToBeProcessed((void *)OldF);
  }
}

/*
  Routines handling the processing stack of files. This stack is special,
  because if a pending file is being placed onto it, the file must be removed
  from the list of pending files. The standard stack handling routines don't
  handle this peculiarity.
*/

void InitStackOfFilesToBeProcessed(void)
{
  InitStackOfFiles(stkFileProcessing);
}

bool PushFileToBeProcessed(TFileInModule *File)
{
  if (File!=NULL) RemovePendingFile(File);
  return(PushFileOntoStack(stkFileProcessing,File));
}

void PushAllFilesInList(TArray *L)
{
  uint16 i;

  for (i=0;i<L->Cnt;i++) {
    PushFileToBeProcessed(L->Arr[i]);
  }
}

void PushSourceFileOfHeaderIfExists(TFileInModule *F)
{
  TFileInModule *NewF;

  if (F->Ext->Assoc1==NULL) return;
  if (F->Module->Location==locWorkDir && F->Ext->Type==ftHeader) {
    NewF=RegisterFileSplitted(F->Module->Name,F->Ext->Assoc1->Ext);
    F->Assoc1=NewF;
    if (NewF!=NULL) {
      if (NewF->TimeStamp!=0) PushFileToBeProcessed(NewF);
    }
  }
}

TFileInModule *PopFileToBeProcessed(void)
{
  return(PopFileOffStack(stkFileProcessing));
}

void CleanupStackOfFilesToBeProcessedAfterUse(void)
{
  CleanupFileStack(stkFileProcessing);
}
