#include <syslib/hhead.h>
#ifndef __h__OBS_FILES__
#define __h__OBS_FILES__

#include <libhs/dictrb.h>
#include <libhs/exparr.h>

#include "config.h"
#include "fileext.h"

typedef char TFileSpec[MAX_FILESPEC_LENGTH+1];
typedef char TFileName[MAX_FILENAME_LENGTH+1];

typedef enum {
  locBinaries=1,
  locNonexecLibraries,
  locWorkDir,
  locHeaders,
  locUnknown
} TLocation;

#define LAST_BINARY_LOC locNonexecLibraries

typedef enum {
  statMayBeOutdated,
  statOutdated,
  statPending,
  statUpToDate
} TFileBuildStatus;

/*** Informations about a module ***/
typedef struct tagTModule {
  /*
    Location of the file.
  */
  TLocation Location;
  /*
    The name of the module.
  */
  const char *Name;
  /*
    The dictionary of files belonging to the module
  */
  TRBDict *Files;
} TModule;

/*** Informations about a file ***/
typedef struct tagTFileInModule {
  /*
    The module this file belongs to. It defines (among other things) the name
    of the file.
  */
  TModule *Module;
  /*
    The extension of the file.
  */
  TFileExtension *Ext;
  /*
    The timestamp of the file (0 if the file does not exist).
  */
  uint32 TimeStamp;
  /*
    The old timestamp of the file (0 if no such timestamp present). This
    timestamp is loaded from the file dependency tree stored in a dependency
    file and is used to check whether a file is up to date.
  */
  uint32 OldTimeStamp;
  /*
    Maximal timestamp. This attribute will hold the largest timestamp among
    the file itself and the files that are its dependences. It is filled up
    by the TraverseHeadersReferencedByFile() call and is used to determine
    which generated files are out of date.
  */
  uint32 MaxTimeStamp;
  /*
    The list of files this one refers to.
  */
  TArray References;
  /*
    The list of libraries this file refers to.
  */
  TArray Libraries;
  /*
    The associated files. A header has the pointer to the source file if any.
    The source code files have here pointers to their respective object
    files.
  */
  struct tagTFileInModule *Assoc1;
  /*
    The owner of the file (the binary this file gets compiled and linked
    into).
  */
  struct tagTFileInModule *Owner;
  /*
    True if another package item can use the file.
  */
  bool Shared;
  /*
    True if the library has been listed for linking. This is used when
    building the list of libraries in the arguments of linker.
  */
  bool Listed;
  /*
    The build status of the file
  */
  TFileBuildStatus Status;
  /*
    The ID assigned to the file by the file enumerator. These IDs are used
    to compress the dependency tree file. 65535 means the file has no ID.
  */
  uint16 ID;
  /*
    Location of the first import of the file. When a file is imported more
    than once, this location is reported as the place where the file was
    already imported.
  */
  uint16 ImportLine;
  uint16 ImportColumn;
} TFileInModule;

typedef struct {
  TFileInModule *Hdr;
  TFileInModule *Src;
  TFileInModule *Obj;
} TModuleFiles;

void InitFiles(void);
void DestroyFiles(void);
TFileInModule *RegisterFileSplittedLoc(
  TLocation Loc,
  const char *Name,
  const char *Ext
);
TFileInModule *RegisterFileSplitted(const char *Name, const char *Ext);
TFileInModule *RegisterFileLoc(TLocation Loc,const char *FileName);
TFileInModule *RegisterFile(const char *FileName);
void UpdateTimeStamp(TFileInModule *F);
TModule *GetModuleOfFile(char *FileName);

#define OF_CHECKONLY 1
#define OF_WRITABLE 2
#define OF_OUTPUT 4

char *ExtractNameFromFile(TFileInModule *F,char *Buffer);
bool DeleteFileInModule(TFileInModule *F);
void GetFilesInModule(TFileInModule *F,TModuleFiles *M);
bool OpenFileInModule(TFileInModule *F,uint16 Flags);
void CloseFileInModule(void);
const char *TypeToStr(TFileType Type);
bool CheckFile(TFileInModule *F,TFileType Type);

void InitStackOfFilesToBeProcessed(void);
bool PushFileToBeProcessed(TFileInModule *File);
void PushAllFilesInList(TArray *L);
void PushSourceFileOfHeaderIfExists(TFileInModule *F);
TFileInModule *PopFileToBeProcessed(void);
void CleanupStackOfFilesToBeProcessedAfterUse(void);

#include "filestk.h"

void InitPendingList(void);
void AddPendingFile(TFileInModule *File);
void PushPendingFiles(void);

void PrintFileName(TFileInModule *F);
void PutFileArg(TFileInModule *F);

#endif
#include <syslib/hfoot.h>
