#include <syslib/types.h>
#define __INTERFACE__
#include "filespec.h"

#include <libhs/strings.h>

#include "config.h"
#include "errors.h"

static bool CheckDirectoryNameValid(const char *Name,bool IsDir)
{
  int16 Len;
  char Ch;

  Len=0;
  do {
    Ch=*(Name++);
    if (Ch!='\0') {
      Len++;
      if (Len>MAX_FILENAME_LENGTH) {
        if (IsDir) {
          TokError("Directory name too long");
        } else {
          TokError("Name too long");
        }
        return(false);
      }
      if (Ch>='a' && Ch<='z') continue;
      if (Ch<'0' || Ch>'9') break;
    }
  } while (Ch!='\0');
  if (Len==0) {
    if (IsDir) {
      TokError("Empty directory name in the path");
    } else {
      TokError("Name is empty");
    }
  }
  if (Ch>='A' && Ch<='Z') {
    if (IsDir) {
      TokError("Uppercase characters in directory name not allowed");
    } else {
      TokError("Uppercase characters in name not allowed");
    }
    return(false);
  }
  if (Ch!='\0') {
    if (IsDir) {
      TokError("Invalid character in directory name");
    } else {
      TokError("Invalid character in name");
    }
    return(false);
  }
  return(true);
}

bool CheckNameValid(const char *Name)
{
  if (Name[0]>='0' && Name[0]<='9') {
    TokError("Name must begin with a letter");
    return(false);
  }
  return(CheckDirectoryNameValid(Name,false));
}

const char *FindModuleBaseName(const char *Name)
{
  const char *Result;
  char Ch;
  Result=Name;
  do {
    Ch=*(Name++);
    if (Ch=='/') Result=Name;
  } while (Ch!='\0');
  return(Result);
}

static int16 FindSlash(const char *FileSpec)
{
  const char *Ptr;
  char Ch;
  Ptr=FileSpec;
  do {
    Ch=*Ptr;
    if (Ch=='/' || Ch=='\0') return(Ptr-FileSpec);
    Ptr++;
  } while (true);
}

static char *DeleteLastComponent(char *Buffer,char *BufPtr)
{
  char Ch;

  while (BufPtr>Buffer) {
    Ch=*(--BufPtr);
    if (Ch=='/') return(BufPtr);
  }
  return(Buffer);
}

/*** Check and canonize a module file name ***/
/*
  "FileName" points to a file name of a module file with a path but without
  an extension. The "Buffer" points to a memory buffer at least
  MAX_FILESPEC_LENGTH long.

  This call canonizes the provided evaulates the "." and ".." subdirectory
  specifiers and removes them along with any path components that are not
  parts of the final file specification due to a ".." specified later in
  the path. For example "foo/../bar/./baz/../../quuz/./file" will become
  "/quuz/file".

  It also checks whether the file name and its path adhere to the limitations
  imposed upon them. Currently all names must be at least 8 character long
  and may contain only letters and digits. An additional limitation is that
  the file name must start with a letter (this is because the file name is
  used as an identifier and identifiers cannot begin with a digit). If any
  of these limitations is violated or if the resulting path is longer than
  MAX_FILESPEC_LENGTH, the function returns NULL and the content of the
  output buffer is undefined.

  The "Buffer" and "FileName" arguments can point to the same buffer provided
  that the "FileName" points to a portion of the buffer that is located after
  the place "FileName" points to. Otherwise buffer overflow probably leading
  to a crash will occur.
*/
char *CheckAndCanonizeModuleFileName(char *Buffer,const char *FileName)
{
  bool IsDirectory;
  int16 Pos;
  char *BufPtr;
  char *Component;

  BufPtr=Buffer;
  while (FileName[0]!='\0') {
    Pos=FindSlash(FileName);
    IsDirectory=(FileName[Pos]=='/');
    if (IsDirectory) {
      if (Pos==1 && FileName[0]=='.') {
        FileName+=2;
        continue;
      } else if (Pos==2 && FileName[0]=='.' && FileName[1]=='.') {
        if (BufPtr==Buffer) {
          TokError("File specification path crosses root of the project");
          return(NULL);
        } else {
          BufPtr=DeleteLastComponent(Buffer,BufPtr-1);
          FileName+=3;
          continue;
        }
      }
    }
    if (Pos+IsDirectory>MAX_FILESPEC_LENGTH-(BufPtr-Buffer)) {
      TokError("File specification too long");
      return(NULL);
    }
    Component=BufPtr;
    BufPtr=MemCpy(BufPtr,FileName,Pos);
    *BufPtr='\0';
    if (IsDirectory) {
      if (!CheckDirectoryNameValid(Component,true)) return(NULL);
      *(BufPtr++)='/';
      FileName++;
    } else {
      if (!CheckNameValid(Component)) return(false);
    }
    FileName+=Pos;
  }
  assert(*BufPtr=='\0');
  return(BufPtr);
}
