#include <syslib/types.h>
#define __INTERFACE__
#include "filestk.h"

#include "filetype.h"

/*
  Generic file stacks

  These file stacks are used to traverse the file dependency tree. The stack
  handling routinnes ensure that no file will be pushed twice in one stack
  session. The stack session is started by InitStackOfFiles() and ended by
  CleanupFileStack().
*/

typedef struct tagTStackInfo {
  opaqueTFileInModule *FirstFile;
  opaqueTFileInModule *UsedFiles;
} TStackInfo;

static TStackInfo Stacks[MAX_STACKS];

#define FILE_STACK_BOTTOM ((void *)&Stacks)

void InitStackOfFiles(uint8 Number)
{
  TStackInfo *S;

  assert(Number<MAX_STACKS);
  S=&Stacks[Number];
  assert(S->FirstFile==NULL);
  S->FirstFile=FILE_STACK_BOTTOM;
  S->UsedFiles=FILE_STACK_BOTTOM;
}

bool PushFileOntoStack(uint8 Number,TFileInModule *File)
/*
  Pushes a file into the stack. Returns "true" if the file was pushed or
  "false" if the file has already entered the stack after last
  (re)initialization.
*/
{
  opaqueTFileInModule *oFile=(void *)File;
  TStackInfo *S;
  bool Result;

  assert(Number<MAX_STACKS);
  S=&Stacks[Number];
  assert(S->FirstFile!=NULL);
  if (oFile==NULL) return(false);
  Result=(oFile->StackNext[Number]==NULL);
  if (Result) {
    oFile->StackNext[Number]=S->FirstFile;
    S->FirstFile=oFile;
  }
  return(Result);
}

TFileInModule *PopFileOffStack(uint8 Number)
{
  opaqueTFileInModule *File;
  TStackInfo *S;

  assert(Number<MAX_STACKS);
  S=&Stacks[Number];
  assert(S->FirstFile!=NULL);
  File=S->FirstFile;
  if (File==FILE_STACK_BOTTOM) return(NULL);
  S->FirstFile=File->StackNext[Number];
  File->StackNext[Number]=S->UsedFiles;
  S->UsedFiles=File;
  return((void *)File);
}

static void Cleanup(opaqueTFileInModule *File,uint8 Number)
{
  opaqueTFileInModule *NextFile;

  while (File!=FILE_STACK_BOTTOM) {
    NextFile=File->StackNext[Number];
    File->StackNext[Number]=NULL;
    File=NextFile;
  }
}

void CleanupFileStack(uint8 Number)
/*
  Reinitialize the stack after using. This should be done as soon as the
  stack usage is terminated to avoid strange interactions with other parts
  of the system. The stack may be not completely empty; this can happen when
  errors are detected during rebuild.
*/
{
  TStackInfo *S;

  assert(Number<MAX_STACKS);
  S=&Stacks[Number];
  Cleanup(S->FirstFile,Number);
  Cleanup(S->UsedFiles,Number);
  if (__DEBUG__) {
    S->FirstFile=NULL;
    S->UsedFiles=NULL;
  }
}
