#include <syslib/types.h>
#define __INTERFACE__
#include "pkgdeps.h"

#include <libhs/strings.h>
#include <libhs/diskio.h>
#include <libhs/filebuf.h>
#include <libhs/exparr.h>
#include <libhs/memory.h>

#include "errors.h"
#include "filestk.h"
#include "files.h"

/**************************
*** The file enumerator ***
**************************/

#define MAX_ENUMERATED_FILES (124*256)
#define ENUMERATED_FILES_IN_BLOCK 7936

#define BLOCK_CNT (MAX_ENUMERATED_FILES/ENUMERATED_FILES_IN_BLOCK)

static bool UseTheTable;
static TFileInModule **Files[BLOCK_CNT];
static int16 FileCnt;

static void InitializeFilesTable(bool TableInUse)
{
  uint16 i;

  UseTheTable=TableInUse;
  InitStackOfFiles(stkIdentifiedFiles);
  FileCnt=0;
  for (i=0;i<BLOCK_CNT;i++) {
    Files[i]=NULL;
  }
}

/*
  This procedure tries to add the file into the table of identified files.
  If successfull, it places the ID of the file into the apropiate component
  of the structure. The file must have no ID assigned.
  If the [UseTheTable] variable is set to false, the file is not added into
  the table, only ID is assigned to it.
*/
static void AddFileIntoTable(TFileInModule *F)
{
  int16 BlockNum;
  int16 FileNum;
  TFileInModule **Block;

  if (F->ID!=65535) return;
  if (FileCnt==MAX_ENUMERATED_FILES) return;
  if (UseTheTable) {
    BlockNum=FileCnt/ENUMERATED_FILES_IN_BLOCK;
    FileNum=FileCnt%ENUMERATED_FILES_IN_BLOCK;
    Block=Files[BlockNum];
    if (Block==NULL) {
      Block=GetMem(ENUMERATED_FILES_IN_BLOCK*sizeof(*Block));
      if (Block==NULL) {
        PlainError("Memory allocation problem during file ID creation");
        return;
      }
      Files[BlockNum]=Block;
    }
    Block[FileNum]=F;
  }
  F->ID=FileCnt;
  FileCnt++;
  PushFileOntoStack(stkIdentifiedFiles,F);
}

static void CleanupFilesTable(void)
{
  void *Block;
  uint16 i;
  TFileInModule *F;

  for (i=0;i<BLOCK_CNT;i++) {
    Block=Files[i];
    if (Block!=NULL) FreeMem(Block);
  }
  while (true) {
    F=PopFileOffStack(stkIdentifiedFiles);
    if (F==NULL) break;
    F->ID=65535;
  }
  CleanupFileStack(stkIdentifiedFiles);
}

/****************************************
*** The dependency file info handling ***
****************************************/

static void BuildDependencyTreeFileSpec(TPackage *P,char *Result)
{
  char *Ptr;

  Ptr=StrCpy(Result,P->Name);
  Ptr=StrCpy(Ptr,".dep");
}

static bool AllocateAndOpenDependencyTreeFile(TPackage *P)
{
  TFileSpec Buffer;
  bool Result;

  BuildDependencyTreeFileSpec(P,Buffer);
  Result=AllocateAndOpenInputFile(Buffer);
  if (Result) {
    PushFileName();
    SetupFileName(Buffer);
  } else {
    IOResult();
  }
  return(Result);
}

static TFileInModule *FetchFile(char *EndChar)
{
  TFileInModule *F;
  TFileInModule **Block;
  TLocation Location;
  TFileSpec Buffer;
  char *Ptr;
  char Ch;
  uint16 FileID;

  Location=ReadByte();
  if (Location>7) {
    FileID=((Location-8)%124)*256+ReadByte();
    if (FileID>=FileCnt) {
      PlainError("File with an invalid ID");
      return(NULL);
    }
    *EndChar=' ';
    if (Location>=124+8) *EndChar='\007';
    Block=Files[FileID/ENUMERATED_FILES_IN_BLOCK];
    assert(Block!=NULL);
    F=Block[FileID%ENUMERATED_FILES_IN_BLOCK];
    assert(F!=NULL);
    return(F);
  }
  if (Location==0 || Location>=locUnknown) {
    PlainError("File with an invalid location");
    return(NULL);
  }
  Ptr=Buffer;
  do {
    if (Ptr-Buffer>=MAX_FILESPEC_LENGTH) {
      PlainError("File specification too long");
      return(NULL);
    }
    Ch=ReadByte();
    *(Ptr++)=Ch;
  } while (Ch!='\0' && Ch!=' ' && Ch!='\007');
  *(Ptr-1)='\0';
  F=RegisterFileLoc(Location,Buffer);
  if (F!=NULL) {
    AddFileIntoTable(F);
    *EndChar=Ch;
  } else {
    PlainError("Error registering\1");
    PutArg(Buffer);
  }
  return(F);
}

static bool ReadDepTreeContent(void)
{
  bool Good;
  uint32 Size;
  uint32 Pos;
  uint32 TimeStamp;
  TFileInModule *File;
  TFileInModule *Other;
  char EndChar;
  TArrayHandle H;
  bool Referencies;

  Good=true;
  Size=GetInputFileSize();
  do {
    Pos=GetInputFilePos();
    if (Pos-Size<6) break;
    TimeStamp=Read32LE();
    File=FetchFile(&EndChar);
    if (File==NULL) return(false);
    Referencies=true;
    File->OldTimeStamp=TimeStamp;
    OpenArrayEx(&File->References,&H);
    while (EndChar==' ') {
      Other=FetchFile(&EndChar);
      if (Other==NULL) return(false);
      if (Other->Ext->Type==ftBinary && Referencies) {
        CloseArray(&H);
        OpenArrayEx(&File->Libraries,&H);
        Referencies=false;
      }
      if (!PutItem(&H,Other)) {
        PlainError("Error recording\1as a dependency of\2");
        PutFileArg(Other);
        PutFileArg(File);
        Good=false;
      }
    }
    CloseArray(&H);
  } while (true);
  if (Pos<Size) {
    PlainError("Garbage at the end of dependency tree file ignored");
    Good=false;
  }
  return(Good);
}

void ReadDepTree(TPackage *P)
/*
  Read the dependency lists. It simply reads the names of the files in the
  file with the saved dependency tree and assembles the file information
  records and the dependency lists in them.

  The format of the file is very simple. The file is divided into rows. The
  first 4 bytes is the timestamp of the file, in little endian format.
  Immediately after it there is a list of space-separated filenames. Each
  filename contains a TLocation code as its first byte and the filename in
  the remaining bytes. The first filename in this list identifies the file to
  which the timestamp belongs. The remaining names are the dependency files
  of that one the timestamp belongs to.

  When an open failure occurs, the procedure silently exits and reads
  nothing. This is because in most cases the open failure is due to the fact
  that the file does not exist. However when a read failure occurs, the
  IOResult variable in the FILEBUF module will be set and is inspected and
  reported to the user.
*/
{
  if (AllocateAndOpenDependencyTreeFile(P)) {
    InitializeFilesTable(true);
    if (!ReadDepTreeContent()) {
      PlainError("Recreating the dependency tree file");
      DestroyFiles();
      DestroyExtensions();
      InitExtensions();
      InitFiles();
    }
    if (!CloseAndFreeInputFile()) {
      CheckIOResultAndReport("Error closing dependency tree file");
    }
    CleanupFilesTable();
    PopFileName();
  }
}



typedef enum {
  bsEmpty,
  bsFileOnly,
  bsFileAndCharacter
} TBufferStatus;

static TBufferStatus BufferStatus;
static TFileInModule *PreparedFile;
static char Character;

static void InitializeBuffer(void)
{
  BufferStatus=bsEmpty;
}

static void FlushBuffer(void)
{
  uint16 ID;
  TFileInModule *F;

  if (BufferStatus==bsEmpty) return;
  F=PreparedFile;
  assert(BufferStatus==bsFileAndCharacter);
  assert(Character=='\007' || Character==' ');
  if (F->ID==65535) {
    WriteByte(F->Module->Location);
    WriteText(F->Module->Name);
    WriteByte('.');
    WriteText(F->Ext->Ext);
    WriteByte(Character);
    AddFileIntoTable(F);
  } else {
    assert(F->ID<MAX_ENUMERATED_FILES);
    assert(F->ID<FileCnt);
    ID=F->ID+8*256;
    if (Character=='\007') ID+=MAX_ENUMERATED_FILES;
    WriteByte(ID/256);
    WriteByte(ID%256);
  }
  BufferStatus=bsEmpty;
}

static void SendFileName(TFileInModule *F)
{
  assert(BufferStatus==bsEmpty || BufferStatus==bsFileAndCharacter);
  FlushBuffer();
  BufferStatus=bsFileOnly;
  PreparedFile=F;
}

static void SendByte(char Byte)
{
  if (BufferStatus==bsFileOnly) {
    assert(Byte=='\007' || Byte==' ');
    BufferStatus=bsFileAndCharacter;
    Character=Byte;
  } else {
    FlushBuffer();
    WriteByte(Byte);
  }
}

static void SendLongInt(uint32 LongInt)
{
  assert(BufferStatus==bsEmpty || BufferStatus==bsFileAndCharacter);
  FlushBuffer();
  Write32LE(LongInt);
}

static void SendFileDeps(TFileInModule *F)
{
  uint16 i;
  TFileInModule *DF;

  SendLongInt(F->OldTimeStamp);
  SendFileName(F);
  for (i=0;i<F->References.Cnt;i++) {
    SendByte(' ');
    DF=F->References.Arr[i];
    assert(DF->Ext->Type==ftHeader);
    SendFileName(DF);
    PushFileToBeProcessed(DF);
  }
  for (i=0;i<F->Libraries.Cnt;i++) {
    SendByte(' ');
    DF=F->Libraries.Arr[i];
    assert(DF->Module->Location<=LAST_BINARY_LOC);
    assert(DF->Ext->Type==ftBinary);
    SendFileName(DF);
  }
  SendByte('\007');
}

static void SendFilesInTheStack(void)
{
  TFileInModule *F;
  TFileInModule *AssocF;

  do {
    F=PopFileToBeProcessed();
    if (F!=NULL) {
      SendFileDeps(F);
      AssocF=F->Assoc1;
      if (AssocF!=NULL && AssocF->TimeStamp>0) {
        PushFileToBeProcessed(F->Assoc1);
      }
    }
  } while (F!=NULL);
}

static void PushPackageItemDeps(TPackageItem *I)
{
  PushFileToBeProcessed(I->Output);
  PushAllFilesInList(&I->Sources);
  PushAllFilesInList(&I->Headers);
}

static void WriteDepTreeCore(TPackage *P)
{
  uint16 i;

  InitializeBuffer();
  InitializeFilesTable(false);
  InitStackOfFilesToBeProcessed();
  for (i=0;i<P->Items.Cnt;i++) {
    PushPackageItemDeps(P->Items.Arr[i]);
  }
  SendFilesInTheStack();
  assert(BufferStatus==bsEmpty || BufferStatus==bsFileAndCharacter);
  FlushBuffer();
  CleanupFilesTable();
  CleanupStackOfFilesToBeProcessedAfterUse();
}

static bool AllocateAndCreateDependencyTreeFile(TPackage *P)
{
  TFileSpec Buffer;
  bool Result;

  BuildDependencyTreeFileSpec(P,Buffer);
  Result=AllocateAndCreateOutputFile(Buffer);
  CheckIOResultAndReport("Error creating dependency tree file");
  return(Result);
}

void WriteDepTree(TPackage *P)
{
  if (!AllocateAndCreateDependencyTreeFile(P)) return;
  WriteDepTreeCore(P);
  CheckIOResultAndReport("Error writing dependency tree file");
  CloseAndFreeOutputFile();
  CheckIOResultAndReport("Error closing dependency tree file");
}

void EraseDepTreeFile(TPackage *P,bool Verbose)
{
  TFileSpec Buffer;

  BuildDependencyTreeFileSpec(P,Buffer);
  if (!Remove(Buffer)) {
    if (Verbose) SysError("Error erasing dependency tree file");
  }
}
