#include <syslib/types.h>
#define __INTERFACE__
#include "pkginst.h"

#include <syslib/kernel.h>
#include <syslib/pdfsys.h>

#include <libhs/strings.h>
#include <libhs/output.h>
#include <libhs/libman.h>
#include <libhs/diskio.h>
#include <libhs/filebuf.h>
#include <libhs/exec.h>
#include <libhs/errno.h>

#include "files.h"
#include "mkidir.h"
#include "errors.h"

#define ERASE_BUFFER_SIZE 8192
#define ERASE_FILES_COUNT 512

static void EraseFilesInADirectory(const char *DirSpec)
{
  int16 Dir;
  TDirItem I;
  char Buffer[2*(MAX_FILESPEC_LENGTH+1)];
  char *Ptr;
  bool Result;
  uint16 Len;
  bool Terminated;
  char EraseBuffer[ERASE_BUFFER_SIZE];
  uint16 EBufSize;
  char *EBufPtr;
  char *ErasedFiles[ERASE_FILES_COUNT];
  uint16 EFileCnt;

  /*
    Prepare the directory specification into the buffer
  */
  Buffer[sizeof(Buffer)-1]='\0';
  Ptr=StrNCpy(Buffer,DirSpec,sizeof(Buffer));
  assert(Ptr>Buffer);
  assert(*(Ptr-1)!='\\' && *(Ptr-1)!='/');
  /*
    Check whether the directory specification is short enough to fit into
    our buffer.
  */
  if (Buffer[sizeof(Buffer)-1]!='\0') {
    PlainError("Dir spec too long:\1");
    PutArg(DirSpec);
    return;
  }
  /*
    Place the path delimiter at the end of the directory specification. In
    the main eraser loop the files to be erased will be placed after this
    directory specification delimiter to form their absolute specifications
    to be passed to the Remove() call.
  */
  *(Ptr++)='/';
  /*
    Erasing files from the directory is somewhat complicated, because it is
    not safe to read a directory and at the same time change it by erasing
    files from it. Therefore we open the directory, read as many items as
    possible, close the directory, erasee the items fetched and repeat until
    we erase everything we wanted to erase.
  */
  do {
    /*
      (Re)initialize the buffers.
    */
    EBufSize=0;
    EFileCnt=0;
    /*
      Read the directory and store as many files to be erased into the
      buffers as possible.
    */
    Dir=OpenDir(DirSpec);
    if (Dir<0) {
      SysError("Unable to open directory\1");
      PutArg(DirSpec);
      break;
    }
    while (true) {
      Result=ReadItem(Dir,&I);
      Terminated=true;
      if (!Result) break;
      if (I.ID==-1) break;
      Terminated=false;
      Len=StrLen(I.Name)+1;
      if (EFileCnt==ERASE_FILES_COUNT || EBufSize+Len>ERASE_BUFFER_SIZE) {
        ErasedFiles[EFileCnt++]=I.Name;
        break;
      }
      EBufPtr=EraseBuffer+EBufSize;
      ErasedFiles[EFileCnt++]=EBufPtr;
      MemCpy(EBufPtr,I.Name,Len);
      EBufSize+=Len;
    }
    Close(Dir);
    /*
      Now erase files which we fetched.
    */
    while (EFileCnt>0) {
      EBufPtr=ErasedFiles[--EFileCnt];
      StrNCpy(Ptr,EBufPtr,sizeof(Buffer)-(Ptr-Buffer));
      if (Buffer[sizeof(Buffer)-1]!='\0') {
        Buffer[sizeof(Buffer)-1]='\0';
        PlainError("File spec name too long:\1");
        PutArg(Buffer);
        Terminated=true;
        break;
      }
      if (!Remove(Buffer)) {
        SysError("Unable to remove\1");
        PutArg(Buffer);
        Terminated=true;
        break;
      }
    }
    /*
      The "Terminated" variable will be set to "true" when there is nothing
      to be deleted anymore.
    */
  } while (!Terminated);
  /*
    If we get here, we completed the task without error.
  */
}

static bool InstallFileCore(TFileInModule *F,const char *DestFile)
{
  upint8 Buffer[16384];
  bool Result;
  uint32 FileSize;
  uint16 SizeToTransfer;

  if (!OpenFileInModule(F,0)) return(false);
  if (!Remove(DestFile)) {
    if (GetErrNo()!=fseNOENT) return(false);
  }
  Result=AllocateAndCreateOutputFile(DestFile);
  if (Result) {
    FileSize=GetInputFileSize();
    while (FileSize>0) {
      SizeToTransfer=sizeof(Buffer);
      if (SizeToTransfer>FileSize) SizeToTransfer=FileSize;
      if (!ReadBlock(Buffer,SizeToTransfer)) {
        Result=false;
        break;
      }
      if (!WriteBlock(Buffer,SizeToTransfer)) {
        Result=false;
        break;
      }
      FileSize-=SizeToTransfer;
    }
    CloseAndFreeOutputFile();
  }
  CloseFileInModule();
  return(Result);
}

static void InstallFile(TFileInModule *F,const char *DestFile)
{
  if (!InstallFileCore(F,DestFile)) {
    SysError("Unable to install\1");
    PutFileArg(F);
  }
}

static char *ExtractLibraryName(TFileInModule *F,char *Buffer)
{
  char *Ptr;

  Ptr=StrCpy(Buffer,F->Module->Name);
  return(Ptr);
}

static void IssueLibraryLoadOnFile(TFileInModule *F)
{
  TFileSpec Buffer;

  ExtractNameFromFile(F,Buffer);
  if (!Load(Buffer)) {
    SysError("Unable to load library\1");
    PutArg(Buffer);
  }
}

static void InstallPackageItem(TPackageItem *I)
{
  char Buffer[2*(MAX_FILESPEC_LENGTH+1)];
  char *Ptr;
  uint16 i;
  TFileInModule *F;

  Print("Installing ");
  PrintFileName(I->Output);
  NewLine();
  if (I->Sources.Cnt>0) {
    Ptr=GetBinaryDirectory(Buffer,sizeof(Buffer));
  } else {
    Ptr=GetNonexecLibraryDirectory(Buffer,sizeof(Buffer));
  }
  ExtractNameFromFile(I->Output,Ptr);
  InstallFile(I->Output,Buffer);
  if (I->Headers.Cnt>0) {
    IssueLibraryLoadOnFile(I->Output);
    Ptr=GetHeaderDirectory(Buffer,sizeof(Buffer));
    Ptr=ExtractLibraryName(I->Output,Ptr);
    if (CreateDirectoryForInstall(Buffer)!=0) return;
    EraseFilesInADirectory(Buffer);
    *(Ptr++)='/';
    for (i=0;i<I->Headers.Cnt;i++) {
      F=I->Headers.Arr[i];
      ExtractNameFromFile(F,Ptr);
      InstallFile(F,Buffer);
    }
  }
}

void InstallPackage(TPackage *P)
{
  TPackageItem *I;
  uint16 i;

  for (i=0;i<P->Items.Cnt;i++) {
    I=P->Items.Arr[i];
    if ((I->Type&M_IS_PRIVATE)==0) InstallPackageItem(I);
  }
}
