#include <syslib/types.h>
#define __INTERFACE__
#include "pkgscan.h"

#include <libhs/strings.h>
#include <libhs/filebuf.h>
#include <libhs/memory.h>
#include <libhs/exparr.h>
#include <libhs/expbuf.h>
#include <libhs/output.h>
#include <libhs/exec.h>

#include "args.h"
#include "errors.h"
#include "files.h"
#include "scanbase.h"
#include "scanmain.h"
#include "pkgbuild.h"
#include "filelib.h"

static TFileInModule *CurrentBinary;

static void PrepareBinaryForScan(TFileInModule *Binary)
{
  assert(Binary!=NULL);
  assert(CurrentBinary==NULL);
  CurrentBinary=Binary;
}

static TFileInModule *GetCurrentBinary(void)
{
  assert(CurrentBinary!=NULL);
  return(CurrentBinary);
}

static void FinalizeBinaryAfterScan(void)
{
  assert(CurrentBinary!=NULL);
  if (__DEBUG__) CurrentBinary=NULL;
}

static bool FileNeedsRescan(TFileInModule *F)
{
  assert(F!=NULL);
  if (F->Module->Location!=locHeaders) return(true);
  return(MemCmp(F->Module->Name,"syslib",6)!=0);
}

typedef struct {
  TFileInModule *File;
  uint16 Pos;
} TTraversalStackItem;

typedef enum {
  ssVisiting,
  ssGoingDown,
  ssGoingUp
} TStackStatus;

static void CheckFileReference(TFileInModule *From,TFileInModule *To)
{
  TFileInModule *Bin;

  Bin=GetCurrentBinary();
  if (To->Owner==NULL) {
    To->Owner=Bin;
    if (From->Owner!=NULL) To->Owner=From->Owner;
  } else {
    if (To->Owner!=Bin) {
      if (!To->Shared) {
        PlainError("Illegall reference of\1in\2");
        PutFileArg(To);
        PutFileArg(From);
      }
    }
  }
}

static TRebuildStatus TraverseHeadersReferencedByFileCore(TFileInModule *F)
{
  TTraversalStackItem Stack[MAX_HEADER_NESTING_LEVEL+1];
  uint16 Level;
  TTraversalStackItem *Cur;
  TTraversalStackItem *Prev;
  TFileInModule *File;
  TFileInModule *PrevFile;
  TStackStatus Status;
  TRebuildStatus Result;
  uint16 LastErrorCount;

  assert(F!=NULL);
  Level=1;
  Stack[0].File=F;
  Stack[0].Pos=0;
  Prev=Stack-1;
  Cur=Stack;
  Status=ssVisiting;
  if (F->Ext->Type==ftSource) PushFileOntoStack(stkFilesToLink,F);
  while (Level>0) {
    if (Status==ssVisiting) {
      Status=ssGoingUp;
      File=Cur->File;
      if (!PushFileOntoStack(stkIncludedHeaders,File)) continue;
      PushSourceFileOfHeaderIfExists(File);
      Status=ssGoingDown;
      if (File->Ext->Translator!=NULL) Status=ssGoingUp;
      if (File->Status==statMayBeOutdated) {
        if (File->Ext->Translator!=NULL) {
          if (!ClearDirectory) {
            Result=CheckAndRegenerate(File,false);
            if (Result!=rsSuccessfullRebuild) return(Result);
          }
          continue;
        }
        assert(File->MaxTimeStamp==0);
        File->MaxTimeStamp=File->TimeStamp;
        PrevFile=GetCurrentBinary();
        if (Level>1) PrevFile=Prev->File;
        CheckFileReference(PrevFile,File);
        File->Status=statUpToDate;
        if (File->TimeStamp!=File->OldTimeStamp) {
          File->Status=statOutdated;
          LastErrorCount=ErrorCount;
          if (FileNeedsRescan(File)) {
            if (DisplayRescans) {
              Print("Rescanning ");
              PrintFileName(File);
              NewLine();
            }
            ScanFile(File);
          }
          if (ErrorCount!=LastErrorCount) return(rsErrorsWhileRebuild);
          File->OldTimeStamp=File->TimeStamp;
        }
      }
    } else if (Status==ssGoingDown)  {
      if (Cur->Pos==Cur->File->References.Cnt) {
        Status=ssGoingUp;
      } else {
        if (Level>MAX_HEADER_NESTING_LEVEL) {
          PlainError("Headers nested too deeply in\1");
          PutFileArg(F);
          while (Level>0) {
            PlainError("Nested from\1");
            PutFileArg(Cur->File);
            Level--;
            Cur--;
          }
          return(rsErrorsWhileRebuild);
        }
        Level+=1;
        Cur++;
        Prev++;
        Cur->File=Prev->File->References.Arr[Prev->Pos++];
        Cur->Pos=0;
        Status=ssVisiting;
      }
    } else {
      if (Level>1) {
        if (Prev->File->MaxTimeStamp<Cur->File->MaxTimeStamp) {
          Prev->File->MaxTimeStamp=Cur->File->MaxTimeStamp;
        }
        if (Cur->File->Status==statOutdated) {
          Prev->File->Status=statOutdated;
        }
      }
      Level-=1;
      Cur--;
      Prev--;
      Status=ssGoingDown;
    }
  }
  return(rsSuccessfullRebuild);
}

static TRebuildStatus TraverseHeadersReferencedByFile(TFileInModule *F)
{
  TRebuildStatus Result;

  InitStackOfFiles(stkIncludedHeaders);
  Result=TraverseHeadersReferencedByFileCore(F);
  CleanupFileStack(stkIncludedHeaders);
  return(Result);
}

static TRebuildStatus TraverseAllFilesInList(TArray *L)
{
  uint16 i;
  TRebuildStatus Result;

  Result=rsSuccessfullRebuild;
  for (i=0;i<L->Cnt;i++) {
    Result=TraverseHeadersReferencedByFile(L->Arr[i]);
    if (Result!=rsSuccessfullRebuild) break;
  }
  return(Result);
}

static TRebuildStatus ProcessFile(TFileInModule *F)
{
  TRebuildStatus Result;

  Result=TraverseHeadersReferencedByFile(F);
  return(Result);
}

static TRebuildStatus ProcessFilesOnTheStack(void)
{
  TFileInModule *F;
  TRebuildStatus Result;

  Result=rsSuccessfullRebuild;
  do {
    F=PopFileToBeProcessed();
    if (F!=NULL) {
      Result=ProcessFile(F);
      if (Result!=rsSuccessfullRebuild) break;
    }
  } while (F!=NULL);
  return(Result);
}

static TRebuildStatus CompileAndLinkFile(TFileInModule *F)
{
  TRebuildStatus Result;

  assert(F->Ext->Type==ftSource);
  if (ClearDirectory) return(rsSuccessfullRebuild);
  Result=CheckAndBuildSource(F);
  return(Result);
}

typedef struct {
  uint16 ArgCnt;
  TBuffer ArgText;
  TBufferHandle BH;
  TBuffer LibList;
  uint32 OutputTimeStamp;
  uint32 NewestTimeStamp;
} TLinkerArguments;

static void NoMemForLinkerArgument(void)
{
  PlainError("Out of memory creating argument for linker");
}

static bool AddStringToArgs(TLinkerArguments *A,const void *Data,uint16 Size)
{
  char Tmp;
  bool NameOK;
  char ChSize;

  if (Size>255) {
    PlainError("Linker argument too long");
    return(false);
  }
  NameOK=ExpandBuffer(&A->BH,Size+2);
  if (NameOK) {
    ChSize=Size;
    NameOK=PutData(&A->BH,&ChSize,1);
    assert(NameOK);
    NameOK=PutData(&A->BH,Data,Size);
    assert(NameOK);
    Tmp='\0';
    NameOK=PutData(&A->BH,&Tmp,1);
    assert(NameOK);
  }
  if (!NameOK) {
    NoMemForLinkerArgument();
  }
  if (A->ArgCnt==65535) {
    PlainError("Too many arguments for linker");
    return(false);
  }
  A->ArgCnt++;
  return(NameOK);
}

static void PushLibrariesFromFile(TFileInModule *F)
{
  TFileInModule *L;
  uint16 i;

  for (i=0;i<F->Libraries.Cnt;i++) {
    L=F->Libraries.Arr[i];
    PushFileOntoStack(stkLibraries,L);
  }
}

static bool AddFileToArgs(TFileInModule *F,TLinkerArguments *A)
{
  TFileSpec NameBuf;
  char *NameEnd;

  NameEnd=ExtractNameFromFile(F,NameBuf);
  if (NameEnd==NULL) return(false);
  return(AddStringToArgs(A,NameBuf,NameEnd-NameBuf));
}

static char **BuildArgList(TLinkerArguments *A)
{
  char **Result;
  char *Ptr;
  uint16 i;

  Result=GetMem((A->ArgCnt+1)*sizeof(char *));
  if (Result==NULL) {
    PlainError("No memory for argument pointer table");
  } else {
    Ptr=(void *)A->ArgText.Ptr;
    Result[A->ArgCnt]=NULL;
    for (i=0;i<A->ArgCnt;i++) {
      Result[i]=Ptr+1;
      Ptr+=2+*(Ptr);
    }
  }
  return(Result);
}

static uint16 ProcessFilesInTheStack(
  TStackID StkID,
  const char *Option,
  TLinkerArguments *A
)
{
  TFileInModule *F;
  TFileInModule *FObj;
  TRebuildStatus Result;
  bool NameOK;

  Result=rsSuccessfullRebuild;
  do {
    F=PopFileOffStack(StkID);
    if (F!=NULL) {
      if (F->MaxTimeStamp>A->NewestTimeStamp) {
        A->NewestTimeStamp=F->MaxTimeStamp;
      }
      assert(F->Ext->Type==ftSource);
      Result=CompileAndLinkFile(F);
      if (Result==rsSuccessfullRebuild) {
        FObj=RegisterFileSplitted(F->Module->Name,F->Ext->Assoc1->Ext);
        if (Option!=NULL) {
          NameOK=AddStringToArgs(A,Option,StrLen(Option));
          if (!NameOK) Result=rsErrorsWhileRebuild;
          Option=NULL;
        }
        NameOK=AddFileToArgs(FObj,A);
        if (!NameOK) Result=rsErrorsWhileRebuild;
      }
      if (Result==rsSuccessfullRebuild) {
        PushLibrariesFromFile(F);
      }
    }
  } while (F!=NULL && Result==rsSuccessfullRebuild);
  return(Result);
}

static void BuildDependencyLibraryList(TLinkerArguments *A,TPackageItem *I)
{
  TFileInModule *F;
  TArrayHandle LL;
  TBufferHandle LH;
  uint16 Size;
  char Arg[7];
  char *Ptr;
  uint16 TempSize;

  CreateArrayEx(&I->Output->Libraries,&LL);
  CreateBufferEx(&A->LibList,&LH);
  while (true) {
    F=PopFileOffStack(stkLibraries);
    if (F==NULL) break;
    if (!PutItem(&LL,F)) {
      PlainError("Unable to record library\1into\2");
      PutFileArg(F);
      PutFileArg(I->Output);
    }
    Size=StrLen(F->Module->Name);
    if (!LinkStatically) PutData(&LH,F->Module->Name,Size+1);
  }
  CloseBuffer(&LH);
  CloseArray(&LL);
  if (A->LibList.Size>65535) {
    PlainError("Dependency library list too large");
    return;
  }
  TempSize=A->LibList.Size;
  if (TempSize>0) {
    Arg[6]='\0';
    Arg[0]='-';
    Arg[1]='A';
    for (Ptr=Arg+5;Ptr>Arg+1;Ptr--) {
      *Ptr=(TempSize%16)+'A';
      TempSize/=16;
    }
    if (!AddStringToArgs(A,Arg,6)) {
      PlainError("No memory for OBS area size argument");
      return;
    }
  }
}

static void PutDependencyLibraryList(TLinkerArguments *A,TFileInModule *F)
{
  uint32 Size;
  upint32 Pos[2];

  if (A->LibList.Size==0) return;
  if (!OpenFileInModule(F,OF_WRITABLE)) {
    PlainError("Unable to open\1");
    PutFileArg(F);
    return;
  }
  Size=GetInputFileSize();
  SeekInputFile(Size-8);
  ReadBlock(Pos,8);
  Size-=Pos[1]+4;
  Size-=Pos[0];
  SeekOutputFile(Size);
  WriteBlock(A->LibList.Ptr,A->LibList.Size);
  CloseFileInModule();
}

void BuildLibraryArgs(TPackageItem *I,TLinkerArguments *A)
{
  TFileInModule *F;
  TFileInModule *OF;
  const char *Ptr;
  bool Continue;
  bool AllListed;
  bool WereListed;
  uint16 i;
  TStackID StackOut;
  TStackID StackIn;
  TStackID Temp;

  /*
    Write the proper argument denoting that libraries are coming and how to
    link the binary against them (whether statically or dynamically).
  */
  Ptr="-d";
  if (LinkStatically) Ptr="-l";
  if (!AddStringToArgs(A,Ptr,2)) {
    PlainError("Out of memory placing library list starting mark");
    return;
  }
  /*
    Gather all libraries into a stack.
  */
  InitStackOfFilesToBeProcessed();
  InitStackOfFiles(stkLibraries);
  PushAllFilesInList(&I->Output->Libraries);
  while (true) {
    F=PopFileToBeProcessed();
    if (F==NULL) break;
    assert(F->Ext->Type==ftBinary);
    FetchDependencyLibraries(F);
    PushFileOntoStack(stkLibraries,F);
    PushAllFilesInList(&F->Libraries);
  }
  CleanupStackOfFilesToBeProcessedAfterUse();
  /*
    Build the list of libraries. There is a strict ordering rule imposed to
    them: if a library has dependencies, it must be listed before all its
    dependencies. The problem here is that using stack it is easy to build a
    list where a library is listed after all its dependencies, which is the
    exact reversal of what we need. The solution is simple: instead of
    sending the libraries to the argument list here
  */
  StackOut=stkLibraries;
  StackIn=stkUsedLibraries;
  InitStackOfFilesToBeProcessed();
  do {
    Continue=false;
    WereListed=false;
    while (true) {
      F=PopFileOffStack(StackOut);
      if (F==NULL) break;
      AllListed=true;
      for (i=0;i<F->Libraries.Cnt;i++) {
        OF=F->Libraries.Arr[i];
        if (!OF->Listed) {
          AllListed=false;
          break;
        }
      }
      if (AllListed) {
        F->Listed=true;
        WereListed=true;
        PushFileToBeProcessed(F);
      } else {
        if (!Continue) {
          Continue=true;
          InitStackOfFiles(StackIn);
        }
        PushFileOntoStack(StackIn,F);
      }
    }
    CleanupFileStack(StackOut);
    if (Continue) {
      if (WereListed) {
        Temp=StackOut;
        StackOut=StackIn;
        StackIn=Temp;
      } else {
        PlainError("Mutually depending libraries present");
        CleanupFileStack(StackIn);
        Continue=false;
      }
    }
  } while (Continue);
  /*
    Clear the "listed" flags for all libraries involved in this
  */
  while (true) {
    F=PopFileToBeProcessed();
    if (F==NULL) break;
    F->Listed=false;
    if (!AddFileToArgs(F,A)) {
      PlainError("Out of memory placing\1into argument list");
      PutFileArg(F);
    }
  }
  CleanupStackOfFilesToBeProcessedAfterUse();
}

static void PrepareSystemLibsLink(TLinkerArguments *A)
{
  const char *Ptr;

  if (!LinkSystemLibsDifferently) return;
  Ptr="-l";
  if (LinkStatically) Ptr="-d";
  if (!AddStringToArgs(A,Ptr,2)) {
    PlainError("Out of memory placing library list starting mark");
    return;
  }
}

static TRebuildStatus ProcessFilesToBeCompiledAndLinked(TPackageItem *I)
{
  TLinkerArguments A;
  TRebuildStatus Result;
  bool NameOK;
  char **Args;
  bool CleanupLibraryList;

  CleanupLibraryList=false;
  InitStackOfFiles(stkLibraries);
  A.ArgCnt=0;
  CreateBufferEx(&A.ArgText,&A.BH);
  Result=rsSuccessfullRebuild;
  NameOK=AddStringToArgs(&A,"@:/link",7);
  if (NameOK) {
    NameOK=AddFileToArgs(I->Output,&A);
  } else {
    PlainError("Out of memory placing the linker name to arguments");
    Result=rsErrorsWhileRebuild;
  }
  if (Result==rsSuccessfullRebuild) {
    A.OutputTimeStamp=I->Output->TimeStamp;
    A.NewestTimeStamp=0;
    Result=ProcessFilesInTheStack(stkFilesToLink,NULL,&A);
  }
  if (Result==rsSuccessfullRebuild) {
    Result=ProcessFilesInTheStack(stkFilesToLinkIntoLibrary,"-L",&A);
  }
  if (Result==rsSuccessfullRebuild) {
    BuildDependencyLibraryList(&A,I);
    CleanupLibraryList=true;
  }
  CleanupFileStack(stkLibraries);
  if (Result==rsSuccessfullRebuild) {
    BuildLibraryArgs(I,&A);
    PrepareSystemLibsLink(&A);
  }
  CloseBuffer(&A.BH);
  if (Result==rsSuccessfullRebuild) {
    Args=BuildArgList(&A);
    if (Args!=NULL) {
      if (A.OutputTimeStamp<A.NewestTimeStamp) {
        Print("Linking ");
        PrintFileName(I->Output);
        NewLine();
        if (Call((const char *const*)Args,NULL)!=0) {
          Result=rsErrorsWhileRebuild;
        } else {
          PutDependencyLibraryList(&A,I->Output);
          UpdateTimeStamp(I->Output);
        }
      }
      FreeMem(Args);
    } else {
      Result=rsErrorsWhileRebuild;
    }
  }
  DestroyBuffer(&A.ArgText);
  if (CleanupLibraryList) DestroyBuffer(&A.LibList);
  return(Result);
}

static void FetchFilesIntoLibraryStack(void)
{
  TFileInModule *F;

  while (true) {
    F=PopFileOffStack(stkFilesToLink);
    if (F==NULL) break;
    PushFileOntoStack(stkFilesToLinkIntoLibrary,F);
  }
}

static TRebuildStatus ScanPackageItem(TPackageItem *I)
{
  TRebuildStatus Result;

  I->Status=stBuilding;
  PrepareBinaryForScan(I->Output);
  InitStackOfFilesToBeProcessed();
  InitStackOfFiles(stkFilesToLink);
  InitStackOfFiles(stkFilesToLinkIntoLibrary);
  Result=TraverseAllFilesInList(&I->Headers);
  if (Result==rsSuccessfullRebuild) {
    Result=ProcessFilesOnTheStack();
  }
  if (Result==rsSuccessfullRebuild) {
    FetchFilesIntoLibraryStack();
    PushAllFilesInList(&I->Sources);
    Result=ProcessFilesOnTheStack();
  }
  CleanupStackOfFilesToBeProcessedAfterUse();
  if (Result==rsSuccessfullRebuild && !ClearDirectory) {
    Result=ProcessFilesToBeCompiledAndLinked(I);
  }
  FinalizeBinaryAfterScan();
  CleanupFileStack(stkFilesToLink);
  CleanupFileStack(stkFilesToLinkIntoLibrary);
  if (I->Status==stBuilding) {
    if (Result==rsPendingRebuild) {
      I->Status=stPending;
    } else {
      I->Status=stDone;
    }
  }
  return Result;
}

uint16 ProcessItems(TPackage *P,uint16 FlagsToScan)
{
  uint16 i;
  TPackageItem *I;
  TRebuildStatus Result;
  bool ScanIt;

  Result=rsSuccessfullRebuild;
  for (i=0;i<P->Items.Cnt;i++) {
    I=P->Items.Arr[i];
    if (FlagsToScan) {
      ScanIt=(I->Type & FlagsToScan)>0;
    } else {
      ScanIt=(I->Type & M_ALL_FLAGS)==0;
    }
    if (ScanIt) {
      Result=ScanPackageItem(I);
      if (Result!=rsSuccessfullRebuild) break;
    }
  }
  return(Result);
}

void ScanPackage(TPackage *P)
{
  TRebuildStatus Result;

  /*
    The order is following: first scan and build any translators, then any
    libraries and finally any programs. This order causes the least number of
    "interrupted builds", which are builds that could not be performed due to
    a dependency being out-of-date. In this stage the most of the dependency
    extraction is performed.
  */
  Result=ProcessItems(P,M_IS_GENERATOR);
  if (Result==rsSuccessfullRebuild) Result=ProcessItems(P,M_IS_LIBRARY);
  if (Result==rsSuccessfullRebuild) Result=ProcessItems(P,0);
}
