#include <syslib/types.h>
#define __INTERFACE__
#include "scanbase.h"

#include <libhs/strings.h>
#include <libhs/exparr.h>
#include <libhs/dictrb.h>
#include <libhs/filebuf.h>

#include "config.h"
#include "files.h"
#include "errors.h"
#include "filespec.h"
#include "filestk.h"
#include "fileread.h"

static TFileInModule *PreparedFile;
static TArrayHandle DependencyFilesHandle;

void SubstituteUnderscoresForSlashes(char *NameStart)
{
  char Ch;
  do {
    Ch=*NameStart;
    if (Ch=='/') *NameStart='_';
    NameStart++;
  } while (Ch!='\0');
}

static void SetCurrentDirectory(TFileInModule *F)
{
  const char *FileSpec;
  const char *BaseName;
  char *Ptr;

  FileSpec=F->Module->Name;
  assert(StrLen(FileSpec)<=MAX_FILESPEC_LENGTH);
  BaseName=FindModuleBaseName(FileSpec);
  Ptr=MemCpy(CurrentDir,FileSpec,BaseName-FileSpec);
  *Ptr='\0';
}

bool PrepareFileForScan(TFileInModule *File)
{
  assert(File!=NULL);
  assert(PreparedFile==NULL);
  if (!OpenFileInModule(File,0)) {
    CheckIOResultAndReport("Unable to open file");
    SetupFileName(NULL);
    return(false);
  }
  InitFileReader(true);
  PreparedFile=File;
  OpenArrayEx(&File->References,&DependencyFilesHandle);
  ClearOpenedArray(&DependencyFilesHandle);
  InitStackOfFiles(stkLibraries);
  InitStackOfFiles(stkAlreadyImportedHeaders);
  SetCurrentDirectory(File);
  return(true);
}

void RegisterHeaderImport(TFileInModule *File)
{
  assert(File!=NULL);
  if (PushFileOntoStack(stkAlreadyImportedHeaders,File)) {
    File->ImportLine=TokLine;
    File->ImportColumn=TokColumn;
  } else {
    TokError("File already imported at line\1, column\2");
    PutNumArg(File->ImportLine);
    PutNumArg(File->ImportColumn+1);
  }
}

TFileInModule *GetCurrentlyScannedFile(void)
{
  return(PreparedFile);
}

const char *GetCurrentFileName(void)
{
  assert(PreparedFile!=NULL);
  return(GetCurrentlyScannedFile()->Module->Name);
}

bool CheckCurrentFileHasHeader(void)
{
  TFileInModule *F;
  const char *Name;
  const char *Ext;

  assert(PreparedFile!=NULL);
  assert(PreparedFile->Ext!=NULL);
  assert(PreparedFile->Ext->Type==ftSource);
  Name=PreparedFile->Module->Name;
  Ext=PreparedFile->Ext->Assoc2->Ext;
  F=RegisterFileSplitted(Name,Ext);
  if (F==NULL) return(false);
  return(F->TimeStamp!=0);
}

static TFileInModule *RegisterOrComplain(
  TLocation Loc,
  const char *Name,
  const char *Ext
)
{
  TFileInModule *F;

  F=RegisterFileSplittedLoc(Loc,Name,Ext);
  if (F==NULL) {
    TokError("Unable to register \"\11.\12\"");
    PutArg(Name);
    PutArg(Ext);
  }
  return(F);
}

TFileInModule *RegisterDependencyFile(
  TLocation Loc,
  const char *Name,
  const char *Ext
)
{
  TFileInModule *File;

  File=RegisterOrComplain(Loc,Name,Ext);
  if (File!=NULL) {
    if (File->Ext->Type!=ftHeader) {
      TokError("File\1is not a header file");
      File=NULL;
    } else {
      if (!PutItem(&DependencyFilesHandle,File)) {
        TokError("Out of memory recording referenced file\1");
        PutFileArg(File);
        File=NULL;
      }
    }
  }
  return(File);
}

TFileInModule *FindDependencyLibrary(const char *Name,const char *Ext)
{
  TFileInModule *File;

  File=RegisterOrComplain(locBinaries,Name,Ext);
  if (File!=NULL) {
    if (File->TimeStamp==0) {
      File=RegisterOrComplain(locNonexecLibraries,Name,Ext);
    }
  }
  return(File);
}

void RegisterDependencyLibrary(const char *Name,const char *Ext)
{
  TFileInModule *File;

  File=FindDependencyLibrary(Name,Ext);
  if (File!=NULL) {
    PushFileOntoStack(stkLibraries,File);
  }
}

void FinalizeFileAfterScan(void)
{
  TArrayHandle DependencyLibrariesHandle;
  TFileInModule *F;

  assert(PreparedFile!=NULL);
  CloseFileInModule();
  CloseArray(&DependencyFilesHandle);
  OpenArrayEx(&PreparedFile->Libraries,&DependencyLibrariesHandle);
  ClearOpenedArray(&DependencyLibrariesHandle);
  do {
    F=PopFileOffStack(stkLibraries);
    if (F!=NULL) {
      if (!PutItem(&DependencyLibrariesHandle,F)) {
        TokError("Out of memory recording referenced library\1");
        PutFileArg(F);
      }
    }
  } while (F!=NULL);
  CloseArray(&DependencyLibrariesHandle);
  CleanupFileStack(stkAlreadyImportedHeaders);
  CleanupFileStack(stkLibraries);
  SetupFileName(NULL);
  if (__DEBUG__) PreparedFile=NULL;
}
