#include <syslib/types.h>
#define __INTERFACE__
#include "scancp.h"

/*
  This module parses the C compatibility P files. Currently it does some
  basic checks of the structure files and create its dependencies.
*/

#include <libhs/strings.h>
#include <libhs/output.h>

#include "config.h"
#include "fileread.h"
#include "filespec.h"
#include "errors.h"
#include "scanbase.h"

static TFileInModule *ParseImport(
  TLocation Loc,
  char *Cmd,
  uint16 Len,
  bool Active
)
{
  char *Ext;
  TFileInModule *F;
  char *FileSpec;
  char FileName[2*(MAX_FILESPEC_LENGTH+1)];
  char *Ptr;

  assert(Len==StrLen(Cmd));
  /*
    Strip out the trailing quotation mark.
  */
  Cmd[Len-1]='\0';
  /*
    Strip out the extension of the module. If the module has no extension,
    reject it. Otherwise put the pointer to the extension into "Ext".
  */
  Ext=Cmd+Len-1;
  while (Ext>Cmd && *(--Ext)!='.');
  if (Ext==Cmd) {
    if (Loc==locWorkDir) {
      FileError("Malformed local module reference");
    } else {
      FileError("Malformed external module reference");
    }
    return(NULL);
  }
  *(Ext++)='\0';
  /*
    Register the file and add it into the list of dependency files. Correct the Prepend
    the current working directory to the file.
  */
  FileSpec=Cmd+1;
  if (Loc==locWorkDir) {
    assert(StrLen(CurrentDir)<=MAX_FILESPEC_LENGTH);
    Ptr=StrCpy(FileName,CurrentDir);
    FileName[sizeof(FileName)-1]='\0';
    StrNCpy(Ptr,FileSpec,sizeof(FileName)-(Ptr-FileName));
    if (FileName[sizeof(FileName)-1]!='\0') {
      TokError("File specification too long");
      return(NULL);
    }
    FileSpec=FileName;
  }
  F=RegisterDependencyFile(Loc,FileSpec,Ext);
  if (F!=NULL) RegisterHeaderImport(F);
  return(F);
}

void ParseGlobalImport(char *Cmd,uint16 Len)
{
  char Ch;
  char *Ptr;
  TFileInModule *File;
  TFileExtension *Ext;

  Ptr=Cmd;
  do {
    Ch=*(++Ptr);
  } while (Ch!='\0' && Ch!='/');
  if (Ch=='\0') {
    TokError("Malformed external library reference");
    return;
  }
  File=ParseImport(locHeaders,Cmd,Len,true);
  if (File!=NULL) {
    *Ptr='\0';
    assert(File->Ext!=NULL);
    Ext=File->Ext;
    assert(Ext->Type==ftHeader);
    Ext=Ext->Assoc1;
    assert(Ext->Type==ftSource);
    Ext=Ext->Assoc1;
    assert(Ext->Type==ftCompiledModule);
    Ext=Ext->Assoc1;
    assert(Ext->Type==ftBinary);
    if (Ext!=MachineCodeExt || StrCmp(Cmd+1,"syslib")!=0) {
      RegisterDependencyLibrary(Cmd+1,Ext->Ext);
    }
  }
}

void ParseLocalImport(char *Cmd,uint16 Len,bool Active)
{
  ParseImport(locWorkDir,Cmd,Len,Active);
}

static void ParseImportFileCommandArgument(char *Cmd)
{
  uint16 Len;

  /*
    There are two types of #include commands.

    The first is #include <{filename}>. The {filename} is then expected to be
    in the form {library}/{module}.h and the {library} is then added to the
    list of dependency libraries.

    The second is #include "{filename}". The {filename} is then expected to
    be in the form {module}.h and {module} is expected to be a name of local
    module and thus this module is added to the list of dependency modules.
  */
  Len=StrLen(Cmd);
  if (Cmd[0]=='<' && Cmd[Len-1]=='>') {
    ParseGlobalImport(Cmd,Len);
  } else if (Cmd[0]=='"' && Cmd[Len-1]=='"') {
    ParseLocalImport(Cmd,Len,true);
  } else {
    TokError("Malformed #include directive argument");
  }
}

static bool IsIdentifier(const char *Buffer)
{
  char Ch;
  bool NumbersAllowed;

  NumbersAllowed=false;
  while (true) {
    Ch=*Buffer++;
    if (NumbersAllowed) {
      if (Ch=='\0') return(true);
      if (Ch>='0' && Ch<='9') continue;
    }
    NumbersAllowed=true;
    if (Ch!='_') {
      if (Ch<'A' || (Ch>'Z' && Ch<'a') || Ch>'z') break;
    }
  }
  return(false);
}

static bool ScanCLine(void)
{
  char Ch;
  char Buffer[128];

  /*
    Look at the first character in the file. If zero is read, it is treated
    as EOF condition. Normal C sources do not contain zero bytes.
  */
  Ch=GetChar();
  if (Ch=='\0') return(false);
  /*
    If there is a nonblank, put it back.
  */
  GoNonwhitespaceBack();
  /*
    Read the first word into the buffer.
  */
  ReadWordIntoBuf(Buffer,128);
  /*
    P language does not have preprocessor. However P sources may contain
    ESPM build directives, which are used to control how the program
    is packed up from its pieces.
    In C compatible mode we use the preprocessor directives as the ESPM build
    directives. The only directive that we parse here is #include,
    which gets the "import" semantics under ESPM. No conditional compilation
    is allowed in the P language.
    Also pay attention to the "#endif" directive, which must be at the end
    of the header file.
  */
  if (!StrCmp(Buffer,"#include")) {
    if (TokColumn==0) {
      ReadWordIntoBuf(Buffer,128);
      ParseImportFileCommandArgument(Buffer);
    }
  } else if (!StrCmp(Buffer,"#endif")) {
    return(false);
  } else if (!StrCmp(Buffer,"#define")) {
    ReadWordIntoBuf(Buffer,128);
    if (!IsIdentifier(Buffer)) {
      FileError("Invalid ESPM constant definition directive");
    }
  } else if (Buffer[0]=='#') {
    FileError("Invalid ESPM build directive\1");
    PutArg(Buffer);
  }
  /*
    Skip the rest of the line.
  */
  ReadToNL();
  /*
    Return that no EOF was encountered.
  */
  return(true);
}

static bool CheckWordSequence(const char **Strings)
{
  int16 i;
  const char *String;
  char Num[2];
  char Buf[MAX_FILESPEC_LENGTH+16];

  i=0;
  do {
    String=Strings[i++];
    if (String==NULL) break;
    ReadWordIntoBuf(Buf,MAX_FILESPEC_LENGTH+16);
    if (StrCmp(Buf,String)) {
      FileError("Word \11 is\2, expected\3");
      Num[0]=i+'0';
      Num[1]='\0';
      PutArg(Num);
      PutArg(Buf[0]=='\n'?"(newline)":Buf);
      PutArg(String[0]=='\n'?"(newline)":String);
      return(false);
    }
    if (String[0]=='\n') ReadToNL();
  } while (true);
  return(true);
}

static bool CheckWordSequenceIfNotInstalled(const char **Strings)
{
  TFileInModule *F;
  const char *String;
  char Buffer[128];

  F=GetCurrentlyScannedFile();
  if (F->Module->Location==locWorkDir) {
    return(CheckWordSequence(Strings));
  } else do {
    String=*(Strings++);
    if (String==NULL) return(true);
    ReadWordIntoBuf(Buffer,128);
    if (String[0]=='\n') ReadToNL();
  } while (true);
}

static const char *AddSystemHeader(const char *Hdr)
{
  char Buf[32];
  char *Ptr;

  assert(StrLen(Hdr)<32);
  Ptr=StrCpy(Buf,Hdr);
  ParseGlobalImport(Buf,Ptr-Buf);
  return(Hdr);
}

void ScanCCompatiblePSource(void)
{
  char Buf[MAX_FILESPEC_LENGTH+16];
  char Buf1[MAX_FILESPEC_LENGTH+16];
  char *Ptr;
  const char *Strings[10];
  const char **SPtr;
  bool ValidInput;
  const char *Name;

  SetupCCommentFiltering();
  SPtr=Strings;
  *(SPtr++)="#include";
  *(SPtr++)=AddSystemHeader("<syslib/types.h>");
  *(SPtr++)="\n";
  if (CheckCurrentFileHasHeader()) {
    *(SPtr++)="#define";
    *(SPtr++)="__INTERFACE__";
    *(SPtr++)="\n";
    *(SPtr++)="#include";
    Ptr=Buf;
    *(Ptr++)='"';
    Name=GetCurrentFileName();
    assert(StrLen(Name)<=MAX_FILESPEC_LENGTH);
    Ptr=StrCpy(Ptr,FindModuleBaseName(Name));
    Ptr=StrCpy(Ptr,".h\"");
    StrCpy(Buf1,Buf);
    ParseLocalImport(Buf1,Ptr-Buf,true);
    *(SPtr++)=Buf;
    *(SPtr++)="\n";
  }
  *(SPtr++)=NULL;
  ValidInput=CheckWordSequence(Strings);
  if (ValidInput) {
    while (ScanCLine());
  }
}

static bool CheckHeaderProlog(TFileInModule *F)
{
  TFileInModule *B;
  char HdrSym[MAX_FILESPEC_LENGTH+32];
  char *Ptr;
  char *NameStart;
  const char *Strings[10];
  const char **SPtr;
  int16 RemSize;

  /*
    Prepare the header file control symbol to be compared in the header file
    start.
  */
  assert(F->Owner!=NULL);
  B=F->Owner;
  Ptr=StrCpy(HdrSym,"__h__");
  RemSize=MAX_FILESPEC_LENGTH+32-MAX_FILENAME_LENGTH;
  Ptr=StrNToUpr(Ptr,B->Module->Name,RemSize);
  *(Ptr++)='_';
  NameStart=Ptr;
  RemSize=MAX_FILESPEC_LENGTH+32-3-(Ptr-HdrSym);
  Ptr=StrNToUpr(Ptr,GetCurrentFileName(),RemSize);
  SubstituteUnderscoresForSlashes(NameStart);
  StrCpy(Ptr,"__");
  /*
    Build the prolog description for the check. The first line in any C
    compatible P header must contain "#include <syslib/hhead.h>", the next
    one "#ifndef __h__<header>__", and the third line "#define
    __h__<header>__". Then lines of the header file follow and the
    last nonempty line must contain "#endif". The <header> here is the name
    of the header file converted into capital letters.
  */
  SPtr=Strings;
  *(SPtr++)="#include";
  *(SPtr++)=AddSystemHeader("<syslib/hhead.h>");
  *(SPtr++)="\n";
  *(SPtr++)="#ifndef";
  *(SPtr++)=HdrSym;
  *(SPtr++)="\n";
  *(SPtr++)="#define";
  *(SPtr++)=HdrSym;
  *(SPtr++)="\n";
  *(SPtr++)=NULL;
  return(CheckWordSequenceIfNotInstalled(Strings));
}

static bool CheckHeaderEpilog(void)
{
  const char *Strings[10];
  const char **SPtr;

  ReadToNL();
  SPtr=Strings;
  *(SPtr++)="#include";
  *(SPtr++)=AddSystemHeader("<syslib/hfoot.h>");
  *(SPtr++)=NULL;
  return(CheckWordSequenceIfNotInstalled(Strings));
}

void ScanCCompatiblePHeader(void)
{
  TFileInModule *F;

  SetupCCommentFiltering();
  F=GetCurrentlyScannedFile();
  if (!CheckHeaderProlog(F)) return;
  while (ScanCLine());
  CheckHeaderEpilog();
}
