#include <syslib/types.h>
#define __INTERFACE__
#include "bench.h"

#include <syslib/kernel.h>

#include <libhs/strings.h>
#include <libhs/thread.h>
#include <libhs/numconv.h>

#include "settings.h"
#include "errors.h"
#include "screen.h"

static uint32 Counter;
static uint32 Cur;
static uint32 Last;
static uint32 MinCur;
static uint32 MaxCur;
static uint16 Time;
static uint16 EllapsedTime;

static int16 BenchLock;
static int16 BenchActiveCond;
static int16 StatsOutOfDateCond;
static bool ThreadActive;

static int16 ThreadCnt;
static int16 AllThreadsExittedCond;

static void ThreadStarted(void)
{
  Lock(BenchLock);
  ThreadCnt++;
  Unlock(BenchLock);
}

static void ThreadExitted(void)
{
  Lock(BenchLock);
  ThreadCnt--;
  if (ThreadCnt==0) SetCond(AllThreadsExittedCond);
  Unlock(BenchLock);
}

static int16 WorkingThreadCnt;
static int16 AllThreadsIdleCond;

static void ThreadWorkStart(void)
{
  Lock(BenchLock);
  WorkingThreadCnt++;
  Unlock(BenchLock);
}

static void ThreadWorkStop(void)
{
  Lock(BenchLock);
  WorkingThreadCnt--;
  if (WorkingThreadCnt==0) SetCond(AllThreadsIdleCond);
  Unlock(BenchLock);
}

static void ShowStatistics(void)
{
  char Line1[41];
  char Line2[41];
  uint32 AvgCur;

  Lock(BenchLock);
  FormatDWord(Time,4,' ',Line1);
  FormatDWord(Counter,10,' ',Line1+5);
  FormatDWord(Cur,10,' ',Line1+16);
  AvgCur=0;
  if (EllapsedTime>0) AvgCur=Counter/EllapsedTime;
  FormatDWord(AvgCur,10,' ',Line1+27);
  MemSet(Line2,5,' ');
  FormatDWord(MinCur,10,' ',Line2+5);
  FormatDWord(MaxCur,10,' ',Line2+16);
  Unlock(BenchLock);
  Line1[4]=' ';
  Line1[15]=' ';
  Line1[26]=' ';
  Line2[15]=' ';
  PrintXY(1, ScrHeight/16-2, 7, Line1);
  PrintXY(1, ScrHeight/16-1, 7, Line2);
}

static void UpdateStatistics(void)
{
  while (Time>0) {
    WaitCond(StatsOutOfDateCond);
    ShowStatistics();
    ResetCond(StatsOutOfDateCond);
  }
}

static int16 ShowStatisticsThreadMain(void *Data)
{
  ThreadStarted();
  do {
    WaitCond(BenchActiveCond);
    if (ThreadActive) {
      ThreadWorkStart();
      UpdateStatistics();
      Lock(BenchLock);
      InProgress=false;
      ResetCond(BenchActiveCond);
      Unlock(BenchLock);
      ThreadWorkStop();
    }
  } while (ThreadActive);
  ThreadExitted();
  return(0);
}

static void WatchCounter(void)
{
  while (Time>0) {
    SetCond(StatsOutOfDateCond);
    KCALL__SUSPEND(1000000);
    Lock(BenchLock);
    Time--;
    EllapsedTime++;
    Cur=Counter-Last;
    if (MinCur>Cur) MinCur=Cur;
    if (MaxCur<Cur) MaxCur=Cur;
    Last=Counter;
    Unlock(BenchLock);
  }
}

static int16 BenchmarkThreadMain(void *Data)
{
  ThreadStarted();
  do {
    WaitCond(BenchActiveCond);
    if (ThreadActive) {
      ThreadWorkStart();
      WatchCounter();
      Lock(BenchLock);
      InProgress=false;
      ResetCond(BenchActiveCond);
      Unlock(BenchLock);
      ThreadWorkStop();
    }
  } while (ThreadActive);
  ThreadExitted();
  return(0);
}

void InitializeBenchmarking(void)
{
  int16 WatcherThreadID;
  int16 StatsUpdaterThreadID;

  ThreadActive=true;
  BenchLock=CreateLock();
  if (BenchLock==-1) Fatal("Unable to create benchmark thread lock");
  AllThreadsExittedCond=CreateCond();
  if (AllThreadsExittedCond==-1) {
    Fatal("Unable to create threads exit conditional");
  }
  AllThreadsIdleCond=CreateCond();
  if (AllThreadsIdleCond==-1) {
    Fatal("Unable to create threads idling conditional");
  }
  BenchActiveCond=CreateCond();
  if (BenchActiveCond==-1) {
    Fatal("Unable to create benchmark activity conditional");
  }
  StatsOutOfDateCond=CreateCond();
  if (StatsOutOfDateCond==-1) {
    Fatal("Unable to create statistics outdate conditional");
  }
  Time=1;
  WatcherThreadID=CreateNewThread(BenchmarkThreadMain,NULL);
  StatsUpdaterThreadID=CreateNewThread(ShowStatisticsThreadMain,NULL);
  if (WatcherThreadID==-1) Fatal("Unable to create benchmarking thread");
  if (StatsUpdaterThreadID==-1) {
    Fatal("Unable to create statistics updater thread");
  }
}

void StartBenchmark(void)
{
  Lock(BenchLock);
  Time=TestDuration+1;
  Counter=0;
  MinCur=0xFFFFFFFF;
  MaxCur=0;
  Last=0;
  EllapsedTime=0;
  Unlock(BenchLock);
  InProgress=true;
  ResetCond(AllThreadsIdleCond);
  ResetCond(StatsOutOfDateCond);
  SetCond(BenchActiveCond);
}

void AddToCounter(uint32 Count)
{
  Lock(BenchLock);
  Counter+=Count;
  Unlock(BenchLock);
}

void StopBenchmark(void)
{
  Time=0;
  SetCond(StatsOutOfDateCond);
  WaitCond(AllThreadsIdleCond);
}
