#include <syslib/types.h>
/*
  Font generator program for OTBT
*/

#include <syslib/env.h>

#include <libhs/strings.h>
#include <libhs/filebuf.h>
#include <libhs/output.h>

#define MAX_LINE 128
#define OUTPUT_WIDTH (70-18)

/*** Input ***/
static char Line[MAX_LINE+1];   /* input line */
static uint16 LineLen;          /* length of the line */
/*** Output ***/
static int16 OutputPos;         /* position on output line */
static bool OutputOpen;         /* true if output file is opened */
/*** Error reporting ***/
static uint16 LineNum;          /* input line number */
static bool WasError;           /* error flag */
/*** RLE encoding ***/
static uint16 Count;            /* repeat count */
static uint16 Byte;             /* repeated byte */

/*** Output an error message ***/
static void Msg(char *Msg)
{
  if (OutputOpen) DeactivateCurrentOutputFile();
  ActivateErr();
  if (LineNum>0) {
    PrintNum(LineNum,5);
    Print(": ");
  }
  PrintLn(Msg);
  DeactivateErr();
  if (OutputOpen) ActivateCurrentOutputFile();
  WasError=true;
}

/*** Input line reading ***/
static void ReadLine(void)
{
  char Ch;
  bool IgnoreBlanks;
  bool Comment;
  char *Ptr;

  /*
    Read
  */
  do {
    /*
      Count the line. This must be done here, because the line reading
      process.
    */
    LineNum++;
    /*
      Read the line into the static line buffer. Collapse the multiple blanks
      into single space. The leading blanks are trimmed by simply pretending
      the blank collapsing routine that we're just outputted a space at the
      beginning of the line. Also comments are stripped here.
    */
    Comment=false;
    IgnoreBlanks=true;
    Ptr=Line;
    do {
      Ch=ReadByte();
      if (Ch!='\0') {
        /*
          Check for comments. Characters in comment are ignored (including
          the comment mark). Other characters will be processed. Notice that
          even if we know that there will be no characters outputted anymore
          in this loop, we can't exit the loop now. We need to find the end
          of line, otherwise the comment would load as the next line instead
          of the actual content of that line.
        */
        if (Ch=='#') Comment=true;
        if (!Comment) {
          /*
            The line character processing. Here the blank collapsing
            routinne resides.
          */
          if (Ch==' ' || Ch=='\t' || Ch=='\r' || Ch=='\n') {
            if (!IgnoreBlanks) {
              *(Ptr++)=' ';
              IgnoreBlanks=true;
            }
          } else {
            IgnoreBlanks=false;
            *(Ptr++)=Ch;
          }
          /*
            Check if the line is too long. If yes, output an error message
            and strip the line. This check will avoid the "buffer overrun"
            problems. Notice that the size of the input line is checked
            after the blanks are collapsed and the leading blanks and
            trailing comments are stripped, so huge amount of blanks or a
            long comment on a source line will not hurt the font compilation
            process.
          */
          if (Ptr==Line+MAX_LINE+1) {
            Ptr--;
            Msg("Line too long");
            Comment=true;
          }
        }
      }
    } while (Ch!='\n' && Ch!='\0');
    /*
      Trim trailing space. We don't need to check if there is more than one
      space or other blanks, because the blank collapsing routinne already
      did this job. We must beware to the condition when the line is empty
      to avoid "buffer underrun" problems.
    */
    if ((Ptr>Line) && *(Ptr-1)==' ') Ptr--;
    /*
      Terminate the line with the zero byte.
    */
    *Ptr='\0';
    /*
      The empty lines will be ignored and only nonempty lines will be
      allowed to exit this procedure and enter the font processing code. But
      we must beware to the EOF condition, because it we don't and there is
      an empty line on EOF, the program would enter an infinite loop.
    */
  } while ((Line[0]=='\0') && (Ch!='\0'));
  /*
    Compute the length of the line and store it into the "LineLen" global
    variable.
  */
  LineLen=Ptr-Line;
}

/*** Start a new line with the code ***/
static void StartOutputCodeLine(void)
{
  Print("  streami8u ");
}

/*** End the line with the code ***/
static void EndOutputCodeLine(void)
{
  Print(" endstream\n");
}

/*** Output file writing ***/
/*
  Write a byte to the output stream. The byte stream is encoded in such a way
  that the encoded output will form a "C" string.
*/
static void WriteOneByte(uint8 Byte)
{
  int16 Width;

  /*
    Update the position on the output line according to the outputted
    character and it's format.
  */
  Width=1+(Byte>9)+(Byte>99)+(OutputPos>0);
  /*
    If the line is getting too long, break it.
  */
  if (OutputPos>OUTPUT_WIDTH) {
    EndOutputCodeLine();
    StartOutputCodeLine();
    OutputPos=Width;
  } else {
    if (OutputPos>0) Print(",");
    OutputPos+=Width;
  }
  /*
    Output the character.
  */
  PrintNum(Byte,1);
}

/*** Flush RLE packed bytes ***/
static void FlushBytes(void)
{
  uint8 OutCnt;
  uint8 Boundary;

  while (Count>0) {
    Boundary=2;
    if (Byte>0) Boundary++;
    OutCnt=31+Boundary;
    if (OutCnt>Count) OutCnt=Count;
    Count-=OutCnt;
    if (OutCnt>=Boundary) {
      OutCnt=(OutCnt-Boundary)*4+1;
      if (Byte>0) OutCnt+=2;
      WriteOneByte(OutCnt);
      if (Byte>0) WriteOneByte(Byte);
    } else {
      for (;OutCnt>0;OutCnt--) WriteOneByte(Byte);
    }
  }
}

/*** Send an unpacked byte ***/
static void SendByte(uint8 What)
{
  if (What!=Byte) {
    FlushBytes();
    Byte=What;
  }
  Count++;
}

/*** Compile character shape line ***/
static void CompileCharacterShapeLine(void)
{
  char *Ptr;
  int16 i;
  int16 Byte;
  bool Invalid;

  Invalid=false;
  Ptr=Line;
  Byte=0;
  for (i=0;i<8;i++) {
    Byte*=2;
    switch (*Ptr) {
      case '1': Byte++;
      case '0': break;
      default: {
        Invalid=true;
        return;
      }
    }
    Ptr++;
  }
  if (Invalid || (Byte&1)>0) {
    Msg("Invalid character line");
    return;
  }
  SendByte(Byte);
}

/*** Dump the header comment for the generated file ***/
void PrintHeadingComment(char *SrcName)
{
  Print(
    "/*\n"
    "  This file was automatically generated by FONTGEN. "
    "DO NOT EDIT! Make changes\n"
    "  to the source of this file instead.\n"
    "  Source: "
  );
  PrintLn(SrcName);
  Print(
    "*/\n\n"
  );
}

/*** The main program ***/
int16 Main(void)
{
  int16 argc;
  char * const *argv;
  bool Bad;
  int16 i;
  char Ch;
  bool Result;

  /*
    Get the argument informations
  */
  argc=GetArgCnt();
  argv=GetArgs();
  /*
    Check if required arguments were given on command line. If not, write
    the usage and exit with error status set to true.
  */
  if (argc!=3) {
    Print(
      "FONTGEN for OTBT 1.0.0: Font generator program for OTBT\n\n"
      "USAGE: fontgen <source> <output-header> <output-source>\n"
    );
    return(true);
  }
  /*
    Try to open the input file. If this fails, exit with error status set
    to true.
  */
  Result=AllocateAndOpenInputFile(argv[0]);
  if (!Result) {
    Msg("Unable to open font source file");
    return(true);
  }
  /*
    Read the first line. It must specify an identifier.
  */
  ReadLine();
  if (Line[0]=='\0') {
    Msg("Font name expected");
  } else {
    Bad=false;
    for (i=0;!Bad && Line[i]!='\0';i++) {
      Ch=Line[i];
      Bad=(Ch<'A' || (Ch>'Z' && Ch<'a') || Ch>'z');
      if (Bad && i>0) Bad=Ch<'0' || Ch>'9';
    }
    if (Bad) Msg("Invalid font name");
  }
  /*
    Build the header file for the generated font.
  */
  Result=AllocateAndCreateOutputFile(argv[1]);
  if (!Result) {
    Print("Unable to create compiled font header file");
    return(true);
  }
  ActivateCurrentOutputFile();
  PrintHeadingComment(argv[0]);
  Print(
    "extdatablock("
  );
  Print(Line);
  Print(
    ");\n"
  );
  DeactivateCurrentOutputFile();
  CloseAndFreeOutputFile();
  if (IOResult()!=0) {
    Msg("Error closing compiled font header file");
    return(true);
  }
  /*
    Try to create the output file. If this fails, also return the error
    status set to true. In fact the output from the compiler is binary, but
    we need to generate a C include source file, so we will open the output
    file in text mode.
  */
  Result=AllocateAndCreateOutputFile(argv[2]);
  if (!Result) {
    Msg("Unable to create compiled font file");
    return(true);
  }
  ActivateCurrentOutputFile();
  OutputOpen=true;
  /*
    Write the header of the output file.
  */
  PrintHeadingComment(argv[0]);
  Print(
    "#include <syslib/types.h>\n\n"
    "datablock("
  );
  Print(Line);
  Print(
    ")\n"
  );
  StartOutputCodeLine();
  /*
    Read the rest of the file and send the data bytes to the output
  */
  do {
    ReadLine();
    if (Line[0]!='\0') CompileCharacterShapeLine();
  } while (Line[0]!='\0');
  FlushBytes();
  /*
    Write the footer
  */
  EndOutputCodeLine();
  Print(
    "enddatablock\n"
  );
  /*
    Close the files
  */
  LineNum=0;
  DeactivateCurrentOutputFile();
  CloseAndFreeOutputFile();
  OutputOpen=false;
  if (IOResult()!=0) Msg("Error closing output file");
  CloseAndFreeInputFile();
  if (IOResult()!=0) Msg("Error closing input file");
  /*
    Return the success flag
  */
  return(WasError?1:0);
}
