#include <syslib/types.h>
#define __INTERFACE__
#include "screen.h"

#include <syslib/traps.h>
#include <syslib/kernel.h>
#include <syslib/pdbase.h>
#include <syslib/pdkeys.h>
#include <syslib/pdfsys.h>
#include <syslib/pdfileio.h>
#include <syslib/pdgfxcon.h>

#include <libhs/thread.h>
#include <libhs/memory.h>
#include <libhs/diskio.h>
#include <libhs/gfxcon.h>
#include <libhs/errno.h>
#include <libhs/strings.h>

#include "settings.h"
#include "font.gi"

static upint8 *CharLine;
static uint32 Colors[16];

#define CHARS_ON_LINE 160

static upint16 TextScreen[10240];

staticdatablock(ColorDefs)
  streamstr
    "\077"        /* 15 = white */
    "\066"        /* 14 = yellow */
    "\055"        /* 13 = light magenta */
    "\044"        /* 12 = light red */
    "\033"        /* 11 = light cyan */
    "\022"        /* 10 = light green */
    "\011"        /*  9 = light blue */
    "\070"        /*  8 = dark grey */
    "\007"        /*  7 = light grey */
    "\024"        /*  6 = brown */
    "\005"        /*  5 = magenta */
    "\004"        /*  4 = red */
    "\003"        /*  3 = cyan */
    "\002"        /*  2 = green */
    "\001"        /*  1 = blue */
                  /*  0 = black (the terminating zero is here) */
  endstream
enddatablock

/*
  Build the EGA translation. This table is used
*/
static void BuildColorTranslation(void)
{
  upint8 *Table;
  uint16 i;
  uint16 Color;
  uint16 Red;
  uint16 Green;
  uint16 Blue;

  Table=ColorDefs();
  i=16;
  while ((i--)>0) {
    Color=*(Table++);
    Red  =((Color&04)/04)*85*2+((Color&040)/040)*85;
    Green=((Color&02)/02)*85*2+((Color&020)/020)*85;
    Blue =((Color&01)/01)*85*2+((Color&010)/010)*85;
    if (ScrPixelSize>=3) {
      Colors[i]=(Red<<16)+(Green<<8)+Blue;
    } else if (ScrPixelSize==2) {
      Colors[i]=((Red>>3)<<11)+((Green>>2)<<5)+(Blue>>3);
    } else if (ScrPixelSize==1) {
      Colors[i]=((Blue>>6)<<6)+((Green>>5)<<3)+(Red>>5);
    }
  }
}

bool SetTheScreenMode(void)
{
  void *NewCharLine;
  int16 VidMem;

  Reconfigure();
  NewCharLine=ReallocMem(CharLine,ScrWidth*ScrPixelSize*16);
  if (NewCharLine==NULL) {
    SetErrNo(lheNOMEM);
    return(false);
  }
  CharLine=NewCharLine;
  VidMem=SetScreenMode(ScrMode);
  if (VidMem!=-1) BuildColorTranslation();
  return(VidMem!=-1);
}

static upint8 *Bytes;
static uint8 Count;
static uint8 Byte;

static void OpenFont(void)
{
  Count=0;
  Bytes=GiveFont();
}

static void Fetch(void)
{
  Byte=*(Bytes++);
  Count=1;
  if ((Byte&1)>0) {
    Count=(Byte>>2)+2;
    if ((Byte&2)>0) {
      Byte=*(Bytes++);
      Count+=1;
    } else {
      Byte=0;
    }
  }
}

static void SkipBytes(uint16 Cnt)
{
  uint8 SkipCnt;

  while (Cnt) {
    SkipCnt=Count;
    if (SkipCnt>Cnt) SkipCnt=Cnt;
    Count-=SkipCnt;
    Cnt-=SkipCnt;
    if (Count==0) Fetch();
  }
}

static uint8 GetByte(void)
{
  if (Count==0) Fetch();
  Count--;
  return(Byte);
}

static void GenChar(upint8 *Buf,uint8 Width,uint8 Attr,char Ch)
{
  uint8 Lines[16];
  int16 i,j;
  uint8 Line;
  uint32 FgColor;
  uint32 BgColor;
  uint32 Color;

  assert(Width>=8);
  assert(Attr>=0 && Attr<=255);
  for (i=0;i<16;i++) Lines[i]=0;
  if (Ch>'\040' && Ch<'\177') {
    OpenFont();
    SkipBytes((Ch-'!')*16);
    for (i=0;i<16;i++) Lines[i]=GetByte();
  } else {
    switch (Ch) {
      case 0: {
        Lines[0]=255;
        Lines[1]=128;
        Lines[15]=128;
        break;
      }
      case 1: {
        Lines[0]=255;
        Lines[1]=1;
        Lines[15]=1;
        break;
      }
      case 2: {
        Lines[15]=255;
        Lines[1]=128;
        Lines[0]=128;
        break;
      }
      case 3: {
        Lines[15]=255;
        Lines[1]=1;
        Lines[0]=1;
        break;
      }
    }
    for (i=2;i<15;i++) Lines[i]=Lines[1];
  }
  BgColor=Colors[Attr>>4];
  FgColor=Colors[Attr%16];
  for (i=0;i<16;i++) {
    Line=Lines[i];
    for (j=0;j<8;j++) {
      Color=((Line&128)>0)?FgColor:BgColor;
      switch (ScrPixelSize) {
        case 1: {
          *(Buf++)=Color;
          break;
        }
        case 2: {
          *((upint16 *)Buf)=Color;
          Buf+=2;
          break;
        }
        case 3: {
          *(Buf++)=Color&255;
          *(Buf++)=(Color>>8)&255;
          *(Buf++)=Color>>16;
          break;
        }
        case 4: {
          *((upint32 *)Buf)=Color;
          Buf+=4;
          break;
        }
        default: {
          ReportBug("Invalid pixel size in current video mode");
          break;
        }
      }
      Line=(Line&127)<<1;
    }
    Buf+=(Width-8)*ScrPixelSize;
  }
}

void PrintXY(uint8 x,uint8 y,uint8 Attr,const char *Text)
{
  uint8 Ch;
  int16 Size;
  upint16 *Ptr;
  int16 Width;
  uint16 ChAttr;

  /*
    Write the content of the line onto the screen.
  */
  assert(y<ScrHeight/16);
  Ptr=TextScreen+y*CHARS_ON_LINE+x;
  Width=ScrWidth/8;
  do {
    Ch=*(Text++);
    if (Ch!='\0') {
      assert(x<Width);
      x++;
      *(Ptr++)=Attr*256+Ch;
    }
  } while (Ch!='\0');
  /*
    NOTE:
    Calls on video memory seeks are not allowed to fail except for conditions
    such as attempt to write at the end of the video memory or I/O from/to
    an invalid region of memory. All these situations are bugs.
  */
  Size=16*ScrWidth;
  OpenScreenRegion(0,16*y,ScrWidth,16);
  x=0;
  Ptr=TextScreen+y*CHARS_ON_LINE;
  for (x=0;x<Width;x++) {
    ChAttr=*(Ptr++);
    GenChar(CharLine+x*8*ScrPixelSize,ScrWidth,ChAttr/256,ChAttr%256);
  }
  ScreenWrite(CharLine,Size);
  UpdateScreen();
}

void ClearScreen(void)
{
  uint16 Size;
  uint16 i;
  int16 Width;
  uint8 Height;

  Size=ScrWidth*16;
  Width=ScrWidth/8;
  for (i=0;i<Width;i++) {
    GenChar(CharLine+i*8*ScrPixelSize,ScrWidth,0,32);
  }
  OpenScreenRegion(0,0,ScrWidth,ScrHeight);
  Height=ScrHeight/16;
  for (i=0;i<Height;i++) ScreenWrite(CharLine,Size);
  ScreenWrite(CharLine,ScrWidth*(ScrHeight%16));
  MemSet(TextScreen,sizeof(TextScreen),0);
  UpdateScreen();
}

void DisplayCorners(uint8 Attr)
{
  int16 i;
  int16 Width;
  uint16 Size;

  Size=ScrWidth*16;
  Width=ScrWidth/8;
  for (i=1;i<Width-1;i++) {
    GenChar(CharLine+i*8*ScrPixelSize,ScrWidth,0,32);
  }
  GenChar(CharLine,ScrWidth,Attr,0);
  GenChar(CharLine+(ScrWidth-8)*ScrPixelSize,ScrWidth,Attr,1);
  OpenScreenRegion(0,0,ScrWidth,16);
  ScreenWrite(CharLine,Size);
  GenChar(CharLine,ScrWidth,Attr,2);
  GenChar(CharLine+(ScrWidth-8)*ScrPixelSize,ScrWidth,Attr,3);
  OpenScreenRegion(0,ScrHeight-16,ScrWidth,16);
  ScreenWrite(CharLine,Size);
  UpdateScreen();
}

uint16 ReadKey()
{
  uint16 Key;

  do {
    Key=RawReadKey();
    if ((Key&MK_RELEASED)==0) return(Key);
  } while (true);
}
