"""
Debugger for M4 macros.
Requires OSHS terminal
"""

import re

def LogIt(What):
  Log.append(What)

def Define(Name,Expansion):
  global ParsedInputString
  try:
    Old=Macros[Name]
  except KeyError:
    Old=None
  if type(Old) is type([]):
    Old[-1]=Expansion
  else:
    Macros[Name]=Expansion
  ParsedInputString=None

def Undefine(Name):
  global ParsedInputString
  try:
    del Macros[Name]
    ParsedInputString=None
  except KeyError:
    LogIt("Attempt to undefine an undefined macro")
    pass

def PushDef(Name,Expansion):
  global ParsedInputString
  try:
    Old=Macros[Name]
  except KeyError:
    Old=None
  if type(Old) is not type([]):
    if Old is None:
      Old=[]
      ParsedInputString=None
    else:
      Old=[Old]
    Macros[Name]=Old
  Old.append(Expansion)

def PopDef(Name):
  global ParsedInputString
  try:
    Old=Macros[Name]
    Old.pop()
    if len(Old)==0:
      del Macros[Name]
      ParsedInputString=None
  except KeyError:
    LogIt("Attempt to pop from an undefined macro")
    pass

def ValueToInt(Value):
  for Ch in Value:
    if not Ch in "0123456789+-":
      raise ValueError
  return int(Value)

def Divert(NewDiv):
  global Diversion
  if Stack:
    LogIt("'divert' not allowed here")
  try:
    Diversion=ValueToInt(NewDiv)
  except ValueError:
    LogIt("Non-numeric argument to 'divert'")

def SendToStdOut(Output):
  global OutBuffer
  OutBuffer+=Output

def SendOutput(Output):
  if Output=="":
    return
  if Diversion==0 or Stack:
    SendToStdOut(Output)
  elif Diversion>0:
    try:
      Diversions[Diversion]+=Output
    except KeyError:
      Diversions[Diversion]=Output

def Undivert(DivNum):
  if DivNum is Ellipsis:
    List=Diversions.keys()
    List.sort()
    for DivNum in List:
      Undivert(DivNum)
  try:
    DivNum=ValueToInt(DivNum)
  except ValueError:
    LogIt("Nonnumeric argument to 'undivert'")
    return
  if DivNum<=0:
    LogIt("Invalid diversion number in 'undivert'")
    return
  if DivNum==Diversion:
    LogIt("Attempt to undivert current diversion ignored")
    return
  try:
    SendOutput(Diversions[DivNum])
    del Diversions[DivNum]
  except KeyError:
    LogIt("Undiverting an empty diversion")
    pass

def Shift(*Args):
  Args=Args[1:]
  Result=[]
  for Arg in Args:
    Result.append(QuoteL+Arg+QuoteR)
  return ','.join(Result)

def IfElse(*Args):
  if len(Args)==1:
    return ""
  if len(Args)==2:
    LogIt("Too few arguments to 'ifelse'")
    return ""
  if len(Args)%3==2:
    LogIt("Excess arguments to 'ifelse' ignored")
    Args=Args[:-1]
  Index=0
  while Index+1<len(Args):
    if Args[Index]==Args[Index+1]:
      return Args[Index+2]
    else:
      Index+=3
  try:
    return Args[Index]
  except IndexError:
    return ""

def IfDef(Name,IfStr,ElseStr):
  if Name in Macros:
    return IfStr
  else:
    return ElseStr

Table=[]
for Ch in xrange(256):
  Table.append(chr(Ch))

def Translit(String,WhatReplace,ReplaceWith):
  T=Table[:]
  Num=0
  for Ch in WhatReplace:
    New=ReplaceWith[Num]
    Num+=1
    T[ord(Ch)]=New
  return String.translate("".join(T))

"""
Builtin macro list

The tuples which act like the declarations have the function as 0. element,
the number of arguments as 1. element and the defaults for the arguments as
the rest. If the number of arguments is negative, the macro is recognized
with the arguments only (and the count of the arguments is the absolute value
of the count element). If the count of the defaults (DCount) is less than
the count of arguments for the macro (ACount), the first ACount-DCount
arguments of the macro are mandatory.

If the 1. element of the tuple is Ellipsis, the macro can have any number of
arguments. The 2. element is 0 if the macro is not recognized without
arguments or 1 if it is.
"""
Builtins={
  "define":(Define,-2,""),
  "undefine":(Undefine,-1),
  "pushdef":(PushDef,-2,""),
  "popdef":(PopDef,-1),
  "divert":(Divert,1,"0"),
  "undivert":(Undivert,1,Ellipsis),
  "shift":(Shift,Ellipsis,1),
  "ifelse":(IfElse,Ellipsis,0),
  "ifdef":(IfDef,-3,""),
  "translit":(Translit,-3,""),
}

def InitializeM4():
  global Macros,QuoteL,QuoteR,Diversion,Diversions,OutBuffer
  global Log,InputString,Input,ParsedInputString,Expanding,Stack
  Log=[]
  Macros={}
  for Key in Builtins.keys():
    Macros["m4_"+Key]=Builtins[Key]
  QuoteL="["
  QuoteR="]"
  Diversion=-1
  Diversions={}
  Input=InputString
  ParsedInputString=None
  Expanding=None
  Stack=[]
  OutBuffer=''

def InstallM4Input(InputFiles):
  global InputString
  InputString=""
  for FileName in InputFiles:
    InputString+=open(FileName).read()

WordExpr=re.compile(r"([_a-zA-Z][_a-zA-Z0-9]*|\(|\))")
ArgExpr=re.compile(r"\$([0-9]+|\#|\*|\@)")
CommentExpr=re.compile(r"(\n#[^\n]*\n)")

def ParseQuotesCore(String):
  """
  Parse the string into the tree of quotations
  """
  String=String.split(QuoteL)
  Result=[]
  Stack=[]
  for Str in String:
    if Stack:
      List=Str.split(QuoteR,len(Stack))
      Str=None
      for Part in List:
        if not Str is None:
          if Str:
            Result.append(Str)
          Res=Stack.pop()
          Res.append(Result)
          Result=Res
        Str=Part
    if Str:
      Result.append(Str)
    Stack.append(Result)
    Result=[]
  Result=Stack[0]
  if len(Stack)>1:
    Result.append(None)
  return Result

def ParseQuotes(String):
  List=CommentExpr.split(String)
  Result=[]
  Valid=0
  for Str in List:
    Valid=1-Valid
    if Valid:
      Data=ParseQuotesCore(Str)
      Result.extend(Data)
    else:
      Last=''
      if Result and type(Result[-1]) is type(""):
        Last=Result.pop()
      Result.append(Last+Str)
  return Result

def UnparseQuotes(String):
  Result=""
  for Str in String:
    if type(Str) is type([]):
      Str=QuoteL+UnparseQuotes(Str)+QuoteR
    Result+=Str
  return Result

def GetMacroDefinition(Macro):
  Result=Macros[Macro]
  if type(Result) is type([]):
    Result=Result[-1]
  return Result

def ConvertNonParam(String,EndPos):
  Expansion=GetMacroDefinition(String)
  if type(Expansion) is type(()):
    ArgCnt=Expansion[1]
    if ArgCnt is Ellipsis:
      Unrecognized=not Expansion[2]
    else:
      Unrecognized=ArgCnt<0
    if Unrecognized:
      return String
  return (String,EndPos,EndPos,[])

def GetQuotedLen(Quoted):
  Result=len(QuoteL+QuoteR)
  for Str in Quoted:
    if type(Str) is type([]):
      Result+=GetQuotedLen(Str)
    else:
      Result+=len(Str)
  return(Result)

def Len(Str):
  if type(Str) is type(""):
    return len(Str)
  else:
    return GetQuotedLen(Str)

def FindMacrosInSplitted(String,Start=0):
  Result=[]
  Index=0
  Position=0
  while Index<len(String):
    Str=String[Index]
    Index+=1
    Position+=Len(Str)
    while 1:
      if type(Str) is type([]):
        break
      if not Str in Macros:
        break
      if Index>=len(String)-2:
        Str=ConvertNonParam(Str,Start+Position)
        break
      if String[Index]!='' or String[Index+1]!='(':
        Str=ConvertNonParam(Str,Start+Position)
        break
      Index+=2
      Position+=1
      Starting=Position
      Level=1
      Args=[]
      while Level:
        if Index==len(String):
          break
        S=String[Index]
        Position+=Len(S)
        Index+=1
        if S=="(":
          Level+=1
        elif S==")":
          Level-=1
        if Level==0:
          Args=FindMacrosInSplitted(Args,Starting)
        else:
          Args.append(S)
      else:
        Str=(Str,Starting+Start,Position+Start,Args)
        break
      break
    Result.append(Str)
  return Result

def FindMacros(String):
  Result=[]
  for Str in String:
    if type(Str) is type(""):
      Str=WordExpr.split(Str)
      Result.extend(Str)
    else:
      Result.append(Str)
  return FindMacrosInSplitted(Result)

def ParseInputString():
  global ParsedInput,ParsedInputString
  if Expanding is None and not ParsedInputString is Input:
    ParsedInput=FindMacros(ParseQuotes(Input))
    ParsedInputString=Input

def PushStatus():
  Stack.append((Input,ParsedInput,OutBuffer))

def PopStatus():
  global Input,ParsedInput,Expanding,OutBuffer,ParsedInputString
  (Input,ParsedInput,OutBuffer)=Stack.pop()
  ParsedInputString=Input
  Expanding=1

def ExpandMacro(Name,Args):
  Def=GetMacroDefinition(Name)
  if type(Def) is type(()):
    Func=Def[0]
    ArgCnt=Def[1]
    if not ArgCnt is Ellipsis:
      ArgCnt=abs(ArgCnt)
      if len(Args)<ArgCnt:
        Args.extend(Def[1+ArgCnt-len(Args):])
      if len(Args)<ArgCnt:
        LogIt("Mandatory arguments missing")
        return ""
    Result=Func(*Args)
    if Result is None:
      Result=""
    Result=[Result]
  else:
    Def=ArgExpr.split(Def)
    Result=[]
    Arg=0
    for Part in Def:
      if Arg:
        Part=(Part,)
      Result.append(Part)
      Arg=1-Arg
  return Result

def ExpandArgExpr(Expr,Macro,Args):
  if Expr=="*":
    return ','.join(Args)
  elif Expr=="#":
    return str(len(Args))
  elif Expr=="@":
    A=[]
    for Arg in Args:
      A.append(QuoteL+Arg+QuoteR)
    return ','.join(A)
  else:
    Expr=int(Expr)
    if Expr==0:
      return Macro
    else:
      try:
        return Args[Expr-1]
      except IndexError:
        LogIt("Unpassed argument referenced")
        return ""

def SingleStep():
  global Input,Expanding,OutBuffer,ParsedInputString,CurrentArg
  ParseInputString()
  if not ParsedInput:
    Result=OutBuffer
    if not Stack:
      return
    PopStatus()
    Result=ParseQuotes(Result)
    Args=[]
    Arg=''
    DeleteSpaces=1
    for Item in Result:
      if type(Item) is type([]):
        Arg+=UnparseQuotes(Item)
        DeleteSpaces=0
      else:
        Item=Item.split(",")
        if DeleteSpaces:
          Item[0]=Item[0].lstrip()
        for Index in xrange(1,len(Item)):
          Item[Index]=Item[Index].lstrip()
        Arg+=Item[0]
        if len(Item)>1:
          Item[0]=Arg
          Args.extend(Item[:-1])
          Arg=Item[-1]
        DeleteSpaces=(Arg=="")
    Args.append(Arg)
    I=ParsedInput[0]
    ParsedInput[0]=I[:3]+(Args,)
    return
  SomethingDone=0
  while ParsedInput:
    Inst=ParsedInput[0]
    if Expanding==0:
      Start=Inst[1]
      End=Inst[2]
      if Start==End:
        Expanding=1
        return
      PushStatus()
      Input=Input[Start:End-1]
      ParsedInputString=None
      Expanding=None
      OutBuffer=""
      return
    if Expanding==1:
      Result=ExpandMacro(Inst[0],Inst[3])
      ParsedInput[0]=Inst[:3]+(Result,Inst[3])
      Expanding=2
      CurrentArg=1
      return
    if Expanding==2:
      Args=Inst[3]
      if CurrentArg<len(Args):
        Args[CurrentArg]=ExpandArgExpr(Args[CurrentArg][0],Inst[0],Inst[4])
        CurrentArg+=2
        return
      End=Inst[2]
      Result=''.join(Args)
      if Result!='' or End!=0:
        Input=Result+Input[End:]
      if Input is ParsedInputString:
        del ParsedInput[0]
      Expanding=None
      return
    if type(Inst) is type("") or type(Inst) is type([]):
      if type(Inst) is type([]):
        SomethingDone=1
        Inst=UnparseQuotes(Inst)
        if Stack:
          Inst=QuoteL+Inst+QuoteR
      if Inst:
        SomethingDone=1
        SendOutput(Inst)
    else:
      if SomethingDone:
        return
      Expanding=0
      return
    del ParsedInput[0]
