## -------------------------- ##
##   About the declarations   ##
## -------------------------- ##

# All GPS source files use special signatures to indicate the calling format
# of the macros being declared. These special signatures are quoted by
# special lines, which contain these characters (in the order as mentioned):
# a hash mark, a space, 73 dashes, a space, a hasmark and finally the
# newline. Between these lines each line contains a prototype of one macro.
#
# The macro prototype starts with the name of the macro. The name may contain
# only the english letters, digits and underscores and must start with a
# letter.
#
# After the name there is the list of comma separated arguments. The list is
# enclosed in brackets and the arguments declare. Each argument 
#
# In the following examples assume these macros:
#
# define(`abc',`ABC')
#
# - string
#   A string. In the definition of the macro each reference to such argument
#   must be doublequoted. In the invocation of the macro any quoted text at
#   the place of this argument must be a raw string. In such a raw string no
#   already defined macros are allowed.
#
#   Example:
#   define(`FOO',`bar $1')dnl Wrong: $1 must be doublequoted
#   define(`FOO',``bar `$1'')
#
#   FOO(abc)
#   =>ABC
#   FOO(`abc')
#   =>abc
#   FOO(``abc'')
#   =>`abc'
#   FOO(```abc''')
#   =>``abc''
#
# - script
#   A script. In the definition of the macro each reference to such an
#   argument must be singlequoted. In the invocation of the macro a text must
#   be doublequoted.
#
#   Example:
#   define(`FOO',`bar `$1'')dnl Wrong: $1 must not be doublequoted
#   define(`FOO',`bar $1')
#
#   FOO(abc)
#   =>ABC
#   FOO(`abc')
#   =>ABC
#   FOO(``abc'')
#   =>abc
#   FOO(```abc''')
#   =>`abc'
#
# - newname
#   A name of a new macro. When the macro is invoked, the content of this
#   argument must be constant and must be a name. A prototype must be
#   supplied for each such new name. The prototype may be included either in
#   the prototype block of the declarator (in the form DECL declares
#   NAME(ARGS), where DECL is the comma separated list of names of the
#   argument which form the declaration of the NAME, NAME if the name of the
#   'newname' argument and ARGS is the optional description of the arguments
#   of NAME) or each usage of the macro must be preceeded by the prototype
#   block stating the prototype for NAME

## ----------------------------------- ##
##   Additional branching constructs   ##
## ----------------------------------- ##

# Both 'm4_ifval' and 'm4_ifset' tests against the empty string.  The
# difference is that 'm4_ifset' is specialized on macros.
#
# In case of arguments of macros, eg $[1], it makes little difference.
# In the case of a macro 'FOO', you don't want to check 'm4_ifval(FOO,
# TRUE)', because if 'FOO' expands with commas, there is a shifting of
# the arguments.  So you want to run 'm4_ifval([FOO])', but then you just
# compare the *string* 'FOO' against '', which, of course fails.
#
# So you want a variation of 'm4_ifset' that expects a macro name as $[1].
# If this macro is both defined and defined to a non empty value, then
# it runs TRUE etc.

# ------------------------------------------------------------------------- #
# m4_ifval(COND:string, [TRUE:script], [FALSE:script])
# ------------------------------------------------------------------------- #
# If COND is not the empty string, expand IF-TRUE, otherwise IF-FALSE.
# Comparable to m4_ifdef.
m4_define([m4_ifval], [m4_ifelse([$1], [], [$3], [$2])])

# ------------------------------------------------------------------------- #
# m4_ifset(MACRO:oldname, [TRUE:script], [FALSE:script])
# ------------------------------------------------------------------------- #
# If MACRO has no definition, or of its definition is the empty string,
# expand IF-FALSE, otherwise IF-TRUE.
m4_define([m4_ifset],[m4_ifdef(
  [$1],
  [m4_ifelse(m4_defn([$1]), [], [$3], [$2])],
  [$3]
)])

## -------------------------- ##
##   Additional misc macros   ##
## -------------------------- ##

m4_define([m4__for],[m4_ifelse(
  $#, 0, [],
  $#, 1, [m4_ifelse($1,[],[],[m4__for__function([$1])])],
  [m4__for__function([$1])][m4__for(m4_shift($@))]dnl
)])

m4_define([m4_for],
[dnl
m4_pushdef([m4__for__function],[$2($][1)])dnl
m4__for($1)dnl
m4_popdef([m4__for__function])dnl
])

# m4__cat(STRING, ...)
# --------------------
# This internal macro concatenates all its arguments (in sequence, of
# course). It is used in the advanced macro construction kit, which is useful
# for making your macros more structured and readable by dropping unnecessary
# dnl's and having them indented properly and consistently.
m4_define([m4__cat],[m4_ifelse(
  $#, 0, [],
  $#, 1, [[$1[]]],
  [[$1[]][]m4__cat(m4_shift($@))]dnl
)])

# m4__textcat(STRING, ...)
# ------------------------
m4_define([m4__textcat],[m4_ifelse(
  $#, 0, [],
  $#, 1, [[$1]],
  [[$1]m4__textcat(m4_shift($@))]dnl
)])

# m4_prog(NAME, COMMANDS)
# -------------------
# Define a m4 program. This macro differs from m4_define in that fashion that
# the second parameter is a comma separated list of quoted commands, in which
# the whitespace between a comma (or at the beginning of the string) and the
# command itself is discarded. To discard whitespace after the last command,
# place a comma after it.

m4_define([m4_prog],[m4_define([$1],m4__cat($2))])

m4_define([m4_quote],[[[$1]]])
m4_define([m4_unquote],[$1])
m4_define([m4__qcat],[m4_quote(m4__cat($1))])
m4_define([m4__qtcat],[m4_quote(m4__textcat($1))])

m4_define([m4_subprog],[[m4_define([$1],]m4_quote(m4__cat($2))[)]])

# m4_if(NAME, CASES)
# ------------------
# Works as m4_ifelse except the second argument is a comma-separated list of
# commands. Unlike m4_ifelse, can be nicely formatted.
m4_define([m4__if],[m4_ifelse(
  $#, 4, [[$1,$2,]m4__qcat([$3])],
  $#, 5, [[$1,$2,]m4__qcat([$3])[,]m4__qcat([$4])],
  [[$1,$2,]m4__qcat([$3])[,]m4__if(m4_shift(m4_shift(m4_shift($@))))]dnl
)])

m4_define([m4_if],[[m4_ifelse(]m4__if($@)[)]])

m4_define([NL],[
])
m4_define([m4_text],[m4__qtcat([$@])])
