#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_DEPENDENCIES 16
#define MAX_FILES 64

FILE *File;
int WasEOF;
int VerboseMakefile;

void PrintErr(const char *Msg)
{
  fputs(Msg,stderr);
}

void Error(const char *Msg)
{
  fprintf(stderr,"%s\n",Msg);
  exit(255);
}

void OpenFile(const char *Name,const char *Mode)
{
  File=fopen(Name,Mode);
  WasEOF=0;
  if (File==NULL) {
    PrintErr("Unable to open ");
    Error(Name);
  }
}

void Close(FILE *What)
{
  if (fclose(What)!=0) {
    Error("Unable to close a file");
  }
}

void CloseFile(void)
{
  Close(File);
}

char ReadChar(void)
{
  int Byte;

  if (WasEOF) return('\0');
  Byte=fgetc(File);
  if (Byte==EOF) {
    WasEOF=1;
    Byte='\0';
  }
  return(Byte);
}

void SkipRestOfLine(void)
{
  char Ch;

  while (!WasEOF) {
    Ch=ReadChar();
    if (Ch=='\n') break;
  }
}

typedef struct tagTFileInfo {
  char Name[12];
  struct tagTFileInfo *DepList[MAX_DEPENDENCIES];
  int DepCount;
  struct tagTFileInfo *Next;
} TFileInfo;

TFileInfo Files[MAX_FILES];
int FileCnt;
int CurrentFile;

void Initialize(void)
{
  FileCnt=0;
  CurrentFile=0;
}

TFileInfo *CreateFile(const char *Name)
{
  TFileInfo *I;
  int i;
  for (i=0;i<FileCnt;i++) {
    I=&Files[i];
    if (strcmp(Name,I->Name)==0) return(I);
  }
  if (FileCnt==MAX_FILES) Error("Too many files");
  I=&Files[FileCnt];
  FileCnt++;
  I->Name[11]='\0';
  strncpy(I->Name,Name,12);
  if (I->Name[11]!='\0') Error("File name too long");
  I->DepCount=0;
  I->Next=NULL;
  return(I);
}

void CreateFileIfExists(const char *Name)
{
  FILE *Test;
  Test=fopen(Name,"rt");
  if (Test!=NULL) {
    Close(Test);
    CreateFile(Name);
  }
}

void AnalyzeFileContent(void)
{
  const char *Ptr;
  char *WPtr;
  char Ch;
  char NameBuf[12];
  TFileInfo *C;
  TFileInfo *I;

  C=&Files[CurrentFile];
  while (!WasEOF) {
    Ptr="#include \"";
    while ((*Ptr)!='\0') {
      Ch=ReadChar();
      if (WasEOF || (*Ptr)!=Ch) break;
      Ptr++;
    }
    if ((*Ptr)=='\0') {
      WPtr=NameBuf;
      for (;;) {
        Ch=ReadChar();
        if (WasEOF || Ch=='"') break;
        if (WPtr==NameBuf+11) Error("#include file name too long");
        *(WPtr++)=Ch;
      }
    }
    if (Ch=='"') {
      *WPtr='\0';
      if (WPtr<NameBuf+2) {
        PrintErr("Filename is too short: ");
        Error(NameBuf);
      }
      if (*(WPtr-2)!='.' || *(WPtr-1)!='h') {
        PrintErr("File is not a header file: ");
        Error(NameBuf);
      }
      I=CreateFile(NameBuf);
      if (C->DepCount==MAX_DEPENDENCIES) {
        PrintErr("Too many dependencies for ");
        Error(C->Name);
      }
      C->DepList[C->DepCount++]=I;
      *(WPtr-1)='c';
      CreateFileIfExists(NameBuf);
    }
    if (Ch!='\n') SkipRestOfLine();
  }
}

void AnalyzeCurrentFile(void)
{
  TFileInfo *I;

  I=&Files[CurrentFile];
  OpenFile(I->Name,"rt");
  AnalyzeFileContent();
  CloseFile();
}

void AnalyzeFiles(void)
{
  while (CurrentFile<FileCnt) {
    AnalyzeCurrentFile();
    CurrentFile++;
  }
}

void Print(const char *Msg)
{
  fputs(Msg,File);
}

static void NewRule(void)
{
  Print("\n\n");
}

static void StartNewCommand(int Echo)
{
  Print("\n\t");
  if (!Echo) Print("@");
}

static void NewCommand(void)
{
  StartNewCommand(VerboseMakefile);
}

void PrintMessageCommand(const char *Prefix,const char *Name)
{
  if (!VerboseMakefile) {
    StartNewCommand(0);
    Print("echo ");
    Print(Prefix);
    Print(Name);
  }
}

void PrintRestOfRule(TFileInfo *I)
{
  TFileInfo *D;
  int i;

  Print(": ");
  Print(I->Name);
  for (i=0;i<I->DepCount;i++) {
    Print(" ");
    D=I->DepList[i];
    Print(D->Name);
  }
  PrintMessageCommand("compiling makegen/",I->Name);
  NewCommand();
  Print("$(CC) -c ");
  Print(I->Name);
}

void OutputObjectFiles(int OnlyList)
{
  TFileInfo *I;
  int i;
  int Len;

  for (i=0;i<FileCnt;i++) {
    I=&Files[i];
    Len=strlen(I->Name);
    if (I->Name[Len-1]=='c') {
      I->Name[Len-1]='o';
      if (OnlyList) {
        Print(" ");
      } else {
        NewRule();
      }
      Print(I->Name);
      I->Name[Len-1]='c';
      if (!OnlyList) PrintRestOfRule(I);
    }
  }
}

void GenMakeFile(void)
{
  OpenFile("makefile","wt");
  NewRule();
  Print("makegen:");
  OutputObjectFiles(1);
  PrintMessageCommand("linking makegen/makegen","");
  NewCommand();
  Print("$(LINK) -o makegen");
  OutputObjectFiles(1);
  OutputObjectFiles(0);
  Print("\n");
  CloseFile();
}

int main(int argc,const char **argv)
{
  const char *Arg;

  Arg=argv[1];
  VerboseMakefile=Arg!=NULL && Arg[0]=='-' && Arg[1]=='v' && Arg[2]=='\0';
  Initialize();
  CreateFile("makegen.c");
  AnalyzeFiles();
  GenMakeFile();
  return(0);
}
