#include <stdio.h>
#include <string.h>

#include "types.h"
#include "filein.h"

#define MAX_FILENAME_SIZE 128
#define MAX_LINE_SIZE 128

static char FName[MAX_FILENAME_SIZE];
static bool ReportFileName;

bool WereErrors;
char Line[MAX_LINE_SIZE];

static int LineNum;
static char *ColPtr;

static int TokLine;
static int TokCol;

static bool SkipComments;
static bool LineEndIsComment;

void Fatal(const char *Msg)
{
  ReportFileReadingError(Msg);
}

static void ReportErrorEx(const char *Msg,int LineNum,int ColNum)
{
  WereErrors=true;
  if (ReportFileName) {
    ReportFileName=false;
    fprintf(stderr,"In file '%s'\n",FName);
  }
  fprintf(stderr,"L %5d  ",LineNum);
  if (ColNum!=-1) {
    fprintf(stderr,"C %3d  ",ColNum);
  } else {
    fprintf(stderr,"       ");
  }
  fprintf(stderr,"%s\n",Msg);
}

void ReportErrorPos(const char *Msg,int Col)
{
  ReportErrorEx(Msg,LineNum,Col);
}

static void ReportErrorLO(const char *Msg)
{
  ReportErrorPos(Msg,-1);
}

int GetCurrentPos(void)
{
  return(ColPtr-Line+1);
}

int GetTokenPos(void)
{
  return(TokCol);
}

void LineError(const char *Msg)
{
  ReportErrorPos(Msg,-1);
}

void ReportError(const char *Msg)
{
  ReportErrorPos(Msg,GetCurrentPos());
}

void TokError(const char *Msg)
{
  ReportErrorEx(Msg,TokLine,GetTokenPos());
}

bool OpenFile(const char *FileName)
{
  int Result;

  Result=OpenTheFile(FileName);
  if (Result) {
    strncpy(FName,FileName,MAX_FILENAME_SIZE-1);
    FName[MAX_FILENAME_SIZE-1]='\0';
    ReportFileName=true;
    LineNum=0;
  }
  SkipComments=false;
  return(Result);
}

bool OpenCFile(const char *FileName)
{
  int Result;

  Result=OpenFile(FileName);
  SkipComments=true;
  LineEndIsComment=false;
  return(Result);
}

void ClearComments(void)
{
  char Ch;
  bool IsHalf;

  ColPtr=Line;
  IsHalf=false;
  while (true) {
    Ch=*ColPtr;
    if (Ch=='\0') break;
    if (LineEndIsComment) {
      if (IsHalf) LineEndIsComment=(Ch!='/');
      IsHalf=(Ch=='*');
      if (Ch!='\t') *ColPtr=' ';
    } else {
      if (IsHalf) LineEndIsComment=(Ch=='*');
      IsHalf=(Ch=='/');
      if (LineEndIsComment) {
        if (*(ColPtr+1)=='/') ReportError("Invalid comment");
      }
    }
    ColPtr++;
  }
}

void FetchLine(void)
{
  Line[MAX_LINE_SIZE-1]='\0';
  ReadLine(Line,MAX_LINE_SIZE);
  if (Line[MAX_LINE_SIZE-1]!='\0') {
    ReportErrorLO("Line too long; truncating ...");
    Line[MAX_LINE_SIZE-1]='\0';
  }
  LineNum++;
  if (SkipComments) ClearComments();
  ColPtr=Line;
}

char GetByte(void)
{
  char Ch;

  Ch=*ColPtr;
  if (Ch=='\0') Ch='\n';
  if (Ch!='\n') ColPtr++;
  return(Ch);
}

void GoBack(void)
{
  ColPtr--;
}

int IsBlank(char Ch)
{
  return(Ch=='\t' || Ch=='\r' || Ch==' ');
}

bool IsIDChar(char Ch)
{
  return(
    (Ch=='_') ||
    (Ch>='0' && Ch<='9') ||
    (Ch>='A' && Ch<='Z') ||
    (Ch>='a' && Ch<='z')
  );
}

void StripBlanksAndComments(void)
{
  char *Ptr;
  char Ch;

  Ptr=Line;
  while (true) {
    Ch=*Ptr;
    if (Ch=='#' || Ch=='\0') break;
    Ptr++;
  }
  if (Ptr>Line) {
    while (Ptr>Line) {
      Ch=*(--Ptr);
      if (!IsBlank(Ch)) break;
    }
    Ptr++;
  }
  *Ptr='\0';
}

void SkipBlanks(void)
{
  char Ch;
  
  do {
    Ch=GetByte();
  } while (IsBlank(Ch));
  if (Ch!='\n') GoBack();
}

char *ReadWordIntoBuf(char *Buf,unsigned int BufSize)
{
  char Ch;
  char *Ptr;

  Ptr=Buf;
  SkipBlanks();
  TokLine=LineNum;
  TokCol=ColPtr-Line+1;
  do {
    Ch=GetByte();
    if (Ch==' ' || Ch=='\t' || Ch=='\n' || Ch=='\0') break;
    if (BufSize>0) {
      if ((--BufSize)==1) {
        TokError("Token too long");
	continue;
      }
      *(Buf++)=Ch;
    }
  } while (1);
  *Buf='\0';
  return(Buf);
}

bool EndOfFile(void)
{
  return(IsEOF() && *ColPtr=='\0');
}

void CloseFile(void)
{
  CloseTheFile();
}
