#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "types.h"
#include "files.h"
#include "utils.h"
#include "fileread.h"
#include "dictrb.h"

bool IsValid8dot3Char(char Ch)
{
  return((Ch>='0' && Ch<='9') || (Ch>='a' && Ch<='z'));
}

bool CheckCorrect8dot3(const char *FileName)
{
  int NameLen;
  int ExtLen;
  char Ch;

  NameLen=0;
  while (true) {
    Ch=*(FileName++);
    if (!IsValid8dot3Char(Ch)) break;
    NameLen++;
  }
  if (Ch!='.' && Ch!='\0') return(false);
  ExtLen=0;
  if (Ch=='.') {
    while (IsValid8dot3Char(*(FileName++))) ExtLen++;
    if (ExtLen==0) return(false);
  }
  return(NameLen>0 && NameLen<9 && ExtLen<3);
}

#define stkFilesToBeProcessed 0
#define stkDependencyFiles 1
#define stkIncludedFiles 2
#define stkAccessedFiles 2

static TRBDict *FileDict;
static TFileInfo *StkTop[MAX_FILE_STACKS];
static TFileInfo *StkUsed[MAX_FILE_STACKS];

static void InitFileStack(int StackNum)
{
  assert(StkTop[StackNum]==NULL);
  StkTop[StackNum]=(void *)StkTop;
  StkUsed[StackNum]=(void *)StkTop;
}

static bool PushFileOntoStack(int StackNum,TFileInfo *F)
{
  bool Result;

  assert(StkTop[StackNum]!=NULL);
  Result=F->StkNext[StackNum]==NULL;
  if (Result) {
    F->StkNext[StackNum]=StkTop[StackNum];
    StkTop[StackNum]=F;
  }
  return(Result);
}

static TFileInfo *PopFileOffStack(int StackNum)
{
  TFileInfo *F;

  assert(StkTop[StackNum]!=NULL);
  F=StkTop[StackNum];
  if (F==(void *)StkTop) {
    F=NULL;
  } else {
    StkTop[StackNum]=F->StkNext[StackNum];
    F->StkNext[StackNum]=StkUsed[StackNum];
    StkUsed[StackNum]=F;
  }
  return(F);
}

static void CleanupFileStack(int StackNum)
{
  TFileInfo *F;
  TFileInfo *Next;

  assert(StkTop[StackNum]==(void *)StkTop);
  StkTop[StackNum]=NULL;
  F=StkUsed[StackNum];
  while (F!=(void *)StkTop) {
    Next=F->StkNext[StackNum];
    F->StkNext[StackNum]=NULL;
    F=Next;
  }
}

/*********************************
*** File registration handling ***
*********************************/

void InitializeFiles(void)
{
  FileDict=CreateDict();
  if (FileDict==NULL) Fatal("Unable to create dictionary of files");
  InitFileStack(stkFilesToBeProcessed);
}

TFileInfo *RegisterFile(const char *FileName)
{
  TFileInfo *F;
  char *Name;
  bool Bad;
  int i;

  F=SearchKey(FileDict,FileName);
  if (F==NULL) {
    F=malloc(sizeof(*F));
    Name=StrDup(FileName);
    Bad=true;
    if (Name!=NULL) {
      Bad=(AddItemWithKey(FileDict,Name,F)!=rbSUCCESS);
      if (Bad) free(Name);
    }
    if (Bad) {
      free(F);
      F=NULL;
    } else {
      F->FileName=Name;
      F->DepCount=0;
      F->DepList=NULL;
      F->Source=NULL;
      for (i=0;i<MAX_FILE_STACKS;i++) F->StkNext[i]=NULL;
    }
  }
  return(F);
}

bool RegisterFileToBeProcessed(TFileInfo *F)
{
  assert(F!=NULL);
  return(PushFileOntoStack(stkFilesToBeProcessed,F));
}

TFileInfo *GetNextFileToBeProcessed(void)
{
  return(PopFileOffStack(stkFilesToBeProcessed));
}

void ReinitializeFileProcessingStack(void)
{
  CleanupFileStack(stkFilesToBeProcessed);
  InitFileStack(stkFilesToBeProcessed);
}

void InitFileAccessStack(void)
{
  InitFileStack(stkAccessedFiles);
}

void RegisterAccessedFile(TFileInfo *F)
{
  PushFileOntoStack(stkAccessedFiles,F);
}

TFileInfo *GetNextAccessedFile(void)
{
  return(PopFileOffStack(stkAccessedFiles));
}

void DestroyFileAccessStack(void)
{
  CleanupFileStack(stkAccessedFiles);
}

/***********************************
*** File dependency registration ***
***********************************/

static TFileInfo *Current;

void ActivateFile(TFileInfo *F)
{
  Current=F;
  InitFileStack(stkDependencyFiles);
  InitFileStack(stkIncludedFiles);
}

bool RegisterDependencyFile(TFileInfo *F,bool Active)
{
  bool Result;

  if (Active) {
    Result=PushFileOntoStack(stkDependencyFiles,F);
    if (Result) Current->DepCount++;
  }
  Result=PushFileOntoStack(stkIncludedFiles,F);
  if (Result) PopFileOffStack(stkIncludedFiles);
  return(Result);
}

static void ReportDepListProblem(TFileInfo *F)
{
  char Msg[256];

  strcpy(Msg,"Unable to create dependency file list for ");
  strcpy(Msg+strlen(Msg),F->FileName);
  Fatal(Msg);
}

void DeactivateFile(void)
{
  TFileInfo *F;
  int i;

  if (Current->DepCount>0) {
    Current->DepList=malloc(Current->DepCount*sizeof(*Current->DepList));
    if (Current->DepList==NULL) ReportDepListProblem(Current);
    for (i=0;i<Current->DepCount;i++) {
      F=PopFileOffStack(stkDependencyFiles);
      assert(F!=NULL);
      Current->DepList[i]=F;
    }
  }
  CleanupFileStack(stkDependencyFiles);
  CleanupFileStack(stkIncludedFiles);
  Current=NULL;
}

static bool FileExists(const char *FileName)
{
  FILE *F;
  bool Result;

  F=fopen(FileName,"r");
  Result=(F!=NULL);
  if (Result) {
    if (fclose(F)!=0) {
      TokError("Error closing the file");
    }
  }
  return(Result);
}

TFileInfo *RegisterModuleIdentifiedByHeader(char *Name)
{
  TFileInfo *F;
  TFileInfo *OF;
  int Len;
  bool IsDefsGh;

  Len=strlen(Name);
  if (Len>12 || !CheckCorrect8dot3(Name)) {
    TokError("File name not 8.3 compatible");
    return(NULL);
  }
  IsDefsGh=(strcmp(Name,"defs.gh")==0);
  if (!IsDefsGh) {
    if (Name[Len-1]!='h' || Name[Len-2]!='.') { 
      TokError("File is not a header file");
      return(NULL);
    }
    if (!FileExists(Name)) {
      TokError("Header does not exist");
      return(NULL);
    }
  }
  F=RegisterFile(Name);
  if (F==NULL) {
    TokError("Error recording this file as dependency");
    return(NULL);
  }
  if (IsDefsGh)  return(F);
  if (!RegisterFileToBeProcessed(F)) return(F);
  Name[Len-1]='c';
  if (FileExists(Name)) {
    OF=RegisterFile(Name);
    if (OF!=NULL) {
      F->Source=OF;
      assert(RegisterFileToBeProcessed(OF));
    }
  }
  return(F);
}
