#include <string.h>

#include "types.h"
#include "fileread.h"
#include "filein.h"
#include "files.h"
#include "dictrb.h"
#include "prepsym.h"
#include "modseq.h"

#define MAX_CONDITIONAL_STACK_DEPTH 32

typedef enum {
  /*
    The code line is not in a block.
  */
  stNormal,
  /*
    The code line is in an "if" block.
  */
  stIf,
  /*
    The code line is in an "elif" block.
  */
  stElif,
  /*
    The code line is in an "else" block.
  */
  stElse,
  /*
    The line is in the IfDef block.
  */
  stIfDef
} TPrepStatus;

typedef struct {
  /*
    True means the code is in effect, False means the code is being stripped
    out by the preprocessor. If this is False for a line, the files mentioned
    on the line (in an #include directive) are not considered dependencies of
    the current file but they are internally recorded for scan as well. The
    reason for this is that I want the whole source of the library being
    checked for errors.
  */
  bool Active;
  /*
    Status of the preprocessing engine.
  */
  TPrepStatus Status;
  /*
    After an #elif with a condition that is evaulated to be true this is set
    to True. This causes the following #elif blocks and the #else block to be
    ignored.
  */
  bool ForceIgnore;
  /*
    In some circumstances the symbolspace changes are not allowed.
  */
  bool SymbolspaceFrozen;
} TDefineStackItem;

static bool Active;
static TPrepStatus Status;
static bool ForceIgnore;
static bool SymbolspaceFrozen;
static int DefStackTop;
static TDefineStackItem DefStack[MAX_CONDITIONAL_STACK_DEPTH];

static void ReinitializePreprocessor(void)
{
  Status=stNormal;
  ForceIgnore=false;
}

static void PushPrepStatus(void)
{
  TDefineStackItem *S;

  if (DefStackTop==MAX_CONDITIONAL_STACK_DEPTH) {
    LineError("Conditionals nested too deeply");
    return;
  }
  S=&DefStack[DefStackTop++];
  S->Active=Active;
  S->Status=Status;
  S->ForceIgnore=ForceIgnore;
  S->SymbolspaceFrozen=SymbolspaceFrozen;
}

static void PopPrepStatus(void)
{
  TDefineStackItem *S;

  if (DefStackTop==0) {
    LineError("#elif/#else/#endif without #if");
    return;
  }
  S=&DefStack[--DefStackTop];
  Active=S->Active;
  Status=S->Status;
  ForceIgnore=S->ForceIgnore;
  SymbolspaceFrozen=S->SymbolspaceFrozen;
}

static bool CheckNotInSpecialConditionBlock(void)
{
  if (Active==-1) {
    LineError("Directive not allowed in special conditional");
    return(false);
  }
  return(true);
}

static void ProcessModuleReference(char *HeaderName)
{
  TFileInfo *F;

  /*
    Register the module and obtain an unique structure for it. Use this
    structure to ensure the file was #included only once. If not, report
    an error and ignore the directive.

    If the file is not "defs.gh" (which has no source code module),
    determine the source code module file name, register it and push its
    unique structure onto the stack of files to be processed. But before
    the registration ensure the file really exists.
  */
  F=RegisterModuleIdentifiedByHeader(HeaderName);
  if (F==NULL) return;
  /*
    If the current line is active, record the file unique structure as a
    dependency of the file currently being processed. If this fails, the
    file is already registered, so report an error in that case.
  */
  if (!RegisterDependencyFile(F,Active)) {
    TokError("File imported more than once");
  }
  /*
    There are special requirements on the sequence of the module import
    commands in this package which must be checked. These checks are
    performed in a procedure from a different module.
  */
  CheckModuleImportSeq(F);
  /*
    Done.
  */
}

static void ParseImportFileCommandArgument(char *Cmd)
{
  unsigned int Len;
  char *Ext;

  /*
    The #include statements are not allowed in the special condition blocks.
    If it is in such a block, it will be reported and ignored.
  */
  if (!CheckNotInSpecialConditionBlock()) return;
  /*
    There are two types of #include commands.

    The first is #include <{filename}>. In SYSLIB they import platform
    specific header files about which we have absolutely no information.
    Also the format of the {filename} string is completely undefined.
    Therefore they are ignored here. We don't use library tracking for this
    library; that is the job for GPS.

    The second is #include "{filename}". The {filename} is then expected to
    be in the form {module}.h and {module} is expected to be a name of a
    local module. The file mentioned here and the source code file it belongs
    to it are recorded for analysis. However the file can be added into the
    dependency list of the current file only when the line on which the
    directive resides is active. This way we can rescan the whole SYSLIB
    source code but avoid pieces of the source which handle platforms other
    then the one we are building for.
  */
  Len=strlen(Cmd);
  if (Cmd[0]=='<' && Cmd[Len-1]=='>') {
    return;
  } else if (Cmd[0]=='"' && Cmd[Len-1]=='"') {
    /*
      Strip out the trailing quotation mark.
    */
    Cmd[Len-1]='\0';
    /*
      Check if the file is a header file. Either the file is "defs.gh" or
      the extension must be ".h" since the source code is written in C.
      The "defs.gh" file is a special autogenerated file that is #include-d
      by most if not all of the files in the SYSLIB source. It does not
      induce additional modules being added to the dependencies but it must
      be recorded as a dependency.
    */
    if (strcmp(Cmd+1,"defs.gh")!=0) {
      Ext=Cmd+Len-3;
      if (strcmp(Ext,".h")!=0) {
        TokError("Malformed local module reference");
        return;
      }
    }
    /*
      Process the module identified by the header we just checked.
    */
    ProcessModuleReference(Cmd+1);
  } else {
    LineError("Malformed #include directive");
  }
}

static void CheckSymbolspaceNotFrozen(void)
{
  if (Active<-1) {
    ReportError("Symbolspace is frozen here");
  }
}

static bool CheckIfSpecialSymbol(const char *Symbol)
{
  return(
    !strcmp(Symbol,"NULL") ||
    !strcmp(Symbol,"__DEBUG__") || !strcmp(Symbol,"__NO_DEBUG__") ||
    !strcmp(Symbol,"__INTERFACE__") || !strcmp(Symbol,"extern")
  );
}

static void ParseDefine(void)
{
  char Buffer[128];
  char Value[128];

  ReadWordIntoBuf(Buffer,128);
  if (Buffer[0]!='_') return;
  if (CheckIfSpecialSymbol(Buffer)) return;
  CheckNotInSpecialConditionBlock();
  CheckSymbolspaceNotFrozen();
  ReadWordIntoBuf(Value,128);
  DefineTempSymbol(Buffer,Value,Active);
}

static void ParseUndef(void)
{
  char Buffer[128];

  ReadWordIntoBuf(Buffer,128);
  if (Buffer[0]!='_') return;
  if (CheckIfSpecialSymbol(Buffer)) return;
  CheckNotInSpecialConditionBlock();
  CheckSymbolspaceNotFrozen();
  UndefineTempSymbol(Buffer,Active);
}

static void CheckEOL(void)
{
  char Buf[2];

  ReadWordIntoBuf(Buf,2);
  if (Buf[0]!='\0') {
    ReportError("Extra characters on line");
  }
}

static int ParseCondition(void)
{
  char Symbol[128];
  char Operand[4];
  char Value[128];
  bool Negate;
  bool Result;

  ReadWordIntoBuf(Symbol,128);
  if (!strcmp(Symbol,"__SPECIAL__")) return(-1);
  ReadWordIntoBuf(Operand,4);
  if (Operand[0]=='\0') {
    Negate=(Symbol[0]=='!');
    Result=CheckSymbol(Negate?Symbol+1:Symbol,NULL);
  } else {
    Negate=!strcmp(Operand,"!=");
    if (!Negate) {
      if (strcmp(Operand,"==")) ReportError("Invalid operator");
    }
    ReadWordIntoBuf(Value,128);
    Result=CheckSymbol(Symbol,Value);
  }
  CheckEOL();
  if (Negate) Result=!Result;
  return(Result);
}

static void ParseCondBlockStart(bool IsElif)
{
  int Result;
  bool ActualResult;

  /*
    Parse the condition. This must be done even when the code is not active
    because the condition parser checks for errors and displays them.
  */
  Result=ParseCondition();
  if (Result==-1) {
    ActualResult=-1;
    if (IsElif) ReportError("Special conditions not allowed in #elif");
  } else {
    ActualResult=!ForceIgnore && Result;
  }
  ForceIgnore=ForceIgnore || Result;
  PushPrepStatus();
  if (ActualResult==-1) {
    Active=-1;
  } else {
    Active=Active && ActualResult;
  }
  ReinitializePreprocessor();
}

static void NotAllowed(void)
{
  ReportError("Directive not allowed here");
}

static void ParseIf(void)
{
  /*
    Some sanity checks
  */
  CheckNotInSpecialConditionBlock();
  /*
    The status must be "Normal" here and it changes to "if".
  */
  if (Status!=stNormal) {
    NotAllowed();
    return;
  }
  Status=stIf;
  /*
    The rest is common to #if and #elif.
  */
  ParseCondBlockStart(false);
}

static bool CheckIfElif(void)
{
  if (Status!=stIf && Status!=stElif) {
    NotAllowed();
    return(false);
  }
  return(true);
}

static void ParseElif(void)
{
  /*
    Some sanity checks
  */
  CheckNotInSpecialConditionBlock();
  /*
    Pop the previous status
  */
  PopPrepStatus();
  /*
    Update the status.
  */
  if (!CheckIfElif()) return;
  Status=stElif;
  /*
    The rest is common to #if and #elif.
  */
  ParseCondBlockStart(true);
}

static void ParseElse(void)
{
  bool NewActivity;

  /*
    Some sanity checks
  */
  CheckNotInSpecialConditionBlock();
  /*
    The #else block is active when the #if block isn't and vice versa.
    Determine the activity of the block before the information about the
    activity of the previous block is lost.
  */
  NewActivity=!Active;
  /*
    Update the status variable of the preprocessor on the top of the stack
    and setup the activity flag. Remember that the block is not active when
    the ForceIgnore flag is set.
  */
  PopPrepStatus();
  if (Status!=stIfDef) {
    if (!CheckIfElif()) return;
  }
  Status=stElse;
  PushPrepStatus();
  Active=!ForceIgnore && NewActivity;
  ReinitializePreprocessor();
}

static void ParseEndif(void)
{
  /*
    Simply pop the previous status.
  */
  PopPrepStatus();
  /*
    Update the status.
  */
  Status=stNormal;
  /*
    True conditions in the conditionals are not ignored anymore.
  */
  ForceIgnore=false;
}

static void ParseIfNDef(void)
{
  char Symbol[128];

  if (Status!=stNormal) {
    NotAllowed();
    return;
  }
  Status=stIfDef;
  ReadWordIntoBuf(Symbol,128);
  if (!CheckIfSpecialSymbol(Symbol)) {
    ReportError("Invalid symbol reference");
  };
  PushPrepStatus();
  ReinitializePreprocessor();
  SymbolspaceFrozen=true;
}

bool ScanCLine(void)
{
  char Buffer[128];

  /*
    If the EOF flag is set, return with an EOF condition.
  */
  if (IsEOF()) return(false);
  /*
    Fetch next line.
  */
  FetchLine();
  /*
    Read the first word into the buffer.
  */
  ReadWordIntoBuf(Buffer,128);
  /*
    Check whether there are any blank characters at the beginning of the
    line. If so, we don't process the line.
  */
  if (GetTokenPos()!=1) {
    if (Buffer[0]=='#') {
      TokError("Preprocessor directives must be at column 1");
    }
  }
  /*
    We are interested in the most preprocessor directives that can appear in
    the file. The most interesting one is the "#include", which in fact means
    "import". However the conditional compilation directives are also
    interesting, because they (among other things) determine if the #includes
    have any effects.
  */
  if (!strcmp(Buffer,"#include")) {
    ReadWordIntoBuf(Buffer,128);
    ParseImportFileCommandArgument(Buffer);
  } else if (!strcmp(Buffer,"#define")) {
    ParseDefine();
  } else if (!strcmp(Buffer,"#undef")) {
    ParseUndef();
  } else if (!strcmp(Buffer,"#if")) {
    ParseIf();
  } else if (!strcmp(Buffer,"#elif")) {
    ParseElif();
  } else if (!strcmp(Buffer,"#else")) {
    ParseElse();
  } else if (!strcmp(Buffer,"#endif")) {
    ParseEndif();
  } else if (!strcmp(Buffer,"#ifdef") || !strcmp(Buffer,"#ifndef")) {
    ParseIfNDef();
  } else if (Buffer[0]=='#') {
    ReportError("Malformed preprocessor directive");
  }
  /*
    Return that no EOF was encountered.
  */
  return(true);
}

static bool OpenAndReportIfFailed(TFileInfo *F)
{
  char Msg[4096];
  bool Result;

  Result=OpenCFile(F->FileName);
  if (!Result) {
    strcpy(Msg,"Unable to open ");
    strcpy(Msg+strlen(Msg),F->FileName);
    ReportError(Msg);
  }
  return(Result);
}

static void InitTempSymbols(void)
{
  ClearTempSymbols();
  DefineTempSymbol("__t__SYSLIB_internals__","1",true);
}

static void InitializePreprocessor(void)
{
  DefStackTop=0;
  Active=true;
  SymbolspaceFrozen=false;
  ReinitializePreprocessor();
  InitTempSymbols();
  InitModuleImportSequenceTester();
}

static char *PutWord(char *Ptr,const char *Word)
{
  const char *Src;
  int Len;

  Ptr+=strlen(Ptr);
  if (Word[0]=='\0') {
    Src="(newline)";
  } else {
    Src=Word;
  }
  Len=strlen(Src);
  if (Len>80) {
    Len=80;
    Ptr[80]='\0';
  }
  strncpy(Ptr,Src,Len);
  return(Ptr+Len);
}

static void ReportNotTheSame(int Num,const char *Act,const char *Exp)
{
  char Buf[4096];
  char *Ptr;

  strcpy(Buf,"Word ");
  Ptr=Buf+strlen(Buf);
  *(Ptr++)=Num+'0';
  strcpy(Ptr," is ");
  Ptr=PutWord(Ptr,Act);
  strcpy(Ptr,", expected ");
  Ptr=PutWord(Ptr,Exp);
  *Ptr='\0';
  ReportError(Buf);
}

static int CheckWordSequence(const char **Strings,bool AllowSpecial)
{
  int i;
  const char *String;
  char Buf[512];

  FetchLine();
  i=0;
  do {
    String=Strings[i++];
    if (String==NULL) break;
    ReadWordIntoBuf(Buf,512);
    if (strcmp(Buf,String)) {
      if (i==2 && AllowSpecial) {
        if (!strcmp(Buf,"__SPECIAL__")) return(2);
      }
      ReportNotTheSame(i,Buf,String);
      return(0);
    }
    if (String[0]=='\0') FetchLine();
  } while (1);
  return(1);
}

static char *StrNToUpr(char *Dest,const char *Src,int Max)
{
  char Ch;

  Max--;
  while (Max>0) {
    Ch=*(Src++);
    if (Ch=='\0') break;
    Max--;
    if (Ch>='a' && Ch<='z') Ch-=32;
    *(Dest++)=Ch;
  }
  *(Dest)='\0';
  return Dest;
}

static void ScanTheHeader(const char *Name)
{
  char Buf[512];
  char HdrSym[128];
  char *Ptr;
  const char *NewLineStr;
  const char *Strings[10];
  const char **SPtr;
  int Result;
  int MaxLen;
  int Len;

  /*
    Prepare the header file control symbol to be compared in the header file
    start.
  */
  strcpy(HdrSym,"__h__SYSLIB_");
  Ptr=HdrSym+strlen(HdrSym);
  MaxLen=128-2-(Ptr-HdrSym);
  Len=strlen(Name)-1;
  if (Len>MaxLen) Len=MaxLen;
  Ptr=StrNToUpr(Ptr,Name,Len);
  strcpy(Ptr,"__");
  /*
    Analyze the C source code. All C headers contain C source code. However
    it is not so easy than the .c files. The first line in any header must
    contain "#ifndef __h__<header>__", and the second line
    "#define __h__<header>__". Then lines of the header file follow and the
    last nonempty line must contain "#endif". The <header> here is the name
    of the header file converted into capital letters.
  */
  NewLineStr="";
  SPtr=Strings;
  *(SPtr++)="#ifndef";
  *(SPtr++)=HdrSym;
  *(SPtr++)=NewLineStr;
  *(SPtr++)="#define";
  *(SPtr++)=HdrSym;
  *(SPtr++)=NewLineStr;
  *(SPtr++)=NULL;
  Result=CheckWordSequence(Strings,true);
  if (Result!=1) return;
  PushPrepStatus();
  while (DefStackTop>0) {
    if (!ScanCLine()) break;
  }
  Buf[0]='\0';
  do {
    ReadWordIntoBuf(Buf,512);
    if (Buf[0]!='\0') break;
    FetchLine();
  } while (!EndOfFile());
  if (DefStackTop>0 || (Buf[0]!='\0'))  {
    ReportError("Nonblank lines after the header footer present");
  }
}

static void AnalyzeCSourceContent(void)
{
  while (ScanCLine());
}

static bool GetAndProcessOneFile(void)
{
  TFileInfo *F;
  int Len;

  F=GetNextFileToBeProcessed();
  if (F==NULL) return(false);
  ActivateFile(F);
  if (OpenAndReportIfFailed(F)) {
    InitializePreprocessor();
    Len=strlen(F->FileName);
    if (F->FileName[Len-1]=='h') {
      ScanTheHeader(F->FileName);
    } else {
      AnalyzeCSourceContent();
    }
    CloseFile();
  }
  DeactivateFile();
  return(true);
}

void AnalyzeFiles(void)
{
  while (GetAndProcessOneFile());
}
