#include <string.h>
#include <stdlib.h>

#include "types.h"
#include "fileread.h"
#include "dictrb.h"
#include "utils.h"
#include "filewrt.h"

#define MAX_ITEM_SIZE 15

static int DashCnt;
static bool EndOfBlock;
static bool EndOfList;

static void ReadNextLine(void)
{
  char Ch;
  int Len;
  char *Ptr;

  FetchLine();
  /*
    Strip trailing blanks
  */
  Len=strlen(Line);
  Ptr=Line+Len;
  while (Ptr>Line) {
    Ch=*(--Ptr);
    if (!IsBlank(Ch)) {
      Ptr++;
      break;
    }
  }
  *Ptr='\0';
  /*
    Count dashes in the line if it is all-dash line
  */
  DashCnt=0;
  Ptr=Line;
  do {
    Ch=*(Ptr++);
    if (Ch=='\0') break;
    if (Ch!='-') {
      DashCnt=0;
      break;
    }
    DashCnt++;
  } while (true);
}

static void CheckEOF(void)
{
  if (EndOfFile()) ReportError("Unexpected EOF in the CDF file");
}

static void ReadBlockLine(void)
{
  if (EndOfList) ReportError("Unexpected end of block in the CDF file");
  CheckEOF();
  ReadNextLine();
  EndOfList=(DashCnt==24) || EndOfFile();
  EndOfBlock=(DashCnt==12) || EndOfList;
}

static void FindBlockList(void)
{
  do {
    CheckEOF();
    ReadNextLine();
  } while (DashCnt!=24);
  EndOfBlock=0;
  EndOfList=0;
}

/*** Value dictionary handling ***/
/*
  This adds some special properties to standard RBDict value handling, for
  example quick program termination on problems with memory.
*/

TRBDict *ValueDict;
TRBDict *ItemDict;
TRBDict *SymbolDict;
TRBDict *TempSymbolDict;
TRBDict *LibDict;
TRBDict *LibArgDict;
TRBDict *LibSetDict;

static void CreateDictionaries(void)
{
  ValueDict=CreateDict();
  ItemDict=CreateDict();
  SymbolDict=CreateDict();
  TempSymbolDict=NULL;
  LibDict=CreateDict();
  LibArgDict=CreateDict();
  LibSetDict=CreateDict();
  if (
    !ValueDict || !ItemDict || !SymbolDict || 
    !LibDict || !LibArgDict || !LibSetDict
  ) {
    Fatal("No memory for dictionaries");
  }
}

static const char *Add(TRBDict *Dict,const char *Key,void *Value)
{
  int Result;

  Key=StrDup(Key);
  if (Key==NULL) Fatal("Out of memory creating name for the item");
  Result=AddItemWithKey(Dict,Key,Value);
  if (Result!=rbSUCCESS) {
    if (Result==rbNOMEM) Fatal("Out of memory adding item to dictionary");
    return(NULL);
  }
  return(Key);
}

/*** The configuration items and their values ***/
/*
  The configuration items are described in one dictionary and the values
  belonging to them are described in the other dictionary. The structures
  that hold the informations about the values are described below.
*/

#define MAX_VALUES_FOR_ITEM 256

typedef struct tagTValue {
  const char *Name;
  int Value;
  struct tagTValue *Next;
} TValue;

typedef enum {
  itBoolean,
  itInclusive,
  itExclusive,
  itText
} TItemType;

typedef struct tagTConfigItem {
  const char *Name;
  TItemType Type;
  int CountOfValidValues;
  TValue *ValidValues[MAX_VALUES_FOR_ITEM];
  struct tagTConfigItem *Next;
  bool Seen;
  int CountOfCurrentValues;
  TValue *CurrentValues[MAX_VALUES_FOR_ITEM];
} TConfigItem;

TConfigItem *FirstConfigItem;
TConfigItem *I;
TConfigItem DummyConfigItem;
TValue *FirstValue;
TValue *V;
TValue DummyValue;
int CurValue;

static void TypeNotInBrackets(void)
{
  ReportError("The configuration item type must be enclosed in brackets");
}

static void ClearSeenFlags(void)
{
  TConfigItem *I;
  I=FirstConfigItem;
  while (I!=NULL) {
    I->Seen=false;
    I=I->Next;
  }
}

static TConfigItem *CreateConfigItem(void)
{
  TConfigItem *I;

  I=malloc(sizeof(*I));
  if (I) {
    I->Name=NULL;
    I->Next=FirstConfigItem;
    I->CountOfValidValues=0;
    I->CountOfCurrentValues=0;
    FirstConfigItem=I;
  }
  return(I);
}

static bool ReadIdentifier(char *Name,const char *IName)
{
  char ErrMsg[128];
  char *MPtr;
  const char *Ptr;
  char Ch;
  bool ForInclusive;

  ForInclusive=(IName[0]=='-');
  if (ForInclusive) IName++;
  strcpy(ErrMsg,IName);
  MPtr=ErrMsg+strlen(ErrMsg);
  *(MPtr++)=' ';
  Name[MAX_ITEM_SIZE]='\0';
  ReadWordIntoBuf(Name,MAX_ITEM_SIZE+2);
  if (Name[MAX_ITEM_SIZE]!='\0') {
    Name[MAX_ITEM_SIZE]='\0';
    strcpy(MPtr,"too long");
    TokError(ErrMsg);
    return(true);
  };
  Ptr=Name;
  if (ForInclusive && *(Ptr)=='-') Ptr++;
  for (;;) {
    Ch=*(Ptr++);
    if (Ch=='\0') break;
    if (!IsIDChar(Ch)) {
      strcpy(MPtr,"must be an identifier");
      TokError(ErrMsg);
      return(true);
    }
  }
  return(false);
}

static void ProcessHeaderLine(void)
{
  char Ch;
  char Name[MAX_ITEM_SIZE+2];
  char Type[MAX_ITEM_SIZE+2];
  TItemType TypeID;
  int Len;
  bool Bad;

  /*
    Read the name of the configuration item. It must be an identifier.
  */
  Bad=ReadIdentifier(Name,"Item name");
  /*
    Read the type specifier.
  */
  Type[MAX_ITEM_SIZE]='\0';
  ReadWordIntoBuf(Type,MAX_ITEM_SIZE+2);
  if (Type[MAX_ITEM_SIZE]!='\0') {
    TokError("Item type too long");
    Bad=true;
  };
  Len=strlen(Type);
  if (Type[0]!='(' || Type[Len-1]!=')') {
    TypeNotInBrackets();
    Bad=true;
  } else {
    Type[Len-1]='\0';
  }
  /*
    We must be at the end of the line and looking at the closing bracket.
  */
  Ch=GetByte();
  if (Ch!='\n') {
    ReportError("Extra characters on line");
    Bad=true;
  }
  /*
    Decode the item type. Set it to "boolean" if not recognized.
  */
  if (!strcmp(Type+1,"Exclusive")) {
    TypeID=itExclusive;
  } else if (!strcmp(Type+1,"Inclusive")) {
    TypeID=itInclusive;
  } else if (!strcmp(Type+1,"Text")) {
    TypeID=itText;
  } else {
    TypeID=itBoolean;
    if (!Bad && strcmp(Type+1,"Boolean")) {
      ReportError("Unrecognized configuration item type");
    }
  }
  /*
    Create an item structure and write it into the dictionary.
  */
  I=CreateConfigItem();
  if (I==NULL) {
    Fatal("Out of memory creating the item structure");
  }
  I->Name=Add(ItemDict,Name,I);
  if (I->Name==NULL) ReportError("Item defined more than once");
  I->Type=TypeID;
  I->CountOfValidValues=0;
}

static bool CheckIfDescriptionParagraph(void)
/*
  Check if a paragraph belongs to the description. The input is the first
  line of the paragraph.
*/
{
  char *Ptr;
  char Ch;

  /*
    Empty lines signal the end of the description. So is the end of the
    block.
  */
  if (Line[0]=='\0' || EndOfBlock) return(0);
  /*
    Check the potential first line of a description paragraph. Such a line
    must have exactly one space between the first and the second word.
  */
  Ptr=Line;
  do {
    Ch=*(Ptr++);
    if (Ch=='\0') return(1);
  } while (!IsBlank(Ch));
  return Ch==' ' && !IsBlank(*Ptr);
}

static void SkipDescriptionParagraph(void)
/*
  Read a description paragraph and conver it into list of words separated by
  spaces. This call assumes that the paragraph it is about to read is a
  description paragraph.
*/
{
  do {
    ReadBlockLine();
  } while (Line[0]!='\0');
}

static void SkipDescriptionText(void)
{
  while (CheckIfDescriptionParagraph()) {
    SkipDescriptionParagraph();
    ReadBlockLine();
  }
}

static void EnumerateValue(void)
{
  if (CurValue<32767) {
    V->Value=CurValue++;
  } else {
    Fatal("Too many total values");
  }
}

static TValue *CreateValue(void)
{
  TValue *V;

  V=malloc(sizeof(*V));
  if (V) {
    V->Name=NULL;
    V->Next=FirstValue;
    FirstValue=V;
  }
  return(V);
}

static void ReadValueList(void)
{
  char Name[MAX_ITEM_SIZE+2];
  char *Ptr;
  bool Space;

  V=NULL;
  do {
    Ptr=Line;
    Space=IsBlank(Line[0]);
    if (!Space) {
      /*
        Read the name of the value.
      */
      ReadIdentifier(Name,"Value name");
      /*
        Add the value into the dictionary of values. Remember to omit
        the process of creating the new structure for the value when
        the value already exists.
      */
      V=SearchKey(ValueDict,Name);
      if (V==NULL) {
        /*
          Create the value
        */
        V=CreateValue();
        if (V==NULL) {
          Fatal("No memory for value identifier");
        }
	/*
	  Assign the number to the value
	*/
        if (I->Type==itInclusive) {
          V->Value=0;
        } else {
          EnumerateValue();
        }
        /*
          Add the value into the dictionary of values.
        */
        if (V!=&DummyValue) V->Name=Add(ValueDict,Name,V);
      } else {
        if (V->Value<0) {
          ReportError("Value mentioned more than once");
        } else {
          if (V->Value==0) {
            EnumerateValue();
          } else {
            V->Value=-V->Value;
          }
        }
      }
      /*
        Record the value in the current configuration item
      */
      if (I->CountOfValidValues==MAX_VALUES_FOR_ITEM) {
        Fatal("Too many values");
      } else {
        I->ValidValues[I->CountOfValidValues++]=V;
      }
    }
    if (V==NULL) {
      ReportError("Ignored value description without value");
    }
    /*
      Read the next line
    */
    ReadBlockLine();
  } while (Line[0]!=0 && !EndOfBlock);
}

static void ClearItemDefinition(void)
{
  int i;
  TValue *V;

  for (i=0;i<I->CountOfValidValues;i++) {
    V=I->ValidValues[i];
    if (V->Value<0) V->Value=-V->Value;
  }
}

static void ReadOneDescription(void)
{
  /*
    Read the lines until a nonempty one is found
  */
  do {
    ReadBlockLine();
  } while (!EndOfBlock && Line[0]=='\0');
  /*
    The line found is a header line that must contain the name of the
    configuration item and its type. Both of these things must be identifiers
    and the type must be enclosed in brackets.
  */
  ProcessHeaderLine();
  /*
    Next line must be empty. So load the nonempty line and complain when no
    empty line is there.
  */
  ReadBlockLine();
  if (Line[0]!='\0') {
    ReportError("Line after header not empty");
  } else {
    ReadBlockLine();
  }
  /*
    Now read the description text.
  */
  SkipDescriptionText();
  /*
    Read the list of items. The list of items is allowed only for nonboolean
    items.
  */
  if (I->Type!=itBoolean && I->Type!=itText) {
    ReadValueList();
  }
  /*
    Clear the symbol description from the value usage flags.
  */
  ClearItemDefinition();
}

static void CheckEndOfBlock(void)
{
  while (!EndOfBlock) {
    ReadBlockLine();
    if (!EndOfBlock && Line[0]!='\0') ReportError("Ignored line");
  }
}

static void ReadAndProcessDescriptions(void)
{
  /*
    Find the first block list on the way.
  */
  FindBlockList();
  /*
    Read and translate the descriptions until the end of the block list is
    reached.
  */
  do {
    ReadOneDescription();
    CheckEndOfBlock();
  } while (!EndOfList);
}

void ReadDescriptions(void)
{
  CreateDictionaries();
  if (!OpenFile("defs.cds")) Fatal("Unable to find config item description");
  ReadAndProcessDescriptions();
  CloseFile();
}

/*******************************************
*** The configuration description parser ***
*******************************************/

static bool CheckValueValid(TConfigItem *I,TValue *V)
{
  int i;

  for (i=0;i<I->CountOfValidValues;i++) {
    if (I->ValidValues[i]==V) return(true);
  }
  TokError("Value not valid for the config item");
  return(false);
}

static TValue *ReadValue(TConfigItem *I,char *CfgValue,bool ForInclusive)
{
  TValue *V;

  if (ReadIdentifier(CfgValue,"-Config item value")) {
    return(NULL);
  }
  if (CfgValue[0]=='\0') return(NULL);
  if (ForInclusive && CfgValue[0]=='-') CfgValue++;
  V=SearchKey(ValueDict,CfgValue);
  if (V!=NULL) {
    if (!CheckValueValid(I,V)) V=NULL;
  } else {
    TokError("Invalid value");
  }
  return(V);
}

static void ParseConfigLine(void)
{
  char CfgName[MAX_ITEM_SIZE+2];
  char CfgValue[MAX_ITEM_SIZE+2];
  TConfigItem *I;
  TValue *V;
  int Column;
  int i;
  bool Found;
  bool Delete;
  int Col;
  int Len;
  char *Value;

  if (ReadIdentifier(CfgName,"Config item")) return;
  I=SearchKey(ItemDict,CfgName);
  if (I==NULL) {
    TokError("Invalid config item");
    return;
  }
  if (I->Type==itText) {
    SkipBlanks();
    Col=GetCurrentPos();
    Len=strlen(Line+Col);
    Value=malloc(Len+1);
    if (Value!=NULL) {
      memcpy(Value,Line+Col-1,Len+1);
      if (I->CountOfCurrentValues>0) {
        free(I->CurrentValues[0]);
      }
      I->CurrentValues[0]=(void *)Value;
      I->CountOfCurrentValues=1;
    } else {
      TokError("No memory for text item value");
    }
    return;
  }
  if (I->Type==itInclusive) {
    Column=GetTokenPos();
    while (true) {
      V=ReadValue(I,CfgValue,true);
      Delete=(CfgValue[0]=='-');
      if (CfgValue[0]=='\0') break;
      Column=-1;
      if (V==NULL) continue;
      Found=false;
      for (i=0;i<I->CountOfCurrentValues;i++) {
        if (I->CurrentValues[i]==V) {
          Found=true;
          break;
        }
      }
      if (Delete) {
        if (Found) {
          I->CountOfCurrentValues--;
          I->CurrentValues[i]=I->CurrentValues[I->CountOfCurrentValues];
        }
      } else if (!Found) {
        if (I->CountOfCurrentValues==MAX_VALUES_FOR_ITEM) {
          TokError("Too many values");
        } else {
          I->CurrentValues[I->CountOfCurrentValues++]=V;
        }
      }
    };
    if (Column!=-1) {
      ReportErrorPos("Must have at least one value specified",Column);
    }
  } else {
    if (I->Seen) {
      TokError("Config item seen more than once");
      return;
    }
    I->Seen=true;
    if (I->Type==itBoolean) {
      Column=GetTokenPos();
      if (ReadIdentifier(CfgValue,"Config item value")) return;
      if (CfgValue[0]=='\0') {
        ReportErrorPos("Value must be specified",Column);
      } else if (strcmp(CfgValue,"no")==0) {
        I->CountOfCurrentValues=1;
        I->CurrentValues[0]=NULL;
      } else if (strcmp(CfgValue,"yes")==0) {
        I->CountOfCurrentValues=1;
        I->CurrentValues[0]=(void *)"yes";
      } else {
        TokError("This value must be 'yes' or 'no'");
        return;
      }
    } else {
      Column=GetTokenPos();
      I->CountOfCurrentValues=1;
      I->CurrentValues[0]=ReadValue(I,CfgValue,false);
      if (CfgValue[0]=='\0') {
        ReportErrorPos("Value must be specified",Column);
      }
    }
  }
}

void ReportFileNotFound(const char *Msg)
{
  char MsgBuf[128];
  strcpy(MsgBuf,"Unable to open ");
  strcpy(MsgBuf+strlen(MsgBuf),Msg);
  Fatal(MsgBuf);
}

static bool OpenThisFile(const char *FileName,const char *Msg)
{
  if (!OpenFile(FileName)) {
    if (Msg!=NULL) ReportFileNotFound(Msg);
    return(false);
  }
  return(true);
}

#define MAX_LIBRARY_LINKER_ARGS 32
#define MAX_LIBRARIES_PER_CONFIG 16
#define MAX_LIBRARY_DEPENDENCIES 16
#define MAX_LIBRARY_LINKING_ARG_SIZE 4096

#define NEXT_ACTIVE_NULL (void *)OpenThisFile

typedef struct tagTLibraryInfo {
  const char *Name;
  unsigned int ArgCnt;
  const char *Args[MAX_LIBRARY_LINKER_ARGS];
  unsigned int LibDepCnt;
  struct tagTLibraryInfo *LibDeps[MAX_LIBRARY_DEPENDENCIES];
  struct tagTLibraryInfo *Next;
  struct tagTLibraryInfo *NextActive;
} TLibraryInfo;

typedef struct tagTConfigItemSetting {
  const char *Name;
  unsigned int LibCnt;
  TLibraryInfo *LibList[MAX_LIBRARIES_PER_CONFIG];
  struct tagTConfigItemSetting *Next;
} TConfigItemSetting;

TLibraryInfo *FirstLibraryInfo;
TConfigItemSetting *FirstConfigItemSetting;

static TLibraryInfo *SearchLibraryInfo(const char *Name)
{
  return(SearchKey(LibDict,Name));
}

static TLibraryInfo *CreateLibraryInfo(const char *Name)
{
  TLibraryInfo *I;
  bool Invalid;

  I=SearchLibraryInfo(Name);
  if (I==NULL) {
    I=malloc(sizeof(*I));
    if (I!=NULL) {
      I->Name=StrDup(Name);
      Invalid=(I->Name==NULL);
      if (Invalid) {
        LineError("No memory for library ID string");
      } else {
        Invalid=(AddItemWithKey(LibDict,I->Name,I)!=rbSUCCESS);
        if (Invalid) {
          LineError("No memory to add library info into dictionary");
        }
      }
      if (Invalid) {
        free(I);
        I=NULL;
      } else {
        I->ArgCnt=65535;
        I->LibDepCnt=0;
        I->NextActive=NULL;
        I->Next=FirstLibraryInfo;
        FirstLibraryInfo=I;
      }
    } else {
      TokError("No memory for library info");
    }
  }
  return(I);
}

static TConfigItemSetting *SearchConfigItemSetting(const char *Name)
{
  return(SearchKey(LibSetDict,Name));
}

static TConfigItemSetting *CreateConfigItemSetting(const char *Name)
{
  TConfigItemSetting *S;
  bool Invalid;

  S=SearchConfigItemSetting(Name);
  if (S==NULL) {
    S=malloc(sizeof(*S));
    if (S!=NULL) {
      S->Name=StrDup(Name);
      Invalid=(S->Name==NULL);
      if (Invalid) {
        LineError("No memory for config item setting ID string");
      } else {
        Invalid=(AddItemWithKey(LibSetDict,S->Name,S)!=rbSUCCESS);
        if (Invalid) {
          LineError("No memory to add config item setting into dictionary");
        }
      }
      if (Invalid) {
        free(S);
        S=NULL;
      } else {
        S->LibCnt=0;
        S->Next=FirstConfigItemSetting;
        FirstConfigItemSetting=S;
      }
    } else {
      TokError("No memory for config item setting info");
    }
  }
  return(S);
}

static void ParseDriverLibraryDependency(void)
{
  char Name[2*(MAX_ITEM_SIZE+2)];
  char *Value;
  TConfigItem *I;
  TValue *V;
  TConfigItemSetting *S;
  char LibName[MAX_ITEM_SIZE+2];
  TLibraryInfo *LI;

  /*
    Skip the type character
  */
  GetByte();
  /*
    Read the config item and its value and check whether they form a valid
    and supported configuration setting. Currently supported are only
    exclusive and boolean configuration items. After all checks are passed,
    the configuration item and its value will be placed into the "Name"
    buffer next to each other, delimited by a zero byte.
  */
  Value=ReadWordIntoBuf(Name,MAX_ITEM_SIZE+2)+1;
  if (Name[0]=='\0') {
    LineError("No config item name");
    return;
  }
  I=SearchKey(ItemDict,Name);
  if (I==NULL) {
    TokError("Invalid config item");
    return;
  }
  if (I->Type==itExclusive) {
    V=ReadValue(I,Value,false);
    if (V==NULL) return;
  } else if (I->Type==itBoolean) {
    ReadWordIntoBuf(Value,MAX_ITEM_SIZE+2);
    if ((strcmp(Value,"yes")!=0) && (strcmp(Value,"no")!=0)) {
      TokError("Boolean item value can be 'yes' or 'no' only");
      return;
    }
  } else {
    TokError("Unsupported config item type");
    return;
  }
  /*
    Now when we know the config item setting is full and correct, we can
    construct the string for adding the setting into the dictionary. This
    is achieved by replacing the delimiting null byte with a space (spaces
    cannot occur as a part of the config item name or its value).
  */
  *(Value-1)=' ';
  /*
    Create a record that will represent this configuration if it does not
    exist yet. If it does exist, find it for addition of more library
    identifiers. If it cannot be created, the line will be ignored.
  */
  S=CreateConfigItemSetting(Name);
  if (S==NULL) return;
  /*
    Now we can read the list of dependency libraries
  */
  while (true) {
    LibName[MAX_ITEM_SIZE]='\0';
    ReadWordIntoBuf(LibName,MAX_ITEM_SIZE+2);
    if (LibName[0]=='\0') break;
    if (LibName[MAX_ITEM_SIZE]=='\0') {
      if (S->LibCnt==MAX_LIBRARIES_PER_CONFIG) {
        TokError("Too many library identifiers for config item setting");
      } else {
        LI=CreateLibraryInfo(LibName);
        if (LI!=NULL) S->LibList[S->LibCnt++]=LI;
      }
    }
  }
}

static const char *SearchLibraryLinkerArgument(const char *Name)
{
  return(SearchKey(LibArgDict,Name));
}

static const char *RecordLibraryLinkerArgument(const char *Name)
{
  const char *Previous;
  char *Result;

  Previous=SearchLibraryLinkerArgument(Name);
  if (Previous!=NULL) return(Previous);
  Result=StrDup(Name);
  if (Result==NULL) {
    TokError("No memory for library linker arg copy");
  } else {
    if (AddItemWithKey(LibArgDict,Result,Result)!=rbSUCCESS) {
      TokError("No memory for library linker argument");
      free(Result);
      Result=NULL;
    }
  }
  return(Result);
}

static TLibraryInfo *ReadLibraryName(bool Expected)
{
  char LibName[MAX_ITEM_SIZE+2];

  LibName[MAX_ITEM_SIZE]='\0';
  ReadWordIntoBuf(LibName,MAX_ITEM_SIZE+2);
  if (LibName[MAX_ITEM_SIZE]!='\0') return(NULL);
  if (LibName[0]=='\0') {
    if (Expected) TokError("Expected library identifier");
    return(NULL);
  }
  return(CreateLibraryInfo(LibName));
}

static void ParseLibraryLinkerArguments(void)
{
  TLibraryInfo *I;
  char LibArg[MAX_LIBRARY_LINKING_ARG_SIZE+2];
  const char *Arg;

  /*
    Skip the type character
  */
  GetByte();
  /*
    Read the name of the library and find its description or create it if it
    does not exist.
  */
  I=ReadLibraryName(true);
  /*
    Fresh library descriptions are marked by their argument count being set
    to 65535. This is used by the checker to find out which library
    identifiers were not 'declared" by mentioning at least one "!" line for
    it. Since after execution of this procedure the library info will not
    be fresh anymore, I need to place a zero there to indicate no arguments
    were recorded yet.
  */
  if (I->ArgCnt==65535) I->ArgCnt=0;
  /*
    Read the arguments and record them in the library info record.
  */
  while (true) {
    LibArg[MAX_LIBRARY_LINKING_ARG_SIZE]='\0';
    ReadWordIntoBuf(LibArg,MAX_LIBRARY_LINKING_ARG_SIZE+2);
    if (LibArg[0]=='\0') break;
    if (LibArg[MAX_LIBRARY_LINKING_ARG_SIZE]=='\0') {
      if (I->ArgCnt==MAX_LIBRARY_LINKER_ARGS) {
        TokError("Too many library linker arguments in library description");
      } else {
        Arg=RecordLibraryLinkerArgument(LibArg);
        if (Arg!=NULL) I->Args[I->ArgCnt++]=Arg;
      }
    }
  }
}

static void ParseLibraryDependencies(void)
{
  TLibraryInfo *I;
  TLibraryInfo *RI;

  /*
    Skip the type character
  */
  GetByte();
  /*
    Read the name of the library and find its description or create it if it
    does not exist. This first name denotes the library the references are in
    and it must be present.
  */
  I=ReadLibraryName(true);
  /*
    Read referenced libraries and gather them into the array recording the
    library reference array.
  */
  while (true) {
    RI=ReadLibraryName(false);
    if (RI==NULL) break;
    if (I->LibDepCnt==MAX_LIBRARY_DEPENDENCIES) {
      TokError("Too many references in library");
    } else {
      I->LibDeps[I->LibDepCnt++]=RI;
    }
  }
}

static void CheckLibraryNamesValid(void)
{
  char Buffer[MAX_ITEM_SIZE+2+64];
  char *Ptr;
  char *Ptr1;
  TLibraryInfo *I;
  strcpy(Buffer,"Invalid library identifier: \"");
  Ptr=Buffer+strlen(Buffer);
  I=FirstLibraryInfo;
  while (I!=NULL) {
    if (I->ArgCnt==65535) {
      strcpy(Ptr,I->Name);
      Ptr1=Ptr+strlen(Ptr);
      *(Ptr1++)='"';
      *Ptr1='\0';
      LineError(Buffer);
      I->ArgCnt=0;
    }
    I=I->Next;
  }
}

static void BuildAddonLibraryList(void)
{
  TConfigItem *I;
  char Setting[2*(MAX_ITEM_SIZE+2)];
  TRBDict *LibArgUsedDict;
  char *Ptr;
  const char *Val;
  TConfigItemSetting *S;
  TLibraryInfo *LI;
  unsigned int i;
  const char *Arg;
  bool Space;
  TLibraryInfo *RL;
  TLibraryInfo Sentinel;
  TLibraryInfo *Next;
  TLibraryInfo *Prev;
  bool WereSent;
  bool SendIt;

  /*
    Determine which libraries are active in out configuration. Gather them
    into a linked list.
  */
  Sentinel.NextActive=NEXT_ACTIVE_NULL;
  I=FirstConfigItem;
  while (I!=NULL) {
    Setting[0]='\0';
    if (I->Type==itExclusive || I->Type==itBoolean) {
      strcpy(Setting,I->Name);
      Ptr=Setting+strlen(Setting);
      *(Ptr++)=' ';
      assert(I->CountOfCurrentValues==0 || I->CountOfCurrentValues==1);
      Val="";
      if (I->Type==itBoolean) {
        Val="no";
        if (I->CountOfCurrentValues>0) {
          if (I->CurrentValues[0]!=NULL) Val="yes";
        }
      } else {
        if (I->CountOfCurrentValues>0) {
          Val=I->CurrentValues[0]->Name;
        }
      }
      strcpy(Ptr,Val);
    }
    S=NULL;
    if (Setting[0]!='\0') S=SearchConfigItemSetting(Setting);
    if (S!=NULL) {
      for (i=0;i<S->LibCnt;i++) {
        LI=S->LibList[i];
        if (LI->NextActive==NULL) {
          LI->NextActive=Sentinel.NextActive;
          Sentinel.NextActive=LI;
        }
      }
    }
    I=I->Next;
  }
  /*
    Search the list of active libraries and make all dependencies of them
    also active. Note the method the new libraries are placed into the list.
    The new library will be inserted to the position immediately after the
    one we are scanning. This ensures the libraries are traversed deeply
    since once we are done with the current library, the "NextActive" field
    will point to a bunch of newly added libraries.
  */
  Prev=&Sentinel;
  LI=Sentinel.NextActive;
  while (LI!=NEXT_ACTIVE_NULL) {
    for (i=0;i<LI->LibDepCnt;i++) {
      RL=LI->LibDeps[i];
      if (RL->NextActive==NULL) {
        RL->NextActive=LI->NextActive;
        LI->NextActive=RL;
      }
    }
    LI=LI->NextActive;
  }
  /*
    Now traverse the list until all arguments for the libraries are sent to
    the output or only mutually dependent ones are left in the list.
  */
  LibArgUsedDict=CreateDict();
  if (LibArgUsedDict==NULL) {
    LineError("No memory for used arguments dictionary");
    return;
  }
  Space=false;
  while (Sentinel.NextActive!=NEXT_ACTIVE_NULL) {
    Prev=&Sentinel;
    LI=Sentinel.NextActive;
    WereSent=false;
    while (LI!=NEXT_ACTIVE_NULL) {
      Next=LI->NextActive;
      SendIt=true;
      for (i=0;i<LI->LibDepCnt;i++) {
        RL=LI->LibDeps[i];
        if (RL->NextActive!=NULL) {
          SendIt=false;
          break;
        }
      }
      if (SendIt) {
        for (i=0;i<LI->ArgCnt;i++) {
          Arg=LI->Args[i];
          if (SearchKey(LibArgUsedDict,Arg)==NULL) {
            if (Space) Print(" ");
            Print(Arg);
            Space=true;
            AddItemWithKey(LibArgUsedDict,Arg,Setting);
          }
        }
        Prev->NextActive=Next;
        LI->NextActive=NULL;
        WereSent=true;
      } else {
        Prev=LI;
      }
      LI=Next;
    }
    if (!WereSent) break;
  }
  DestroyDict(LibArgUsedDict);
  /*
    If there are some mutually dependent libraries, report that.
  */
  if (Sentinel.NextActive!=NEXT_ACTIVE_NULL) {
    LineError("Mutually dependent libraries detected");
  }
}

bool ReadDefinitions(const char *FileName,const char *Msg,bool AllowLibInfo)
{
  if (!OpenThisFile(FileName,Msg)) return(false);
  ClearSeenFlags();
  while (!EndOfFile()) {
    FetchLine();
    StripBlanksAndComments();
    if (Line[0]=='@') {
      if (AllowLibInfo) {
        ParseDriverLibraryDependency();
      } else {
        LineError("Driver dependency library specifiers not allowed here");
      }
    } else if (Line[0]=='!') {
      if (AllowLibInfo) {
        ParseLibraryLinkerArguments();
      } else {
        LineError("Library linking argument specifiers not allowed here");
      }
    } else if (Line[0]==':') {
      if (AllowLibInfo) {
        ParseLibraryDependencies();
      } else {
        LineError("Library dependency records not allowed here");
      }
    } else if (Line[0]!='\0') ParseConfigLine();
  }
  CloseFile();
  if (AllowLibInfo) CheckLibraryNamesValid();
  return(true);
}

/****************************************
*** The preprocessor symbols handling ***
****************************************/

/*** Parsed informations about a symbol ***/
/*
  This structure is used to store informations parsed out of a config item
  symbol.
*/
typedef struct {
  /*
    The type character from the symbol.
  */
  char Type;
  /*
    The configuration item information record associated with the symbol.
  */
  TConfigItem *I;
  /*
    For inclusive config item symbols this is set to the value descriptor
    matching the value portion of the symbol by the ParseConfigItemSymbol()
    call.

    For exclusive config item this is set to the value descriptor matching
    the value used in the comparision expression of the #[el]if command
    and is set by ParseValueSymbol() after this structure built by
    ParseConfigItemSymbol() and the value from the expression is passed to
    it.

    For other symbols this is unused and is set to NULL.
  */
  TValue *V;
  /*
    Points to a 128 byte buffer where the ~ParseSymbolCore() call stores the
    name of the symbol.
  */
  char *Name;
} TParsedSymbol;

#define svFalse ((void *)(&OpenThisFile))
#define svTrue ((void *)(&ReadDefinitions))

static void InvalidSymbol(void)
{
  ReportError("Malformed symbol referenced");
}

static void ReportErrorAboutMsgPartLen(const char *MsgPart)
{
  char Msg[128];

  strcpy(Msg,"Name of the ");
  strcpy(Msg+strlen(Msg),MsgPart);
  strcpy(Msg+strlen(Msg)," too long");
  ReportError(Msg);
}

static bool ParseSymbol(
  /*
    The symbol to be parsed
  */
  const char *Symbol,
  /*
    The buffer to which the informations from the symbol is to be stored.
    The field "Name" must be filled with the caller with a pointer to a
    128 byte buffer which will be filled with the name parsed out of the
    symbol.
  */
  TParsedSymbol *S,
  /*
    The class of the symbol to be reported in an error message when
    something goes wrong.
  */
  const char *MsgPart
)
{
  int Len;

  Len=strlen(Symbol);
  if (Len<8) {
    InvalidSymbol();
    return(false);
  }
  if (
    Symbol[0]!='_' || Symbol[1]!='_' ||
    Symbol[3]!='_' || Symbol[4]!='_' ||
    Symbol[Len-2]!='_' || Symbol[Len-1]!='_'
  ) {
    InvalidSymbol();
    return(false);
  }
  if (Len>127+7) {
    ReportErrorAboutMsgPartLen(MsgPart);
    return(false);
  }
  S->Type=Symbol[2];
  memcpy(S->Name,Symbol+5,Len-7);
  S->Name[Len-7]='\0';
  return(true);
}

void DefineTempSymbol(const char *Symbol,const char *Value,bool Active)
/*
  Only boolean temporary symbols are allowed in the sources.
*/
{
  void *Val;
  char Name[128];
  TParsedSymbol S;

  /*
    Convert the value to be defined to a boolean. Only temporary symbols can
    be defined and they all are boolean symbols.
  */
  if (!strcmp(Value,"0")) {
    Val=svFalse;
  } else if (!strcmp(Value,"1")) {
    Val=svTrue;
  } else {
    ReportError("Boolean symbols can be set to 0 or 1 only");
    return;
  }
  /*
    Create the dictionary of temporary symbols if it does not exist yet.
  */
  if (TempSymbolDict==NULL) {
    TempSymbolDict=CreateDict();
    if (TempSymbolDict==NULL) {
      ReportError("No memory for the temporary symbol dictionary");
      return;
    }
  }
  /*
    Only temporary symbols can be changed in the build system.
  */
  S.Name=Name;
  if (!ParseSymbol(Symbol,&S,"temporary symbol")) return;
  if (S.Type!='t') {
    ReportError("Only temporary symbols can be changed");
    return;
  }
  /*
    If the activity flag passed to this call is false, don't actually define
    the symbol. Only the sanity checks are performed in that case.
  */
  if (!Active) return;
  /*
    Define the symbol. The symbol must be undefined at the time of defining.
  */
  switch (AddItem(TempSymbolDict,Name,Val)) {
    case rbNOMEM: {
      ReportError("No memory for the temporary symbol");
      break;
    }
    case rbDUPLICATE: {
      ReportError("Attempt to redefine an already defined symbol");
      break;
    }
    default: {
      break;
    }
  }
}

void UndefineTempSymbol(const char *Symbol,bool Active)
{
  char Name[128];
  TParsedSymbol S;

  /*
    Check if the caller really wants to undefine a temporary symbol. Other
    symbols are not allowed to be undefined.
  */
  S.Name=Name;
  if (!ParseSymbol(Symbol,&S,"temporary symbol")) return;
  if (S.Type!='t') {
    ReportError("Only temporary symbols can be undefined");
    return;
  }
  /*
    If the activity flag passed to this call is false, don't actually
    undefine the symbol. Only the sanity checks are performed in that case.
  */
  if (!Active) return;
  /*
    If no temporary symnbol dictionary exists yet, there is nothing to
    undefine. Otherwise delete the symbol from the dictionary. Attempt to
    undefine an already undefined symbol is OK.
  */
  if (TempSymbolDict==NULL) return;
  DeleteItem(TempSymbolDict,Name);
}

static char *GetValueNameFromSymbol(char *ItemName)
{
  char *ValName;

  /*
    Search for the value in the symbol name
  */
  ValName=ItemName;
  while (ValName[1]!='\0') {
    if (ValName[0]=='_' && ValName[1]=='_') {
      ValName[0]='\0';
      return(ValName+2);
    }
    ValName++;
  }
  /*
    Nno value was found in the symbol. Complain.
  */
  ReportError("Inclusive config item without value");
  return(NULL);
}

static bool CheckConfigItemSymbol(
  /*
    The buffer with a parsed symbol. The buffer will be adjusted with the
    additional informations available for config items.
  */
  TParsedSymbol *S
)
{
  char *ValName;
  TItemType T;

  /*
    If the symbol is an inclusive one, it will have th value stored as
    the second part of the symbol name.
  */
  ValName=NULL;
  if (S->Type=='i') {
    ValName=GetValueNameFromSymbol(S->Name);
    if (ValName==NULL) return(false);
  }
  /*
    Search for the item description structure
  */
  S->V=NULL;
  S->I=SearchKey(ItemDict,S->Name);
  if (S->I==NULL) {
    ReportError("Unknown config item");
    return(false);
  }
  /*
    Check if the type of the config item match the type in the symbol
  */
  switch (S->Type) {
    case 'e': T=itExclusive; break;
    case 'i': T=itInclusive; break;
    case 'b': T=itBoolean; break;
    case 'x': T=itText; break;
    default: {
      ReportError("Symbol has invalid type field");
      return(false);
    }
  }
  if (T!=S->I->Type) {
    ReportError("Symbol type does not match config item type");
    return(false);
  }
  /*
    If the symbol is not for an inlusive config item, we are done with it.
  */
  if (T!=itInclusive) return(true);
  /*
    Connect the value back to the name of the configuration item.
  */
  *(ValName-2)='_';
  /*
    Search for the value
  */
  S->V=SearchKey(ValueDict,ValName);
  if (S->V==NULL) {
    ReportError("Unknown value");
    return(false);
  }
  /*
    Check if the value is in the list of valid values for the item
  */
  return(CheckValueValid(S->I,S->V));
}

static bool ParseValueSymbol(
  /*
    The value symbol to be parsed
  */
  const char *Symbol,
  /*
    The buffer to which the value information record from the symbol is to be
    stored. The buffer is expected to be filled with an exclusive parsed
    symbol.
  */
  TParsedSymbol *S
)
{
  char Name[128];
  TParsedSymbol V;

  /*
    Parse out the value name
  */
  V.Name=Name;
  if (!ParseSymbol(Symbol,&V,"value")) return(false);
  if (V.Type!='v') {
    ReportError("Value symbol expected");
    return(false);
  }
  /*
    Search for the value in the dictionary of values.
  */
  S->V=SearchKey(ValueDict,Name);
  if (S->V==NULL) {
    ReportError("Unknown value");
    return(false);
  }
  /*
    Check if the value is valid and return the result of this scan.
  */
  return(CheckValueValid(S->I,S->V));
}

static bool GetBoolSymbolValue(const char *Symbol)
{
  void *Value;
  char ItemName[128];
  TParsedSymbol S;
  int i;

  /*
    The symbol must be a valid build system symbol.
  */
  S.Name=ItemName;
  if (!ParseSymbol(Symbol,&S,"boolean symbol")) return(false);
  /*
    The symbol may be either a temporary symbol (which must be a boolean
    symbol), an inclusive symbol (the inclusive configuration items generate
    a set of boolean symbols) or a boolean configuration item symbol.
  */
  if (S.Type=='t') {
    Value=NULL;
    if (TempSymbolDict) {
      Value=SearchKey(TempSymbolDict,ItemName);
    }
    if (Value==NULL) {
      ReportError("Undefined temporary symbol referenced");
    }
    return(Value==svTrue);
  } else {
    if (!CheckConfigItemSymbol(&S)) return(false);
    if (S.Type=='b') {
      return(S.I->CurrentValues[0]!=NULL);
    } else if (S.Type=='i') {
      for (i=0;i<S.I->CountOfCurrentValues;i++) {
        if (S.I->CurrentValues[i]==S.V) return(true);
      }
      return(false);
    } else {
      ReportError("Symbol is not a boolean symbol");
      return(false);
    }
  }
}

static bool CheckValuedSymbol(const char *Symbol,const char *Value)
{
  char ItemName[128];
  TParsedSymbol S;

  /*
    Check the symbol and get the configuration item name
  */
  S.Name=ItemName;
  if (!ParseSymbol(Symbol,&S,"config item")) return(false);
  /*
    The symbol must be a config item symbol
  */
  if (!CheckConfigItemSymbol(&S)) return(false);
  /*
    Only exclusive symbols may hold a value.
  */
  if (S.I->Type!=itExclusive) {
    ReportError("Only exclusive symbols can hold a value");
    return(false);
  }
  /*
    Parse the value symbol and check if it is correct for the config item
    symbol.
  */
  if (!ParseValueSymbol(Value,&S)) return(false);
  /*
    If no value found, the config item will not match anything.
  */
  if (S.I->CountOfCurrentValues<1) return(false);
  /*
    Check the value
  */
  return(S.I->CurrentValues[0]==S.V);
}

bool CheckSymbol(const char *Symbol,const char *Value)
{
  if (Value==NULL) {
    return(GetBoolSymbolValue(Symbol));
  } else {
    return(CheckValuedSymbol(Symbol,Value));
  }
}

void ClearTempSymbols(void)
{
  if (TempSymbolDict) {
    DestroyDict(TempSymbolDict);
    TempSymbolDict=NULL;
  }
}

static void PrintSymbolName(TConfigItem *I,TValue *V)
{
  Print("__");
  if (I!=NULL) {
    switch (I->Type) {
      case itExclusive: Print("e"); break;
      case itInclusive: Print("i"); break;
      case itBoolean: Print("b"); break;
      case itText: Print("x"); break;
      default: {
        assert(false);
      }
    }
    Print("__");
    Print(I->Name);
  } else {
    Print("v");
  }
  Print("__");
  if (V!=NULL) {
    Print(V->Name);
    Print("__");
  }
}

void PrintNumber(int Number)
{
  char Data[6];
  char *Ptr;

  assert(Number>=0);
  Data[5]='\0';
  Ptr=Data+5;
  do {
    *(--Ptr)=Number%10+'0';
    Number/=10;
  } while (Number!=0);
  Print(Ptr);
}

void DumpValueSymbolContent(void)
{
  TValue *V;

  V=FirstValue;
  while (V!=NULL) {
    Print("#undef ");
    PrintSymbolName(NULL,V);
    Print("\n#define ");
    PrintSymbolName(NULL,V);
    Print(" ");
    PrintNumber(V->Value);
    NewLine();
    V=V->Next;
  }
}

void DumpConfigurationSymbolsContent(void)
{
  TConfigItem *I;
  int i;

  I=FirstConfigItem;
  while (I!=NULL) {
    if (I->Type==itInclusive) {
      for (i=0;i<I->CountOfValidValues;i++) {
        Print("#undef ");
        PrintSymbolName(I,I->ValidValues[i]);
        NewLine();
      }
    } else {
      Print("#undef ");
      PrintSymbolName(I,NULL);
      NewLine();
    }
    switch (I->Type) {
      case itExclusive: {
        if (I->CountOfCurrentValues!=0) {
          assert(I->CountOfCurrentValues==1);
          Print("#define ");
          PrintSymbolName(I,NULL);
          Print(" ");
          PrintSymbolName(NULL,I->CurrentValues[0]);
          NewLine();
        }
        break;
      }
      case itInclusive: {
        for (i=0;i<I->CountOfCurrentValues;i++) {
          Print("#define ");
          PrintSymbolName(I,I->CurrentValues[i]);
          Print(" 1\n");
        }
        break;
      }
      case itBoolean: {
        if (I->CountOfCurrentValues!=0) {
          assert(I->CountOfCurrentValues==1);
          if (I->CurrentValues[0]!=NULL) {
            Print("#define ");
            PrintSymbolName(I,NULL);
            Print(" 1\n");
          }
        }
        break;
      }
      case itText: {
        if (I->CountOfCurrentValues!=0) {
          assert(I->CountOfCurrentValues==1);
          Print("#define ");
          PrintSymbolName(I,NULL);
          Print(" \"");
          Print((char *)I->CurrentValues[0]);
          Print("\"\n");
        }
        break;
      }
      default: {
        assert(false);
      }
    }
    I=I->Next;
  }
}

void WriteDefinitionsAsCCode(void)
{
  OpenOutputFile("defs.ghn");
  Print(
    "/*\n"
    "  !!! DO NOT EDIT !!!\n\n"
    "  This file is automatically regenerated each time the project is "
    "rebuilt.\n"
    "  Therefore any changes made to it are lost. To force your "
    "modifications to\n"
    "  the configuration of the build process use 'config.usr' file instead. "
    "See\n"
    "  'defs.cds' for more informations. This file is generated using "
    "template\n"
    "  file 'defs.cds', data from GPS and file 'config.usr'\n"
    "*/\n\n"
    "#undef __INTERFACE__\n"
    "#undef __SPECIAL__\n"
    "#define __SPECIAL__\n"
    "#undef __t__SYSLIB_internals__\n"
    "#define __t__SYSLIB_internals__ 1\n"
  );
  DumpValueSymbolContent();
  DumpConfigurationSymbolsContent();
  Print(
    "#undef __x__ADDON_LIBS__\n"
    "#define __x__ADDON_LIBS__ \""
  );
  BuildAddonLibraryList();
  Print("\"\n");
  CloseOutputFile();
}
