#include <string.h>

#include "types.h"
#include "fileread.h"
#include "files.h"

typedef enum {
  kUnknown,
  kSYSTEM,
  kPROGRAM,
  kTRANSLATOR,
  kMODULEGEN,
  kLIBRARY,
  kAUTHOR,
  kLICENSE,
  kCREDIT,
  kTITLE,
  kCOPYRIGHT,
  kBRIEF,
  kDESC,
  kSOURCE,
  kHEADER,
  kSHARED,
  kTEST,
  kEOF
} TKeyWord;

static bool WasSystem;
static bool WasLibrary;

static char UpCase(char Ch)
{
  if (Ch>='a' && Ch<='z') Ch-=32;
  return(Ch);
}

#define RESERVED_WORD_LEN_LIMIT 16

static TKeyWord RecognizeWord(const char *W,const char *L)
{
  char BufferWord[RESERVED_WORD_LEN_LIMIT+1];
  char BufferIn[RESERVED_WORD_LEN_LIMIT+1];
  char *Ptr;
  TKeyWord WordID;

  /*
    Copy the reserved word into the reserved word input buffer. If case
    insensitivity is requested, convert the command to uppercase since the
    reserved word values are.
  */
  Ptr=strncpy(BufferIn,W,RESERVED_WORD_LEN_LIMIT+1);

  /*
    If the potential reserved word name is longer than the longest possible
    reserved word, return "unrecognized" value, which is represented as zero.
    This check is here to avoid "buffer overrun" security hole bugs.
  */
  if (strlen(Ptr)>RESERVED_WORD_LEN_LIMIT) return(kUnknown);

  /*
    Go through the reserved word list and search for the reserved word.
  */

  WordID=1;
  do {

    /*
      Read a command into the command compare buffer and convert to uppercase.
    */

    Ptr=BufferWord;
    do {
      *Ptr=UpCase(*L);
      L++;
      Ptr++;
    } while (*L>'Z');
    *Ptr='\0';

    /*
      Compare the command already read to the one given by the caller. If they
      match, return apropiate number, which is stored in "WordID".
    */

    if (!strcmp(BufferIn,BufferWord)) return(WordID);

    /*
      Update "WordID" and loop again.
    */

    WordID++;

  } while (*L);

  return(kUnknown);
}


static TKeyWord ReadKeyword(void)
{
  char KeyWordStr[RESERVED_WORD_LEN_LIMIT+1];

  ReadWordIntoBuf(KeyWordStr,RESERVED_WORD_LEN_LIMIT+1);
  return(RecognizeWord(KeyWordStr,
    "System" "Program" "Translator" "Modulegen" "Library"
    "Author" "License" "Credit" "Title"
    "Copyright"
    "Brief" "Desc" "Source" "Header" "Shared" "Test"
  ));
}

static void AnalyzeHeaderList(void)
{
  char Buffer[14];
  TFileInfo *F;

  while (true) {
    Buffer[13]='\0';
    ReadWordIntoBuf(Buffer,14);
    if (Buffer[0]=='\0') break;
    F=RegisterModuleIdentifiedByHeader(Buffer);
    if (F!=NULL) {
      if (!RegisterDependencyFile(F,true)) {
        TokError("File referenced more than once");
      }
    }
  }
}

static void AnalyzeDOPLine(void)
{
  TKeyWord OpCode;

  OpCode=ReadKeyword();
  switch (OpCode) {
    case kSYSTEM:{
      if (WasLibrary) {
        TokError("This keyword must appear before LIBRARY");
      } else if (WasSystem) {
        TokError("Keyword mentioned more than once");
      }
      WasSystem=true;
      break;
    }
    case kSOURCE:
    case kSHARED:
    case kTEST:
    case kPROGRAM:
    case kTRANSLATOR:
    case kMODULEGEN: {
      TokError("This keyword is not allowed in SYSLIB");
      break;
    }
    case kLIBRARY: {
      if (!WasSystem) {
        TokError("The SYSTEM keyword must appear somewhere above");
      } else if (WasLibrary) {
        TokError("Keyword mentioned more than once");
      }
      WasLibrary=true;
      break;
    }
    case kHEADER: {
      if (!WasLibrary) {
        TokError("This may appear only after LIBRARY keyword");
      }
      AnalyzeHeaderList();
      break;
    }
    case kAUTHOR:
    case kLICENSE:
    case kCREDIT:
    case kTITLE:
    case kCOPYRIGHT:
    case kBRIEF:
    case kDESC: {
      break;
    }
    default: {
      TokError("Invalid keyword");
      break;
    }
  }
}

TFileInfo Dop;

void ReadPackageDescription(const char *FileName)
{
  WasSystem=false;
  WasLibrary=false;
  if (!OpenFile(FileName)) {
    Fatal("Unable to read package description file");
  }
  Dop.FileName="syslib.dop";
  ActivateFile(&Dop);
  while (!EndOfFile()) {
    FetchLine();
    StripBlanksAndComments();
    if (Line[0]!='\0') AnalyzeDOPLine();
  }
  DeactivateFile();
}

void PrepareFileProcessing(void)
{
  int i;

  ReinitializeFileProcessingStack();
  for (i=0;i<Dop.DepCount;i++) RegisterFileToBeProcessed(Dop.DepList[i]);
}
