#include "defs.gh"
#include "types.h"
#include "strings.h"
#include "env.h"
#include "traps.h"
#include "ltwd.h"

#undef int
#undef unsigned
#include "filein.h"

#define __INTERFACE__
#include "cfgfile.h"
#undef __INTERFACE__

#define MAX_CONFIG_LINE_SIZE 2048
#define MAX_CONFIG_ITEMS 1024

char CodeBaseDir[INTERNAL_DIR_SPEC_SIZE+1];
char InternalDir[INTERNAL_DIR_SPEC_SIZE+1];
static char *Config[MAX_CONFIG_ITEMS];
static uint16 ConfigItems;

static bool IsWhiteSpace(char Ch)
{
  return((Ch=='\t') || (Ch=='\n') || (Ch=='\r') || (Ch==' '));
}

static void ProcessLine(char *Line)
{
  int16 Len;
  int16 Pos;
  char Ch;
  bool WhiteSpace;

  /*
    Remove the leading whitespace
  */
  do {
    WhiteSpace=IsWhiteSpace(*Line);
    if (WhiteSpace) Line++;
  } while (WhiteSpace);
  /*
    Remove any trailing whitespace
  */
  Len=StrLen(Line)-1;
  do {
    WhiteSpace=IsWhiteSpace(Line[Len]);
    if (WhiteSpace) Len--;
  } while (WhiteSpace && Len>0);
  Line[Len+1]='\0';
  /*
    Find the "=" sign. Report a fatal error if none is found.
  */
  Pos=0;
  do {
    Ch=Line[Pos];
    if (Ch!='=') Pos++;
  } while ((Ch!='=') && (Ch!='\0'));
  if (Ch=='\0') Fatal("Malformed configuration line");
  /*
    Copy the line
  */
  Line=StrDup(Line);
  Config[ConfigItems++]=Line;
  /*
    Split the line at the "=" sign to two pieces. The first one is the name
    of the configuration item and the second one is the value. The position
    of the sign is still in "Pos".
  */
  Line[Pos]='\0';
}

static void SetupDirectorySpec(const char *Dir,char *Buf)
{
  char *Ptr;
  char Ch;

  Ptr=StrNCpy(Buf,Dir,INTERNAL_DIR_SPEC_SIZE+1);
#if __e__FSPEC_TYPE__ == __v__msdos__
  Ch='\\';
#elif __e__FSPEC_TYPE__ == __v__unix__
  Ch='/';
#endif
  if (Ptr==CodeBaseDir || *(Ptr-1)!=Ch) {
    if (Ptr<CodeBaseDir+INTERNAL_DIR_SPEC_SIZE+1) *(Ptr++)=Ch;
  }
}

static void SetupCodeBaseDirectory(void)
{
  const char *RootSpec;
#if __e__CODEBASE_DIR__ == __v__linux__
  char *Ptr;
#endif

  CodeBaseDir[INTERNAL_DIR_SPEC_SIZE]='\0';
  RootSpec=GetEnv("OSHSCODEBASEDIR");
  if (RootSpec==NULL) {
#if __e__CODEBASE_DIR__ == __v__linux__
    Ptr=StrNCpy(CodeBaseDir,GetEnv("HOME"),INTERNAL_DIR_SPEC_SIZE+1);
    if (Ptr==CodeBaseDir || *(Ptr-1)!='/') {
      if (Ptr<CodeBaseDir+INTERNAL_DIR_SPEC_SIZE+1) *(Ptr++)='/';
    }
    StrNCpy(Ptr,"oshs/",INTERNAL_DIR_SPEC_SIZE+1-(Ptr-CodeBaseDir));
#elif __e__CODEBASE_DIR__ == __v__msdos__
    StrCpy(CodeBaseDir,"C:\\OSHS\\");
#else
    Fatal("OSHSCODEBASEDIR not set and no default provided");
#endif
  } else {
    SetupDirectorySpec(RootSpec,CodeBaseDir);
  }
}

void ReadConfig(void)
{
  char Line[MAX_CONFIG_LINE_SIZE];
  const char *RootSpec;
  char *Ptr;
  uint16 Size;

  /*
    Determine the specification of the configuration file
  */
  InternalDir[INTERNAL_DIR_SPEC_SIZE]='\0';
  RootSpec=GetEnv("OSHSINTERNALDIR");
  if (RootSpec==NULL) {
#if __e__INTERNAL_DIR__ == __v__linux__
    Ptr=StrNCpy(InternalDir,GetEnv("HOME"),INTERNAL_DIR_SPEC_SIZE+1);
    if (Ptr==InternalDir || *(Ptr-1)!='/') {
      if (Ptr<InternalDir+INTERNAL_DIR_SPEC_SIZE+1) *(Ptr++)='/';
    }
    StrNCpy(Ptr,".oshs",INTERNAL_DIR_SPEC_SIZE+1-(Ptr-InternalDir));
#elif __e__INTERNAL_DIR__ == __v__msdos__
    StrCpy(InternalDir,"C:\\OSHSINT");
#elif __e__INTERNAL_DIR__ == __v__optional__
    return;
#else
    Fatal("OSHSINTERNALDIR not set and no default provided");
#endif
  } else {
    SetupDirectorySpec(RootSpec,InternalDir);
  }
  if (InternalDir[INTERNAL_DIR_SPEC_SIZE]!='\0') {
    Fatal("OSHS internal directory specification too long");
  }
  Ptr=StrCpy(Line,InternalDir);
  StrCpy(Ptr,"/oshs.cfg");
  /*
    Initialize the configuration
  */
  ConfigItems=0;
  /*
    Prepare the internal directory if it does not exist yet. This must be
    done here since the rest of the procedure tries to touch it.
  */
  PrepareInternalDirectory();
  /*
    Open the file. If no configuration file is found, silently return.
  */
  if (!OpenTheFile(Line)) return;
  /*
    Read the lines and process them
  */
  do {
    Size=ReadLine(Line,sizeof(Line));
    if (Line[0]=='#' || Line[0]=='\0') continue;
    if (Size==sizeof(Line)) Fatal("Configuration line too long");
    ProcessLine(Line);
  } while (Size>0);
}

void SetupConfigFileVariables(void)
{
  /*
    Setup the codebase directory now. This is done here so it is possible
    to override this setting in the config file.
  */
  SetupCodeBaseDirectory();
}
