#include "defs.gh"
#include "types.h"
#include "memory.h"
#include "kernel.h"
#include "pdfsys.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pddev.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "disklett.h"
#include "diskdumb.h"
#include "disknull.h"
#include "diskmain.h"
#include "errno.h"
#include "strings.h"
#if  __e__INET_DRV__ != __v__none__
#include "inet.h"
#endif

#define __INTERFACE__
#include "devmain.h"
#undef __INTERFACE__

#define MAX_DEVICES 16

typedef struct {
  POpenProc Open;
  char Name[65];
  TDeviceType Type;
} TDeviceInfo;

static TDeviceInfo *Devices;
static int16 Current;

uint8 AddDevice(
  TDeviceType Type,
  char *Name,
  POpenProc OpenProc
)
{
  TDeviceInfo *Dev;

  assert(StrLen(Name)<64);
  assert(OpenProc!=NULL);
  Dev=Devices;
  while (Dev->Open!=NULL) {
    Dev++;
    assert(Dev-Devices<MAX_DEVICES);
  }
  Dev->Type=Type;
  StrCpy(Dev->Name,Name);
  Dev->Open=OpenProc;
  return (Dev-Devices);
}

void InitDevices(void)
{
  int16 i;
  TDeviceInfo *Dev;

  Current=MAX_DEVICES-1;
  Devices=KMEM__malloc(MAX_DEVICES*sizeof(*Devices));
  Dev=Devices;
  for (i=0;i<MAX_DEVICES;i++) {
    Dev->Open=NULL;
    Dev++;
  }
  AddNullFullZero();
#if  __e__INET_DRV__ != __v__none__
  AddInternetDevices();
#endif
}

static int16 GenericStat(void *fd,TFileInfo *Info)
{
  TGenericDeviceHandle *D;
  TDeviceInfo *N;

  D=fd;
  N=D->DevNode;
  Info->CreatedTime=(uint64)-1;
  Info->ModifiedTime=(uint64)-1;
  Info->Owner=65535;
  Info->Group=65535;
  StrToUpr(Info->Description,N->Name);
  Info->Size=0;
  Info->Attr=fmPKTFIFO;
  Info->ID=(uint32)-1;
  return(sizeof(*Info));
}

static int16 DevOpenCore(uint8 DevNum,TOpenReport *R)
{
  int16 Result;
  TDeviceInfo *Dev;
  TGenericDeviceHandle *H;

  /*
    Attempt to find the device in the device table. If the device name is not
    found in the table, return eNOENT.
  */
  if (DevNum>=MAX_DEVICES) return(fseNODEV);
  Dev=Devices+DevNum;
  if (Dev->Open==NULL) return(fseNODEV);
  /*
    Open the device using the correct driver. Also fill in the device name
    for the generic device node info dumper.
  */
  Result=Dev->Open(R);
  if (Result==0 && R->Handler==NULL) {
    H=R->Data;
    if (H->H.Stat==NULL) H->H.Stat=GenericStat;
    H->DevNode=Dev;
  }
  return(Result);
}

static int16 DevOpen(uint8 DevNum,TOpenDeviceReply *OR)
{
  int16 Handle;
  int16 Result;
  TOpenReport R;

  Handle=CreateChannelNoLock((void *)&R.Channel);
  if (Handle<0) return(Handle);
  Result=DevOpenCore(DevNum,&R);
  if (Result==0) {
    if (R.Handler==NULL) {
      InitializeOpenFileChannel(R.Channel,R.Data,R.Close,R.Destroy);
    } else {
      InitializeChannel(R.Channel,R.Handler,R.Data,R.Close,R.Destroy);
    }
    OR->Handle=Handle;
    Result=sizeof(*OR);
  } else {
    CloseCoreNoLock(Handle);
  }
  return(Result);
}

int16 FindDevice(TDevInfoReply *Out)
{
  TDeviceInfo *Dev;

  do {
    Current++;
    if (Current==MAX_DEVICES) {
      Current--;
      return(0);
    }
    Dev=Devices+Current;
  } while (Dev->Open==NULL);
  Out->ID=Current;
  Out->Type=Dev->Type;
  MemSet(Out->InfoString,sizeof(Out->InfoString),0);
  StrCpy(Out->InfoString,Dev->Name);
  return(sizeof(*Out));
}

int16 HandleDevices(void *Msg, int16 Size, void *Out, void *PrivateData)
{
  TOpenDeviceCmd *C;

  if (Size<1) return(beREQLEN);
  C=Msg;
  switch (C->Command) {
    case cmdFindFirstDevice: {
      Current=-1;
      /* FALL THROUGH */
    }
    case cmdFindNextDevice: {
      return(FindDevice(Out));
    }
    case cmdOpenDevice: {
      if (Size<2) return(beREQLEN);
      return(DevOpen(C->Device,Out));
    }
    default: {
      return beINVCMD;
    }
  }
}


void CreateDevicesHandle(void)
{
  void *Channel;

  assert(CreateChannel(&Channel)==hndDEVICES);
  InitializeChannel(Channel,HandleDevices,NULL,DumbClose,DumbDestroy);
}
