#include "defs.gh"
#include "types.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "diskdir.h"
#include "disklett.h"
#include "diskspec.h"
#include "traps.h"
#include "memory.h"
#include "screen.h"
#include "exec.h"
#include "pddev.h"
#include "devmain.h"
#include "strings.h"

#define __INTERFACE__
#include "diskmain.h"
#undef __INTERFACE__

typedef int16 TReqHandler(void *Req,void *Out, void *PrivateData);

static int16 HRead(void *Req,void *Out, void *PrivateData)
{
  TReadCmd *R;
  TGenericFileHandle *H;
  int32 Result;
  TIOReply *O;

  R=Req;
  H=PrivateData;
  O=Out;
  if (R->Size<0) return(beRANGE);
  Result=H->Read(PrivateData,R->Ptr,R->Size);
  if (Result<0) {
    assert(Result>=-0x8000L);
    return(Result);
  }
  O->Size=Result;
  return(sizeof(*O));
}

static int16 HWrite(void *Req,void *Out, void *PrivateData)
{
  TWriteCmd *R;
  TGenericFileHandle *H;
  int32 Result;
  TIOReply *O;

  R=Req;
  H=PrivateData;
  O=Out;
  if (R->Size<0) return(beRANGE);
  Result=H->Write(PrivateData,R->Ptr,R->Size);
  if (Result<0) {
    assert(Result>=-0x8000);
    return(Result);
  }
  O->Size=Result;
  return(sizeof(*O));
}

static int16 HReadNextMessage(void *Out, void *PrivateData)
{
  TGenericFileHandle *H;
  TDataCmd *O;
  int16 Size;
  int32 Result;
  bool Destroy;

  H=PrivateData;
  O=Out;
  do {
    Destroy=true;
    if (H->RemainingSize==0) {
      break;
    }
    O->Command=cmdDATA;
    Size=MAX_MESSAGE_SIZE-1;
    if (Size>H->RemainingSize) Size=H->RemainingSize;
    Result=H->Read(PrivateData,O->Data,Size);
    if (Result<=0) {
      assert(Result>=-0x8000L);
      break;
    }
    H->RemainingSize-=Result;
    Result++;
    Destroy=false;
  } while (false);
  if (Destroy) {
    InstallPostHandler(H->Channel,NULL);
    Result=0;
  }
  return(Result);
}

static int16 HReadClassic(void *Req,void *Out, void *PrivateData)
{
  TReadCmd *R;
  TGenericFileHandle *H;

  R=Req;
  H=PrivateData;
  if (R->Size<0) return(beRANGE);
  H->RemainingSize=R->Size;
  InstallPostHandler(H->Channel,HReadNextMessage);
  return(HReadNextMessage(Out,PrivateData));
}

static int16 HWriteClassic(void *Req,void *Out, void *PrivateData)
{
  TDataCmd *R;
  TGenericFileHandle *H;
  int32 Result;
  TIOReply *O;
  int16 Size;

  R=Req;
  H=PrivateData;
  O=Out;
  assert(H->PacketSize>1);
  Size=H->PacketSize-1;
  Result=H->Write(PrivateData,R->Data,Size);
  if (Result<0) {
    assert(Result>=-0x8000);
    return(Result);
  }
  O->Size=Result;
  return(sizeof(*O));
}

static int16 HSeek(void *Req,void *Out, void *PrivateData)
{
  TSeekCmd *R;
  TGenericFileHandle *H;
  int32 Position;
  int16 How;

  R=Req;
  H=PrivateData;
  if (R->Page>0 || R->Position>0x7FFFFFFF) return(beNOTSUP);
  Position=R->Position;
  switch (R->Direction) {
    case dirABSOLUTE: How=seekSET; break;
    case dirBACKWARD: Position=-Position; /* FALL THROUGH */
    case dirFORWARD: How=seekCUR; break;
    default: return(beINVAL);
  }
  return(H->Seek(PrivateData,Position,How));
}

static int16 HTell(void *Req,void *Out, void *PrivateData)
{
  TGenericFileHandle *H;
  TTellReply *R;
  int16 Result;

  H=PrivateData;
  R=Out;
  Result=H->Tell(PrivateData,R);
  if (Result==0) Result=sizeof(TTellReply);
  return(Result);
}

static int16 HStat(void *Req,void *Out, void *PrivateData)
{
  TGenericFileHandle *H;
  int16 Result;

  H=PrivateData;
  Result=H->Stat(PrivateData,Out);
  if (Result==0) Result=sizeof(TFileInfo);
  return(Result);
}

static int16 HTruncate(void *Req,void *Out, void *PrivateData)
{
  TGenericFileHandle *H;
  int16 Result;

  H=PrivateData;
  Result=H->Truncate(PrivateData);
  return(Result);
}

static int16 HChMod(void *Req,void *Out, void *PrivateData)
{
  TAttrManipCmd *R;
  TGenericFileHandle *H;
  int16 Result;

  R=Req;
  H=PrivateData;
  Result=H->ChMod(PrivateData,R->Value);
  return(Result);
}

static int16 HChOwn(void *Req,void *Out, void *PrivateData)
{
  TAttrManipCmd *R;
  TGenericFileHandle *H;
  int16 Result;

  R=Req;
  H=PrivateData;
  Result=H->ChOwn(PrivateData,R->Value);
  return(Result);
}

static int16 HChGrp(void *Req,void *Out, void *PrivateData)
{
  TAttrManipCmd *R;
  TGenericFileHandle *H;
  int16 Result;

  R=Req;
  H=PrivateData;
  Result=H->ChGrp(PrivateData,R->Value);
  return(Result);
}

static int16 HandleOpenFile(void *Msg,int16 Size,void *Out,void *PrivateData)
{
  upint8 Command;
  TReqHandler *Handler;
  int16 ReqSize;
  TGenericFileHandle *H;

  if (Size<1) return(beREQLEN);

  H=PrivateData;
  if (GetPostHandler(H->Channel)!=NULL) {
    ReportBug("Message pipe deadlock would occur");
  }
  Command=((TNoParamCmd *)Msg)->Command;
  ReqSize=1;
  switch (Command) {
    case cmdREAD: {
      ReqSize=sizeof(TReadCmd);
      Handler=HRead;
      break;
    }
    case cmdWRITE: {
      ReqSize=sizeof(TWriteCmd);
      Handler=HWrite;
      break;
    }
    case cmdSEEK: {
      ReqSize=sizeof(TSeekCmd);
      Handler=HSeek;
      break;
    }
    case cmdTELL: {
      Handler=HTell;
      break;
    }
    case cmdFSTAT: {
      Handler=HStat;
      break;
    }
    case cmdFCHOWN:
    case cmdFCHGRP:
    case cmdFCHMOD: {
      ReqSize=sizeof(TAttrManipCmd);
      switch (Command) {
        case cmdFCHMOD: Handler=HChMod; break;
        case cmdFCHOWN: Handler=HChOwn; break;
        case cmdFCHGRP: Handler=HChGrp; break;
      }
      break;
    }
    case cmdTRUNCATE: {
      Handler=HTruncate;
      break;
    }
    case cmdREADANDSEND: {
      ReqSize=sizeof(TReadAndSendCmd);
      Handler=HReadClassic;
      break;
    }
    case cmdDATA: {
      H->PacketSize=Size;
      ReqSize=2;
      Handler=HWriteClassic;
      break;
    }
    default: {
      return(beINVCMD);
    }
  }
  if (Size<ReqSize) return(beREQLEN);
  return(Handler(Msg,Out,PrivateData));
}

void InitializeOpenFileChannel(
  void *Channel,
  void *Data,
  TCloseProc *Closer,
  TDestroyProc *Destroyer
)
{
  TGenericFileHandle *H;

  H=Data;
  H->Channel=Channel;
  InitializeChannel(Channel,HandleOpenFile,Data,Closer,Destroyer);
}

static void *CreatePipeHandle(int16 fd)
{
  TOpenReport R;
  int16 Result;
  void *Channel;

  PipeToChannel(fd,&R);
  if (R.Data==NULL) Fatal("Error creating pipe for standard I/O handle");
  Result=CreateChannel(&Channel);
  assert(Result==fd);
  InitializeOpenFileChannel(Channel,R.Data,R.Close,R.Destroy);
  return(R.Data);
}

static int16 HOpen(void *Req,void *Out, void *PrivateData)
{
  TOpenCmd *C;
  TOpenReply *R;
  TOpenReport Rep;
  int16 Handle;
  TFileSpec SpecBuf;
  TEmulatedDisk *Drive;
  int16 ErrNo;
  TRootStatus *S;
  TGenericFileHandle *H;

  C=Req;
  R=Out;
  S=PrivateData;
  /*
    First things first: create the channel and bail out if none could be
    created. Once the filesystem object open request is carried out, the
    filesystem may be modified and I don't want to return an error code
    in that case.
  */
  Handle=CreateChannelNoLock((void *)&Rep.Channel);
  if (Handle<0) return(Handle);
  /*
    Convert the file specification. If it is too long, return failure
  */
  Drive=ConvertFileSpecEx(S,C->FileSpec,SpecBuf);
  if (Drive==NULL) {
    CloseCoreNoLock(Handle);
    return(S->Result);
  }
  /*
    Attempt to open the file
  */
  Rep.Handler=NULL;
  switch (Drive->Type) {
    case dtDIRECT:ErrNo=DirectOpen(SpecBuf,C->Mode,&Rep); break;
    default: {
      ReportBug("Invalid drive type flag");
      return(-127);
    }
  }
  if (ErrNo!=0) {
    CloseCoreNoLock(Handle);
    return(ErrNo);
  }
  /*
    Initialize the reader status used by the classic file reader. Fill in the
    pointer to the channel and deactivate the "classic read in progress"
    flag.
  */
  H=Rep.Data;
  H->Channel=Rep.Channel;
  /*
    Allocate new channel for the opened file. The file handle structure is
    created by the opening call and the pointer is filled into the NPD
    pointer variable. The channel representing the file will use the file I/O
    protocol which is handled by the HandleOpenFile() call. Beware that
    channels are already locked at this point. If the channel creation call
    fails, the (already opened) file must be closed using the closer
    provided by the opening call. The channel creator returns a handle
    if successfull and an error code if failed so on success write the handle
    into the output message buffer and return 2.
  */
  if (Rep.Handler==NULL) {
    InitializeOpenFileChannel(Rep.Channel,Rep.Data,Rep.Close,Rep.Destroy);
  } else {
    InitializeChannel(
      Rep.Channel,
      Rep.Handler,
      Rep.Data,
      Rep.Close,
      Rep.Destroy
    );
  }
  *((upint16 *)Out)=Handle;
  Handle=2;
  return(Handle);
}

static void ProcessItem(TDirItem *I,TGenericDirHandle *H)
{
  if (H->MsDosProcessing) {
    StrToLwr(I->Name,I->Name);
  }  
}

static int32 HReadDir(void *dd,void *Buf,int32 Size)
{
  TGenericDirHandle *H;
  int32 Result;
  TDirItem I;

  H=dd;
  /*
    Fetch the item using the lower level item getting call.
  */
  Result=H->ReadDir(dd,&I);
  if (Result<=0) return(Result);
  /*
    Process the fetched item. Special processing is needed for example for
    MS-DOS manesystems.
  */
  ProcessItem(&I,H);
  /*
    Return the data and the success flag.
  */
  if (Size>sizeof(I)) Size=sizeof(I);
  MemCpy(Buf,&I,Size);
  return(Size);
}

static int16 HOpenDir(void *Req,void *Out, void *PrivateData)
{
  TOpenCmd *C;
  TOpenReply *R;
  TOpenReport Rep;
  int16 Handle;
  TFileSpec SpecBuf;
  TEmulatedDisk *Drive;
  int16 ErrNo;
  TRootStatus *S;
  TGenericDirHandle *H;
  void *Channel;

  C=Req;
  R=Out;
  S=PrivateData;
  /*
    Convert the file specification. If it is too long, return failure
  */
  Drive=ConvertFileSpecEx(S,C->FileSpec,SpecBuf);
  if (Drive==NULL) return(S->Result);
  /*
    Sanity checks for the open mode flags. Directories can be open as read
    only, not created when not existing. Synchronized access flag and
    appendance flag are ignored for directories.
  */
  if ((C->Mode & oWRONLY) != 0) return(beINVAL);
  if ((C->Mode & oRDONLY) == 0) return(beINVAL);
  if ((C->Mode & oeTRUNC) != 0) return(beINVAL);
  if ((C->Mode & onCREAT) != 0) return(beINVAL);
  /*
    Attempt to open the file
  */
  switch (Drive->Type) {
    case dtDIRECT:ErrNo=DirectOpenDir(SpecBuf,&Rep); break;
    default: {
      ReportBug("Invalid drive type flag");
      return(-127);
    }
  }
  if (ErrNo!=0) return(beUNKERR);
  assert(Rep.Data!=NULL);
  H=Rep.Data;
  H->H.Read=HReadDir;
  /*
    Setup the filename conversion if the namesystem of the drive is set
    to MS-DOS. MS-DOS nnamesystems use capital letters for their file names.
  */
  H->MsDosProcessing=(Drive->FSType==fsMSDOS);
  /*
    When "action when existing" is set to "report error" and the directory
    was opened successfully, the call must return fseEXIST. This is used to
    check if a particular file specification specifies a directory.
  */
  Handle=0;
  if ((C->Mode & oeOPEN) == 0) Handle=fseEXIST;
  /*
    Allocate new channel for the opened file. The file handle structure is
    created by the opening call and the pointer is filled into the NPD
    pointer variable. The channel representing the file will use the file I/O
    protocol which is handled by the HandleOpenFile() call. Beware that
    channels are already locked at this point. If the channel creation call
    fails, the (already opened) file must be closed using the closer
    provided by the opening call. The channel creator returns a handle
    if successfull and an error code if failed so on success write the handle
    into the output message buffer and return 2.
  */
  if (Handle==0) {
    Handle=CreateChannelNoLock(&Channel);
    if (Handle>=0) {
      InitializeOpenFileChannel(
        Channel,
        Rep.Data,
        Rep.Close,
        Rep.Destroy
      );
      *((upint16 *)Out)=Handle;
      Handle=2;
    }
  }
  if (Handle<0) {
    Rep.Close(Rep.Data);
    Rep.Destroy(Rep.Data);
  }
  return(Handle);
}

static int16 HIStat(void *Req,void *Out, void *Root)
{
  TIStatCmd *C;
  uint32 Number;
  int16 Result;
  uint8 DriverID;

  C=Req;
  Number=C->ID;
  DriverID=Number >> 24;
  Number&=0x00FFFFFF;
  switch (DriverID) {
    case dtDIRECT:Result=DirectReadFileInfo(Number,Out); break;
    default: Result=beINVAL; break;
  }
  return(Result);
}

static int16 HRemove(void *Req,void *Out, void *Root)
{
  TManipCmd *C;
  TRootStatus *S;
  TFileSpec SpecBuf;
  TEmulatedDisk *Drive;
  int16 Result;

  C=Req;
  S=Root;
  Drive=ConvertFileSpecEx(Root,C->FileSpec,SpecBuf);
  if (Drive==NULL) return(S->Result);
  switch (Drive->Type) {
    case dtDIRECT: Result=DirectRemove(SpecBuf); break;
    default: Result=beINVAL; break;
  }
  return(Result);
}

static int16 HRename(void *Req,void *Out, void *Root)
{
  TManipCmd *C;
  TRootStatus *S;
  TFileSpec SrcSpec;
  TFileSpec DstSpec;
  TEmulatedDisk *SrcDrive;
  TEmulatedDisk *DstDrive;
  int16 Result;
  char *Spec;

  C=Req;
  S=Root;
  Spec=C->FileSpec;
  SrcDrive=ConvertFileSpecEx(Root,Spec,SrcSpec);
  if (SrcDrive==NULL) return(S->Result);
  do {
    Spec++;
  } while (*(Spec-1)!='\0');
  DstDrive=ConvertFileSpecEx(Root,Spec,DstSpec);
  if (DstDrive==NULL) return(S->Result);
  if (SrcDrive!=DstDrive) return(fseRENADRV);
  switch (SrcDrive->Type) {
    case dtDIRECT: Result=DirectRename(SrcSpec,DstSpec); break;
    default: Result=beINVAL; break;
  }
  return(Result);
}

static int16 HMkDir(void *Req,void *Out, void *Root)
{
  TManipCmd *C;
  TRootStatus *S;
  TFileSpec SpecBuf;
  TEmulatedDisk *Drive;
  int16 Result;

  C=Req;
  S=Root;
  Drive=ConvertFileSpecEx(Root,C->FileSpec,SpecBuf);
  if (Drive==NULL) return(S->Result);
  switch (Drive->Type) {
    case dtDIRECT: Result=DirectMkDir(SpecBuf); break;
    default: Result=beINVAL; break;
  }
  return(Result);
}

static int16 HRmDir(void *Req,void *Out, void *Root)
{
  TManipCmd *C;
  TRootStatus *S;
  TFileSpec SpecBuf;
  TEmulatedDisk *Drive;
  int16 Result;

  C=Req;
  S=Root;
  Drive=ConvertFileSpecEx(Root,C->FileSpec,SpecBuf);
  if (Drive==NULL) return(S->Result);
  switch (Drive->Type) {
    case dtDIRECT: Result=DirectRmDir(SpecBuf); break;
    default: Result=beINVAL; break;
  }
  return(Result);
}

static int16 HGetDrv(void *Req,void *Out, void *Root)
{
  TDriveInfo *D;
  TRootStatus *S;
  TEmulatedDisk *Drive;

  D=Out;
  S=Root;
  D->Name=S->CurrentDrive;
  Drive=GetDrive(S,D->Name);
  assert(Drive!=NULL);
  D->Type=(Drive->Type!=0)?Drive->FSType+1:0;
  return(sizeof(*D));
}

static int16 HSetDrv(void *Req,void *Out, void *Root)
{
  TSetDrvCmd *C;
  TRootStatus *S;
  TEmulatedDisk *Drive;

  C=Req;
  S=Root;
  Drive=GetDrive(S,C->Name);
  if (Drive==NULL) return(fseINVDRV);
  if (Drive->Type==0) return(fseINVDRV);
  S->CurrentDrive=C->Name;
  return(0);
}

static int16 HandleRoot(void *Msg, int16 Size, void *Out, void *PrivateData)
{
  upint8 Command;
  TReqHandler *Handler;
  int16 ReqSize;
  int16 Scan;
  upint8 *Ptr;

  if (Size<1) return(beREQLEN);

  Command=((TNoParamCmd *)Msg)->Command;
  ReqSize=1;
  Scan=0;
  Handler=NULL;
  switch (Command) {
    case cmdOPENFILE: {
      ReqSize=2;
      Handler=HOpen;
      Scan=1;
      break;
    }
    case cmdOPENDIR: {
      ReqSize=2;
      Handler=HOpenDir;
      Scan=1;
      break;
    }
    case cmdSETDRV: {
      ReqSize=2;
      Handler=HSetDrv;
      break;
    }
    case cmdISTAT: {
      ReqSize=sizeof(TIStatCmd);
      Handler=HIStat;
      break;
    }
    case cmdREMOVE:
    case cmdRENAME:
    case cmdRMDIR:
    case cmdMKDIR: {
      Scan=1;
      /* FALL THROUGH */
    }
    case cmdGETDRV: {
      ReqSize=1;
      switch (Command) {
        case cmdREMOVE: Handler=HRemove; break;
        case cmdRENAME: Handler=HRename;Scan++; break;
        case cmdRMDIR: Handler=HRmDir; break;
        case cmdMKDIR: Handler=HMkDir; break;
        case cmdGETDRV: Handler=HGetDrv; break;
      }
      break;
    }
    default: {
      return(beINVCMD);
    }
  }
  if (Size<ReqSize) return(beREQLEN);
  while (Scan>0) {
    Ptr=Msg;
    Ptr+=ReqSize;
    Size-=ReqSize;
    while (Size>0 && *Ptr!='\0') {
      Ptr++;
      Size--;
    }
    if (Size==0) return(beREQLEN);
    Scan--;
    Ptr++;
    Size--;
  }
  return(Handler(Msg,Out,PrivateData));
}

static TRootStatus *CreateRootHandle(void)
{
  TRootStatus *Data;
  int16 Result;
  void *Channel;

  Data=KMEM__malloc(sizeof(*Data));
  if (Data==NULL) {
    Fatal("Not enough memory for root directory status");
  }
  InitDisks(Data);
  Result=CreateChannel(&Channel);
  assert(Result==hndROOTFS);
  InitializeChannel(Channel,HandleRoot,Data,DumbClose,DumbDestroy);
  return(Data);
}

void CreateStandardHandles(void)
{
  TRootStatus *Root;
  void *Channel;

  Channel=CreatePipeHandle(0);
  CreatePipeHandle(1);
  CreatePipeHandle(2);
  Root=CreateRootHandle();
  CreateExecHandle(Root);
  CreateDevicesHandle();
  CreateScreenHandle();
  SetupSpecialWriteHandling(Channel,Root);
}
