#include "defs.gh"
#include "types.h"
#include "traps.h"
#include "strings.h"
#include "memory.h"
#include "kernel.h"
#include "pdgfxcon.h"

#define __INTERFACE__
#include "gcwdata.h"
#undef __INTERFACE__

/*
  A helper function to extract the shift size from a mask.
*/
static uint8 GetShift(uint32 Mask)
{
  uint8 Result;

  assert(Mask!=0);
  Result=0;
  while ((Mask&(1<<Result))==0) Result++;
  return(Result);
}

/*
  A helper function to extract the drop size from a mask. Drop size is count
  of bits that must be thrown away from the 8 bit number.
*/
static uint8 GetDrop(uint32 Mask,uint8 Shift)
{
  uint8 Result;

  Result=0;
  while ((Mask&(1<<(Result+Shift)))!=0) Result++;
  assert(Result<=8);
  return(8-Result);
}

/*
  Build precomputed values for the format. Expexts format structure with the
  color masks filled in.
*/
void BuildColorFormat(TColorFormat *F)
{
  F->RShift=GetShift(F->RMask);
  F->RDrop=GetDrop(F->RMask,F->RShift);
  F->GShift=GetShift(F->GMask);
  F->GDrop=GetDrop(F->GMask,F->GShift);
  F->BShift=GetShift(F->BMask);
  F->BDrop=GetDrop(F->BMask,F->BShift);
}

void UnpackColor(TColorFormat *F,uint32 Pixel)
{
  F->R=(Pixel&F->RMask)>>F->RShift<<F->RDrop;
  F->G=(Pixel&F->GMask)>>F->GShift<<F->GDrop;
  F->B=(Pixel&F->BMask)>>F->BShift<<F->BDrop;
}

uint32 PackColor(TColorFormat *F)
{
  uint32 Result;

  Result =((F->R>>F->RDrop)<<F->RShift)&F->RMask;
  Result+=((F->G>>F->GDrop)<<F->GShift)&F->GMask;
  Result+=((F->B>>F->BDrop)<<F->BShift)&F->BMask;
  return(Result);
}

uint32 ConvertPixel(TColorFormat *In,TColorFormat *Out,uint32 Pixel)
{
  UnpackColor(In,Pixel);
  Out->R=In->R;
  Out->G=In->G;
  Out->B=In->B;
  return(PackColor(Out));
}

void ResetPixelsToFillIfNeeded(TColorConversionInfo *I)
{
  if (I->PixelsToFill==0) {
    I->PixelsToFill=I->PixelsOnLine;
    I->OutPtr+=I->LinePaddingBytes;
    I->LineNumber++;
  }
}

static bool IsTheRGBLayoutTheSame(TColorConversionInfo *I)
{
  return(
    (I->In.RMask==I->Out.RMask) &&
    (I->In.GMask==I->Out.GMask) &&
    (I->In.BMask==I->Out.BMask)
  );
}

static uint32 ReadPixel(TColorConversionInfo *I)
{
  uint32 Pixel;

  switch (I->In.BitsPerPixel) {
    case 8: {
      Pixel=*(upint8 *)(I->InPtr++);
      break;
    }
    case 16: {
      Pixel=*(upint16 *)(I->InPtr);
      I->InPtr+=2;
      break;
    }
    case 24: {
#if __b__LITTLE_ENDIAN__
      Pixel=*(upint16 *)(I->InPtr);
      I->InPtr+=2;
      Pixel+=(*(upint8 *)(I->InPtr++))*65536;
#else
      Pixel=(*(upint8 *)(I->InPtr++))*65536;
      Pixel+=*(upint16 *)(I->InPtr);
      I->InPtr+=2;
#endif
      break;
    }
    case 32: {
      Pixel=*(upint32 *)(I->InPtr);
      I->InPtr+=4;
      break;
    }
    default: {
      ReportBug("Invalid bits per pixel setting in input pixel format");
      break;
    }
  }
  Pixel&=I->In.RMask|I->In.GMask|I->In.BMask;
  return(Pixel);
}

static void WritePixel(TColorConversionInfo *I,uint32 Pixel)
{
  switch (I->Out.BitsPerPixel) {
    case 8: {
      *(upint8 *)(I->OutPtr++)=Pixel;
      break;
    }
    case 16: {
      *(upint16 *)(I->OutPtr)=Pixel;
      I->OutPtr+=2;
      break;
    }
    case 24: {
#if __b__LITTLE_ENDIAN__
      *(upint16 *)(I->OutPtr)=Pixel;
      I->OutPtr+=2;
      *(upint8 *)(I->OutPtr++)=Pixel/65536;
#else
      *(upint8 *)(I->OutPtr++)=Pixel/65536;
      *(upint16 *)(I->OutPtr)=Pixel;
      I->OutPtr+=2;
#endif
      break;
    }
    case 32: {
      *(upint32 *)(I->OutPtr)=Pixel;
      I->OutPtr+=4;
      break;
    }
    default: {
      ReportBug("Invalid bits per pixel setting in output pixel format");
      break;
    }
  }
}

void ConvertPixels(TColorConversionInfo *I)
{
  bool SameRGBLayout;
  bool SamePixelSize;
  uint32 Pixel;
  uint8 PixelSize;
  uint16 PixelsToTransfer;
  uint32 Size;

  /*
    Determine which portions of the formats are the same
  */
  SameRGBLayout=IsTheRGBLayoutTheSame(I);
  SamePixelSize=(I->In.BitsPerPixel==I->Out.BitsPerPixel);
  /*
    Store the starting position in the buffer
  */
  if (!I->UnprocessedPixels) {
    I->UnprocessedPixels=true;
    I->SLine=I->LineNumber;
    I->SPixelsToFill=I->PixelsToFill;
    I->SOutPtr=I->OutPtr;
  }
  /*
    Update the count of free pixels
  */
  assert(I->FreePixels>=I->PixelCount);
  I->FreePixels-=I->PixelCount;
  /*
    If the two formats are exactly the same and there is no padding
    required, simply copy the data. If padding is required but the two
    formats are exactly the same, copy the data line by line and insert
    the required paddings.
  */
  if (SameRGBLayout && SamePixelSize) {
    PixelSize=I->In.BitsPerPixel / 8;
    if (I->LinePaddingBytes==0) {
      Size=PixelSize * I->PixelCount;
      MemCpy(I->OutPtr,I->InPtr,Size);
      I->InPtr+=Size;
      I->OutPtr+=Size;
      I->PixelCount+=I->PixelsOnLine - I->PixelsToFill;
      I->PixelsToFill=I->PixelsOnLine - (I->PixelCount % I->PixelsOnLine);
      I->LineNumber+=I->PixelCount / I->PixelsOnLine;
      I->PixelCount=0;
    } else {
      while (I->PixelCount>0) {
        PixelsToTransfer=I->PixelsToFill;
        if (PixelsToTransfer>I->PixelCount) PixelsToTransfer=I->PixelCount;
        Size=PixelsToTransfer * PixelSize;
        MemCpy(I->OutPtr,I->InPtr,Size);
        I->InPtr+=Size;
        I->OutPtr+=Size;
        I->PixelsToFill-=PixelsToTransfer;
        ResetPixelsToFillIfNeeded(I);
        I->PixelCount-=PixelsToTransfer;
      }
    }
  }
  /*
    Convert the pixels
  */
  for (;I->PixelCount>0;I->PixelCount--) {
    /*
      Read the pixel from the source image.
    */
    Pixel=ReadPixel(I);
    /*
      Convert it.
    */
    if (!SameRGBLayout) Pixel=ConvertPixel(&I->In,&I->Out,Pixel);
    /*
      Store the converted pixel into the destination image.
    */
    WritePixel(I,Pixel);
    /*
      Skip the padding bytes at the output and track the line number.
    */
    I->PixelsToFill--;
    ResetPixelsToFillIfNeeded(I);
  }
}

static void ClearMaskOutputBuffer(TColorConversionInfo *I)
{
  I->MaskOutBuf=0;
  if (I->MaskType==mtLeftmostIsLSB) {
    I->MaskOutBit=1;
  } else {
    I->MaskOutBit=128;
  }
}

void InitMaskProcessing(TColorConversionInfo *I)
{
  I->MaskType=mtLeftmostIsLSB;
  I->MaskBits=0;
  ClearMaskOutputBuffer(I);
}

static void FlushMaskBits(TColorConversionInfo *I)
{
  uint8 EmptyValue;
  EmptyValue=(I->MaskType==mtLeftmostIsLSB?1:128);
  if (I->MaskOutBit!=EmptyValue) {
    *(I->MaskPtr++)=I->MaskOutBuf;
    ClearMaskOutputBuffer(I);
  }
}

static void ResetMaskBufferIfNeeded(TColorConversionInfo *I)
{
  if (I->PixelsToFill==0) {
    FlushMaskBits(I);
    I->MaskPtr+=I->MaskPaddingBytes;
  }
}

int16 ConvertPixelsOfMaskedBitmap(TColorConversionInfo *I)
{
  bool SameRGBLayout;
  uint32 Pixel;
  uint8 PixelSize;
  void *Temp;
  uint32 PixelSet;
  uint32 PixelReset;
  uint32 MaskPixel;

  /*
    If the mask type is "pixmap", we need to determine the "Set" and
    "Reset" pixels.
  */
  if (I->MaskType==mtPixmap) {
    I->Out.R=I->PixelSetR;
    I->Out.G=I->PixelSetG;
    I->Out.B=I->PixelSetB;
    PixelSet=PackColor(&I->Out);
    I->Out.R=I->PixelResetR;
    I->Out.G=I->PixelResetG;
    I->Out.B=I->PixelResetB;
    PixelReset=PackColor(&I->Out);
  }
  /*
    Determine if the pixel formats are the same.
  */
  SameRGBLayout=IsTheRGBLayoutTheSame(I);
  /*
    Store the starting position in the buffer
  */
  if (!I->UnprocessedPixels) {
    I->UnprocessedPixels=true;
    I->SLine=I->LineNumber;
    I->SPixelsToFill=I->PixelsToFill;
    I->SOutPtr=I->OutPtr;
  }
  /*
    Compute the size of the pixel in bytes.
  */
  PixelSize=I->In.BitsPerPixel/8;
  /*
    Convert the pixels. Pixels in the data are stored in blocks. Each block
    starts with a mask byte. If a bit is set to 0, the pixel is transparent,
    otherwise it is opaque. After the mask byte the pixel values of the
    opaque pixels come. Only opaque pixels have their values specified in
    this list so the total number of pixel values that follow is equal to
    the number of bits set in the mask byte. The shortest block is the
    mask byte of 0x00 being alone, the longest block is the byte 0xFF
    followed by eight pixel values. The last block must have bits of its mask
    byte that don't have their representantive pixels in the destination
    bitmap set to zero.
  */
  while (I->InputSize>0 && I->PixelCount>0) {
    /*
      If the input mask buffer is empty, fetch a new one.
    */
    if (I->MaskBits==0) {
      I->MaskBits=8;
      I->MaskBuffer=*(upint8 *)(I->InPtr++);
      I->InputSize--;
      if (I->InputSize==0) break;
    }
    /*
      Fetch the color for the next pixel. If the mask says a pixel is
      transparent, a device dependent value (chosen by the driver) will be
      stored in the pixmap. Remember also to copy the mask bit to the mask
      output buffer.
    */
    if (I->MaskBuffer>127) {
      if (I->InputSize<PixelSize) return(screPARTPIX);
      Pixel=ReadPixel(I);
      I->InputSize-=PixelSize;
      if (!SameRGBLayout) Pixel=ConvertPixel(&I->In,&I->Out,Pixel);
      I->MaskOutBuf|=I->MaskOutBit;
      MaskPixel=PixelSet;
    } else {
      Pixel=I->TransparentPixel;
      MaskPixel=PixelReset;
    }
    /*
      Discard the mask input bit just processed.
    */
    I->MaskBuffer=(I->MaskBuffer&127)*2;
    I->MaskBits--;
    /*
      Advance to the next mask output bit.
    */
    switch (I->MaskType) {
      case mtLeftmostIsLSB: {
        if (I->MaskOutBit==128) {
          FlushMaskBits(I);
        } else {
          I->MaskOutBit*=2;
        }
        break;
      }
      case mtLeftmostIsMSB: {
        I->MaskOutBit/=2;
        if (I->MaskOutBit==0) FlushMaskBits(I);
        break;
      }
      case mtPixmap: {
        Temp=I->OutPtr;
        I->OutPtr+=I->PixmapMaskDelta;
        WritePixel(I,MaskPixel);
        I->OutPtr=Temp;
        break;
      }
      default: {
        ReportBug("Invalid mask type in transfer control structure");
      }
    }
    /*
      Write the pixel.
    */
    WritePixel(I,Pixel);
    /*
      Skip the padding bytes at the output and track the line number.
    */
    I->PixelCount--;
    I->PixelsToFill--;
    if (I->MaskType<=mtLeftmostIsMSB) ResetMaskBufferIfNeeded(I);
    ResetPixelsToFillIfNeeded(I);
  }
  /*
    Everything went OK.
  */
  return(0);
}

/*
  Pixmap info control structure handling.
*/

void PixIDInitialize(TPixmapInfoList *List)
{
  int16 i;

  for (i=0;i<256;i++) List->Block[i]=NULL;
  List->FirstMounted=NULL;
}

TPixmapInfo *PixIDGetInfo(TPixmapInfoList *List,uint16 ID)
{
  TPixmapInfoBlock *B;
  TPixmapInfo *I;

  B=List->Block[ID/256];
  if (B==NULL) return(NULL);
  I=&B->Info[ID%256];
  if (I->Pixmap==NULL) I=NULL;
  return(I);
}

TPixmapInfo *PixIDAllocate(TPixmapInfoList *List,uint16 ID)
{
  TPixmapInfoBlock *B;
  TPixmapInfo *I;

  B=List->Block[ID/256];
  if (B==NULL) {
    B=KMEM__calloc(sizeof(*B));
    List->Block[ID/256]=B;
  }
  if (B==NULL) return(NULL);
  I=&B->Info[ID%256];
  if (I->Pixmap==NULL) B->CountInBlock++;
  I->Pixmap=NO_PIXMAP;
  I->ScreenX=65535;
  return(I);
}

int16 PixIDFree(TPixmapInfoList *List,uint16 ID)
{
  TPixmapInfoBlock *B;
  TPixmapInfo *I;

  B=List->Block[ID/256];
  if (B==NULL) return(screBPXID);
  I=&B->Info[ID%256];
  if (I->Pixmap==NULL) return(screBPXID);
  if (B->CountInBlock==1) {
    KMEM__free(B);
    List->Block[ID/256]=NULL;
  } else {
    I->Pixmap=NULL;
    B->CountInBlock--;
  }
  return(0);
}

void RegisterMountedPixmap(TPixmapInfoList *List,TPixmapInfo *I)
{
  TPixmapInfo *Prev;

  assert(I->ScreenX!=65535);
  if (List->FirstMounted==NULL) {
    I->PrevMounted=I;
    I->NextMounted=I;
  } else {
    Prev=List->FirstMounted->PrevMounted;
    I->PrevMounted=Prev;
    List->FirstMounted->PrevMounted=I;
    I->NextMounted=List->FirstMounted;
    Prev->NextMounted=I;
  }
  List->FirstMounted=I;
}

void UnregisterMountedPixmap(TPixmapInfoList *List,TPixmapInfo *I)
{
  TPixmapInfo *Prev;
  TPixmapInfo *Next;

  assert(I->ScreenX==65535);
  if (I->NextMounted==I) {
    List->FirstMounted=NULL;
  } else {
    Prev=I->PrevMounted;
    Next=I->NextMounted;
    if (List->FirstMounted==I) List->FirstMounted=Next;
    Prev->NextMounted=Next;
    Next->PrevMounted=Prev;
  }
}
