#include "defs.gh"
#include "types.h"
#include "kernel.h"
#include "pdgfxcon.h"
#include "strings.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "diskdumb.h"
#include "diskmain.h"
#include "errno.h"
#include "scrvars.h"
#include "memory.h"
#include "thread.h"
#include "time.h"
#include "gcwdata.h"

#if __e__SCREEN_DRV__ == __v__SDL__
#include "gcwsdl.h"
#elif __e__SCREEN_DRV__ == __v__windows__
#include "gcww32.h"
#elif __e__SCREEN_DRV__ == __v__X11__
#include "gcwx11.h"
#endif


static int32 KeyboardRead(void *fd,void *Buf,int32 Size)
{
  upint16 Key;

  UnlockChannels();
  Key=ScrRawReadKey();
  LockChannels();
  if (Key==65535) return(0);
  if (Size>2) Size=2;
  MemCpy(Buf,&Key,Size);
  return(Size);
}

static int16 KeyboardClose(void *fd)
{
  return(ScrCloseKeyboard());
}

int16 ScrExecOpenKeyboard(const void *Cmd,void *Out,void *PD)
{
  TGenericFileHandle *H;
  TOpenKeyboardReply *R;
  int16 Result;
  void *Channel;
  int16 Handle;

  R=Out;
  H=CreateFileHandleData(sizeof(*H));
  if (H==NULL) return(beNOMEM);
  UnlockChannels();
  Result=ScrOpenKeyboard();
  LockChannels();
  if (Result<0) return(Result);
  Handle=CreateChannelNoLock(&Channel);
  if (Handle<0) return(Handle);
  InitializeOpenFileChannel(Channel,H,KeyboardClose,DumbDestroy);
  H->Read=KeyboardRead;
  R->Handle=Handle;
  return(sizeof(*R));
}

typedef struct {
  uint32 RemainingPixels;
  uint16 LineCount;
  uint16 LineSize;
  uint8 PixelSize;
  TColorConversionInfo PTInfo;
  bool InstantUpdate;
  /*
    The list of IDs and their respective pixmaps. This list is an array
    indexed by the IDs that is two-levelled to avoid excessive memory
    consumption when there are little or no pixmaps in use.
  */
  TPixmapInfoList List;
  /*
    When the transfer is directed to an offscreen pixmap, the screen update
    must be avoided even in an instant update mode cause the screen is not
    changing on the writes at this case. This is not the same as the
    "InstantUpdate" flag, because the "InstantUpdate" flag must be kept set
    to "true" even when an offscreen pixmap is written to.
  */
  bool UpdateActive;
  /*
    Rectangle to be updated.
  */
  uint16 MinX;
  uint16 MinY;
  uint16 MaxX;
  uint16 MaxY;
} TVideoMemoryHandle;

void *UpdateRequired;
void *RectLock;
void *Updater;
TVideoMemoryHandle *CurrentData;

static void ClearRectangleToRedraw(TVideoMemoryHandle *Data)
{
  Data->MinX=65535;
  Data->MinY=65535;
  Data->MaxX=0;
  Data->MaxY=0;
}

static void RecordDirtyRectangle(
  TVideoMemoryHandle *Data,
  uint16 X,
  uint16 Y,
  uint16 Width,
  uint16 Height
) {
  if (Data->InstantUpdate) {
    ScrUpdateRectangle(X,Y,Width,Height);
  } else {
    if (Data->MinX>=X) Data->MinX=X;
    if (Data->MinY>=Y) Data->MinY=Y;
    X+=Width;
    Y+=Height;
    if (Data->MaxX<=X) Data->MaxX=X;
    if (Data->MaxY<=Y) Data->MaxY=Y;
  }
}

static void InitializeRectangleRedraw(TVideoMemoryHandle *Data)
{
  ClearRectangleToRedraw(Data);
}

static int16 FlushPixels(TVideoMemoryHandle *Data)
{
  uint16 StartLocY;
  uint16 LineCount;
  int16 Result;

  if (Data->PTInfo.UnprocessedPixels) {
    Result=ScrFlushPixels(&Data->PTInfo);
    if (Result!=0) return(Result);
    if (Data->UpdateActive) {
      StartLocY=Data->PTInfo.SLine;
      LineCount=Data->PTInfo.LineNumber-StartLocY;
      RecordDirtyRectangle(Data,0,StartLocY,Data->LineSize,LineCount);
    }
    Data->PTInfo.UnprocessedPixels=false;
    Data->PTInfo.FreePixels=0;
    if (Data->RemainingPixels==0) {
      Result=ScrDestroyAuxiliaryTransferInfo(&Data->PTInfo);
      if (Result!=0) return(Result);
    }
  }
  return(0);
}

static void InitializeScreenBufferTransfers(TVideoMemoryHandle *Data)
{
  Data->RemainingPixels=0;
  Data->PTInfo.UnprocessedPixels=false;
}

static int16 ExecOpenScreenRegion(const void *Msg,void *Buf,void *fd)
{
  TVideoMemoryHandle *Data;
  const TOpenScreenRegionCommand *Cmd;
  int16 Result;

  Data=fd;
  Cmd=Msg;
  if (Data->RemainingPixels!=0) return(screREGWRT);
  if (Cmd->X>=Data->LineSize || Cmd->Y>=Data->LineCount) return(beRANGE);
  if (Cmd->Width>Data->LineSize-Cmd->X) return(beRANGE);
  if (Cmd->Height>Data->LineCount-Cmd->Y) return(beRANGE);
  if (Cmd->Width!=Data->LineSize) return(beNOTSUP);
  Data->PTInfo.PixelsOnLine=Data->LineSize;
  Data->RemainingPixels=Cmd->Width*Cmd->Height;
  Data->PTInfo.LineNumber=Cmd->Y;
  Data->PTInfo.PixelsToFill=Data->PTInfo.PixelsOnLine;
  Data->PTInfo.FreePixels=0;
  Data->PTInfo.DestPixmap=ScrGetScreenPixmap();
  Data->PTInfo.MaskOutBit=0;
  Data->UpdateActive=true;
  Result=ScrPrepareAuxiliaryTransferInfo(&Data->PTInfo);
  if (Result!=0) Data->RemainingPixels=0;
  return(Result);
}

static void DestroyPixmap(TPixmapInfo *I)
{
  ScrDestroyPixmap(I->Pixmap);
}

static int16 CreatePixmapCore(
  const TCreatePixmapCommand *Cmd,
  TVideoMemoryHandle *Data
) {
  int16 Result;
  TPixmapInfo *I;

  I=PixIDAllocate(&Data->List,Cmd->ID);
  if (I==NULL) return(beNOMEM);
  if (I->Pixmap!=NO_PIXMAP) DestroyPixmap(I);
  if (Cmd->Flags & bfEnableTransparent) {
    Data->PTInfo.TransparentPixel=0;
    Result=ScrCreateMaskedPixmap(
      Cmd->Width,Cmd->Height,&Data->PTInfo.DestPixmap
    );
    InitMaskProcessing(&Data->PTInfo);
  } else {
    Result=ScrCreatePixmap(Cmd->Width,Cmd->Height,&Data->PTInfo.DestPixmap);
    Data->PTInfo.MaskOutBit=0;
  }
  if (Result!=0) {
    PixIDFree(&Data->List,Cmd->ID);
    return(Result);
  }
  I->Pixmap=Data->PTInfo.DestPixmap;
  I->Height=Cmd->Height;
  I->Width=Cmd->Width;
  return(0);
}

static int16 ExecCreatePixmap(const void *Msg,void *Buf,void *fd)
{
  TVideoMemoryHandle *Data;
  const TCreatePixmapCommand *Cmd;
  int16 Result;

  Data=fd;
  Cmd=Msg;

  if (!ScrCheckModeSet()) return(screNOMODESET);
  if (Data->RemainingPixels!=0) return(screREGWRT);
  Result=CreatePixmapCore(Cmd,Data);
  if (Result!=0) return(Result);
  Data->RemainingPixels=Cmd->Width*Cmd->Height;
  Data->PTInfo.LineNumber=0;
  Data->PTInfo.PixelsOnLine=Cmd->Width;
  Data->PTInfo.PixelsToFill=Cmd->Width;
  Data->PTInfo.FreePixels=0;
  Data->UpdateActive=false;
  Result=ScrPrepareAuxiliaryTransferInfo(&Data->PTInfo);
  if (Result!=0) Data->RemainingPixels=0;
  return(Result);
}

static int16 ExecCreateOutPixmap(const void *Msg,void *Buf,void *fd)
{
  TVideoMemoryHandle *Data;
  const TCreateOutPixmapCommand *Cmd;
  TCreatePixmapCommand Cmd1;

  Data=fd;
  Cmd=Msg;

  if (!ScrCheckModeSet()) return(screNOMODESET);
  if (Data->RemainingPixels!=0) return(screREGWRT);
  Cmd1.ID=Cmd->ID;
  Cmd1.Width=Cmd->Width;
  Cmd1.Height=Cmd->Height;
  Cmd1.Flags=0;
  return(CreatePixmapCore(&Cmd1,Data));
}

typedef struct {
  const upint8 *Code;
  uint16 Size;
  bool Error;
} TDrawInstructionList;

uint8 FetchByte(TDrawInstructionList *L)
{
  if (L->Size>0) {
    L->Size--;
    return(*(L->Code++));
  } else {
    L->Error=true;
    return(0);
  }
}

static int16 SaveMountedPixmapContent(TPixmapInfo *I)
{
  int16 Result;
  void *Screen;

  Screen=ScrGetScreenPixmap();
  Result=ScrTransferRectangle(
    Screen,I->ScreenX,I->ScreenY,I->Width,I->Height,
    I->Pixmap,0,0
  );
  return(Result);
}

static int16 ExecuteMountPixmap(
  TVideoMemoryHandle *Data,
  uint16 ID,
  uint16 X,
  uint16 Y
) {
  int16 Result;
  TPixmapInfo *I;
  TPixmapInfo *OI;
  void *Screen;
  bool AlreadyMounted;

  I=PixIDGetInfo(&Data->List,ID);
  if (I==NULL) return(screBPXID);
  /*
    Check if the destination mounting point is completely within the window.
    This is not as easy as one might think. Summing the X coordinate with
    the width or the Y coordinate with the height of the destination
    rectangle could cause a silently ignored integer overflow resulting in a
    number small enough to cause the condition to "not realize there is a
    problem". To avoid this I will chech whether the coordinates and
    the sizes are small enough by comparing them to the limits alone.
    Only after this condition is satisfied I will try to sum them together.
    If any of these conditions realize a problem, the code will bail out
    with the screMRECTOUT error code.
  */
  if (
    (I->Width>Data->LineSize || I->Height>Data->LineCount) ||
    (X>Data->LineSize || Y>Data->LineCount)
  ) {
    return(screMRECTOUT);
  }
  if (X+I->Width>Data->LineSize) return(screMRECTOUT);
  if (Y+I->Height>Data->LineCount) return(screMRECTOUT);
  /*
    Check whether the pixmap
  */
  OI=Data->List.FirstMounted;
  if (OI!=NULL) {
    do {
      do {
        if (OI==I) break;
        if (I->ScreenX+I->Width<=OI->ScreenX) break;
        if (OI->ScreenX+OI->Width<=I->ScreenX) break;
        if (I->ScreenY+I->Height<=OI->ScreenY) break;
        if (OI->ScreenY+OI->Height<=I->ScreenY) break;
        return(screMOVERLAP);
      } while (false);
    } while (OI!=Data->List.FirstMounted);
  }
  /*
    Determine if the pixmap was already mounted. Handling already mounted
    pixmaps is a bit different than the unmounted ones.
  */
  AlreadyMounted=I->ScreenX!=65535;
  /*
    If the pixmap was already mounted, transfer the content from the
    previous mounting position on the screen to the pixmap store. Moving
    the content of the screen directly from the source location to the
    destination location may not work because these destinations are
    specifically allowed to overlap.
  */
  if (AlreadyMounted) {
    Result=SaveMountedPixmapContent(I);
    if (Result!=0) return(Result);
  }
  /*
    Perform the transfer.
  */
  Screen=ScrGetScreenPixmap();
  Result=ScrTransferRectangle(I->Pixmap,0,0,I->Width,I->Height,Screen,X,Y);
  if (Result!=0) return(Result);
  I->ScreenX=X;
  I->ScreenY=Y;
  RecordDirtyRectangle(Data,X,Y,I->Width,I->Height);
  /*
    If the pixmap is freshly mounted, register it in the list of mounted
    pixmaps.
  */
  if (!AlreadyMounted) RegisterMountedPixmap(&Data->List,I);
  /*
    Everything went OK.
  */
  return(0);
}

static int16 ExecuteUnmountPixmap(
  TVideoMemoryHandle *Data,
  uint16 ID
) {
  int16 Result;
  TPixmapInfo *I;

  I=PixIDGetInfo(&Data->List,ID);
  if (I==NULL) return(screBPXID);
  if (I->ScreenX==65535) return(screNOTMNT);
  Result=SaveMountedPixmapContent(I);
  if (Result!=0) return(Result);
  I->ScreenX=65535;
  UnregisterMountedPixmap(&Data->List,I);
  return(0);
}

static int16 ExecuteRectangleTransfer(
  TVideoMemoryHandle *Data,
  uint16 SrcID,
  uint16 SrcX,
  uint16 SrcY,
  uint16 Width,
  uint16 Height,
  uint16 DstID,
  uint16 DstX,
  uint16 DstY
) {
  int16 Result;
  TPixmapInfo *SrcI;
  TPixmapInfo *DstI;
  void *Src;
  void *Dst;
  bool Update;

  SrcI=PixIDGetInfo(&Data->List,SrcID);
  DstI=PixIDGetInfo(&Data->List,DstID);
  if (SrcI==NULL || DstI==NULL) return(screBPXID);
  Src=SrcI->Pixmap;
  if (SrcI->ScreenX!=65535) {
    Src=ScrGetScreenPixmap();
    SrcX+=SrcI->ScreenX;
    SrcY+=SrcI->ScreenY;
  }
  Update=false;
  Dst=DstI->Pixmap;
  if (DstI->ScreenX!=65535) {
    Dst=ScrGetScreenPixmap();
    DstX+=DstI->ScreenX;
    DstY+=DstI->ScreenY;
    Update=true;
  }
  if (Width==65535) {
    Width=SrcI->Width;
    Height=SrcI->Height;
  }
  if (
    (Width>SrcI->Width || Height>SrcI->Height) ||
    (SrcX>SrcI->Width || SrcY>SrcI->Height)
  ) {
    return(screSRECTOUT);
  }
  if (SrcX+Width>SrcI->Width) return(screSRECTOUT);
  if (SrcY+Height>SrcI->Height) return(screSRECTOUT);
  if (
    (Width>DstI->Width || Height>DstI->Height) ||
    (SrcX>DstI->Width || SrcY>DstI->Height)
  ) {
    return(screDRECTOUT);
  }
  if (SrcX+Width>DstI->Width) return(screDRECTOUT);
  if (SrcY+Height>DstI->Height) return(screDRECTOUT);
  Result=ScrTransferRectangle(Src,SrcX,SrcY,Width,Height,Dst,DstX,DstY);
  if (Result!=0) return(Result);
  if (Update) {
    RecordDirtyRectangle(Data,DstX,DstY,Width,Height);
  }
  return(Result);
}

static int16 ExecDraw(const upint8 *Code,uint16 Size,void *fd)
{
  TDrawInstructionList L;
  int16 Result;
  uint8 OpCode;
  uint16 SrcID;
  uint16 SrcX;
  uint16 SrcY;
  uint16 SrcW;
  uint16 SrcH;
  uint16 DstID;
  uint16 DstX;
  uint16 DstY;
  uint8 Flags1;
  uint8 Flags2;

  L.Code=Code;
  L.Size=Size;
  L.Error=false;
  Result=0;
  do {
    OpCode=FetchByte(&L);
    switch (OpCode) {
      case drwinsMountPixmap: {
	SrcID=FetchByte(&L);
	SrcID+=FetchByte(&L)*256;
	SrcX=FetchByte(&L);
	SrcX+=FetchByte(&L)*256;
	SrcY=FetchByte(&L);
	SrcY+=FetchByte(&L)*256;
        break;
      }
      case drwinsUnmountPixmap: {
	SrcID=FetchByte(&L);
	SrcID+=FetchByte(&L)*256;
        break;
      }
      default: {
        if (
	  (OpCode<drwinsTransferRectangle) ||
	  (OpCode>=drwinsTransferRectangle+0xF8)
	) {
	  return(screINVINS);
	}
	SrcID=256*(OpCode-drwinsTransferRectangle)+FetchByte(&L);
	DstX=FetchByte(&L);
	DstY=FetchByte(&L);
	Flags1=FetchByte(&L);
	DstX+=(Flags1 & 31)*256;
	Flags2=FetchByte(&L);
	DstY+=(Flags2 & 31)*256;
	DstID=FetchByte(&L);
	DstID+=FetchByte(&L)*256;
	SrcX=0;
	SrcY=0;
        SrcW=65535;
	OpCode=drwinsTransferRectangle;
	break;
      }
    }
    if (L.Error) return(screINCINS);
    switch (OpCode) {
      case drwinsMountPixmap: {
        Result=ExecuteMountPixmap(fd,SrcID,SrcX,SrcY);
	break;
      }
      case drwinsUnmountPixmap: {
        Result=ExecuteUnmountPixmap(fd,SrcID);
	break;
      }
      default: {
        Result=ExecuteRectangleTransfer(
          fd,
          SrcID,SrcX,SrcY,SrcW,SrcH,
          DstID,DstX,DstY
        );
      }
    }
  } while (L.Size>0 && Result==0);
  return(Result);
}

static int16 ExecDestroyPixmaps(const upint16 *IDList,int16 Count,void *fd)
{
  TVideoMemoryHandle *Data;
  TPixmapInfo *I;
  int16 ID;
  int16 Result;

  Result=0;
  Data=fd;
  if (!ScrCheckModeSet()) return(screNOMODESET);
  while (Count>0) {
    ID=*(IDList++);
    I=PixIDGetInfo(&Data->List,ID);
    if (I==NULL) {
      Result=screBPXID;
    } else {
    }
  }
  return(Result);
}

static int32 WriteRawPixmap(void *fd,const void *Buf,int32 Size)
{
  TVideoMemoryHandle *Data;
  uint32 RemSize;
  int16 Result;

  Data=fd;
  /*
    Adjust the size of the data to be written so the data that don't fit
    into the range that is still to be written will be ignored.
  */
  if (Size>Data->RemainingPixels) Size=Data->RemainingPixels;
  if (Size==0) return(0);
  /*
    Copy the data into the video memory using the bulk pixel conversion call
    to transfer pixels to the screen driver.
  */
  Data->PTInfo.InPtr=Buf;
  RemSize=Size;
  while (RemSize>0) {
    /*
      If there are no free pixels in the driver's buffer, ask the driver for
      new buffer.
    */
    if (Data->PTInfo.FreePixels==0) {
      assert(Data->PTInfo.PixelsToFill==Data->PTInfo.PixelsOnLine);
      Result=FlushPixels(Data);
      assert(Result<=0);
      if (Result<0) return(Result);
      Data->PTInfo.MaskPtr=NULL;
      Result=ScrExposeBuffer(&Data->PTInfo);
      if (Result<0) return(Result);
      assert(Data->PTInfo.FreePixels%Data->PTInfo.PixelsOnLine == 0);
    }
    /*
      Transfer pixels to the screen buffer
    */
    assert(Data->PTInfo.FreePixels>0);
    Data->PTInfo.PixelCount=Data->PTInfo.FreePixels;
    if (Data->PTInfo.PixelCount>RemSize) Data->PTInfo.PixelCount=RemSize;
    RemSize-=Data->PTInfo.PixelCount;
    ConvertPixels(&Data->PTInfo);
  }
  /*
    Update the amount of pixels to be filled.
  */
  Data->RemainingPixels-=Size;
  /*
    If there are no more pixels to fill, flush any pixels that are in the
    buffer.
  */
  if (Data->RemainingPixels==0) FlushPixels(Data);
  /*
    Return the count of pixels written.
  */
  return Size;
}

static int32 WriteMaskedPixmap(void *fd,const void *Buf,int32 Size)
{
  TVideoMemoryHandle *Data;
  uint32 TransferSize;
  uint32 TotalTransferSize;
  int16 Result;
  uint32 PixelCount;

  Data=fd;
  /*
    Copy the data into the video memory using the bulk pixel conversion call
    to transfer pixels to the screen driver.
  */
  TotalTransferSize=0;
  Data->PTInfo.InPtr=Buf;
  Data->PTInfo.InputSize=Size;
  while (Data->PTInfo.InputSize>0) {
    /*
      If there are no free pixels in the driver's buffer, ask the driver for
      new buffer.
    */
    if (Data->PTInfo.FreePixels==0) {
      assert(Data->PTInfo.PixelsToFill==Data->PTInfo.PixelsOnLine);
      FlushPixels(Data);
      Data->PTInfo.MaskPaddingBytes=0;
      Result=ScrExposeBuffer(&Data->PTInfo);
      if (Result<0) return(Result);
      assert(Data->PTInfo.FreePixels%Data->PTInfo.PixelsOnLine == 0);
    }
    /*
      Transfer pixels to the bitmap buffer and to the bitmap mask buffer.
    */
    assert(Data->PTInfo.FreePixels>0);
    PixelCount=Data->PTInfo.FreePixels;
    if (PixelCount>Data->RemainingPixels) PixelCount=Data->RemainingPixels;
    Data->PTInfo.PixelCount=PixelCount;
    TransferSize=Data->PTInfo.InputSize;
    Result=ConvertPixelsOfMaskedBitmap(&Data->PTInfo);
    if (Result!=0) return(Result);
    TransferSize-=Data->PTInfo.InputSize;
    TotalTransferSize+=TransferSize;
    PixelCount-=Data->PTInfo.PixelCount;
    Data->RemainingPixels-=PixelCount;
  }
  /*
    If there are no more pixels to fill, flush any pixels that are in the
    buffer.
  */
  if (Data->RemainingPixels==0) FlushPixels(Data);
  /*
    Return the count of pixels written.
  */
  return(TotalTransferSize);
}

static int32 VidWrite(void *fd,const void *Buf,int32 Size)
{
  TVideoMemoryHandle *Data;

  Data=fd;
  if (Data->PTInfo.MaskOutBit!=0) {
    return(WriteMaskedPixmap(fd,Buf,Size));
  } else {
    return(WriteRawPixmap(fd,Buf,Size));
  }
}

static void SetupColorFormat(const TSetModeCommand *C, TColorFormat *F)
{
  F->BitsPerPixel=0;
  switch (C->PixelSize) {
    case scrpx8bit: {
      F->BitsPerPixel=8;
      F->BMask=0xE0;
      F->GMask=0x38;
      F->RMask=0x07;
      break;
    }
    case scrpx16bit: {
      F->BitsPerPixel=16;
      F->RMask=0xF800;
      F->GMask=0x07E0;
      F->BMask=0x001F;
      break;
    }
    case scrpx32bit: {
      F->BitsPerPixel=8;
      /* FALL THROUGH */
    }
    case scrpx24bit: {
      F->BitsPerPixel+=24;
      F->RMask=0xFF0000;
      F->GMask=0x00FF00;
      F->BMask=0x0000FF;
      break;
    }
  }
  BuildColorFormat(F);
}

static int16 ScrExecScreenModeSet(const void *Cmd,void *Out,void *PD)
{
  const TSetModeCommand *C;
  int16 Result;
  TVideoMemoryHandle *Data;

  C=Cmd;
  Data=PD;
  if (Data->RemainingPixels!=0) return(screREGWRT);
  if (C->Flags & scrflgReserved) return(screBADFLAGS);
  if (C->PixelSize>=scrpxInvalid) return(screBADPXSIZE);
  Result=ScrSetupScreenMode(C);
  if (Result<0) return(Result);
  Data->PixelSize=C->PixelSize+1;
  Data->LineSize=C->Width;
  Data->LineCount=C->Height;
  Data->InstantUpdate=(C->Flags & scrflgUpdateOnDemand)==0;
  SetupColorFormat(C,&Data->PTInfo.In);
  ScrGetCurrentColorFormat(&Data->PTInfo.Out);
  BuildColorFormat(&Data->PTInfo.Out);
  InitializeScreenBufferTransfers(Data);
  InitializeRectangleRedraw(Data);
  return(0);
}

static int16 ScrExecUpdateScreen(const void *Cmd,void *Out,void *PD)
{
  TVideoMemoryHandle *Data;
  uint16 X;
  uint16 Y;
  uint16 Width;
  uint16 Height;

  Data=PD;
  if (!Data->InstantUpdate) {
    X=Data->MinX;
    Y=Data->MinY;
    if (X!=65535) {
      Width=Data->MaxX-X;
      Height=Data->MaxY-Y;
      ScrUpdateRectangle(X,Y,Width,Height);
      ClearRectangleToRedraw(Data);
    }
  }
  return(0);
}

/*
  Create the private datablock used by the channel.
*/
void *CreatePrivateData(void)
{
  TVideoMemoryHandle *Data;

  Data=KMEM__malloc(sizeof(*Data));
  if (Data==NULL) return(NULL);
  Data->LineCount=0;
  Data->RemainingPixels=0;
  PixIDInitialize(&Data->List);
  return(Data);
}

static int16 LongWrite(const void *Msg,void *Out,void *PD)
{
  const TWriteCmd *C;
  TIOReply *R;

  C=Msg;
  R=Out;
  R->Size=VidWrite(PD,C->Ptr,C->Size);
  return(sizeof(*R));
}

static int16 ShortWrite(const void *Msg,int16 Size,void *Out,void *PD)
{
  const TDataCmd *C;
  TIOReply *R;

  C=Msg;
  R=Out;
  R->Size=VidWrite(PD,C->Data,Size-1);
  return(sizeof(*R));
}

typedef int16 (*TExecutor)(const void *Packet,void *Result,void *PD);

int16 HandleScreen(void *Buf,int16 Size,void *Out,void *PD)
{
  TVideoMemoryHandle *Data;
  TExecutor Exec;
  uint16 ReqSize;
  uint8 Cmd;

  Data=PD;
  /*
    Get the command code.
  */
  Cmd=((TNoParamCmd *)Buf)->Command;
  /*
    Determine how to execute the command. The "Exec" pointer will be set to
    the procedure that executes the command in question. The ReqSize will be
    set to the minimal size that a packet must have.
  */
  switch (Cmd) {
    default: {
      return(beINVCMD);
    }
    /*
      File I/O commands are handled by a generic file I/O procedure. The file
      I/O commands must be listed one by one due to future security features
      that may reshape the protocols.
    */
    case cmdWRITE: {
      if (Data->LineCount==0) return(screNOMODESET);
      ReqSize=sizeof(TWriteCmd);
      Exec=LongWrite;
      break;
    }
    case cmdDATA: {
      if (Data->LineCount==0) return(screNOMODESET);
      return(ShortWrite(Buf,Size,Out,PD));
    }
    /*
      Other commands are handled directly
    */
    case scrcmdSetMode: {
      ReqSize=sizeof(TSetModeCommand);
      Exec=ScrExecScreenModeSet;
      break;
    }
    case scrcmdOpenKeyboard: {
      ReqSize=sizeof(TOpenKeyboardCommand);
      Exec=ScrExecOpenKeyboard;
      break;
    }
    case scrcmdUpdateScreen: {
      ReqSize=sizeof(TNoParamCmd);
      Exec=ScrExecUpdateScreen;
      break;
    }
    case scrcmdProbeMode: {
      return(beNOTSUP);
    }
    case scrcmdOpenScreenRegion: {
      ReqSize=sizeof(TOpenScreenRegionCommand);
      Exec=ExecOpenScreenRegion;
      break;
    }
    case scrcmdCreatePixmap: {
      ReqSize=sizeof(TCreatePixmapCommand);
      Exec=ExecCreatePixmap;
      break;
    }
    case scrcmdCreateOutPixmap: {
      ReqSize=sizeof(TCreateOutPixmapCommand);
      Exec=ExecCreateOutPixmap;
      break;
    }
    case scrcmdDraw: {
      if (Size<2) return(screBADSIZE);
      return(ExecDraw(Buf+1,Size-1,PD));
    }
    case scrcmdDestroyPixmaps: {
      if (Size<3 || (Size&1)==0) return(screBADSIZE);
      return(ExecDestroyPixmaps(Buf+1,(Size-1)/2,PD));
    }
  }
  /*
    Check if enough bytes were passed. If not, set the result to BADSIZE and
    ignore the command.
  */
  if (Size<ReqSize) return(screBADSIZE);
  /*
    Execute the command and store the result.
  */
  return(Exec(Buf,Out,PD));
}
