#include "defs.gh"

#include <stdlib.h>
#include <SDL/SDL.h>

#include "types.h"
#include "kernel.h"
#include "pdbase.h"
#include "pdgfxcon.h"
#include "pdkeys.h"
#include "gcwdata.h"
#include "strings.h"

#define __INTERFACE__
#include "gcwsdl.h"
#undef __INTERFACE__

#include "typesnd.h"

bool KeyboardOpen;

#if __b__NUMCAPS_BUG__
/*
  This kludge fixes broken keypress/key release reports from SDL. Normal
  reporta are: when a key is pressed, the system reports the key to be
  pressed and when the key is released, the system reports the key to be
  released. However for NUMLOCK and CAPSLOCK when the key is pressed, the SDL
  reports the key is pressed but when it is released, SDL reports nothing.
  When the key is pressed next time, the SDL reports thet the key is
  released. The problem is fixed by converting these broken keystrokes to
  matching pairs of "keypress"/"keyrelease". The first report from this pair
  is returned to the caller and the next one is stored here until the caller
  requests next keystroke.
*/
uint16 PendingKey;
#endif

/* the video mode surface access */
static SDL_Surface *ScrSurface;

int16 ScrOpenKeyboard(void)
{
  static bool KeyboardInitialized;

  if (!ScrCheckModeSet()) return(screNOMODESET);
  /*
    Ensure the keyboard is initialized
  */
  if (!KeyboardInitialized) {
    if (
      SDL_EnableKeyRepeat(
        SDL_DEFAULT_REPEAT_DELAY,
	SDL_DEFAULT_REPEAT_INTERVAL
      )==-1
    ) {
      return(screKBDINIT);
    }
    KeyboardInitialized=true;
  }
  /*
    Set that the keyboard is opened.
  */
  KeyboardOpen=true;
  /*
    Return success.
  */
  return(0);
}

/*
  This procedure reads a key and returns it's code in low byte and it's
  modifiers in high byte.
*/
uint16 ScrRawReadKey(void)
{
  SDL_Event Event;	
  uint16 Result;

#if __b__NUMCAPS_BUG__
  /*
    If there is a keystroke ready for sending, do so now.
  */
  if (PendingKey) {
    Result=PendingKey;
    PendingKey=0;
    return(Result);
  }
#endif
  /*
    Wait for a keystroke event. Here also the bogus keystroke events (such as
    keystrokes related to NUM LOCK, CAPS LOCK or SHIFTs) are filtered out and
    ignored. Only regular keys are able to break this loop.
  */
  Result=0xFFFF;
  do {
    /*
      If the keyboard is not opened, return a failure. This may mean that
      either the keyboard was not opened yet or was closed during waiting for
      a keyboard event.
    */
    if (!KeyboardOpen) return(0xFFFF);
    /*
      Wait for an event to be processed.
    */
    SDL_WaitEvent(&Event);
    /*
      Process the keypress events.
    */
    if (Event.type==SDL_KEYDOWN || Event.type==SDL_KEYUP) {
      /*
        Scan for keys that are guaranteed to be good.
      */
      switch (Event.key.keysym.sym) {
        case SDLK_BACKSPACE   :Result=RK_BACKSPACE; break;
        case SDLK_TAB         :Result=RK_TAB; break;
        case SDLK_RETURN      :Result=RK_ENTER; break;
        case SDLK_ESCAPE      :Result=RK_ESC; break;
        case SDLK_SPACE       :Result=RK_SPACE; break;
        case SDLK_DELETE      :Result=RK_DELETE; break;
        case SDLK_KP0         :Result=RK_KP0; break;
        case SDLK_KP1         :Result=RK_KP1; break;
        case SDLK_KP2         :Result=RK_KP2; break;
        case SDLK_KP3         :Result=RK_KP3; break;
        case SDLK_KP4         :Result=RK_KP4; break;
        case SDLK_KP5         :Result=RK_KP5; break;
        case SDLK_KP6         :Result=RK_KP6; break;
        case SDLK_KP7         :Result=RK_KP7; break;
        case SDLK_KP8         :Result=RK_KP8; break;
        case SDLK_KP9         :Result=RK_KP9; break;
        case SDLK_KP_PERIOD   :Result=RK_KP_PERIOD; break;
        case SDLK_KP_DIVIDE   :Result=RK_KP_SLASH; break;
        case SDLK_KP_MULTIPLY :Result=RK_KP_STAR; break;
        case SDLK_KP_MINUS    :Result=RK_KP_MINUS; break;
        case SDLK_KP_PLUS     :Result=RK_KP_PLUS; break;
        case SDLK_KP_ENTER    :Result=RK_KP_ENTER; break;
        case SDLK_UP          :Result=RK_UP; break;
        case SDLK_DOWN        :Result=RK_DOWN; break;
        case SDLK_RIGHT       :Result=RK_RIGHT; break;
        case SDLK_LEFT        :Result=RK_LEFT; break;
        case SDLK_INSERT      :Result=RK_INSERT; break;
        case SDLK_HOME        :Result=RK_HOME; break;
        case SDLK_END         :Result=RK_END; break;
        case SDLK_PAGEUP      :Result=RK_PGUP; break;
        case SDLK_PAGEDOWN    :Result=RK_PGDN; break;
        case SDLK_F1          :Result=RK_F1; break;
        case SDLK_F2          :Result=RK_F2; break;
        case SDLK_F3          :Result=RK_F3; break;
        case SDLK_F4          :Result=RK_F4; break;
        case SDLK_F5          :Result=RK_F5; break;
        case SDLK_F6          :Result=RK_F6; break;
        case SDLK_F7          :Result=RK_F7; break;
        case SDLK_F8          :Result=RK_F8; break;
        case SDLK_F9          :Result=RK_F9; break;
        case SDLK_F10         :Result=RK_F10; break;
        case SDLK_F11         :Result=RK_F11; break;
        case SDLK_F12         :Result=RK_F12; break;
        case SDLK_LSHIFT      :Result=RK_LSHIFT; break;
        case SDLK_RSHIFT      :Result=RK_RSHIFT; break;
        case SDLK_LCTRL       :Result=RK_LCTRL; break;
        case SDLK_RCTRL       :Result=RK_RCTRL; break;
        case SDLK_LALT        :Result=RK_LALT; break;
#if __b__ALT_BUG__
        /*
          Buggy X Windows report "MODE SHIFT", when RIGHT ALT is pressed. So
	  we will convert "MODE SHIFT" to "RIGHT ALT".
        */
        case SDLK_MODE:
#endif
        case SDLK_RALT        :Result=RK_RALT; break;
        case SDLK_NUMLOCK     :Result=RK_NUMLOCK; break;
        case SDLK_CAPSLOCK    :Result=RK_CAPSLOCK; break;
        case SDLK_SCROLLOCK   :Result=RK_SCROLLOCK; break;
        default: break;
      }
#if __b__NUMCAPS_BUG__
      /*
        NUM LOCK and CAPS LOCK keypress reports are broken. Convert them into
	two reports, the first is reported now and the second one will be
	prepared to be reported at the next call.
      */
      if (Result==RK_NUMLOCK || Result==RK_CAPSLOCK) {
        PendingKey=Result+MK_RELEASED;
	return(Result);
      }
#endif
      if (Result!=0xFFFF) break;
      /*
        If the key was rejected, try to examine the scan code and recognize
        the key.
      */
      switch (Event.key.keysym.scancode) {
        case 0x31:Result=RK_BACKAPOSTROPHE; break;
        case 0x0A:Result=RK_1; break;
        case 0x0B:Result=RK_2; break;
        case 0x0C:Result=RK_3; break;
        case 0x0D:Result=RK_4; break;
        case 0x0E:Result=RK_5; break;
        case 0x0F:Result=RK_6; break;
        case 0x10:Result=RK_7; break;
        case 0x11:Result=RK_8; break;
        case 0x12:Result=RK_9; break;
        case 0x13:Result=RK_0; break;
        case 0x14:Result=RK_MINUS; break;
        case 0x15:Result=RK_EQUAL; break;
        case 0x33:Result=RK_SLASH; break;
        case 0x17:Result=RK_TAB; break;
        case 0x18:Result=RK_Q; break;
        case 0x19:Result=RK_W; break;
        case 0x1A:Result=RK_E; break;
        case 0x1B:Result=RK_R; break;
        case 0x1C:Result=RK_T; break;
        case 0x1D:Result=RK_Y; break;
        case 0x1E:Result=RK_U; break;
        case 0x1F:Result=RK_I; break;
        case 0x20:Result=RK_O; break;
        case 0x21:Result=RK_P; break;
        case 0x22:Result=RK_LEFTBRACKET; break;
        case 0x23:Result=RK_RIGHTBRACKET; break;
        case 0x24:Result=RK_ENTER; break;
        case 0x26:Result=RK_A; break;
        case 0x27:Result=RK_S; break;
        case 0x28:Result=RK_D; break;
        case 0x29:Result=RK_F; break;
        case 0x2A:Result=RK_G; break;
        case 0x2B:Result=RK_H; break;
        case 0x2C:Result=RK_J; break;
        case 0x2D:Result=RK_K; break;
        case 0x2E:Result=RK_L; break;
        case 0x2F:Result=RK_SEMICOLON; break;
        case 0x30:Result=RK_APOSTROPHE; break;
        case 0x34:Result=RK_Z; break;
        case 0x35:Result=RK_X; break;
        case 0x36:Result=RK_C; break;
        case 0x37:Result=RK_V; break;
        case 0x38:Result=RK_B; break;
        case 0x39:Result=RK_N; break;
        case 0x3A:Result=RK_M; break;
        case 0x3B:Result=RK_COMMA; break;
        case 0x3C:Result=RK_PERIOD; break;
        case 0x3D:Result=RK_SLASH; break;
	case 0x73:Result=RK_LWIN; break;
	case 0x74:Result=RK_RWIN; break;
	case 0x75:Result=RK_MENU; break;
        default: break;
      }
    }
  } while (Result==0xFFFF);
  /*
    If the key was released, notice this in the result
  */
  if (Event.type==SDL_KEYUP) Result+=MK_RELEASED;
  /*
    The result is done now, expedition to the caller can be made.
  */
  return(Result);
}

int16 ScrCloseKeyboard(void)
{
  SDL_Event E;

  /*
    Setup the keyboard to be closed
  */
  KeyboardOpen=false;
  /*
    Send a dummy event to the queue to allow the waiting thread to be
    unblocked and return a failure.
  */
  E.type=SDL_NUMEVENTS-1;
  SDL_PushEvent(&E);
  /*
    Return success.
  */
  return(0);
}

void ScrUpdateRectangle(uint16 X,uint16 Y,uint16 Width,uint16 Height)
{
  SDL_UpdateRect(ScrSurface,X,Y,Width,Height);
}

int16 ScrSetupScreenMode(const TSetModeCommand *C)
{
  static bool SDLInitialized=false;
  int16 Result;
  int32 SurfaceType;
  uint16 LineWidth;
  uint16 BitsPerPixel;
  bool SoftMode;

  /*
    Determine if the mode is a "soft" mode
  */
  SoftMode=(C->Flags & scrflgUpdateOnDemand)>0;
  /*
    Initialize the result variable
  */
  Result=0;
  /*
    Check if the SDL library was successfully initialized.
  */
  if (!SDLInitialized) {
    /*
      Initialize the SDL library
    */
    if (SDL_Init(SDL_INIT_VIDEO)==-1) return(screSCRINIT);
    /*
      Setup the SDL shutdown call as exit procedure
    */
    atexit(SDL_Quit);
    /*
      Set the SDL initialization flag
    */
    SDLInitialized=true;
  }
  /*
    If there is a surface, free it
  */
  if (ScrSurface!=NULL) SDL_FreeSurface(ScrSurface);
  /*
    Determine the type of the surface
  */
  SurfaceType=SDL_HWSURFACE;
  if (SoftMode) SurfaceType=SDL_SWSURFACE;
  /*
    Setup the video mode
  */
  BitsPerPixel=(C->PixelSize<<3)+8;
  ScrSurface=SDL_SetVideoMode(C->Width,C->Height,BitsPerPixel,SurfaceType);
  if (ScrSurface==NULL) return(screMODESUP);
  /*
    If the surface does not needs locking, we are done with the mode. Take
    out the pointer to the video memory. Also check if the number of bytes
    per line is correct. Sometimes the system uses various kinds of paddings
    to adjust the lines; this is not supported by the current version of the
    library.
  */
  LineWidth=C->Width*(C->PixelSize+1);
  if (!SDL_MUSTLOCK(ScrSurface) && LineWidth==ScrSurface->pitch) {
    return(0);
  }
  /*
    Surfaces that need locking are not supported so free the surface and tell
    the caller the mode is not supported.
  */
  SDL_FreeSurface(ScrSurface);
  ScrSurface=NULL;
  if (Result!=0) return(Result);
  return(screMODESUP);
}

bool ScrCheckModeSet(void)
{
  return(ScrSurface!=NULL);
}

void ScrGetCurrentColorFormat(TColorFormat *F)
{
  F->BitsPerPixel=ScrSurface->format->BitsPerPixel;
  if (F->BitsPerPixel==8) {
    F->RMask=0xE0;
    F->GMask=0x1C;
    F->BMask=0x03;
  } else {
    F->RMask=ScrSurface->format->Rmask;
    F->GMask=ScrSurface->format->Gmask;
    F->BMask=ScrSurface->format->Bmask;
  }
}

int16 ScrPrepareAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  return(0);
}

int16 ScrExposeBuffer(TColorConversionInfo *I)
{
  int16 PixelSize;
  upint8 *VideoMemory;
  uint32 PixelNumber;

  VideoMemory=ScrSurface->pixels;
  VideoMemory+=I->LineNumber * ScrSurface->pitch;
  PixelSize=ScrSurface->format->BitsPerPixel / 8;
  VideoMemory+=(I->PixelsOnLine - I->PixelsToFill) * PixelSize;
  I->OutPtr=VideoMemory;
  PixelNumber=ScrSurface->w * I->LineNumber;
  PixelNumber+=ScrSurface->w - I->PixelsToFill;
  I->FreePixels=(ScrSurface->w * ScrSurface->h) - PixelNumber;
  I->LinePaddingBytes=ScrSurface->pitch - (ScrSurface->w * PixelSize);
  I->MaskPtr=NULL;
  return(0);
}

int16 ScrFlushPixels(TColorConversionInfo *I)
{
  int16 StartLocY;
  int16 CurLocY;
  int16 WinWidth;

  assert(I->UnprocessedPixels);
  WinWidth=ScrSurface->w;
  StartLocY=I->SLine;
  CurLocY=I->LineNumber;
  if (I->PixelsToFill==I->PixelsOnLine) CurLocY--;
  return(0);
}

int16 ScrDestroyAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  return(0);
}

void *ScrGetScreenPixmap(void)
{
  return(ScrSurface);
}

int16 ScrCreatePixmap(uint16 Width,uint16 Height,void **Result)
{
  return(beNOTSUP);
}

int16 ScrCreateMaskedPixmap(uint16 Width,uint16 Height,void **Result)
{
  return(beNOTSUP);
}

int16 ScrTransferRectangle(
  void *SrcPixmap,
  uint16 SrcX,
  uint16 SrcY,
  uint16 Width,
  uint16 Height,
  void *DstPixmap,
  uint16 DstX,
  uint16 DstY
) {
  return(beNOTSUP);
}

void ScrDestroyPixmap(void *Pixmap)
{
}
