/***************************************************************************
 Name:          IODriver

 Purpose:       Gets the keyboard and mouse input and processes screen
                output.

 Remarks:       - The act of handling the keyboard and screen I/O is too
                  system dependent to be bound in the application. So
                  dedicated executable is made to handle this task.
                - This program takes the keyboard and mouse input and
                  formats it into system-independent values. Also it takes
                  system-independent display data and displays them on the
                  screen.
***************************************************************************/

/*** INCLUDE FILES ***/

#include <time.h>
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>

#include "types.h"
#include "kernel.h"
#include "pdgfxcon.h"
#include "pdkeys.h"
#include "thread.h"
#include "thrwin.h"
#include "gcwdata.h"
#include "pdbase.h"
#include "memory.h"

#define __INTERFACE__
#include "gcww32.h"
#undef __INTERFACE__

#include "typesnd.h"

/*--------------------------------------------------------------------------
 Code:          Input handling

 Purpose:       This code handles the input devices and provides raw output
                to the OSHS emulation layer.
--------------------------------------------------------------------------*/

/*** InputVars: The input variables ***/
/*
  The size of the input buffer in records. If the buffer gets full, other
  incoming events are discarded.
*/
#define INPUT_BUFFER_SIZE 128
/*
  The keyboard input buffer. The [KeybLock] is used to synchronize the
  access. The [KeybDataReady] is used to wait if a key is available in the
  buffer. The rest belongs to the buffer itself.
*/
static bool KeybEnabled;
static HANDLE KeybReady;
static HANDLE KeybLock;
static uint16 KeybFirst;
static uint16 KeybLast;
static uint16 KeybEvents[INPUT_BUFFER_SIZE];
/*
  Windows keyboard driver has a bug: when right alt is pressed or
  released and the current keyboard layout is not "En", then Windows
  generates fake left-CTRL keystrokes that cannot be distinguished from the
  real ones by examining the content of the messages carrying them. Only
  one difference is and that these fake "left Ctrl" messages appear before
  the real "right Alt" messages in one burst of messages. Therefore we need
  to examine the bursts of messages (the burst boundaries are determined
  with the help of PeekMessage() Windows API) and if we find a "left Ctrl"
  followed by "right Alt" in the same message burst and if the two
  keystrokes are both either keypresses or keyreleases, we discard the
  "left Ctrl". These two variables below are used to buffer the keystrokes
  in the message bursts so we can discard the "left Ctrl" when we detect
  it is a fake one.
*/
static WORD LastKey;
static bool LastReleased;
/*
  The same as the above but for the mouse this time.
*/
static bool MouseEnabled;
static HANDLE MouseReady;
static HANDLE MouseLock;
static uint16 MouseFirst;
static uint16 MouseLast;
static TMouseEvent MouseEvents[INPUT_BUFFER_SIZE];
static TMouseEvent LastEvent;

uint8 ScrOpenKeyboard(void)
{
  ThrLock(KeybLock);
  KeybFirst=0;
  KeybLast=0;
  KeybEnabled=true;
  ResetEvent(KeybReady);
  ThrUnlock(KeybLock);
  return(0);
}

static WORD TranslateKeyStroke(WORD KeyStroke,DWORD SuppInfo)
{
  uint8 ScanCode;

  ScanCode=(SuppInfo>>16)&255;
  if ((SuppInfo&(1<<24))>0) {
    switch (ScanCode) {
      case 0x47 : return(RK_HOME);
      case 0x48 : return(RK_UP);
      case 0x49 : return(RK_PGUP);
      case 0x4B : return(RK_LEFT);
      case 0x4D : return(RK_RIGHT);
      case 0x4F : return(RK_END);
      case 0x50 : return(RK_DOWN);
      case 0x51 : return(RK_PGDN);
      case 0x52 : return(RK_INSERT);
      case 0x53 : return(RK_DELETE);
      case 0x1C : return(RK_KP_ENTER);
      case 0x1D : return(RK_RCTRL);
      case 0x38 : return(RK_RALT);
      case 0x35 : return(RK_KP_SLASH);
      case 0x45 : return(RK_NUMLOCK);
      case 0x5B : return(RK_LWIN);
      case 0x5C : return(RK_RWIN);
      case 0x5D : return(RK_MENU);
    }
  } else {
    switch (ScanCode) {
      case 0x47 : return(RK_KP7);
      case 0x48 : return(RK_KP8);
      case 0x49 : return(RK_KP9);
      case 0x4B : return(RK_KP4);
      case 0x4C : return(RK_KP5);
      case 0x4D : return(RK_KP6);
      case 0x4F : return(RK_KP1);
      case 0x50 : return(RK_KP2);
      case 0x51 : return(RK_KP3);
      case 0x52 : return(RK_KP0);
      case 0x53 : return(RK_KP_PERIOD);
      case 0x1C : return(RK_ENTER);
      case 0x36 : return(RK_RSHIFT);
      case 0x1D : return(RK_LCTRL);
      case 0x38 : return(RK_LALT);
      case 0x37 : return(RK_KP_STAR);
      case 0x4A : return(RK_KP_MINUS);
      case 0x4E : return(RK_KP_PLUS);
      case 0x01 : if (KeyStroke==0x1B) return(RK_ESC); break;
      case 0x3B : return(RK_F1);
      case 0x3C : return(RK_F2);
      case 0x3D : return(RK_F3);
      case 0x3E : return(RK_F4);
      case 0x3F : return(RK_F5);
      case 0x40 : return(RK_F6);
      case 0x41 : return(RK_F7);
      case 0x42 : return(RK_F8);
      case 0x43 : return(RK_F9);
      case 0x44 : return(RK_F10);
      case 0x57 : return(RK_F11);
      case 0x58 : return(RK_F12);
      case 0x0E : return(RK_BACKSPACE);
      case 0x0F : return(RK_TAB);
      case 0x39 : return(RK_SPACE);
      case 0x29 : return(RK_BACKAPOSTROPHE);
      case 0x02 : return(RK_1);
      case 0x03 : return(RK_2);
      case 0x04 : return(RK_3);
      case 0x05 : return(RK_4);
      case 0x06 : return(RK_5);
      case 0x07 : return(RK_6);
      case 0x08 : return(RK_7);
      case 0x09 : return(RK_8);
      case 0x0A : return(RK_9);
      case 0x0B : return(RK_0);
      case 0x0C : return(RK_MINUS);
      case 0x0D : return(RK_EQUAL);
      case 0x2B : return(RK_BACKSLASH);
      case 0x10 : return(RK_Q);
      case 0x11 : return(RK_W);
      case 0x12 : return(RK_E);
      case 0x13 : return(RK_R);
      case 0x14 : return(RK_T);
      case 0x15 : return(RK_Y);
      case 0x16 : return(RK_U);
      case 0x17 : return(RK_I);
      case 0x18 : return(RK_O);
      case 0x19 : return(RK_P);
      case 0x1A : return(RK_LEFTBRACKET);
      case 0x1B : return(RK_RIGHTBRACKET);
      case 0x1E : return(RK_A);
      case 0x1F : return(RK_S);
      case 0x20 : return(RK_D);
      case 0x21 : return(RK_F);
      case 0x22 : return(RK_G);
      case 0x23 : return(RK_H);
      case 0x24 : return(RK_J);
      case 0x25 : return(RK_K);
      case 0x26 : return(RK_L);
      case 0x27 : return(RK_SEMICOLON);
      case 0x28 : return(RK_APOSTROPHE);
      case 0x2C : return(RK_Z);
      case 0x2D : return(RK_X);
      case 0x2E : return(RK_C);
      case 0x2F : return(RK_V);
      case 0x30 : return(RK_B);
      case 0x31 : return(RK_N);
      case 0x32 : return(RK_M);
      case 0x33 : return(RK_COMMA);
      case 0x34 : return(RK_PERIOD);
      case 0x35 : return(RK_SLASH);
      case 0x2A : return(RK_LSHIFT);
      case 0x3A : return(RK_CAPSLOCK);
      case 0x46 : return(RK_SCROLLOCK);
    }
  }
  return(65535);
}

/*** Write a record into the input buffer ***/
static void WriteOutKey(WORD KeyStroke,bool Release)
{
  WORD NewLast;

  ThrLock(KeybLock);
  if (KeybEnabled) {
    NewLast=KeybLast+1;
    if (NewLast==INPUT_BUFFER_SIZE) NewLast=0;
    if (NewLast!=KeybFirst) {
      if (Release) KeyStroke+=MK_RELEASED;
      KeybEvents[KeybLast]=KeyStroke;
      KeybLast=NewLast;
      SetEvent(KeybReady);
    }
  }
  ThrUnlock(KeybLock);
}

static void FlushKey(void)
{
  if (LastKey!=65535) {
    WriteOutKey(LastKey,LastReleased);
    LastKey=65535;
  }
}

/*** Keyboard gathering hook procedure ***/
/*
  This hook procedure is used to gather the keystrokes from the keyboard.
  It is impossible to use the KEYUP/KEYDOWN messages from the window,
  because they are translated according to the current keyboard layout
  selected by the user but OSHS software expects raw keyboard data here.
  The hook gets the desired raw keyboard data.
*/
static void Keyboard(WPARAM wParam, LPARAM lParam)
{
  WORD KeyStroke;
  bool Released;
  bool Valid;

  KeyStroke=TranslateKeyStroke(wParam,lParam);
  Released=(lParam & 0x80000000)>0;
  Valid=(Released!=LastReleased);
  if (!Valid) Valid=(LastKey!=RK_LCTRL) || (KeyStroke!=RK_RALT);
  if (Valid) FlushKey();
  LastKey=KeyStroke;
  LastReleased=Released;
}

uint16 ScrRawReadKey(void)
{
  uint16 Result;
  HANDLE Handles[2];

  Handles[0]=KeybLock;
  Handles[1]=KeybReady;

  WaitForMultipleObjects(2,Handles,TRUE,INFINITE);

  if (KeybEnabled) {
    Result=KeybEvents[KeybFirst++];
    if (KeybFirst==INPUT_BUFFER_SIZE) {
      KeybFirst=0;
    }
    if (KeybFirst==KeybLast) {
      ResetEvent(KeybReady);
    }
  } else {
    Result=65535;
  }
  ThrUnlock(KeybLock);
  return(Result);
}

bool ScrCloseKeyboard(void)
{
  ThrLock(KeybLock);
  KeybEnabled=false;
  SetEvent(KeybReady);
  ThrUnlock(KeybLock);
  return(true);
}

uint8 ScrOpenMouse(void)
{
  ThrLock(MouseLock);
  MouseFirst=0;
  MouseLast=0;
  MouseEnabled=true;
  ResetEvent(MouseReady);
  ThrUnlock(MouseLock);
  return(0);
}

static void WriteOutMouse(SHORT MX, SHORT MY, LPARAM fwKeys)
{
  TMouseEvent *r;
  WORD NewLast;
  BYTE Buttons;
  int32 DeltaX;
  int32 DeltaY;

  /*
    Lock the mouse input buffer.
  */
  ThrLock(MouseLock);
  /*
    If the mouse is not enabled, unlock the mouse input buffer and return
  */
  if (!MouseEnabled) {
    ThrUnlock(MouseLock);
    return;
  }
  /*
    Translate the buttons
  */
  Buttons=0;
  if (fwKeys&MK_LBUTTON) Buttons|=mLeftButton;
  if (fwKeys&MK_RBUTTON) Buttons|=mRightButton;
  if (fwKeys&MK_MBUTTON) Buttons|=mMiddleButton;
  /*
    Convert the absolute mouse coordinates returned by Windows into the
    relative ones expected by OSHS software.
  */
  DeltaX=MX-LastEvent.DeltaX;
  DeltaY=MY-LastEvent.DeltaY;
  /*
    Check if the event really is worth adding to the queue. It must be
    either the mouse motion or button change.
  */
  if ((Buttons==LastEvent.Buttons) && (DeltaX==0) && (DeltaY==0)) {
    ThrUnlock(MouseLock);
    return;
  }
  /*
    Copy the current event into the last event variable.
  */
  LastEvent.DeltaX=MX;
  LastEvent.DeltaY=MY;
  LastEvent.Buttons=Buttons;
  /*
    Write the mouse event. If there is no place for it, ignore it.
  */
  NewLast=MouseLast+1;
  if (NewLast==INPUT_BUFFER_SIZE) NewLast=0;
  if (NewLast!=MouseFirst) {
    r=&MouseEvents[MouseLast];
    MouseLast=NewLast;
    r->DeltaX=DeltaX;
    r->DeltaY=DeltaY;
    r->Buttons=Buttons;
  }
  /*
    Update the event so readers can proceed and unlock the buffer.
  */
  SetEvent(MouseReady);
  ThrUnlock(MouseLock);
}

void ScrRawReadMouse(TMouseEvent *Rec)
{
  TMouseEvent *r;
  HANDLE Handles[2];

  /*
    Wait for both, the lock and the event
  */
  Handles[0]=MouseLock;
  Handles[1]=MouseReady;
  WaitForMultipleObjects(2,Handles,TRUE,INFINITE);
  /*
    If the mouse is enabled, read the mouse event and clear the condition
    variable when there are no more events left. Otherwise simply return
    a meaningless value indicating error.
  */
  if (MouseEnabled) {
    r=&MouseEvents[MouseFirst++];
    if (MouseFirst==INPUT_BUFFER_SIZE) {
      MouseFirst=0;
    }
    Rec->DeltaX=r->DeltaX;
    Rec->DeltaY=r->DeltaY;
    Rec->Buttons=r->Buttons;
    if (MouseFirst==MouseLast) {
      ResetEvent(MouseReady);
    }
  } else {
    Rec->Buttons=255;
  }
  /*
    Unlock the buffer.
  */
  ThrUnlock(MouseLock);
}

uint8 ScrCloseMouse(void)
{
  ThrLock(MouseLock);
  MouseEnabled=false;
  SetEvent(MouseReady);
  ThrUnlock(MouseLock);
  return(0);
}

/*--------------------------------------------------------------------------
 Code:          Window handling

 Purpose:       This code handles the window containing the screen.
--------------------------------------------------------------------------*/

/*** Bitmap info structure ***/
/*
  This structure is created and initialized for the  bitmap transferring
  code. It contains Windows specific code along with the pixel bitmasks.
  The pixel positions are not stored here but in separate variables.
*/
typedef union {
  BITMAPINFO hd;
  struct {
    char _Fill[sizeof(BITMAPINFO)-4];
    DWORD RedBitMask;
    DWORD GreenBitMask;
    DWORD BlueBitMask;
  } Pal;
} TBitMapInfo;

/*
  This type is used to convert the type of the bitmap handle to something
  useful for the layer above this driver. The above layer requires (void *)
  IDs and it is not guaranteed "HDC" type can be safely typecasted to
  (void *).
*/
typedef struct {
  HDC ID;
  HBITMAP PixmapID;
  HDC MaskID;
  HBITMAP MaskPixmapID;
} TBitmap;

/*** WindowVars: The window handling variables ***/
/*
  It is not possible to "change the screen mode" (which involves window
  repositioning and changing) from thread different than the one owning the
  window. If done so, stupid Windows will crash. So a private message is
  going to be used to signal the new mode properties. The actual mode change
  is done by the window handling thread. The thread asking for the mode
  change is suspended and will be resumed once the mode change command is
  completed.
*/
#define WM_SETSCREENMODE 0x7FFF
/*
  The handle of the program itself. This is used when opening any windows.
  I don't understand the purpose of this identifier but Windows demand it
  so lets them have it.
*/
static HINSTANCE hProgram;
/*
  The informations about the video memory bitmap. This is used mainly by
  the Windows only to create bitmaps for drawing on the screen but the
  RGB to pixel conversion routinne also uses the color masks stored here.
*/
/*
  The main window. This window is used as the emulated OSHS screen.
*/
static HWND hWindow;
/*
  The frame dimensions. These are used to compute the location of the window
  on the screen. The window locating calls expect coordinates of the left
  upper corner of the frame but I want the so-called "client area" (the
  place where application's output appear) to be centered (not the entire
  window). I hope these values will not change between reboots.
*/
static int16 FLeft;
static int16 FRight;
static int16 FUp;
static int16 FDown;
/*
  This contains a reference to the physical screen on which the I/O window
  resides. The reference is something that Windows call "device context".
  It is used by the RedrawScreen() procedure to paint the screen image onto
  the screen.
*/
static HDC hScreen;
/*
  The screen (desktop) dimensions. These values are used when the mode is
  changed to compute its new location.
*/
static int16 SWidth;
static int16 SHeight;
static TBitMapInfo VM;
/*
  The current width, height and pixel size of the video mode.
*/
static int16 MWidth;
static int16 MHeight;
static int16 MPixelSize;
/*
  Pointer to the raw virtual video memory of the screen. Programs write into
  this area and when the writing is over, bitmap is created from this
  content and painted onto the window.
*/
static upint8 *VideoMem;
/*
  The handle for the device with the video memory image. The content of this
  device is used to refresh the content of the window when the window was
  overlapped by other windows or so.
*/
static TBitmap hVideoMem;
/*
  The handle for the device with the bitmap to be drawn into the window.
*/
static HDC hBitmap;
/*
  This is the paint structure information. It is used at the end to stop
  the painting session when the I/O driver terminates. The documentation
  says that this is needed and I believe that Windows will crash without
  this action :)
*/
static PAINTSTRUCT ps;
/*
  This lock is used to synchronize the mode change requesting thread with
  the window owning thread. As described above, only the window owner is
  allowed to reposition or resize a window.
*/
static HANDLE WindowReady;
/*
  This flag is used to see if the window is hidden or not. The window is
  hidden at the time of the initialization but will be shown at the time of
  the first attempt to set the mode of the screen.
*/
static BOOL WindowVisible;

/*** Screen refresh ***/
/*
  This procedure is called everytime the I/O window content was disturbed
  and needs a refresh. The screen image will be drawn on the screen or in
  the window (depending on the window's state).
*/
static void RefreshScreen(void)
{
  POINT	ptSize, ptOrg;

  ptSize.x = MWidth;
  ptSize.y = MHeight;
  DPtoLP(hScreen, &ptSize, 1);

  ptOrg.x = 0;
  ptOrg.y = 0;
  DPtoLP(hVideoMem.ID, &ptOrg, 1);

  BitBlt(
    hScreen,
    0, 0,
    ptSize.x, ptSize.y,
    hVideoMem.ID,
    ptOrg.x, ptOrg.y,
    SRCCOPY
  );
}

static void SetupWindowLimits(MINMAXINFO *Info)
{
  /*
    Set the maximal dimensions of the window to something that is surely
    larger than any video mode size supported.
  */
  Info->ptMaxSize.x=4096;
  Info->ptMaxSize.y=4096;
  Info->ptMaxTrackSize.x=4096;
  Info->ptMaxTrackSize.y=4096;
}

/*** Window message handler ***/
/*
  This is a callback procedure which handles the messages sent to the window.
*/
LRESULT CALLBACK HandleMessage(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
  void LogMsg(UINT msg,WPARAM wParam,LPARAM lParam);

  switch (msg) {
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_LBUTTONDOWN: {
      SetCapture(hWindow);
      WriteOutMouse(LOWORD(lParam),HIWORD(lParam),wParam);
      break;
    }
    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP: {
      if (!(wParam & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON))) {
        ReleaseCapture();
      }
      WriteOutMouse(LOWORD(lParam),HIWORD(lParam),wParam);
      break;
    }
    case WM_MOUSEMOVE: {
      WriteOutMouse(LOWORD(lParam),HIWORD(lParam),wParam);
      break;
    }
    case WM_GETMINMAXINFO: {
      SetupWindowLimits((void *)lParam);
      return DefWindowProc(hwnd,msg,wParam,lParam);
    }
    case WM_DESTROY: {
      PostQuitMessage(0);
      break;
    }
    case WM_PAINT: {
      RefreshScreen();
      /*
        This message MUST be processed by the default window procedure.
        Otherwise the message will be posted again and again.
      */
    }
    default: {
      return DefWindowProc(hwnd,msg,wParam,lParam);
    }
  }
  return 0;
}

/*** Window class builder ***/
/*
  Builds and registers a window class to be used by all windows on the
  screen.
*/
static void BuildClass(void)
{
  WNDCLASS wc;

  memset(&wc,0,sizeof(WNDCLASS));
  wc.style = CS_NOCLOSE;
  wc.lpfnWndProc = (WNDPROC)HandleMessage;
  wc.hInstance = hProgram;
  wc.hbrBackground = NULL;
  wc.lpszClassName = "IOWindow";
  wc.lpszMenuName = NULL;
  wc.hCursor = LoadCursor(NULL,IDC_ARROW);
  wc.hIcon = LoadIcon(NULL,IDI_APPLICATION);
  if (!RegisterClass(&wc)) {
    ThrReportError("Unable to register the new window class");
  }
}

/*** Window builder ***/
/*
  Procedure:    BuildWindow

  Purpose:      Builds the I/O driver window and displays it in maximized
                state.

  Return value: When the window was successfully created, the return value
                is nonzero. If the function failed, the return value is
                zero.

  Other output: - [hWindow] will contain the handle of the window.

  Remarks:      - The procedure will return zero after first failure
                  occured.
*/
static void BuildWindow(void)
{
  RECT Win,Client;
  HWND Desktop;

  /* register the class of the window */
  BuildClass();

  /* create the window */
  hWindow=CreateWindow(
    "IOWindow","OSHS application",0,100,100,100,100,NULL,NULL,hProgram,NULL
  );
  if (!hWindow) {
    ThrReportError("Could not create I/O window");
  }
  /*
    Get the frame dimensions. This is so complicated because the
    GetClientRect() has a design bug that prevents us from determining these
    values directly (it returns the "client" coordinates but relative to the
    client area itself so the first two fields of the rectangle structure
    will always be 0 and the rest will contain the dimensions of the client
    area). This code assumes that the left, right and bottom frame have the
    same dimension.
  */
  if (!GetClientRect(hWindow,&Client)) {
    ThrReportError("Unable to get the window position");
  }
  if (!GetWindowRect(hWindow,&Win)) {
    ThrReportError("Unable to get the window area position");
  }
  Win.right-=Win.left;
  Win.bottom-=Win.top;
  FLeft=(Win.right-Client.right) / 2;
  FRight=FLeft;
  FDown=FLeft;
  FUp=Win.bottom-Client.bottom-FDown;
  /*
    Get the dimensions of the desktop itself.
  */
  Desktop=GetDesktopWindow();
  if (Desktop==NULL) {
    ThrReportError("Unable to obtain the desktop window");
  }
  if (!GetWindowRect(Desktop,&Win)) {
    ThrReportError("Unable to obtain the desktop dimensions");
  }
  Win.right-=Win.left;
  Win.bottom-=Win.top;
  SWidth=Win.right;
  SHeight=Win.bottom;
}

/*** Image drawing ***/
/*
  Draws an image on the screen and into the screen image. The [x] and [y]
  contains screen coordinates of the upper left corner of the image to draw
  and [SizeX] and [SizeY] contain image dimensions. The [Data] points to the
  image data.

  - Each row in the image must have even number of pixels. If the [SizeX] is
    odd then each last pixel is ignored.
  - Pixels in the image are 16 bit with the format RRRRRGGGGGGBBBBB.
*/
static void PutBitmapData(
  WORD x,
  WORD y,
  WORD SizeX,
  WORD SizeY,
  void *Data,
  HDC hDestBitmap
) {
  HBITMAP hBitmapData;
  HGDIOBJ hOldGDIObject;

  // Set dimensions of the bitmap
  VM.hd.bmiHeader.biWidth=SizeX;
  VM.hd.bmiHeader.biHeight=-SizeY;

  hBitmapData=CreateDIBitmap(
    hScreen,
    &VM.hd.bmiHeader,
    CBM_INIT,
    Data,
    &VM.hd,
    DIB_RGB_COLORS
  );

  // build memory pseudodevice with the bitmap data
  hOldGDIObject=SelectObject(hBitmap,hBitmapData);

  // transfer the data
  BitBlt(hDestBitmap, x, y, SizeX, SizeY, hBitmap, 0, 0, SRCCOPY);

  // delete the temporary object with the bitmap data
  SelectObject(hBitmap,hOldGDIObject);
  DeleteObject(hBitmapData);
}

void ScrGetCurrentColorFormat(TColorFormat *F)
{
  F->BitsPerPixel=MPixelSize*8;
  F->RMask=VM.Pal.RedBitMask;
  F->GMask=VM.Pal.GreenBitMask;
  F->BMask=VM.Pal.BlueBitMask;
}

int16 ScrPrepareAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  return(0);
}

int16 ScrExposeBuffer(TColorConversionInfo *I)
{
  uint32 PixelNumber;
  uint16 BytesPerLine;
  TBitmap *P;

  I->OutPtr=VideoMem;
  PixelNumber=(MWidth*MHeight);
  PixelNumber-=PixelNumber % I->PixelsOnLine;
  P=I->DestPixmap;
  if (P->MaskID) {
    PixelNumber/=2;
    I->PixelSetR=0;
    I->PixelSetG=0;
    I->PixelSetB=0;
    I->PixelResetR=255;
    I->PixelResetG=255;
    I->PixelResetB=255;
    I->PixmapMaskDelta=PixelNumber*(I->Out.BitsPerPixel/8);
    I->MaskType=mtPixmap;
  }
  I->FreePixels=PixelNumber;
  BytesPerLine=I->PixelsOnLine*(I->Out.BitsPerPixel/8);
  I->LinePaddingBytes=(4 - BytesPerLine%4)&3;
  return(0);
}

int16 ScrFlushPixels(TColorConversionInfo *I)
{
  int16 StartLocY;
  int16 CurLocY;
  int16 Height;
  int16 Width;
  TBitmap *P;
  void *Ptr;

  assert(I->UnprocessedPixels);
  StartLocY=I->SLine;
  CurLocY=I->LineNumber;
  if (I->PixelsToFill==I->PixelsOnLine) CurLocY--;
  Height=CurLocY-StartLocY+1;
  Width=I->PixelsOnLine;
  P=I->DestPixmap;
  PutBitmapData(0,StartLocY,Width,Height,VideoMem,P->ID);
  if (P->MaskID) {
    Ptr=VideoMem+I->PixmapMaskDelta;
    PutBitmapData(0,StartLocY,Width,Height,Ptr,P->MaskID);
  }
  return(0);
}

int16 ScrDestroyAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  return(0);
}

void *ScrGetScreenPixmap(void)
{
  return(&hVideoMem);
}

static int16 CreatePixmapCore(uint16 Width,uint16 Height,TBitmap *P)
{
  P->ID=CreateCompatibleDC(hScreen);
  if (!P->ID) return(beIO);
  P->PixmapID=CreateCompatibleBitmap(hScreen,Width,Height);
  if (!P->PixmapID) {
    DeleteDC(P->ID);
    return(beIO);
  }
  SelectObject(P->ID, P->PixmapID);
  return(0);
}

int16 ScrCreatePixmap(uint16 Width,uint16 Height,void **Result)
{
  TBitmap *P;
  int16 IOResult;

  P=KMEM__malloc(sizeof(*P));
  if (P==NULL) return(beNOMEM);
  IOResult=CreatePixmapCore(Width,Height,P);
  if (IOResult!=0) {
    KMEM__free(P);
    return(IOResult);
  }
  P->MaskID=0;
  P->MaskPixmapID=0;
  *Result=P;
  return(0);
}

int16 ScrCreateMaskedPixmap(uint16 Width,uint16 Height,void **Result)
{
  TBitmap *P;
  int16 IOResult;

  IOResult=ScrCreatePixmap(Width,Height,Result);
  if (IOResult!=0) return(IOResult);
  P=*Result;
  P->MaskID=P->ID;
  P->MaskPixmapID=P->PixmapID;
  IOResult=CreatePixmapCore(Width,Height,P);
  if (IOResult!=0) {
    P->ID=0;
    P->PixmapID=0;
    ScrDestroyPixmap(P);
    return(IOResult);
  }
  return(0);
}

int16 ScrTransferRectangle(
  void *SrcPixmap,
  uint16 SrcX,
  uint16 SrcY,
  uint16 Width,
  uint16 Height,
  void *DstPixmap,
  uint16 DstX,
  uint16 DstY
) {
  TBitmap *Src;
  TBitmap *Dst;

  Src=SrcPixmap;
  Dst=DstPixmap;
  if (Src->MaskID) {
    BitBlt(Dst->ID,DstX,DstY,Width,Height,Src->MaskID,SrcX,SrcY,SRCAND);
    BitBlt(Dst->ID,DstX,DstY,Width,Height,Src->ID,SrcX,SrcY,SRCPAINT);
  } else {
    BitBlt(Dst->ID,DstX,DstY,Width,Height,Src->ID,SrcX,SrcY,SRCCOPY);
  }
  return(0);
}

void ScrDestroyPixmap(void *Pixmap)
{
  TBitmap *P;

  P=Pixmap;
  if (P->ID) DeleteDC(P->ID);
  if (P->PixmapID) DeleteObject(P->PixmapID);
  if (P->MaskID) DeleteDC(P->MaskID);
  if (P->MaskPixmapID) DeleteObject(P->MaskPixmapID);
  KMEM__free(P);
}

/*** Rectange redraw ***/
/*
  This procedure redraws the specified portion of the screen.
*/
void ScrUpdateRectangle(uint16 x,uint16 y,uint16 Width,uint16 Height)
{
  BitBlt(hScreen, x, y, Width, Height, hVideoMem.ID, x, y, SRCCOPY);
}

/*** Screen resizer ***/
/*
  This procedure executes the video mode change request. It may be called
  by the thread owning the window only.
*/
static void ResizeTheWindow(const TSetModeCommand *C)
{
  HGDIOBJ *hOldImage;
  UINT Flags;
  int16 X;
  int16 Y;
  int16 WWidth;
  int16 WHeight;
  void *NewPtr;

  /*
    Store the dimensions of the new video mode to be created. Other code
    needs this information to be available.
  */
  MWidth=C->Width;
  MHeight=C->Height;
  MPixelSize=C->PixelSize+1;
  /*
    Generate the position of the window according to the requested mode
    resolution.
  */
  X=(SWidth-MWidth)/2 - FLeft;
  Y=(SHeight-MHeight)/2 - FUp;
  WWidth=FLeft+MWidth+FRight;
  WHeight=FUp+MHeight+FDown;
  /*
    Resize the window first. The window is still hidden here when this is
    the first call to ResizeTheWindow() so it will not flicker at startup.
  */
  Flags=SWP_NOACTIVATE|SWP_NOCOPYBITS|SWP_NOZORDER;
  if (!SetWindowPos(hWindow,NULL,X,Y,WWidth,WHeight,Flags)) {
    ThrReportError("Unable to move the window");
  }
  /*
    Reallocate the video memory buffer
  */
  NewPtr=realloc(VideoMem,MWidth*MHeight*2+8);
  if (NewPtr) {
    VideoMem=NewPtr;
  } else {
    free(VideoMem);
    ThrReportError("Unable to create video memory buffer");
  }
  /*
    If this is the first request of video mode change, now it is the time
    to show the window on the screen.
  */
  if (!WindowVisible) {
    /*
      Show the window.
    */
    ShowWindow(hWindow,SW_SHOWNORMAL);
    /*
      Initialize the bitmap description. Here only the static items of the
      structure are initialized.
    */
    VM.hd.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
    VM.hd.bmiHeader.biPlanes=1;
    VM.hd.bmiHeader.biCompression=BI_BITFIELDS;
    VM.hd.bmiHeader.biSizeImage=0;
    VM.hd.bmiHeader.biXPelsPerMeter=0;
    VM.hd.bmiHeader.biYPelsPerMeter=0;
    VM.hd.bmiHeader.biClrUsed=0;
    VM.hd.bmiHeader.biClrImportant=0;
    /*
      Start the painting session.
    */
    hScreen=BeginPaint(hWindow,&ps);
    if (!hScreen) {
      ThrReportError("Unable to start painting session");
    }
    /*
      Build memory pseudodevice for the screen image.
    */
    hVideoMem.ID = CreateCompatibleDC(hScreen);
    if (!hVideoMem.ID) {
      ThrReportError("Cannot build access device for the screen image");
    }
    /*
      Create another device context but for the bitmaps that will be drawn
      on the screen during the redrawing requests this time.
    */
    hBitmap=CreateCompatibleDC(hVideoMem.ID);
  }
  /*
    Set the size of the pixel and the bitmasks for the colors in the header
    info.
  */
  VM.hd.bmiHeader.biBitCount=MPixelSize*8;
  if (MPixelSize==2) {
    VM.Pal.RedBitMask=0xF800;
    VM.Pal.GreenBitMask=0x07E0;
    VM.Pal.BlueBitMask=0x001F;
  } else {
    VM.Pal.RedBitMask=0x00FF0000;
    VM.Pal.GreenBitMask=0x0000FF00;
    VM.Pal.BlueBitMask=0x000000FF;
  }
  /*
    Build the screen image. It is built as hardware-dependent bitmap for
    performance reasons.
  */
  hVideoMem.PixmapID=CreateCompatibleBitmap(hScreen,MWidth,MHeight);
  if (!hVideoMem.PixmapID) {
    ThrReportError("Cannot create screen image");
  }
  /*
    Connect the screen image into the memory pseudodevice. The screen image
    cannot be manipulated directly, the
    pseudodevice is used for the access. The previously connected image will
    be placed into hOldImage.
  */
  hOldImage=SelectObject(hVideoMem.ID, hVideoMem.PixmapID);
  if (!hOldImage) {
    ThrReportError("Cannot connect the screen image to the access device");
  }
  /*
    If the window was visible, the "OldImage" refers to an image that must
    be deleted, because it is not required anymore. Otherwise simply set
    this flag (we made the window visible at the beginning).
  */
  if (WindowVisible) {
    DeleteObject(hOldImage);
  } else {
    WindowVisible=TRUE;
  }
  /*
    Refresh the content of the window.
  */
  ScrUpdateRectangle(0,0,MWidth,MHeight);
}

/*** Bitmap destroyer ***/
void DestroyBitmap(void)
{
  if (hVideoMem.PixmapID) DeleteObject(hVideoMem.PixmapID);
  if (hVideoMem.ID) DeleteDC(hVideoMem.ID);
  if (hBitmap) DeleteDC(hBitmap);
  EndPaint(hWindow,&ps);
}

uint8 ScrSetupScreenMode(const TSetModeCommand *C)
{
  if (C->PixelSize!=1 && C->PixelSize!=3) return(screMODESUP);
  ThrLock(WindowReady);
  PostMessage(hWindow,WM_SETSCREENMODE,0,(LPARAM)C);
  ThrLock(WindowReady);
  ThrUnlock(WindowReady);
  return(0);
}

bool ScrCheckModeSet(void)
{
  return(WindowVisible);
}

/*--------------------------------------------------------------------------
 Code:          Main program starter

 Purpose:       This piece of code runs the main program. It is required to
                be in this library so this Windows specific startup is
                completely hidden from the system.
--------------------------------------------------------------------------*/

void InitializeIODriver(void)
{
  /*
    Create the lock used to synchronize the communication about the set mode
    commands.
  */
  WindowReady=ThrCreateLock();
  if (!WindowReady) {
    ThrReportError("Unable to create window ready lock");
  }
  /*
    Initialize the keyboard input buffer.
  */
  LastKey=65535;
  KeybReady=CreateEvent(NULL,TRUE,TRUE,NULL);
  KeybLock=ThrCreateLock();
  if ((!KeybReady) || (!KeybLock)) {
    ThrReportError("Unable to create keyboard synchronization objects");
  }
  /*
    Initialize the mouse input buffer.
  */
  MouseFirst=0;
  MouseLast=0;
  MouseReady=CreateEvent(NULL,TRUE,TRUE,NULL);
  MouseLock=ThrCreateLock();
  if ((!MouseReady) || (!MouseLock)) {
    ThrReportError("Unable to create mouse synchronization objects");
  }
  /*
    Get the handle of the program
  */
  hProgram=GetModuleHandle(NULL);

  /* */
  BuildWindow();
}

void ShutdownIODriver(void)
{
  // shutdown the application
  DestroyBitmap();
}

DWORD WINAPI RunMainThreadOfTheProgram(void *Pointer)
{
  DWORD Code;
  TMainProcedure *Main;

  Main=Pointer;
  Code=Main();
  PostMessage(hWindow,WM_CLOSE,0,0);

  return(Code);
}

/*** Thread builder ***/
/*
  This procedure creates a thread executing from the given adress. It
  returns only on success. In the case of a failure it will terminate with
  error message.
*/
HANDLE BuildThread(LPTHREAD_START_ROUTINE Start,TMainProcedure *Main)
{
  HANDLE Thread;

  Thread=ThrCreateThreadCore(Start,Main);

  if (!Thread) {
    ThrReportError("Unable to create thread");
  }

  return(Thread);
}

int16 RunTheProgram(TMainProcedure *Main)
{
  MSG msg;
  HANDLE Thread;

  /* initialzie IO driver library */
  InitializeIODriver();

  /* build window thread */
  BuildThread(&RunMainThreadOfTheProgram,Main);

  // enter message handling loop
  while (GetMessage(&msg,NULL,0,0)) {
    if (
      (msg.message==WM_KEYUP) || (msg.message==WM_KEYDOWN) ||
      (msg.message==WM_SYSKEYUP) || (msg.message==WM_SYSKEYDOWN)
    ) {
      Keyboard(msg.wParam,msg.lParam);
    } else if (msg.message==WM_SETSCREENMODE) {
      ResizeTheWindow((void *)msg.lParam);
      ThrUnlock(WindowReady);
    } else {
      DispatchMessage(&msg);
    }
    if (!PeekMessage(&msg,NULL,0,0,PM_NOREMOVE)) {
      FlushKey();
    }
  }

  /* Wait for the thread of the program to finish */
  ThrWait(Thread);
  CloseHandle(Thread);

  // shutdown IO driver
  ShutdownIODriver();

  // return the exit code
  return msg.wParam;
}
