#include "defs.gh"

#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#if __b__USE_SHM__
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

typedef Status TX11Status;
#undef Status

#include "types.h"
#include "kernel.h"
#include "traps.h"
#include "pdbase.h"
#include "pdgfxcon.h"
#include "pdkeys.h"
#include "memory.h"
#include "strings.h"
#include "thread.h"
#include "thrmain.h"
#include "gcwconf.h"
#include "gcwdata.h"
#include "init.h"

#define __INTERFACE__
#include "gcwx11.h"
#undef __INTERFACE__

#include "typesnd.h"

/*********************************
*** The window output handling ***
*********************************/


/*
  The lock used to synchronize access to these structures.
*/
static void *ScreenLock;
/*
  So called "graphics context" which is used when transferring data between
  user supplied buffers and the on screen objects such as the background
  pixmap.
*/
static GC GraphicsContext;
/*
  This is some king of video memory format info and is used when pixmaps are
  created for the windows.
*/
static Visual *Vis;
/*
  The connection to the X server which is used to manipulate the windows
  representing the simulated screens.
*/
static Display *GFXDisp;
/*
  Default screen number. This information is mandatory when creating windows.
  I guess that it contains the 'n' from the 'server:n' specification.
*/
static int16 ScreenNum;
/*
  The pixel depth, i.e. count of bits used to represent a single pixel on
  the X server. It is determined at driver initialization time and is
  supposed to not change. It is required when creating the background pixmap
  and when determining additional details about the screen pixel format.
*/
static uint16 ValidDepth;
/*
  The pixel color format used by the X server.
*/
static TColorFormat SF;
/*
  The window that represents the screen.
*/
static Window Win;

#if __b__USE_SHM__
/*
  This is initialized by the driver initialization to 'true' if using the
  "shared memory extension" feature of the X server can be used.
*/
static bool UseSHM;
/*
  This structure is used to manipulate the shared memory block used to
  transport data between the X server and the client.
*/
static XShmSegmentInfo SHMImgInfo;
static XShmSegmentInfo SHMMaskInfo;
/*
  Set to true when something with the shared memory goes wrong. When the
  problem occurs during initialization, the code fallbacks to ordinary X
  server transport (no shared memory piece). During driver runtime no errors
  are expected.
*/
static bool SHMError;
#endif

/*
  Detect the format of the pixels accepted by the server. 
*/
static bool DetectPixelFormat(void)
{
  XVisualInfo Info;
  TX11Status Result;
  XImage *Temp;
  upint8 *Buffer;

  /*
    Try to autodetect the pixel format and fill in the screen pixel format
    structure.
  */
  Result=XMatchVisualInfo(GFXDisp,ScreenNum,ValidDepth,DirectColor,&Info);
  if (Result==0) {
    Result=XMatchVisualInfo(GFXDisp,ScreenNum,ValidDepth,TrueColor,&Info);
  }
  if (Result==0) {
    Result=XMatchVisualInfo(GFXDisp,ScreenNum,ValidDepth,PseudoColor,&Info);
  }
  if (Result==0) {
    Result=XMatchVisualInfo(GFXDisp,ScreenNum,ValidDepth,StaticColor,&Info);
  }
  if (Result==0) {
    /*Fatal("Unable to autodetect the screen pixel format");*/
    return(false);
  }
  Vis=Info.visual;
  SF.RMask=Info.red_mask;
  SF.GMask=Info.green_mask;
  SF.BMask=Info.blue_mask;
  /*
    Detect the pixel size used by the video mode. Sometimes a videomode
    presented as 24 bit videomode is in fact a 32 bit videomode with the
    highest byte unused.
  */
  SF.BitsPerPixel=ValidDepth;
  if (ValidDepth==24) {
    Buffer=KMEM__malloc(64);
    if (Buffer==NULL) return(false);
    Temp=XCreateImage(
      GFXDisp, Vis,
      ValidDepth, ZPixmap, 0,
      (void *)Buffer,
      16, 1,
      32, 0
    );
    if (Temp==NULL) {
      KMEM__free(Buffer);
      return(false);
    }
    if (Temp->bytes_per_line==64) SF.BitsPerPixel=32;
    XDestroyImage(Temp);
  }
  /*
    Build the screen color format variable so that it can be used for pixel
    conversions.
  */
  BuildColorFormat(&SF);
  return(true);
}

#if __b__USE_SHM__
static void CheckSHMAvailable()
{
  const char *DispName;
  bool LocalX11;

  UseSHM=false;
  DispName=XDisplayName(NULL);
  LocalX11=(StrNCmp(DispName,":",1)==0) || (StrNCmp(DispName,"unix:",5)==0);
  if (LocalX11) {
    UseSHM=XShmQueryExtension(GFXDisp);
  }
  XSync(GFXDisp,false);
}

static int16 (*XHandler)(Display *, XErrorEvent *);

static int16 SHMErrorHandler(Display *d, XErrorEvent *e)
{
  if (e->error_code == BadAccess) {
    SHMError=true;
    return(0);
  } else {
    return(XHandler(d,e));
  }
}

static void AllocAndAttachSHM(uint32 MemSize,XShmSegmentInfo *I)
{
  if (!UseSHM) return;
  I->shmid=shmget(
    IPC_PRIVATE,
    MemSize,
    IPC_CREAT | 0777
  );
  SHMError=I->shmid<=0;
  if (!SHMError) {
    I->shmaddr=(char *)shmat(I->shmid,0,0);
    I->readOnly=false;
    SHMError=I->shmaddr==(char *)(-1);
    if (!SHMError) {
      XHandler=XSetErrorHandler(SHMErrorHandler);
      XShmAttach(GFXDisp,I);
      XSync(GFXDisp,true);
      XSetErrorHandler(XHandler);
      if (SHMError) {
        shmdt(I->shmaddr);
      }
    }
    shmctl(I->shmid, IPC_RMID, NULL);
  }
  if (SHMError) UseSHM=false;
}

static void DetachSHM(XShmSegmentInfo *I)
{
  XShmDetach(GFXDisp,I);
  XSync(GFXDisp,false);
  shmdt(I->shmaddr);
}

static void TrySHM(uint32 MemSize)
{
  bool FirstOK;

  AllocAndAttachSHM(MemSize,&SHMImgInfo);
  FirstOK=UseSHM;
  AllocAndAttachSHM(MemSize/8,&SHMMaskInfo);
  if (FirstOK && !UseSHM) DetachSHM(&SHMImgInfo);
}

static void DestroySHM(void)
{
  DetachSHM(&SHMImgInfo);
  DetachSHM(&SHMMaskInfo);
}
#endif

static bool InitializeScreenDriver(void)
{
  ScreenLock=ThrRawCreateLock();
  if (ScreenLock==NULL) {
    return(false);
    /*Fatal("Unable to create screen access lock");*/
  }
  GFXDisp = XOpenDisplay(NULL);
  if (GFXDisp==NULL) {
    return(false);
    /*Fatal("Unable to open GFX display");*/
  }
  ScreenNum = DefaultScreen(GFXDisp);
  ValidDepth=DefaultDepth(GFXDisp,ScreenNum);
  if (!DetectPixelFormat()) {
    return(false);
  }
#if __b__USE_SHM__
  CheckSHMAvailable();
  TrySHM(GCWCFG_TRANSFER_BUFFER_SIZE);
  if (UseSHM) {
    InstallShutdownProc(DestroySHM);
    return(true);
  }
#endif
  return(true);
}

/********************************
*** The event stream handling ***
********************************/

/*
  The connection to the X server which is used to accept events. Attempt to
  accept events directly from the connection stored in GFXDisp would lead
  to crash, because event waiting and windows handling is done in different
  threads and X11 is not thread safe.
*/
static Display *EventDisp;

/*
  Keyboard buffer
*/

#define KEYBOARD_BUFFER_SIZE 1024

static void *KeybReady;
static bool KeyboardReady;
static void *KeybFreeBuf;
static void *KeybLock;
static bool KeyboardOpened;
static int16 KeyIn;
static int16 KeyOut;
static upint16 KeyBuffer[KEYBOARD_BUFFER_SIZE];

static void ScrRawWriteKey(uint16 Key)
{
  bool Done;

  do {
    ThrRawWaitCond(KeybFreeBuf);
    ThrRawLock(KeybLock);
    Done=(KeyIn!=KeyOut) || (!KeyboardReady);
    if (Done) {
      KeyBuffer[KeyIn]=Key;
      KeyIn++;
      KeyboardReady=true;
      ThrRawSetCond(KeybReady);
      if (KeyIn==KEYBOARD_BUFFER_SIZE) KeyIn=0;
      if (KeyOut==KeyIn) ThrRawResetCond(KeybFreeBuf);
    }
    ThrRawUnlock(KeybLock);
  } while (!Done);
}

uint16 ScrRawReadKey(void)
{
  uint16 Result;

  ThrRawWaitCond(KeybReady);
  ThrRawLock(KeybLock);
  if (KeyboardOpened) {
    Result=KeyBuffer[KeyOut];
    KeyOut++;
    ThrRawSetCond(KeybFreeBuf);
    if (KeyOut==KEYBOARD_BUFFER_SIZE) KeyOut=0;
    KeyboardReady=(KeyOut!=KeyIn);
    if (!KeyboardReady) ThrRawResetCond(KeybReady);
  } else {
    Result=0xFFFF;
  }
  ThrRawUnlock(KeybLock);
  return(Result);
}

static uint16 TranslateKey(XEvent *E)
{
  switch (E->xkey.keycode) {
    case   9: return(RK_ESC);
    case  67: return(RK_F1);
    case  68: return(RK_F2);
    case  69: return(RK_F3);
    case  70: return(RK_F4);
    case  71: return(RK_F5);
    case  72: return(RK_F6);
    case  73: return(RK_F7);
    case  74: return(RK_F8);
    case  75: return(RK_F9);
    case  76: return(RK_F10);
    case  95: return(RK_F11);
    case  96: return(RK_F12);
    case 111: return(RK_PRINTSCREEN);
    case  78: return(RK_SCROLLOCK);
    case 110: return(RK_PAUSE);
    case  49: return(RK_BACKAPOSTROPHE);
    case  22: return(RK_BACKSPACE);
    case  10: return(RK_1);
    case  11: return(RK_2);
    case  12: return(RK_3);
    case  13: return(RK_4);
    case  14: return(RK_5);
    case  15: return(RK_6);
    case  16: return(RK_7);
    case  17: return(RK_8);
    case  18: return(RK_9);
    case  19: return(RK_0);
    case  20: return(RK_MINUS);
    case  21: return(RK_EQUAL);
    case  51: return(RK_BACKSLASH);
    case  23: return(RK_TAB);
    case  24: return(RK_Q);
    case  25: return(RK_W);
    case  26: return(RK_E);
    case  27: return(RK_R);
    case  28: return(RK_T);
    case  29: return(RK_Y);
    case  30: return(RK_U);
    case  31: return(RK_I);
    case  32: return(RK_O);
    case  33: return(RK_P);
    case  34: return(RK_LEFTBRACKET);
    case  35: return(RK_RIGHTBRACKET);
    case  36: return(RK_ENTER);
    case  66: return(RK_CAPSLOCK);
    case  38: return(RK_A);
    case  39: return(RK_S);
    case  40: return(RK_D);
    case  41: return(RK_F);
    case  42: return(RK_G);
    case  43: return(RK_H);
    case  44: return(RK_J);
    case  45: return(RK_K);
    case  46: return(RK_L);
    case  47: return(RK_SEMICOLON);
    case  48: return(RK_APOSTROPHE);
    case  50: return(RK_LSHIFT);
    case  52: return(RK_Z);
    case  53: return(RK_X);
    case  54: return(RK_C);
    case  55: return(RK_V);
    case  56: return(RK_B);
    case  57: return(RK_N);
    case  58: return(RK_M);
    case  59: return(RK_COMMA);
    case  60: return(RK_PERIOD);
    case  61: return(RK_SLASH);
    case  62: return(RK_RSHIFT);
    case  37: return(RK_LCTRL);
    case 115: return(RK_LWIN);
    case  64: return(RK_LALT);
    case  65: return(RK_SPACE);
    case 113: return(RK_RALT);
    case 116: return(RK_RWIN);
    case 117: return(RK_MENU);
    case 109: return(RK_RCTRL);
    case 106: return(RK_INSERT);
    case 107: return(RK_DELETE);
    case  97: return(RK_HOME);
    case 103: return(RK_END);
    case  99: return(RK_PGUP);
    case 105: return(RK_PGDN);
    case  77: return(RK_NUMLOCK);
    case 112: return(RK_KP_SLASH);
    case  63: return(RK_KP_STAR);
    case  82: return(RK_KP_MINUS);
    case  79: return(RK_KP7);
    case  80: return(RK_KP8);
    case  81: return(RK_KP9);
    case  86: return(RK_KP_PLUS);
    case  83: return(RK_KP4);
    case  84: return(RK_KP5);
    case  85: return(RK_KP6);
    case  87: return(RK_KP1);
    case  88: return(RK_KP2);
    case  89: return(RK_KP3);
    case  90: return(RK_KP0);
    case  91: return(RK_KP_PERIOD);
    case 108: return(RK_KP_ENTER);
    case 100: return(RK_LEFT);
    case 102: return(RK_RIGHT);
    case  98: return(RK_UP);
    case 104: return(RK_DOWN);
    default: return(0xFFFF);
  }
}

static void ProcessKeyEvent(XEvent *E)
{
  uint16 KeyCode;

  KeyCode=TranslateKey(E);
  if (KeyCode==0xFFFF) return;
  if (E->type==KeyRelease) KeyCode+=MK_RELEASED;
  ScrRawWriteKey(KeyCode);
}

static bool InitializeEventGathering(void)
{
  KeybReady=ThrRawCreateCond();
  KeybFreeBuf=ThrRawCreateCond();
  KeybLock=ThrRawCreateLock();
  if (KeybReady==NULL || KeybFreeBuf==NULL || KeybLock==NULL) {
    return(false);
    /*Fatal("Unable to create keyboard I/O syncronization structures");*/
  }
  EventDisp = XOpenDisplay(NULL);
  if (EventDisp==NULL) {
    return(false);
    /*Fatal("Unable to open display for event gathering");*/
  }
  return(true);
}

static void SelectEvents(Window Win,long EventMask)
{
  XSelectInput(EventDisp, Win, EventMask |
    KeyPressMask | KeyReleaseMask |
    ButtonPressMask | ButtonReleaseMask | PointerMotionMask
  );
}

static void GetNextEvent(XEvent *E)
{
  bool Searching;

  Searching=true;
  do {
    XNextEvent(EventDisp, E);
    switch (E->type) {
      case KeyPress:
      case KeyRelease: {
        ProcessKeyEvent(E);
        break;
      }
      default: {
        Searching=false;
        break;
      }
    }
  } while (Searching);
}

int16 ScrOpenKeyboard(void)
{
  int16 Result;

  ThrRawLock(KeybLock);
  if (KeyboardOpened) {
    Result=beBUSY;
  } else {
    KeyboardOpened=true;
    ThrRawResetCond(KeybReady);
    ThrRawSetCond(KeybFreeBuf);
    KeyIn=0;
    KeyOut=0;
    Result=0;
  }
  ThrRawUnlock(KeybLock);
  return(Result);
}

int16 ScrCloseKeyboard(void)
{
  ThrRawLock(KeybLock);
  KeyboardOpened=false;
  ThrRawSetCond(KeybReady);
  ThrRawUnlock(KeybLock);
  return(beNOTSUP);
}

/**************************************
*** Event gathering thread handling ***
**************************************/

/*
  The ID of the thread which waits for events. Is set to 0 if no thread is
  running yet. Used during driver initialization.
  FIXME: I probably can remove it completely.
*/
static void *Thread;

static int16 RunEventLoop(void *Ptr)
{
  XEvent E;

  /*
    There is no clear way to 
  */
  while(1) {
    GetNextEvent(&E);
  }
  return(0);
}

/*
  Starts the event loop in another thread. This must run in another thread,
  because there is no way how to stop it once the program terminates.
*/
static bool RunEventGatheringThread(void)
{
  if (Thread==NULL) {
    Thread=ThrRawCreate(RunEventLoop,NULL);
  }
  return(Thread!=NULL);
}

/*****************************
*** The common driver code ***
*****************************/

/*
  Initialize the driver. Open a connection to the X server and read some
  basic informations from it. On any errors calls Fatal(), which aborts
  the program with a message.
*/
static bool InitializeDriver(void)
{
  if (!InitializeScreenDriver()) return(false);
  if (!InitializeEventGathering()) return(false);
  return(true);
}

static int16 BuildNewWindow(int16 Width, int16 Height)
{
  /*
    Create the window.
  */
  Win = XCreateSimpleWindow(
    GFXDisp, RootWindow(GFXDisp, ScreenNum),
    0, 0, Width, Height, 0,
    BlackPixel(GFXDisp, ScreenNum),
    WhitePixel(GFXDisp, ScreenNum)
  );
  if (Win==0) return(screSCRINIT);
  /*
    Flush the X11 command buffers so our window will be created before
    continuing. We need the window to be there before we can select events
    for it and so on.
  */
  XSync(GFXDisp,0);
  /*
    Select events to be processed. This must be done here as any changes to
    the list of requested events are ignored by the event gathering machinery
    once it was started.

    Note that we don't accept exposure and structure notify events. This is
    because we will create a pixmap and setup it as a background of the
    window and all drawing operations will be performed on the pixmap.
  */
  SelectEvents(Win,0);
  /*
    Now it is safe to start the event gathering machinery.
  */
  if (!RunEventGatheringThread()) return(screSCRINIT);
  /*
    Success is indicated by a zero.
  */
  return(0);
}

/***************************
*** The main driver code ***
***************************/

/*
  This structure is used to hold the identifier of a pixmap. It is not
  suitable to use the "Pixmap" type directly, because there is no guarantee
  about its castability to (void *) and the layer above this driver requires
  the pixmaps to be referenced by a (void *) typed identifier.
*/
typedef struct {
  Pixmap ID;
  uint16 Width;
  Pixmap MaskID;
  GC MaskGC;
} TPixmap;

/*
  Background pixmap used by the window. The pixmap is stored on the server
  and the server automatically uses its content when it has to redraw exposed
  portions of the screen window. This reduces network bandwidth since it is
  not needed to generate exposure events to the client.
*/
static TPixmap BGPixmap;

/*
  Set screen mode. The screen mode is represented with a window which we
  create here.
*/
static int16 SetVideoMode(uint16 Width,uint16 Height)
{
  XSizeHints H;
  TX11Status Status;
  bool FreshWin;
  int16 Result;

  /*
    Create the window. If something gets wrong, bail out.
  */
  FreshWin=(Win==0);
  if (FreshWin) {
    Result=BuildNewWindow(Width,Height);
    if (Result!=0) return(Result);
  } else {
    XFreePixmap(GFXDisp,BGPixmap.ID);
  }
  /*
    Create graphics context for the window
  */
  GraphicsContext=XCreateGC(GFXDisp, Win, 0, NULL);
  /*
    Everytime something is copied to an offscreen pixmap using XCopyArea(),
    the X server is about to generate GraphicsExpose and/or NoExpose events
    to the window for which the graphics context was created. Turn off this
    feature as the driver doesn't need it and it will cause memory leaks if
    the program is not listening to the events from the window (for example
    when it does not open the keyboard).
  */
  XSetGraphicsExposures(GFXDisp, GraphicsContext, false);
  /*
    Disable window resizing
  */
  H.flags=(
    USPosition | USSize | PPosition | PSize | PMinSize | PMaxSize |
    PResizeInc | PBaseSize
  );
  H.x=0;
  H.y=0;
  H.width=Width;
  H.height=Height;
  H.min_width=Width;
  H.min_height=Height;
  H.max_width=Width;
  H.max_height=Height;
  H.width_inc=0;
  H.height_inc=0;
  H.base_width=Width;
  H.base_height=Height;
  XSetWMNormalHints(GFXDisp,Win,&H);
  /*
    Create the background pixmap.
  */
  BGPixmap.ID=XCreatePixmap(GFXDisp,Win,Width,Height,ValidDepth);
  BGPixmap.Width=Width;
  /*
    Put the pixmap into the background of the window
  */
  Status=XSetWindowBackgroundPixmap(GFXDisp,Win,BGPixmap.ID);
  /*
    Resize the window if not fresh
  */
  if (!FreshWin) XResizeWindow(GFXDisp,Win,Width,Height);
  /*
    Everything went OK (???)
  */
  return(0);
}

bool ScrCheckModeSet(void)
{
  return(BGPixmap.Width>0);
}

static int16 ScrSetupMode(const TSetModeCommand *C)
{
  uint16 Result;

  /*
    Setup the video mode
  */
  Result=SetVideoMode(C->Width,C->Height);
  return(Result);
}

static void DisplayTheWindow(void)
{
  XMapWindow(GFXDisp, Win);
  XSync(GFXDisp, false);
}

int16 ScrSetupScreenMode(const TSetModeCommand *C)
{
  int16 Result;

  /*
    If the screen driver has not been initialized, do so. It is not desirable
    to initialize the driver before any attempts to screen mode setting
    are done, because this would prevent console-only application from being
    able to execute without X.
  */
  if (ScreenLock==NULL) {
    if (!InitializeDriver()) return(screSCRINIT);
  }
  /*
    Synchronize before the screen mode set. This is required when there are
    some small scale screen updates still pending. Trying to tinker with the
    window when there are unprocessed requests may hurt.
  */
  XSync(GFXDisp,false);
  /*
    Execute the screen mode set command.
  */
  ThrRawLock(ScreenLock);
  Result=ScrSetupMode(C);
  ThrRawUnlock(ScreenLock);
  /*
    Place the content of the video memory (an undefined pool of garbage) onto
    the screen. Without this strange artifacts (more video memory seems to
    be updated that really is at program startup) could occur during screen
    updates, which are never perceived on a real screen.
  */
  if (Result>=0) {
    DisplayTheWindow();
  }
  return(Result);
}

void ScrGetCurrentColorFormat(TColorFormat *F)
{
  MemCpy(F,&SF,sizeof(SF));
}

typedef struct {
  XImage *I;
  XImage *MI;
  GC GC;
} TAuxWriteInfo;

int16 ScrPrepareAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  int16 Width;
  XImage *Image;
  void *VidBuffer;
  XImage *MaskImage;
  void *MaskBuffer;
  TAuxWriteInfo *A;
  TPixmap *P;
  uint16 LinesOfMask;
  GC MaskTransferGC;
  int16 Result;

  /*
    Extract input data from the structure.
  */
  Width=I->PixelsOnLine;
  P=I->DestPixmap;
  /*
    Create the temporary auxiliary writing structure.
  */
  A=KMEM__malloc(sizeof(*A));
  if (A==NULL) return(beNOMEM);
  /*
    Create the image for use with the drawing primitives.
  */
  do {
#if __b__USE_SHM__
    if (UseSHM) {
      Image=XShmCreateImage(
        GFXDisp, Vis,
        ValidDepth, ZPixmap,
        SHMImgInfo.shmaddr, &SHMImgInfo,
        Width,2
      );
      break;
    }
#endif
    VidBuffer=KMEM__hugemalloc(GCWCFG_TRANSFER_BUFFER_SIZE);
    if (VidBuffer==NULL) return(beNOMEM);
    Image=XCreateImage(
      GFXDisp, Vis,
      ValidDepth, ZPixmap, 0,
      (char *)VidBuffer,
      Width, 2,
      32, 0
    );
    if (Image==NULL) KMEM__free(VidBuffer);
  } while (false);
  if (Image==NULL) return(beNOMEM);
  /*
    if the pixmap is a masked one, expose the mask buffer as well.
  */
  MaskImage=NULL;
  if (P->MaskGC!=0) {
    do {
      /*
        Create a graphics context for the mask pixmap. Mask pixmaps need
	different graphics contexts than normal ones.
      */
      Result=beIO;
      MaskTransferGC=XCreateGC(GFXDisp,P->MaskID,0,NULL);
      if (MaskTransferGC==0) break;
#if __b__USE_SHM__
      if (UseSHM) {
        MaskImage=XShmCreateImage(
          GFXDisp, Vis,
          1, XYPixmap,
          SHMMaskInfo.shmaddr, &SHMMaskInfo,
          Width,2
        );
        break;
      }
#endif
      Result=beNOMEM;
      MaskBuffer=KMEM__hugemalloc(GCWCFG_TRANSFER_BUFFER_SIZE);
      if (MaskBuffer==NULL) return(beNOMEM);
      MaskImage=XCreateImage(
        GFXDisp, Vis,
        1, XYPixmap, 0,
        (char *)MaskBuffer,
        Width, 2,
        8, 0
      );
      if (MaskImage==NULL) KMEM__free(MaskBuffer);
    } while (false);
    if (MaskImage==NULL) {
      if (MaskTransferGC!=0) XFreeGC(GFXDisp,MaskTransferGC);
      XDestroyImage(Image);
      return(Result);
    }
  }
  /*
    Adjust the image height to fill the whole buffer.
  */
  Image->height=GCWCFG_TRANSFER_BUFFER_SIZE/Image->bytes_per_line;
  if (P->MaskGC!=0) {
    LinesOfMask=(GCWCFG_TRANSFER_BUFFER_SIZE/8)/MaskImage->bytes_per_line;
    if (Image->height>=LinesOfMask) Image->height=LinesOfMask;
    if (LinesOfMask>=Image->height) LinesOfMask=Image->height;
    MaskImage->height=LinesOfMask;
  }
  /*
    Export the output data by writing it to the passed-in structure.
  */
  A->I=Image;
  A->MI=MaskImage;
  A->GC=MaskTransferGC;
  I->OutAuxInfo=A;
  /*
    Return success.
  */
  return(0);
}

int16 ScrExposeBuffer(TColorConversionInfo *I)
{
  uint16 BytesPerLine;
  int16 PixelSize;
  TAuxWriteInfo *A;
  XImage *Image;
  TPixmap *P;

  A=I->OutAuxInfo;
  Image=A->I;
  PixelSize=I->Out.BitsPerPixel / 8;
  I->OutPtr=(void *)Image->data;
  I->FreePixels=(Image->width * Image->height);
  P=I->DestPixmap;
  BytesPerLine=P->Width*PixelSize;
  I->LinePaddingBytes=Image->bytes_per_line-BytesPerLine;
  Image=A->MI;
  if (Image!=NULL) {
    I->MaskPtr=(void *)Image->data;
    BytesPerLine=(Image->width+7)/8;
    I->MaskPaddingBytes=Image->bytes_per_line-BytesPerLine;
  }
  return(0);
}

static void TransferLines(
  XImage *Image,
  uint16 Y,
  uint16 Width,
  uint16 Height,
  GC Context,
  Pixmap DestPixmap
)
{
#if __b__USE_SHM__
  TX11Status Status;
#endif

  ThrRawLock(ScreenLock);
  do {
#if __b__USE_SHM__
    if (UseSHM) {
      Status=XShmPutImage(
        GFXDisp,DestPixmap,Context,
        Image,0,0,0,Y,Width,Height,
        false
      );
      break;
    }
#endif
    XPutImage(
      GFXDisp,DestPixmap,Context,
      Image,0,0,0,Y,Width,Height
    );
  } while (false);
  ThrRawUnlock(ScreenLock);
}

void ScrUpdateRectangle(uint16 X,uint16 Y,uint16 Width,uint16 Height)
{
  ThrRawLock(ScreenLock);
  XClearArea(GFXDisp,Win,X,Y,Width,Height,false);
  XSync(GFXDisp,false);
  ThrRawUnlock(ScreenLock);
}

int16 ScrFlushPixels(TColorConversionInfo *I)
{
  int16 StartLocY;
  int16 CurLocY;
  int16 LineCount;
  TAuxWriteInfo *A;
  XImage *Image;
  XImage *MaskImage;
  TPixmap *P;

  assert(I->UnprocessedPixels);
  StartLocY=I->SLine;
  CurLocY=I->LineNumber;
  LineCount=CurLocY-StartLocY;
  A=I->OutAuxInfo;
  Image=A->I;
  P=I->DestPixmap;
  TransferLines(Image,StartLocY,P->Width,LineCount,GraphicsContext,P->ID);
  MaskImage=A->MI;
  if (MaskImage!=NULL) {
    TransferLines(MaskImage,StartLocY,P->Width,LineCount,A->GC,P->MaskID);
  }
#if __b__USE_SHM__
  if (UseSHM) XSync(GFXDisp,false);
#endif
  return(0);
}

int16 ScrDestroyAuxiliaryTransferInfo(TColorConversionInfo *I)
{
  TAuxWriteInfo *A;
  XImage *Image;
  TPixmap *P;
  GC MaskGC;
  XGCValues V;

  A=I->OutAuxInfo;
  Image=A->I;
  XDestroyImage(Image);
  if (A->MI!=NULL) {
    Image=A->MI;
    XDestroyImage(Image);
    XFreeGC(GFXDisp,A->GC);
    P=I->DestPixmap;
    V.clip_mask=P->MaskID;
    MaskGC=XCreateGC(GFXDisp,Win,GCClipMask,&V);
    XSetGraphicsExposures(GFXDisp, MaskGC, false);
    P->MaskGC=MaskGC;
    if (MaskGC==0) {
      XFreePixmap(GFXDisp,P->ID);
      return(beIO);
    }
  }
  KMEM__free(A);
  return(0);
}

void *ScrGetScreenPixmap(void)
{
  return(&BGPixmap);
}

int16 ScrCreatePixmap(uint16 Width,uint16 Height,void **Result)
{
  TPixmap *P;

  P=KMEM__malloc(sizeof(*P));
  if (P==NULL) return(beNOMEM);
  ThrRawLock(ScreenLock);
  P->ID=XCreatePixmap(GFXDisp,Win,Width,Height,ValidDepth);
  P->Width=Width;
  P->MaskGC=0;
  ThrRawUnlock(ScreenLock);
  *Result=P;
  return(0);
}

int16 ScrCreateMaskedPixmap(uint16 Width,uint16 Height,void **Result)
{
  TPixmap *P;
  Pixmap MaskID;

  MaskID=XCreatePixmap(GFXDisp,Win,Width,Height,1);
  P=KMEM__malloc(sizeof(*P));
  P->ID=XCreatePixmap(GFXDisp,Win,Width,Height,ValidDepth);
  P->Width=Width;
  P->MaskID=MaskID;
  P->MaskGC=GraphicsContext;
  *Result=P;
  return(0);
}

int16 ScrTransferRectangle(
  void *SrcPixmap,
  uint16 SrcX,
  uint16 SrcY,
  uint16 Width,
  uint16 Height,
  void *DstPixmap,
  uint16 DstX,
  uint16 DstY
) {
  TPixmap *Src;
  TPixmap *Dst;
  GC TransferGC;

  Src=SrcPixmap;
  Dst=DstPixmap;
  TransferGC=GraphicsContext;
  if (Src->MaskGC!=0) {
    TransferGC=Src->MaskGC;
    XSetClipOrigin(GFXDisp,TransferGC,DstX,DstY);
  }
  XCopyArea(
    GFXDisp,
    Src->ID,Dst->ID,
    TransferGC,
    SrcX,SrcY,Width,Height,DstX,DstY
  );
  return(0);
}

void ScrDestroyPixmap(void *Pixmap)
{
  TPixmap *P;

  P=Pixmap;
  XFreePixmap(GFXDisp,P->ID);
  if (P->MaskGC!=0) {
    XFreeGC(GFXDisp,P->MaskGC);
    XFreePixmap(GFXDisp,P->MaskID);
  }
  KMEM__free(P);
}
