#include "defs.gh"

#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "types.h"
#include "traps.h"
#include "errno.h"
#include "memory.h"
#include "strings.h"
#include "kernel.h"
#include "pdfsys.h"
#include "pdbase.h"
#include "pdinet.h"
#include "inettype.h"
#if __b__ACCEPT_THREAD__
#include "thread.h"
#include "thrmain.h"
#endif

#define __INTERFACE__
#include "inetlin.h"
#undef __INTERFACE__

#include "typesnd.h"

typedef struct {
  int16 Handle;
#if __b__ACCEPT_THREAD__
  void *Cond;
  void *Lock;
#endif
} TPDSocket;

void *AllocSocketInfo(void)
{
  TPDSocket *Socket;

  Socket=KMEM__malloc(sizeof(TPDSocket));
#if __b__ACCEPT_THREAD__
  if (Socket!=NULL) {
    Socket->Lock=ThrRawCreateLock();
    Socket->Cond=ThrRawCreateCond();
    if (Socket->Lock==NULL || Socket->Cond==NULL) {
      if (Socket->Lock!=NULL) ThrRawDestroyLock(Socket->Lock);
      if (Socket->Cond!=NULL) ThrRawDestroyCond(Socket->Cond);
      KMEM__free(Socket);
      Socket=NULL;
    }
  }
#endif
  return(Socket);
}

void DestroySocketInfo(void *Ptr)
{
#if __b__ACCEPT_THREAD__
  TPDSocket *Socket;

  Socket=Ptr;
  ThrRawDestroyLock(Socket->Lock);
  ThrRawDestroyCond(Socket->Cond);
#else
  if (Ptr==NULL);
#endif
}

static uint8 ReadByte(char **Ptr)
{
  uint16 Result;
  char Ch;
  bool Continue;

  assert((**Ptr)!='.' && (**Ptr)!='\0');
  Result=0;
  do {
    Ch=*((*Ptr)++);
    Continue=Ch!='.' && Ch!='\0';
    if (Continue) {
      assert(Ch>='0' && Ch<='9');
      Result*=10;
      Result+=Ch-'0';
      assert(Result<256);
    }
  } while (Continue);
  return(Result);
}

/*** Convert binary IPv4 adress to OSHS binary adress ***/
/*
  Convers the platform specific binary adress to the OSHS binary adress. OSHS
  specifies certain order of bytes in the adress for both, IPv4 and IPv6
  addresses. This format is required to be platform and network independent.
*/
static void BinIPv4ToOSHSAddr(const void *Addr,THostAddr *Result)
{
  char *Buf;
  uint16 Number;
  int16 i;

  /*
    Convert the adress to the dotted representation
  */
  Buf=inet_ntoa(*(struct in_addr *)Addr);
  /*
    Read the representation and convert it to OSHS bytes. Beware that it is
    not possible to combine the below statements into one single expression.
  */
  Number=ReadByte(&Buf)*256;
  Number+=ReadByte(&Buf);
  (*Result)[6]=Number;
  Number=ReadByte(&Buf)*256;
  Number+=ReadByte(&Buf);
  (*Result)[7]=Number;
  /*
    Fill the rest of the adress with standard values for IPv4 adress embedded
    in IPv6 adress.
  */
  for (i=0;i<5;i++) (*Result)[i]=0;
  (*Result)[5]=0xFFFF;
}

static void WriteByte(uint8 Byte,char **Ptr)
{
  char Buf[3];
  char *B;
  int16 i;

  B=Buf+3;
  i=0;
  do {
    *(--B)=Byte%10 + '0';
    Byte/=10;
    i++;
  } while (Byte>0);
  for (;i>0;i--) *((*Ptr)++)=*(B++);
}

/*** Convert OSHS binary adress to binary IPv4 adress ***/
static void OSHSAddrToBinIPv4(void *Addr,const THostAddr *Host)
{
  char Buf[16];
  char *Ptr;
  uint16 Number;

  /*
    Convert the adress to the dotted representation
  */
  Ptr=Buf;
  Number=(*Host)[6];
  WriteByte(Number/256,&Ptr);
  *(Ptr++)='.';
  WriteByte(Number%256,&Ptr);
  *(Ptr++)='.';
  Number=(*Host)[7];
  WriteByte(Number/256,&Ptr);
  *(Ptr++)='.';
  WriteByte(Number%256,&Ptr);
  *Ptr='\0';
  /*
    From the dotted representation convert the address to the platform
    format.
  */
  assert(inet_aton(Buf,(struct in_addr *)Addr));
}

static int16 TranslateAfterSocket()
{
  switch (errno) {
    case EPROTONOSUPPORT: return(inetePROTOSUP);
    case EAFNOSUPPORT: return(ineteAFSUP);
    case ENFILE: return(beNOMEM);
    case EMFILE: return(ineteTOOMANY);
    case EACCES: return(ineteSACCESS);
#if __SPECIAL__ ENOBUFS != ENOMEM
    case ENOBUFS:
#endif
    case ENOMEM: return(beNOMEM);
    case EINVAL: return(beINVAL);
    default: return(beUNKERR);
  }
}

typedef struct {
  TPDSocket *Server;
  int16 Socket;
  struct sockaddr_in ClientInfo;
  socklen_t AddrLen;
} TServerListenInfo;

static int16 ServerReader(void *Data)
{
  TServerListenInfo *I;
  struct sockaddr *ClientInfoPtr;

  I=Data;
  ClientInfoPtr=(struct sockaddr *)&I->ClientInfo;
  I->Socket=accept(I->Server->Handle,ClientInfoPtr,&I->AddrLen);
  if (I->Socket==-1) I->Socket=TranslateAfterSocket();
#if __b__ACCEPT_THREAD__
  ThrRawSetCond(I->Server->Cond);
#endif
  return(0);
}

int16 InetServerSocketRead(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  TServerListenInfo I;
#if __b__ACCEPT_THREAD__
  bool Valid;
  void *Thread;
#endif

  I.Server=Client->ServerSocket;
  I.AddrLen=sizeof(I.ClientInfo);
#if __b__ACCEPT_THREAD__
  ThrRawLock(I.Server->Lock);
  ThrRawResetCond(I.Server->Cond);
  I.Socket=ineteCERROR;
  Thread=NULL;
  Valid=(I.Server->Handle>=0);
  if (Valid) Thread=ThrRawCreate(ServerReader,&I);
  ThrRawUnlock(I.Server->Lock);
  if (Thread!=NULL) ThrRawWaitCond(I.Server->Cond);
  ThrRawLock(I.Server->Lock);
  if (I.Server->Handle<0) ThrRawKill(Thread);
  ThrRawUnlock(I.Server->Lock);
#else
  ServerReader(&I);
#endif
  if (I.Socket<0) return(I.Socket);
  SockPtr=Client->Socket;
  SockPtr->Handle=I.Socket;

  BinIPv4ToOSHSAddr(&I.ClientInfo.sin_addr,Client->Host);
  *Client->Port=ntohs(I.ClientInfo.sin_port);
  return(0);
}

int16 InetOpenTCPClient(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  bool IPv4;
  struct sockaddr_in Info;
  int16 Result;
  int16 Socket;

  /*
    Check if the address is an IPv4 one
  */
  IPv4=InetCheckIsIPv4(Client->Host);
  if (IPv4) {
    /*
      Translate the address to the platform specific one and fill the Info
      structure to connect to the remote host.
    */
    Info.sin_family = AF_INET;
    Info.sin_port = htons(*Client->Port);
    OSHSAddrToBinIPv4(&Info.sin_addr,(void *)Client->Host);
    /*
      Create the socket
    */
    Socket=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (Socket==-1) return(TranslateAfterSocket());
    /*
      Attempt to connect to the host
    */
    if (connect(Socket,(struct sockaddr *)&Info,sizeof(Info))==-1) {
      switch (errno) {
        case ECONNREFUSED: Result=ineteREFUSED; break;
        case ETIMEDOUT: Result=ineteTIMEOUT; break;
        case ENETUNREACH: Result=ineteUNREACH; break;
        case EADDRINUSE: Result=ineteADDRUSE; break;
#if __SPECIAL__ EACCESS != EPERM
        case EPERM:
#endif
        case EACCES: Result=ineteCACCESS; break;
        default: Result=ineteCERROR; break;
      }
      close(Socket);
      return(Result);
    }
  } else {
    return(inetePROTOSUP);
  }
  SockPtr=Client->Socket;
  SockPtr->Handle=Socket;
  return(0);
}

static int16 InetOpenTCPServerCore(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  int16 Result;
  struct sockaddr_in Info;
  int16 Socket;
  int16 Temp;

  /*
    Fill in the address information structure
  */
  Info.sin_family = AF_INET;
  Info.sin_port = htons(*Client->Port);
  Info.sin_addr.s_addr = INADDR_ANY;
  /*
    Create the socket
  */
  Socket=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (Socket==-1) return(TranslateAfterSocket());
  /*
    Setup the socket to avoid the stupid 1 minute delay of the adress usage
    after the program termination.
  */
  Temp=1;
  if (setsockopt(Socket,SOL_SOCKET,SO_REUSEADDR,&Temp,sizeof(int16)) == -1) {
    close(Socket);
    return(ineteSSETUP);
  }
  /*
    Bind the socket to the address
  */
  if (bind(Socket, (struct sockaddr *)&Info, sizeof(Info)) == -1) {
    switch (errno) {
      case EACCES: Result=ineteADDRACC; break;
      default: Result=beUNKERR; break;
    }
    close(Socket);
    return(Result);
  }
  /*
    Create the connection acceptation queue
  */
  if (listen(Socket,Client->QueueSize)==-1) {
    switch (errno) {
      case EADDRINUSE: Result=ineteADDRUSE; break;
      default: Result=beUNKERR; break;
    }
    close(Socket);
    return(Result);
  }
  /*
    Fill in the socket structure with the socket and return success
  */
  SockPtr=Client->Socket;
  SockPtr->Handle=Socket;
  return(0);
}

int16 InetOpenTCPServer(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  int16 Result;

  SockPtr=Client->Socket;
  Result=InetOpenTCPServerCore(Client);
  return(Result);
}

int32 InetSocketRead(void *Socket,void *Buf,uint16 Size)
{
  TPDSocket *SockPtr;
  int32 Result;

  SockPtr=Socket;
  PrepareSystemCall();
  Result=recv(SockPtr->Handle,Buf,Size,0);
  if (Result==-1) Result=ReadAndTranslateERRNO();
  return(Result);
}

int32 InetSocketWrite(void *Socket,const void *Buf,uint16 Size)
{
  TPDSocket *SockPtr;
  int32 Result;

  SockPtr=Socket;
  PrepareSystemCall();
  Result=send(SockPtr->Handle,Buf,Size,0);
  if (Result==-1) Result=ReadAndTranslateERRNO();
  return(Result);
}

int16 InetSocketClose(void *Socket)
{
  TPDSocket *SockPtr;
  int16 Result;

  SockPtr=Socket;
  PrepareSystemCall();
  Result=close(SockPtr->Handle);
  if (Result==-1) Result=ReadAndTranslateERRNO();
#if __b__ACCEPT_THREAD__
  ThrRawLock(SockPtr->Lock);
  SockPtr->Handle=-1;
  ThrRawSetCond(SockPtr->Cond);
  ThrRawUnlock(SockPtr->Lock);
#endif
  return(Result);
}

int16 InetResolveName(const char *Name,TResolverData *Res)
{
  struct hostent *H;
  uint16 i;
  char *Ptr;

  /*
    Attempt to resolve the name using the Linux specific call. If an error is
    returned, translate it to the error code specified by the INET
    communication protocol.
  */
  H=gethostbyname(Name);
  if (H==NULL) {
    switch (h_errno) {
      case HOST_NOT_FOUND: return(ineteNOHOST);
#if __SPECIAL__ NO_DATA != NO_ADDRESS
      case NO_DATA:
#endif
      case NO_ADDRESS: return(ineteNOADDR);
      case NO_RECOVERY: return(ineteUNSERR);
      case TRY_AGAIN: return(ineteTMPERROR);
      default: return(ineteNRERR);
    }
  }
  /*
    If no adresses were returned, return ineterrNOADDR error
  */
  if (H->h_addr_list[0]==NULL) return(ineteNOADDR);
  /*
    IPv6 is still not supported
  */
  if (H->h_length==16) return(inetePROTOSUP);
  /*
    If the addresses have invalid format, return error. The possible formats
    are 4 (for IPv4 addresses) or 16 (for IPv6 addresses).
  */
  if (H->h_length!=4 && H->h_length!=16) return(ineteADDRFMT);
  /*
    Read and prepare the remaining addresses
  */
  i=0;
  for (;;) {
    Ptr=H->h_addr_list[i++];
    if (Ptr==NULL) break;
    if (Res->PendingCnt==MAX_PENDING_ADRESSES) return(ineteMADDRS);
    BinIPv4ToOSHSAddr(Ptr,Res->Data+Res->PendingCnt);
    Res->PendingCnt++;
  }
  return(0);
}
