#include "defs.gh"

#include <windows.h>
#include <stdlib.h>

#include "types.h"
#include "traps.h"
#include "kernel.h"
#include "pdbase.h"
#include "pdinet.h"
#include "memory.h"
#include "inettype.h"
#include "strings.h"

#define __INTERFACE__
#include "inetwin.h"
#undef __INTERFACE__

#include "typesnd.h"

typedef struct {
  SOCKET Handle;
#if __b__WINSOCK_HACK__
  bool WasFirstIgnoreChar;
#endif
} TPDSocket;

static void Cleanup(void)
{
  WSACleanup();
}

static void InitWinsock(void)
{
  static bool WinsockInitialized=false;
  WSADATA Info;

  if (!WinsockInitialized) {
    if (WSAStartup(MAKEWORD(1,1),&Info)) {
      Fatal("Unable to initialize the Windows Socket API");
    }
    if (atexit(Cleanup)) {
      Fatal("Unable to schedule Windows Socket Cleanup at program exit");
    }
    /*
    if ( (LOBYTE(Info.wVersion)!=1) && (HIBYTE(Info.wVersion)!=1) ) {
      WSACleanup();
      return(NULL);
    }
    */
    WinsockInitialized=true;
  }
}

void *AllocSocketInfo(void)
{
  TPDSocket *Ptr;

  InitWinsock();
  Ptr=KMEM__malloc(sizeof(TPDSocket));
#if __b__WINSOCK_HACK__
  Ptr->WasFirstIgnoreChar=false;
#endif
  return(Ptr);
}

void DestroySocketInfo(void *Ptr)
{
  if (Ptr==NULL);
}

static uint8 ReadByte(char **Ptr)
{
  uint16 Result;
  char Ch;
  bool Continue;

  assert((**Ptr)!='.' && (**Ptr)!='\0');
  Result=0;
  do {
    Ch=*((*Ptr)++);
    Continue=Ch!='.' && Ch!='\0';
    if (Continue) {
      assert(Ch>='0' && Ch<='9');
      Result*=10;
      Result+=Ch-'0';
      assert(Result<256);
    }
  } while (Continue);
  return(Result);
}

/*** Convert binary IPv4 adress to OSHS binary adress ***/
/*
  Convers the platform specific binary adress to the OSHS binary adress. OSHS
  specifies certain order of bytes in the adress for both, IPv4 and IPv6
  addresses. This format is required to be platform and network independent.
*/
static void BinIPv4ToOSHSAddr(const void *Addr,THostAddr *Result)
{
  char *Buf;
  uint16 Number;
  int16 i;

  /*
    Convert the adress to the dotted representation
  */
  Buf=inet_ntoa(*(struct in_addr *)Addr);
  assert(Buf!=NULL);
  /*
    Read the representation and convert it to OSHS bytes. Beware that it is
    not possible to combine the below statements into one single expression.
  */
  Number=ReadByte(&Buf)*256;
  Number+=ReadByte(&Buf);
  (*Result)[6]=Number;
  Number=ReadByte(&Buf)*256;
  Number+=ReadByte(&Buf);
  (*Result)[7]=Number;
  /*
    Fill the rest of the adress with standard values for IPv4 adress embedded
    in IPv6 adress.
  */
  for (i=0;i<5;i++) (*Result)[i]=0;
  (*Result)[5]=0xFFFF;
}

static void WriteByte(uint8 Byte,char **Ptr)
{
  char Buf[3];
  char *B;
  int16 i;

  B=Buf+3;
  i=0;
  do {
    *(--B)=Byte%10 + '0';
    Byte/=10;
    i++;
  } while (Byte>0);
  for (;i>0;i--) *((*Ptr)++)=*(B++);
}

/*** Convert OSHS binary adress to binary IPv4 adress ***/
static void OSHSAddrToBinIPv4(void *Addr,const THostAddr *Host)
{
  char Buf[16];
  char *Ptr;
  uint16 Number;
  uint32 Adr;

  /*
    Convert the adress to the dotted representation
  */
  Ptr=Buf;
  Number=(*Host)[6];
  WriteByte(Number/256,&Ptr);
  *(Ptr++)='.';
  WriteByte(Number%256,&Ptr);
  *(Ptr++)='.';
  Number=(*Host)[7];
  WriteByte(Number/256,&Ptr);
  *(Ptr++)='.';
  WriteByte(Number%256,&Ptr);
  *Ptr='\0';
  /*
    From the dotted representation convert the address to the platform
    format.
  */
  Adr=inet_addr(Buf);
  assert(Adr!=INADDR_NONE);
  *((upint32 *)Addr)=Adr;
}

static uint8 TranslateWinsockError()
{
  switch (WSAGetLastError()) {
    case WSAEPROTONOSUPPORT: return(inetePROTOSUP);
    case WSAEAFNOSUPPORT: return(ineteAFSUP);
    case WSAEMFILE: return(ineteTOOMANY);
    case WSAENOBUFS: return(beNOMEM);
    case WSAECONNREFUSED: return(ineteREFUSED);
    case WSAETIMEDOUT:  return(ineteTIMEOUT);
    case WSAENETUNREACH:  return(ineteUNREACH);
    case WSAEADDRINUSE:  return(ineteADDRUSE);
    case WSAEACCES:  return(ineteCACCESS);
    default: return(beUNKERR);
  }
}

#if __b__WINSOCK_HACK__
static void SendWinsockHackProlog(SOCKET Socket)
{
  uint16 Len;
  upint8 Buffer[256];

  Len=StrLen(WINSOCK_HACK_PROLOG_STR)+1;
  Buffer[0]=WINSOCK_HACK_TAG;
  MemCpy(Buffer+1,WINSOCK_HACK_PROLOG_STR,Len);
  Buffer[++Len]=FIRST_IGNORED_CHAR;
  Buffer[++Len]=SECOND_IGNORED_CHAR;
  send(Socket,Buffer,Len+1,0);
}
#endif

int16 InetServerSocketRead(TClientInfo *Client)
{
  TPDSocket *ServPtr;
  TPDSocket *SockPtr;
  SOCKET Socket;
  struct sockaddr_in ClientInfo;
  int16 AddrLen;

  ServPtr=Client->ServerSocket;
  AddrLen=sizeof(ClientInfo);
  Socket=accept(ServPtr->Handle,(struct sockaddr *)&ClientInfo,&AddrLen);
  if (Socket==INVALID_SOCKET) return(false);
#if __b__WINSOCK_HACK__
  SendWinsockHackProlog(Socket);
#endif
  SockPtr=Client->Socket;
  SockPtr->Handle=Socket;

  BinIPv4ToOSHSAddr(&ClientInfo.sin_addr,Client->Host);
  *Client->Port=ntohs(ClientInfo.sin_port);
  return(true);
}

int16 InetOpenTCPClient(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  bool IPv4;
  struct sockaddr_in Info;
  int16 Result;
  SOCKET Socket;

  /*
    Check if the address is an IPv4 one
  */
  IPv4=InetCheckIsIPv4(Client->Host);
  if (IPv4) {
    /*
      Translate the address to the platform specific one and fill the Info
      structure to connect to the remote host.
    */
    Info.sin_family = AF_INET;
    Info.sin_port = htons(*Client->Port);
    OSHSAddrToBinIPv4(&Info.sin_addr,(void *)Client->Host);
    /*
      Create the socket
    */
    Socket=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (Socket==INVALID_SOCKET) return(TranslateWinsockError());
    /*
      Attempt to connect to the host
    */
    if (connect(Socket,(struct sockaddr *)&Info,sizeof(Info))==SOCKET_ERROR) {
      Result=TranslateWinsockError();
      if (Result==beUNKERR) Result=ineteCERROR;
      closesocket(Socket);
      return(Result);
    }
#if __b__WINSOCK_HACK__
  SendWinsockHackProlog(Socket);
#endif
  } else {
    return(inetePROTOSUP);
  }
  SockPtr=Client->Socket;
  SockPtr->Handle=Socket;
  return(0);
}

int16 InetOpenTCPServer(TClientInfo *Client)
{
  TPDSocket *SockPtr;
  uint8 Result;
  struct sockaddr_in Info;
  SOCKET Socket;
  int16 CResult;

  /*
    Fill in the address information structure
  */
  Info.sin_family = AF_INET;
  Info.sin_port = htons(*Client->Port);
  Info.sin_addr.s_addr = INADDR_ANY;
  /*
    Create the socket
  */
  Socket=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (Socket==INVALID_SOCKET) return(TranslateWinsockError());
  /*
    Bind the socket to the address
  */
  CResult=bind(Socket, (struct sockaddr *)&Info, sizeof(Info));
  if (CResult == SOCKET_ERROR) {
    Result=TranslateWinsockError();
    closesocket(Socket);
    return(Result);
  }
  /*
    Create the connection acceptation queue
  */
  if (listen(Socket,Client->QueueSize)==SOCKET_ERROR) {
    Result=TranslateWinsockError();
    closesocket(Socket);
    return(Result);
  }
  /*
    Fill in the socket structure with the socket and return success
  */
  SockPtr=Client->Socket;
  SockPtr->Handle=Socket;
  return(0);
}

static uint16 ReadWinsockError(void)
{
  switch (WSAGetLastError()) {
    case WSAEACCES: return(ineteSACCESS);
    case WSAEFAULT: return(keFAULT);
    case WSAETIMEDOUT: return(beTIMEOUT);
    case WSAECONNRESET:
    case WSAECONNABORTED:
    case WSAENETRESET:
    case WSAENOTCONN: return(keCLOSED);
    case WSAENOBUFS: return(beNOMEM);
    default: return(beUNKERR);
  }
}

int32 InetSocketRead(void *Socket,void *Buf,uint16 Size)
{
  TPDSocket *SockPtr;
  int32 Result;
#if __b__WINSOCK_HACK__
  upint8 i[2];
#endif

  SockPtr=Socket;
  Result=recv(SockPtr->Handle,Buf,Size,0);
  if (Result==SOCKET_ERROR) Result=ReadWinsockError();
#if __b__WINSOCK_HACK__
  i[0]=FIRST_IGNORED_CHAR;
  i[1]=SECOND_IGNORED_CHAR;
  send(SockPtr->Handle,i,2,0);
#endif
  return(Result);
}

int32 InetSocketWrite(void *Socket,const void *Buf,uint16 Size)
{
  TPDSocket *SockPtr;
  int32 Result;
#if __b__WINSOCK_HACK__
  upint8 Buffer[131];
  const upint8 *InPtr;
  upint8 *Ptr;
  int32 BytesSent;
  uint8 Byte;
#endif

  SockPtr=Socket;
#if __b__WINSOCK_HACK__
  InPtr=Buf;
  Ptr=Buffer;
  BytesSent=0;
  while (Size>0) {
    Byte=*(InPtr++);
    Size--;
    if (SockPtr->WasFirstIgnoreChar) {
      if (Byte==SECOND_IGNORED_CHAR) {
        *(Ptr++)=FIRST_IGNORED_CHAR;
        *(Ptr++)=SECOND_IGNORED_CHAR;
      }
    }
    SockPtr->WasFirstIgnoreChar=(Byte==FIRST_IGNORED_CHAR);
    *(Ptr++)=Byte;
    if ((Size==0 || Ptr-Buffer>=128) && Ptr!=Buffer) {
      Result=send(SockPtr->Handle,Buffer,Ptr-Buffer,0);
      if (Result==SOCKET_ERROR) break;
      BytesSent+=Result;
      Ptr=Buffer;
    }
  }
  if (Result!=SOCKET_ERROR) Result=BytesSent;
#else
  Result=send(SockPtr->Handle,Buf,Size,0);
#endif
  if (Result==SOCKET_ERROR) Result=ReadWinsockError();
  return(Result);
}

int16 InetSocketClose(void *Socket)
{
  TPDSocket *SockPtr;
  int16 Result;

  SockPtr=Socket;
  Result=closesocket(SockPtr->Handle);
  if (Result==SOCKET_ERROR) Result=ReadWinsockError();
  return(Result);
}

uint8 InetResolveName(const char *Name,TResolverData *Res)
{
  struct hostent *H;
  uint16 i;
  char *Ptr;

  /*
    On Windows the socket API must eb initialized first. There is no such
    initialization on Linux.
  */
  InitWinsock();
  /*
    Attempt to resolve the name using the Linux specific call. If an error is
    returned, translate it to the error code specified by the INET
    communication protocol.
  */
  H=gethostbyname(Name);
  if (H==NULL) {
    switch (WSAGetLastError()) {
      case WSAHOST_NOT_FOUND: return(ineteNOHOST);
      case WSANO_DATA: return(ineteNOADDR);
      case WSANO_RECOVERY: return(ineteUNSERR);
      case WSATRY_AGAIN: return(ineteTMPERROR);
      default: return(ineteNRERR);
    }
  }
  /*
    If no adresses were returned, return ineteNOADDR error
  */
  if (H->h_addr_list[0]==NULL) return(ineteNOADDR);
  /*
    IPv6 is still not supported
  */
  if (H->h_length==16) return(inetePROTOSUP);
  /*
    If the addresses have invalid format, return error. The possible formats
    are 4 (for IPv4 addresses) or 16 (for IPv6 addresses).
  */
  if (H->h_length!=4 && H->h_length!=16) return(ineteADDRFMT);
  /*
    Read and prepare the remaining addresses
  */
  i=0;
  for (;;) {
    Ptr=H->h_addr_list[i++];
    if (Ptr==NULL) break;
    if (Res->PendingCnt==MAX_PENDING_ADRESSES) return(ineteMADDRS);
    BinIPv4ToOSHSAddr(Ptr,Res->Data+Res->PendingCnt);
    Res->PendingCnt++;
  }
  return(0);
}
