#ifndef __h__SYSLIB_KERNEL__
#define __h__SYSLIB_KERNEL__

#define hndSTDIN 0
#define hndSTDOUT 1
#define hndSTDERR 2
#define hndROOTFS 3
#define hndEXECDRV 4
#define hndDEVICES 5
#define hndCONSOLE 6

/*** The maximal size of the message ***/
#define MAX_MESSAGE_SIZE (1024+512)

/*** Kernel configuration "calls". ***/
/*
  They are declared as "calls" but in fact they usually are compile-in
  constants.

  All kernel configuration calls are prepended by two underscores and the
  "CFG_" string and they are terminated by two underscores as well. They
  usually take no arguments.
*/
/*
  The extension of the binaries used by the underlying platform. If the
  standard OSHS binary extension is used, this returns NULL.

  The returned extension includes the dot; effectivelyt it is the suffix to
  be added to the name of the file to produce the specification of the file
  used by the platform.

  This is used by the OSHS build system to tweak the up-to-date-ness checks
  so they can cooperate with the platform. This is useful for the development
  platforms.

  Unlike many other systems, OSHS uses a single extensions for both, programs
  and libraries.

  Only few platforms have this property. On other platforms regular ".bcf"
  files are used and programs are executed using the "emu" command which
  loads the SYSLIB library and then passes the binary to it to decompose to
  its portions and then execute the program portion (if one).
*/
const char *KCFG__BINARY_EXTENSION(void);
/*
  Returns "true" if the platform supports shared libraries and "false"
  otherwise. This information is used by OBS to decide whether to put library
  dependency info into the generated binaries.
*/
bool KCFG__SHARED_LIBRARIES_SUPPORTED(void);

/*** The paging info structure ***/
/*
  In this structure the kernel provides useful informations about the
  system enviroment the kernel is running on.
*/
typedef struct tagTSystemInfo {
  /*
    The size of one page, in bytes
  */
  upint16 PageSize;
  /*
    The size of page region, in pages
  */
  upint16 RegionSize;
  /*
    The size of the process address space, in page regions
  */
  upint32 ProcessSize;
  /*
    The size of the kernel reserved region of the process, in pages. Regions
    cannot be mapped into this space. This region is always at the start of
    the process address space.
  */
  upint32 ReservedSize;
  /*
    Properties. See `SysInfoFlags` for more informations.
  */
  upint16 Flags;
} TSystemInfo;

/*** SysInfoFlags: The system feature flags ***/
/*
  These flags show the boolean properties of the underlying hardware.
*/
/*
  If this flag is set, the data-only region is created at the beginning of
  the process adress space. If it is clear, the data-only region is created
  at the end of the process address space.
*/
#define sysDATA_AT_BEGINNING 1
/*
  If this flag is clear, the region reserved for the kernel is at the
  beginning of the address space. If this flag is set, this region is at the
  end.
*/
#define sysKERNEL_AT_END 2

/*** Errors: The error code list ***/
/*
*/
#define keBADHND -1   /* Invalid handle */
#define keNOMEM -2    /* Not enough physical memory for the operation */
#define keNOHND -3    /* No room for new handle */
#define keNOPBK -4    /* Handle is not a valid pageblock handle */
#define keNOPROC -5   /* Handle is not a valid process handle */
#define keNOTSUP -6   /* Feture not supported */
#define keCLOSED -7   /* Blocking call interrupted by handle being closed */
#define keMSGSIZE -8  /* Message has invalid size */
#define keRANGE -9    /* Value out of range */
#define keNOCHN -10   /* Handle is not a valid channel handle */
#define keFAULT -11   /* Fault occured while using user supplied buffer */
#define keMTHREAD -12 /* Too many threads */
#define keNOTHR -13   /* Handle is not a valid thread handle */

/*** System calls ***/
int16 KCALL__SYSINFO(TSystemInfo *Buffer);
int32 KCALL__MKCHAN(void);
int16 KCALL__SEND(int16 Channel,void *Message,int16 Size);
int16 KCALL__RECV(int16 Channel,void *Message,uint16 Size,upint32 *Sender);
int16 KCALL__GIVEHND(int16 Handle, int16 ProcHandle);
int16 KCALL__CLOSE(int16 Handle);
int16 KCALL__MKPROC(uint32 DataSize);
int16 KCALL__MAP(int16 Process, int16 Region, uint32 Location, uint8 Access);
int16 KCALL__MKREG(void);
int16 KCALL__FILL(uint32 FirstPage, uint32 CountOfPages, int16 FillerByte);
int16 KCALL__EMPTY(uint32 FirstPage, uint32 CountOfPages);
int32 KCALL__MAPREGS(int16 ProcessHandle);
uint32 KCALL__GETTID(void);
void KCALL__SUSPEND(uint64 MicroSeconds);
void KCALL__HALT(uint8 ExitCode);

typedef int16 TMainFunction(void *Data);

int16 KCALL__THRNEW(TMainFunction *MainFnc,void *Data,void *SP);
int16 KCALL__THRWAIT(int16 ThreadID);
int16 KCALL__THRKILL(int16 ThreadID);
int32 KCALL__LOCKNEW(void);
int16 KCALL__LOCK(int16 LockID);
int16 KCALL__UNLOCK(int16 LockID);
int16 KCALL__LOCKDEL(int16 LockID);
int32 KCALL__CONDNEW(void);
int16 KCALL__SET(int16 CondID);
int16 KCALL__RESET(int16 CondID);
int16 KCALL__WAIT(int16 CondID);
int16 KCALL__CONDDEL(int16 CondID);

#endif
