#include "defs.gh"
#include "types.h"
#include "kernel.h"
#include "memory.h"
#if __e__THREAD_DRV__ != __v__none__
#include "thread.h"
#include "thrmain.h"
#endif
#include "traps.h"
#include "strings.h"

#define __INTERFACE__
#include "kernmsg.h"
#undef __INTERFACE__

#define CHANNEL_BUFFER_SIZE 2000
#if __b__TIGHT_MEMORY__
#define MAX_HANDLES 1024
#else
#define MAX_HANDLES 32768
#endif

/*
  Some dumb procedures to be used with the channel initializer etc.
*/

int16 DumbClose(void *PrivData)
{
  if (PrivData==NULL);
  return(0);
}

void DumbDestroy(void *PrivData)
{
  if (PrivData==NULL);
}

/*
  The channel information structure.
*/

typedef struct tagTChannel {
  upint8 Buffer[CHANNEL_BUFFER_SIZE];
  int16 MsgSize;
  void *PrivateData;
  /*
    The handler of the channel. If set to NULL, the channel is not
    initialized yet and a ~InitializeChannel() call is required to be made
    with it before first use.
  */
  THandler *Handler;
  TPostHandler *PostHandler;
  TCloseProc *Closer;
  TDestroyProc *Destroyer;
#if __e__THREAD_DRV__ != __v__none__
  uint16 RefCount;
  void *Present;
  void *Absent;
  void *Close;
  bool Closed;
#endif
  bool MsgPresent;
} TChannel;
/*
  The Present and Absent are locks. When there is no message in the
  buffer, the Present lock is locked and Absent lock is unlocked.
  When the writer wants to write a message, it locks Absent first (so
  other writers must wait), writes the message into the buffer and then
  unlocks Present (so readers can get the message). The readers first lock
  Present, then reads the message and then unlocks Absent so a writer
  can write another message.
*/

#if __e__THREAD_DRV__ != __v__none__
static void *ChannelLock;
#endif
static void *Handles[MAX_HANDLES];
static uint16 HandleCnt;
static void **FirstFreeHandle;
static void *CurrentChannel;

/*** Initialize the kernel messaging subsystem ***/
void InitializeChannels(void)
{
#if __e__THREAD_DRV__ != __v__none__
  ChannelLock=ThrRawCreateLock();
  if (ChannelLock==NULL) Fatal("Unable to create channel list lock");
#endif
}

/*** Lock the list of channels ***/
void LockChannels(void)
{
#if __e__THREAD_DRV__ != __v__none__
  ThrRawLock(ChannelLock);
#endif
}

/*** Unlock the list of channels ***/
void UnlockChannels(void)
{
#if __e__THREAD_DRV__ != __v__none__
  ThrRawUnlock(ChannelLock);
#endif
  if (__DEBUG__) CurrentChannel=NULL;
}

/*** Get the channel for which the current handler is executing ***/
/*
  Returns the pointer to channel information structure which can be used
  to adjust the channel operation later.
*/
void *GetCurrentChannel(void)
/*
  NOTES

  - The pointer to the channel information structure is held in a static
    variable and is stored there after the channel list is locked. This means
    that if the handler wants to adjust the channel it is handling, it must
    retrieve this value before it unlocks the channel list. Using this call
    after unlocking the channel list will result in undefined behavior and
    if debugging is turned on, it is reported as a bug (via an assertion
    failure).
*/
{
  assert(CurrentChannel!=NULL);
  return(CurrentChannel);
}

/*** Free the handle ***/
static void FreeHandle(int16 Handle)
{
  assert(Handle>=0 && Handle<32768);
  Handles[Handle]=FirstFreeHandle;
  FirstFreeHandle=Handles+Handle;
}

/*** Creates the channel and returns a a pointer to it ***/
/*
  The nonlocking version of CreateChannel().
*/
int16 CreateChannelNoLock(
  void **Data
)
/*
  NOTES
  
  - This function is the same as ~CreateChannel. The only difference is that
    this function expects the channel list lock to be locked at entry and
    does not unlock the lock at exit.
*/
{
  TChannel *C;
  int16 Handle;

  assert(Data!=NULL);
  /*
    Allocate a handle for the channel.
  */
  if (FirstFreeHandle) {
    Handle=FirstFreeHandle-Handles;
    FirstFreeHandle=*FirstFreeHandle;
  } else {
    if (HandleCnt==MAX_HANDLES) return(keNOHND);
    Handle=HandleCnt++;
  }
  assert(Handle>=0 && Handle<32768);
  /*
    Allocate and initialize the buffer for the channel. On success connect
    the buffer to the handle (via the Handles array), on failure free the
    handle.
  */
  C=KMEM__malloc(sizeof(*C));
  if (C==NULL) {
    FreeHandle(Handle);
    return(keNOMEM);
  } else {
#if __e__THREAD_DRV__ != __v__none__
    C->Present=ThrRawCreateLock();
    C->Absent=ThrRawCreateLock();
    if (C->Present!=NULL && C->Absent!=NULL) {
      C->Close=ThrRawCreateLock();
    }
    if (C->Present==NULL || C->Absent==NULL || C->Close==NULL) {
      if (C->Present!=NULL) ThrRawDestroyLock(C->Present);
      if (C->Absent!=NULL) ThrRawDestroyLock(C->Absent);
      KMEM__free(C);
      FreeHandle(Handle);
      return(keNOMEM);
    }
    ThrRawLock(C->Present);
    C->RefCount=0;
    C->Closed=false;
#endif
    C->MsgPresent=false;
    C->Handler=NULL;
    C->PostHandler=NULL;
    Handles[Handle]=C;
    *Data=(void *)C;
    return(Handle);
  }
}

/*** Creates the channel and returns a handle and a pointer to it ***/
int16 CreateChannel(
  /*
    Pointer to a variable which receives the pointer to the channel
    information structure. This pointer must be passed to InitializeChannel()
    before the channel is used.
  */
  void **Data
  /*
    Returns positive handle for the new channel on success or a negative
    error code on failure.
  */
)
{
  int16 Result;

  LockChannels();
  Result=CreateChannelNoLock(Data);
  UnlockChannels();
  return(Result);
}

/*** Initialize a channel for real work ***/
/*
  This call initializes the channel and prepares it for use. It must be
  called exactly once per channel.
*/
void InitializeChannel(
  /*
    The channel to be initialized.
  */
  void *Channel,
  /*
    The message handler. This procedure will be called everytime a message
    comes into the buffer.
  */
  THandler *Handler,
  /*
    The private data structure. Once it is installed, the
    private data structure is owned by the channel and freed with it once
    the channel is not in use. The content of the PrivateData structure is
    not inspected nor touched by the library in any way. This argument may be
    NULL if the handler uses no private data structure.
  */
  void *PrivateData,
  /*
    The procedure used to initiate the channel destroying. This procedure
    must arrange things so that all threads blocked on the channel will be
    unblocked so they can notice the channel is no longer valid. This
    procedure must not deallocate anything.
  */
  TCloseProc *Closer,
  /*
    The procedure used to release all resources owned by the channel. The
    private data structure, which is owned by the channel buffer structure is
    not freed by this call. Other resources attached to the channel via the
    private data structure must be freed by this call. Also the purpose of
    this call is to wake up any threads waiting on the channel. It may be
    NULL if the channel has no special structures other than the private
    data structure allocated.
  */
  TDestroyProc *Destroyer
)
/*
  NOTES
  
  - This call is not allowed to fail.
*/
{
  TChannel *C;

  C=Channel;
  assert(Channel!=NULL);
  assert(!C->Handler);
  assert(Handler);
  assert(Closer);
  assert(Destroyer);
  C->Handler=Handler;
  C->Closer=Closer;
  C->Destroyer=Destroyer;
  C->PrivateData=PrivateData;
}

/*** Install a posthandler into a channel ***/
/*
*/
void InstallPostHandler(void *Channel,TPostHandler *Handler)
{
  TChannel *C;

  C=Channel;
  C->PostHandler=Handler;
}

/*** Get the currently installed posthandler from the channel ***/
void *GetPostHandler(void *Channel)
{
  TChannel *C;

  C=Channel;
  return(C->PostHandler);
}

static TChannel *HandleToPointer(int16 Handle)
{
  void *Ptr;
  TChannel *C;

  /*
    Check if the handle is valid. An obvious case is when the number is out
    of range, so check it first to get rid of it.
  */
  if ((Handle<0) || (Handle>=HandleCnt)) return(NULL);
  /*
    The non-obvious case is that the number is in range but is a previously
    invalidated handle. In that case a pointer to the next handle table slot
    of the free handles chain is stored at the place of the handle structure
    pointer. Here the fact that all such pointers point into the Handle array
    is exploited.
  */
  C=Handles[Handle];
  Ptr=C;
  if (Ptr>=(void *)Handles && Ptr<(void *)(Handles+MAX_HANDLES)) Ptr=NULL;
  /*
    If the channel is not initialized yet, the handle that is referring to it
    will be treated as invalid. Such a situation can happen for example when
    a thread is listening on an server port; the INET driver creates a
    channel before it enters the wait loop but initializes it after a
    connection is accepted, when another thread tries to use the handle of
    such a spare channel, it should see the "invalid handle" error.
  */
  if (C!=NULL) {
    if (C->Handler==NULL) C=NULL;
  }
  /*
    If the handle is being closed, it is treated as invalid. This must be
    checked here, because now any nonsense pointers are avoided (more
    precisely, converted to NULL). This check is valid only when threads are
    available.
  */
#if __e__THREAD_DRV__ != __v__none__
  if (C!=NULL) {
    if (C->Closed) C=NULL;
  }
#endif
  return(C);
}

#if __e__THREAD_DRV__ != __v__none__
static void HandleIncRef(TChannel *C)
{
  C->RefCount++;
  if (C->RefCount==1) ThrRawLock(C->Close);
}

static void HandleDecRef(TChannel *C)
{
  C->RefCount--;
  if (C->RefCount==0) ThrRawUnlock(C->Close);
}
#endif

int16 SendCore(int16 Channel,void *Message,int16 Size)
{
  TChannel *C;
  int16 Result;

  if (Size<0 || Size>MAX_MESSAGE_SIZE) return(keMSGSIZE);
  LockChannels();
  C=HandleToPointer(Channel);
  CurrentChannel=C;
  if (C==NULL) {
    UnlockChannels();
    return(keNOCHN);
  }
#if __e__THREAD_DRV__ != __v__none__
  HandleIncRef(C);
  ThrRawLock(C->Absent);
  if (C->Closed) {
    ThrRawUnlock(C->Absent);
    HandleDecRef(C);
    UnlockChannels();
    return(keCLOSED);
  }
#else
  if (C->MsgPresent) {
    ReportBug("Program crashes on infinite wait on a full pipe");
    LockChannels();
  }
#endif
  C->MsgSize=C->Handler(Message,Size,C->Buffer,C->PrivateData);
  C->MsgPresent=true;
  Result=0;
#if __e__THREAD_DRV__ != __v__none__
  ThrRawUnlock(C->Present);
  HandleDecRef(C);
  if (C->Closed) Result=keCLOSED;
#endif
  UnlockChannels();
  return(Result);
}

int16 RecvCore(int16 Channel,void *Message,uint16 BufSize,upint32 *Sender)
{
  TChannel *C;
  int16 Size;

  LockChannels();
  C=HandleToPointer(Channel);
  CurrentChannel=C;
  if (C==NULL) {
    UnlockChannels();
    return(keNOCHN);
  }
#if __e__THREAD_DRV__ != __v__none__
  HandleIncRef(C);
  ThrRawLock(C->Present);
  if (C->Closed) {
    ThrRawUnlock(C->Present);
    HandleDecRef(C);
    UnlockChannels();
    return(keCLOSED);
  }
#else
  if (!C->MsgPresent) {
    ReportBug("Program crashes on infinite wait on an empty pipe");
    LockChannels();
  }
#endif
  Size=C->MsgSize;
  assert(Size<=MAX_MESSAGE_SIZE);
  if (Size>0) {
    if (Size>BufSize) Size=BufSize;
    MemCpy(Message,C->Buffer,Size);
  }
  if (C->PostHandler) {
    C->MsgSize=C->PostHandler(C->Buffer,C->PrivateData);
#if __e__THREAD_DRV__ != __v__none__
    ThrRawUnlock(C->Present);
#endif
  } else {
    C->MsgPresent=false;
#if __e__THREAD_DRV__ != __v__none__
    ThrRawUnlock(C->Absent);
#endif
  }
#if __e__THREAD_DRV__ != __v__none__
  HandleDecRef(C);
  if (C->Closed) Size=keCLOSED;
#endif
  UnlockChannels();
  return(Size);
}

int16 CloseCoreNoLock(int16 Channel)
{
  TChannel *C;
  int16 Result;

  C=HandleToPointer(Channel);
  if (C==NULL) return(keBADHND);
  if (C->Handler) {
#if __e__THREAD_DRV__ != __v__none__
    C->Closed=true;
#endif
    Result=0;
    Result=C->Closer(C->PrivateData);
#if __e__THREAD_DRV__ != __v__none__
    UnlockChannels();
    ThrRawLock(C->Close);
    LockChannels();
    if (C->MsgPresent) {
      ThrRawUnlock(C->Absent);
    } else {
      ThrRawUnlock(C->Present);
    }
#endif
    if (C->PrivateData!=NULL) {
      C->Destroyer(C->PrivateData);
      KMEM__free(C->PrivateData);
    }
  } else {
    Result=keBADHND;
  }
  KMEM__free(C);
  FreeHandle(Channel);
  return(Result);
}

int16 CloseCore(int16 Channel)
{
  int16 Result;

  LockChannels();
  Result=CloseCoreNoLock(Channel);
  UnlockChannels();
  return(Result);
}
