#include "defs.gh"
#include "types.h"

/*
  Binary file subsystem main module.

  The binary file subsystem is used by the libtool to handle the binaries.
  It is responsible for these tasks:

  - Extract platform specific portions of a given OSHS binary into given
    directory. Two possible places are: the working directory, from which
    the data is picked up by the platform specific binary tools and the
    "memory" directory, which represents the pool of libraries which are
    ready to use. Also this allows one to specify, whether the static or the
    shared portion of the file should be extracted if the file is a library.
  - Construct platform specific portions of a given OSHS binary. This always
    takes the supplementary files from the working directory since this is
    the place to which the platform specific tools drop their results.
*/

#include "strings.h"
#include "conio.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "diskspec.h"
#include "disklow.h"
#include "diskdir.h"
#include "ltargs.h"
#include "ltwd.h"
#include "pdbase.h"
#include "pdexec.h"

#define __INTERFACE__
#include "ltbin.h"
#undef __INTERFACE__

static void ReportError(const char *Msg)
{
  PrintSTDERR(Msg);
  PrintSTDERR("\n");
}

bool ReadOBSAreaSize(const char *Arg)
/*
  OBS area is a region that is never touched or examined by SYSLIB thus it is
  free for whatever use the OBS has for it. Typically it will contain library
  dependency list so the library dependencies can be retrieved even when the
  cache file that contained them was destroyed. The size of this area is
  limited to 64K.

  The format of the switch that turns this feature on has somewhat strange.
  It uses a hexadecimal number but the digits are the capital letters from
  'A' to 'U' instead of the traditional hexadecimal digits ('0' to '9', then
  'A' to 'F').

  This procedure reads the OBS area size from the switch that was mentioned
  on the linker's commandline and sets the OBSAreaSize global variable to
  that value. The rest of the system (namely the call that assembles the
  resulting binary from the various pieces then picks the value up and
  creates the OBS area.
*/
{
  char Ch;

  if (OBSAreaSize>0) {
    ReportError("OBS area size can be specified only once");
    return(false);
  }
  while (true) {
    Ch=*(Arg++);
    if (Ch=='\0') break;
    if (Ch<'A' || Ch>'P') {
      ReportError("Invalid digit in OBS are size specification");
      return(false);
    }
    if (OBSAreaSize>0xFFF) {
      ReportError("Specified OBS area size is too large");
      return(false);
    }
    OBSAreaSize*=16;
    OBSAreaSize+=Ch-'A';
  }
  if (OBSAreaSize==0) {
    ReportError("OBS area size must be nonzero if specified");
    return(false);
  }
  return(true);
}

static int16 ObtainOutputFileName(const char *const* argv, char *Result)
{
  char *Ptr;
  char Ch;

  Result[MAX_FILESPEC_SIZE]='\0';
  Ptr=StrNCpy(Result,argv[0],MAX_FILESPEC_SIZE+1);
  if (Result[MAX_FILESPEC_SIZE]=='\0') {
    do {
      Ch=*(--Ptr);
    } while (Ch!='.' && Ptr>Result);
    if (Ch=='.') {
      if (Result+MAX_FILESPEC_SIZE-Ptr>4) {
        StrCpy(Ptr+1,"rmc");
      } else {
        Result[MAX_FILESPEC_SIZE]='0';
      }
    }
  }
  if (Result[MAX_FILESPEC_SIZE]!='\0') {
    PlainERR("Output file specification is too large");
    return(1);
  }
  return(0);
}

static void ReportFailure(int16 Error,const char *Action,const char *Msg)
{
  char Line[80];
  char *Ptr;

  Ptr=StrCpy(Line,"Error ");
  Ptr=StrCpy(Ptr,Action);
  Ptr=StrCpy(Ptr,"ing the ");
  Ptr=StrCpy(Ptr,Msg);
  StrCpy(Ptr," file");
  SysERR(Line);
}

static void ReportSeekFailure(int16 Result)
{
  if (Result<0) ReportFailure(Result,"seek","file");
}

static int16 SeekFile(int16 fd,int32 Where,bool Verbose)
{
  int16 Result;

  Result=LowLevelSeek(fd,Where,seekSET);
  assert(Result<=0);
  if (Verbose) ReportSeekFailure(Result);
  return(Result);
}

static int16 SeekFileFromEnd(int16 fd,int32 Where,bool Verbose)
{
  int16 Result;

  Result=LowLevelSeek(fd,-Where,seekEND);
  assert(Result<=0);
  if (Verbose) ReportSeekFailure(Result);
  return(Result);
}

static bool WriteResultFile(int16 fd,const void *Buffer,int16 Size)
{
  int16 Written;

  if (Size>0) {
    Written=LowLevelWrite(fd,Buffer,Size);
    if (Written!=Size) Written=-32768;
    if (Written<0) {
      ReportFailure(Written,"writ","result");
      return(false);
    }
  }
  return(true);
}

static void CloseFile(int16 File)
{
  int16 Result;

  Result=LowLevelClose(File);
  if (Result<0) ReportFailure(Result,"clos","file");
}

static int32 CopyFileContent(
  int16 Dest,
  int16 Source,
  uint32 Bytes,
  bool Verbose
)
{
  int16 TransferSize;
  int16 Size;
  int16 Written;
  uint32 Len;
  char Buffer[16384];

  Len=0;
  do {
    TransferSize=16384;
    if (TransferSize>Bytes) TransferSize=Bytes;
    Size=LowLevelRead(Source,Buffer,TransferSize);
    if (Size<0) {
      if (Verbose) ReportFailure(Size,"read","file");
      return(Size);
    }
    if (Size>0) {
      Written=LowLevelWrite(Dest,Buffer,Size);
      if (Written!=Size) Written=beUNKERR;
      if (Written<0) {
        if (Verbose) ReportFailure(Size,"writ","file");
        return(Written);
      }
    }
    Len+=Size;
    Bytes-=Size;
  } while ((Bytes>0) && (Size>0));
  return(Len);
}

typedef struct {
  char LName[8];
  upint32 Lengths[64];
} TFooter;

typedef struct {
  TFooter F;
  upint32 *Ptr;
} TFooterHandler;

static int16 CopyFileContents(
  int16 fd,
  const char *const* FileList,
  TFooterHandler *F
)
{
  const char *Ptr;
  int16 DataFile;
  int32 CopiedLen;

  do {
    Ptr=*(FileList++);
    if (Ptr==NULL) break;
    DataFile=LowLevelOpen(Ptr,oeOPEN|oRDONLY);
    if (DataFile<0) {
      ReportFailure(DataFile,"open","platform specific binary");
      return(1);
    }
    CopiedLen=CopyFileContent(fd,DataFile,0x7FFFFFFF,true);
    CloseFile(DataFile);
    if (CopiedLen<0) return(CopiedLen);
    *(F->Ptr++)=CopiedLen;
  } while (true);
  return(0);
}

static void CreateGPSRegion(int16 fd,TFooterHandler *F)
{
  upint8 Dummy[8192];
  uint16 Size;
  uint16 ToTransfer;

  Size=OBSAreaSize;
  *(F->Ptr++)=Size;
  while (Size>0) {
    ToTransfer=sizeof(Dummy);
    if (ToTransfer>Size) ToTransfer=Size;
    WriteResultFile(fd,Dummy,ToTransfer);
    Size-=ToTransfer;
  }
}

int16 AssembleResultingBinaryFile(
  /*
    The name of the output file.
  */
  const char *FileName,
  /*
    The list of files with the platform specific code terminated with NULL.
  */
  const char *const* FileList,
  /*
    The file signature to be used
  */
  const char *Signature,
  /*
    The name of the library
  */
  const char *LibName
)
/*
  This call assembles the resulting binary file from its pieces. The content
  of the output file is deliberately overwritten by the platform specific
  stuff. Then the footer of the file which defines the platform and locations
  of various pieces of the file is written. This arrangement was chosen to
  avoid breakage of platform specific mechanisms that handle binary files
  (for example under Linux the resulting library will act as a shared library
  despite the fact that a static version of the same library is embedded into
  it).
*/
{
  int16 Result;
  int16 fd;
  TFooterHandler F;
  uint32 Len;

  /*
    Initialize the footer handler
  */
  F.Ptr=F.F.Lengths;
  /*
    Fill in the library name
  */
  MemSet(F.F.LName,sizeof(F.F.LName),0);
  if (LibName!=NULL) StrNToUpr(F.F.LName,LibName,8);
  /*
    Try to remove the file before opening. This is needed on UNIX when OBS
    is trying to link itself to avoid "text file busy" errors. Errors can be
    safely ignored here as if there are any problems, they will pop up at the
    command that tries to create the file.
  */
  DirectRemove(FileName);
  /*
    Open the output file
  */
  fd=LowLevelOpen(FileName,oeTRUNC|onCREAT|oRDWR);
  if (fd<0) {
    ReportFailure(fd,"creat","result");
    return(1);
  }
  /*
    Copy the files from the file list.
  */
  Result=CopyFileContents(fd,FileList,&F);
  /*
    Create the GPS area region, which will be placed immediately before the
    
  */
  CreateGPSRegion(fd,&F);
  /*
    Write the footer of the file.
  */
  Len=((void *)F.Ptr)-((void *)&F.F)+4;
  *(F.Ptr)=Len;
  WriteResultFile(fd,Signature,4);
  WriteResultFile(fd,&F.F,Len);
  /*
    Finalize and close the file. If errors were detected, delete the created
    but broken resulting file.
  */
  CloseFile(fd);
  if (Result!=0) {
    fd=DirectRemove(FileName);
    if (fd!=0) ReportFailure(fd,"delet","result");
  }
  return(Result);
}

int16 CombineCompilationResults(
  /*
    The arguments passed to the compiler. These are used to obtain the name
    of the file for the result.
  */
  const char *const* argv,
  /*
    The list of files with the platform specific code terminated with NULL.
  */
  const char *const* FileList,
  /*
    The file signature to be used
  */
  const char *Signature
)
/*
  This call assembles the compiled object file from the given pieces. This
  is basically the same as the previous call except that the name of the
  resulting file is derived from the input file.
*/
{
  char FileName[MAX_FILESPEC_SIZE+1];
  int16 Result;

  /*
    Obtain the name of the file to be produced.
  */
  Result=ObtainOutputFileName(argv,FileName);
  if (Result!=0) return(Result);
  /*
    Produce the binary
  */
  return(AssembleResultingBinaryFile(FileName,FileList,Signature,NULL));
}

#define MAX_SIZES 16

static int32 FetchFooter(int16 FD,bool Verbose)
{
  int16 Result;
  upint32 Len;

  Result=SeekFileFromEnd(FD,4,Verbose);
  if (Result<0) return(Result);
  Result=LowLevelRead(FD,&Len,4);
  if (Result!=4) {
    if (Verbose) SysERR("Error reading the footer length");
    if (Result>0) Result=execeFORMAT;
    return(Result);
  }
  if (Len>1024) {
    if (Verbose) PlainERR("The footer length is awfully big");
    return(execeFORMAT);
  }
  return(Len);
}

int16 ExtractLibraryName(int16 FD,char *Buffer,bool Verbose)
{
  uint32 Len;
  int32 Result;

  Buffer[8]='\0';
  Len=FetchFooter(FD,Verbose);
  if (Len<0) return(Len);
  Result=SeekFileFromEnd(FD,Len,Verbose);
  if (Result<0) return(Result);
  Result=LowLevelRead(FD,Buffer,8);
  if (Result!=8) {
    if (Verbose) {
      SysERR("Error reading the library name from the library file");
    }
    if (Result>0) Result=execeFORMAT;
    return(Result);
  }
  return(0);
}

int32 ExtractContent(int16 InFD,int16 OutFD,int16 Number,bool Verbose)
{
  uint32 HdrSize;
  upint32 Sizes[MAX_SIZES];
  uint32 Pos;
  int32 Result;
  int16 i;

  HdrSize=FetchFooter(InFD,Verbose);
  if (HdrSize<0) return(HdrSize);
  if (HdrSize<4+8+(Number+1)*itemsize(Sizes)) {
    if (Verbose) PlainERR("The footer is too small");
    return(execeFORMAT);
  }
  Result=SeekFileFromEnd(InFD,HdrSize-8,Verbose);
  if (Result<0) return(Result);
  Result=LowLevelRead(InFD,Sizes,(Number+1)*itemsize(Sizes));
  if (Result<(Number+1)*itemsize(Sizes)) {
    if (Verbose) ReportFailure(Result,"read","file components sizes");
    if (Result>=0) Result=execeFORMAT;
    return(Result);
  }
  Pos=0;
  for (i=0;i<Number;i++) Pos+=Sizes[i];
  Result=SeekFile(InFD,Pos,Verbose);
  if (Result<0) return(Result);
  Pos=CopyFileContent(OutFD,InFD,Sizes[Number],Verbose);
  if (Pos!=Sizes[Number]) {
    if (Pos>=0) Pos=execeFORMAT;
  }
  return(Pos);
}

static void BuildNumber(uint16 Number,char *Ptr)
{
  int16 i;

  Ptr+=5;
  for (i=0;i<5;i++) {
    *(--Ptr)=Number%10+'0';
    Number/=10;
  }
}

static bool ExtractContentIO(const char *InF,const char *OutF,int16 PartNo)
{
  int16 InFile;
  int16 OutFile;

  InFile=LowLevelOpen(InF,oeOPEN|oRDONLY);
  if (InFile<0) {
    ReportFailure(InFile,"open","input");
    return(false);
  }
  OutFile=LowLevelOpen(OutF,oeTRUNC|onCREAT|oRDWR);
  if (OutFile<0) {
    ReportFailure(OutFile,"creat","output");
    CloseFile(InFile);
    return(false);
  }
  RegisterFileForDeletion(OutF);
  ExtractContent(InFile,OutFile,PartNo,true);
  CloseFile(InFile);
  CloseFile(OutFile);
  return(true);
}

bool ExtractNumberedContents(
  const char **Files,
  int16 PartNum,
  const char *Prefix,
  const char *Ext,
  TArgumentList *List
)
{
  const char *FileName;
  uint16 Number;
  char File[MAX_FILESPEC_SIZE+16];
  char *FilePtr;
  char *Result;

  FilePtr=PutInternalWorkDirSpec(File);
  FilePtr=StrCpy(FilePtr,Prefix);
  StrCpy(FilePtr+6,Ext);
  FilePtr[5]='.';
  Number=0;
  do {
    FileName=*(Files++);
    if (FileName!=NULL) {
      BuildNumber(Number,FilePtr);
      Result=AddArgument(List,File);
      if (Result!=NULL) {
        PlainERR(Result);
        return(false);
      }
      if (!ExtractContentIO(FileName,File,PartNum)) return(false);
      Number+=1;
    }
  } while (FileName!=NULL);
  return(true);
}

int16 ConvertAndOpenFile(TRootStatus *Root,const char *FileSpec,int16 Mode)
{
  char Buffer[MAX_FILESPEC_SIZE+1];

  if (ConvertFileSpec(Root,FileSpec,Buffer)==NULL) {
    return(Root->Result);
  }
  return(LowLevelOpen(Buffer,Mode));
}
