#include "defs.gh"
#include "types.h"
#include "memory.h"
#include "ltwd.h"
#include "strings.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "pdexec.h"
#include "diskdata.h"
#include "disklow.h"
#include "conio.h"

#define __INTERFACE__
#include "ltdeps.h"
#undef __INTERFACE__

/*******************************************
*** Fetching dependency data from a file ***
*******************************************/

static int32 FindDependencyLibraryList(int16 ProgFD)
{
  int16 Result;
  uint32 Sizes[2];
  int32 FilePos;
  int32 ReadSize;

  Result=LowLevelSeek(ProgFD,0,seekEND);
  if (Result<0) return(Result);
  FilePos=LowLevelTell(ProgFD);
  if (FilePos<0) {
    if (FilePos<-32768) FilePos=beUNKERR;
    return(FilePos);
  }
  FilePos-=8;
  Result=LowLevelSeek(ProgFD,FilePos,seekSET);
  if (Result<0) return(Result);
  ReadSize=LowLevelRead(ProgFD,Sizes,8);
  if (ReadSize!=8) {
    if (ReadSize<-32768 || ReadSize>=0) ReadSize=beUNKERR;
    return(ReadSize);
  }
  if (Sizes[0]>65535) return(execeFORMAT);
  FilePos-=Sizes[0];
  FilePos-=Sizes[1]-4;
  Result=LowLevelSeek(ProgFD,FilePos,seekSET);
  if (Result<0) return(Result);
  return(Sizes[0]);
}

int16 FetchDependencyInfo(int16 Handle,TDependencyInfo **Info)
{
  int32 Size;
  int32 Result;
  TDependencyInfo *I;

  Size=FindDependencyLibraryList(Handle);
  if (Size<0) {
    assert(Size>=-32768);
    return(Size);
  }
  I=KMEM__malloc(Size+2);
  if (I!=NULL) {
    I->Length=Size;
    Result=LowLevelRead(Handle,I->Info,Size);
    if (Result!=Size) {
      if (Result>=0) Result=beUNKERR;
      KMEM__free(I);
      I=NULL;
    } else {
      Result=0;
    }
  }
  *Info=I;
  return(Result);
}

/***********************************
*** The library dependency table ***
***********************************/
/*
  The design of this table is so complicated to avoid relying on 32-bit
  support while allowing having many libraries loaded. Some 16-bit platforms
  cannot handle single arrays larger than 64K.
*/

#define MAX_LOADED_LIBRARIES 8192

typedef struct tagTDependencyInfoSlot {
  upint8 LibName[8];
  TDependencyInfo *Info;
  struct tagTDependencyInfoSlot *Next;
} TDependencyInfoSlot;

#define INFOS_IN_BLOCK 128
#define INFO_BLOCK_COUNT (MAX_LOADED_LIBRARIES/INFOS_IN_BLOCK)

typedef TDependencyInfoSlot TDependencyInfoBlock[INFOS_IN_BLOCK];

typedef struct {
  uint16 Count;
  TDependencyInfoBlock *Blocks[INFO_BLOCK_COUNT];
} TDependencyInfoTable;

static TDependencyInfoTable *Table = NULL;
static bool DepTableLoaded = false;
static bool DepTableChanged = false;

static bool CreateDependencyInfoTableStructures(void)
{
  Table=KMEM__malloc(sizeof(*Table));
  if (Table!=NULL) {
    Table->Count=0;
  }
  return(Table!=NULL);
}

static TDependencyInfoSlot *GetSlot(uint16 Number)
{
  uint16 BlockNum;
  TDependencyInfoBlock *B;
  uint16 SlotNum;
  TDependencyInfoSlot *S;

  assert(Table!=NULL);
  assert(Number<Table->Count);
  BlockNum=Number/INFOS_IN_BLOCK;
  B=Table->Blocks[BlockNum];
  SlotNum=Number%INFOS_IN_BLOCK;
  S=&(*B)[SlotNum];
  return(S);
}

static TDependencyInfoSlot *CreateSlot(uint16 *Number)
{
  uint16 BlockNum;
  TDependencyInfoSlot *S;

  if (Table==NULL) {
    CreateDependencyInfoTableStructures();
    if (Table==NULL) return(NULL);
  }
  *Number=Table->Count;
  if ((Table->Count%INFOS_IN_BLOCK)==0) {
    BlockNum=*Number/INFOS_IN_BLOCK;
    Table->Blocks[BlockNum]=KMEM__malloc(sizeof(*Table->Blocks[0]));
  }
  Table->Count++;
  S=GetSlot(*Number);
  S->Info=NULL;
  S->Next=NULL;
  return(S);
}

void DestroyDependencyInfoTableStructures(void)
{
  uint16 i;
  TDependencyInfoSlot *S;
  uint16 BlockCnt;

  if (Table!=NULL) {
    for (i=0;i<Table->Count;i++) {
      S=GetSlot(i);
      if (S->Info!=NULL) KMEM__free(S->Info);
    }
    BlockCnt=(Table->Count+INFOS_IN_BLOCK-1)/INFOS_IN_BLOCK;
    for (i=0;i<BlockCnt;i++) {
      KMEM__free(Table->Blocks[i]);
    }
    KMEM__free(Table);
    Table=NULL;
  }
}

/**********************************
*** The library dependency file ***
**********************************/

#define DEP_FILE_BUFFER_SIZE 8192

static int16 Handle;
static upint8 *Buffer;
static uint16 BufSize;
static uint16 BufPos;
static bool Writing;
static bool EOFHit;

static void CleanDepFileBuffer(void)
{
  if (Buffer==NULL) {
    Buffer=KMEM__malloc(DEP_FILE_BUFFER_SIZE);
  }
  BufSize=0;
  BufPos=0;
  EOFHit=false;
}

static void PutDependencyFileSpec(char *Place)
{
  char *Ptr;

  Ptr=PutInternalMemDirSpec(Place);
  Ptr=StrCpy(Ptr,"libdep.dat");
}

static int16 CreateDependencyFile(void)
{
  TFileSpec Buffer;

  CleanDepFileBuffer();
  Writing=true;
  PutDependencyFileSpec(Buffer);
  Handle=LowLevelOpen(Buffer,oeTRUNC+onCREAT+oWRONLY);
  return(Handle);
}

static int16 OpenDependencyFile(void)
{
  TFileSpec Buffer;

  CleanDepFileBuffer();
  Writing=false;
  PutDependencyFileSpec(Buffer);
  Handle=LowLevelOpen(Buffer,oeOPEN+oRDONLY);
  return(Handle);
}

static uint8 ReadByte(void)
{
  int32 ReadSize;

  assert(!Writing);
  if (EOFHit) return(0);
  if (BufPos>=BufSize) {
    ReadSize=LowLevelRead(Handle,Buffer,DEP_FILE_BUFFER_SIZE);
    if (ReadSize<=0) {
      EOFHit=true;
      return(0);
    }
    assert(ReadSize<=DEP_FILE_BUFFER_SIZE);
    BufSize=ReadSize;
    BufPos=0;
  }
  return(Buffer[BufPos++]);
}

static void Flush(void)
{
  int32 WriteSize;

  assert(Writing);
  WriteSize=LowLevelWrite(Handle,Buffer,BufSize);
  if (WriteSize!=BufSize) {
    PrintSTDERRLn("Error writing library dependency file");
  }
  BufSize=0;
}

static void WriteByte(uint8 Byte)
{
  assert(Writing);
  if (BufSize==DEP_FILE_BUFFER_SIZE) Flush();
  Buffer[BufSize++]=Byte;
}

static void CloseDependencyFile(void)
{
  if (Writing) Flush();
  LowLevelClose(Handle);
}

/******************************************************
*** Reading and writing the library dependency file ***
******************************************************/

static bool ReadOneDependencyRecord(void)
{
  upint8 LibName[8];
  upint8 *Ptr;
  uint8 Ch;
  TDependencyInfoSlot *S;
  bool Continue;
  uint16 Length;
  uint16 Number;

  /*
    Read the name of the file
  */
  MemSet(LibName,8,0);
  Ptr=LibName;
  do {
    Ch=ReadByte();
    if (Ch==0) break;
    Continue=Ch<128;
    if (Ptr==LibName+8) return(false);
    *(Ptr++)=Ch & 127;
  } while (Continue);
  /*
    If we hit EOF and we haven't read something yet, it is OK and the
    dependency reading needs to terminate with no error. Otherwise the file
    is bad and should be erased.
  */
  if (EOFHit) return(Ptr==LibName);
  /*
    Read the length of the dependency info
  */
  Length=ReadByte();
  Length+=256*ReadByte();
  /*
    Create the structures
  */
  S=CreateSlot(&Number);
  if (S==NULL) return(false);
  S->Info=KMEM__malloc(Length+2);
  if (S->Info==NULL) return(false);
  MemCpy(S->LibName,LibName,8);
  S->Info->Length=Length;
  /*
    Read the dependency information block
  */
  Ptr=S->Info->Info;
  while (Length>0) {
    *(Ptr++)=ReadByte();
    Length--;
  }
  /*
    Return success/failure flag depending on whether we hit EOF.
  */
  return(!EOFHit);
}

static void LoadDependencyInfo(void)
{
  int16 DepFile;

  /*
    If the table was already loaded skip this action. Otherwise set the
    dependency table load flag.
  */
  if (DepTableLoaded) return;
  DepTableLoaded=true;
  /*
    Open the dependency tree file. If not found, silently exit.
  */
  DepFile=OpenDependencyFile();
  if (DepFile<0) return;
  /*
    Read the content and create the structures in memory. If an error occurs,
    erase any unfinished data from memory and display an error message. Also
    set the modification flag so the next sync will erase the broken file
    with fresh data.
  */
  while (!EOFHit) {
    if (!ReadOneDependencyRecord()) {
      PrintSTDERRLn("Error reading library dependency info");
      DestroyDependencyInfoTableStructures();
      DepTableChanged=true;
      return;
    }
  }
  /*
    Done.
  */
  CloseDependencyFile();
}

static void StoreDependencySlot(TDependencyInfoSlot *S)
{
  upint8 *Ptr;
  int16 i;
  int16 Len;
  uint16 Length;

  /*
    Write the name of the library
  */
  Ptr=S->LibName+8;
  while (*(Ptr-1)=='\0') Ptr--;
  Len=Ptr-S->LibName;
  Ptr=S->LibName;
  for (i=0;i<Len-1;i++) WriteByte(*(Ptr++));
  WriteByte(*Ptr+128);
  /*
    Write the dependency data size
  */
  Length=S->Info->Length;
  WriteByte(Length % 256);
  WriteByte(Length / 256);
  /*
    Write the dependency data
  */
  Ptr=S->Info->Info;
  while (Length>0) {
    WriteByte(*(Ptr++));
    Length--;
  }
  /*
    Done.
  */
}

static void StoreDependencyInfo(void)
{
  int16 DepFile;
  uint16 i;
  TDependencyInfoSlot *S;

  /*
    Create a new file with the dependency informations
  */
  DepFile=CreateDependencyFile();
  if (DepFile<0) {
    PrintSTDERRLn("Error writing library dependency info");
    return;
  }
  /*
    Go through the dependency information slots and store them into
  */
  for (i=0;i<Table->Count;i++) {
    S=GetSlot(i);
    StoreDependencySlot(S);
  }
  /*
    Done.
  */
  CloseDependencyFile();
}

void SyncDependencyInfo(void)
{
  if (DepTableChanged) {
    StoreDependencyInfo();
    DepTableChanged=false;
  }
}

/**************************************************
*** Serving the library dependency file content ***
**************************************************/

TDependencyInfo *GetDependencyInfoCore(const char *LibName,uint16 *Num)
{
  uint8 LName[8];
  uint16 i;
  uint16 Len;
  TDependencyInfoSlot *S;
  int16 Result;

  /*
    Ensure the dependency info table is loaded.
  */
  LoadDependencyInfo();
  /*
    If no table was created yet, return NULL (we don't have any slots when
    the table of slots does not exist yet).
  */
  if (Table==NULL) return(NULL);
  /*
    Prepare the library name for search
  */
  Len=StrLen(LibName);
  assert(Len<=8);
  if (Len<8) Len++;
  MemSet(LName,8,0);
  MemCpy(LName,LibName,Len);
  /*
    Search for the library name. Currently we use linear search. An array
    called Numbers is prepared to participate in making binary search for
    this but currently it is not worth the effort as we have only a few
    libraries to be concerned with.
  */
  for (i=0;i<Table->Count;i++) {
    S=GetSlot(i);
    Result=MemCmp(LName,S->LibName,8);
    if (Result==0) break;
  }
  /*
    If we found something, Result will be set to zero. Then set the slot
    number if the client requested it (by setting the 'Num' argument to
    non-NULL) and return the pointer. Otherwise return NULL.
  */
  if (Result==0) {
    if (Num!=NULL) *Num=i;
    return(S->Info);
  } else {
    return(NULL);
  }
}

TDependencyInfo *GetDependencyInfo(const char *LibName)
{
  return(GetDependencyInfoCore(LibName,NULL));
}

void PutDependencyInfo(const char *LibName,TDependencyInfo *I)
{
  TDependencyInfo *Old;
  TDependencyInfoSlot *S;
  uint16 SlotNum;
  bool Changed;

  assert(StrLen(LibName)<=8);
  Old=GetDependencyInfoCore(LibName,&SlotNum);
  if (Old!=NULL) {
    Changed=(Old->Length != I->Length);
    if (!Changed) Changed=(MemCmp(I->Info,Old->Info,Old->Length)!=0);
    if (Changed) {
      DepTableChanged=true;
      S=GetSlot(SlotNum);
      KMEM__free(S->Info);
      S->Info=I;
    }
  } else {
    S=CreateSlot(&SlotNum);
    if (S==NULL) return;
    MemSet(S->LibName,8,0);
    StrNCpy(S->LibName,LibName,8);
    S->Info=I;
    DepTableChanged=true;
  }
}

/************************
*** The library stack ***
************************/

#define LIB_STACK_BOTTOM ((TDependencyInfoSlot *)PutDependencyInfo)

static TDependencyInfoSlot *StackTop;
static TDependencyInfoSlot *UsedFiles;

void InitializeLibraryStack(void)
{
  StackTop=LIB_STACK_BOTTOM;
  UsedFiles=LIB_STACK_BOTTOM;
}

void PushLibraryInfo(const char *LibName)
{
  uint16 Num;
  TDependencyInfoSlot *S;

  assert(StackTop!=NULL);
  if (GetDependencyInfoCore(LibName,&Num)==NULL) return;
  S=GetSlot(Num);
  if (S->Next==NULL) {
    S->Next=StackTop;
    StackTop=S;
  }
}

TDependencyInfo *PopLibraryInfo(void)
{
  TDependencyInfoSlot *S;

  assert(StackTop!=NULL);
  if (StackTop==LIB_STACK_BOTTOM) return(NULL);
  S=StackTop;
  StackTop=S->Next;
  S->Next=UsedFiles;
  UsedFiles=S;
  return(S->Info);
}

void CleanupLibraryStack(void)
{
  TDependencyInfoSlot *S;

  assert(StackTop==LIB_STACK_BOTTOM);
  while (UsedFiles!=LIB_STACK_BOTTOM) {
    S=UsedFiles;
    UsedFiles=S->Next;
    S->Next=NULL;
  }
  StackTop=NULL;
  UsedFiles=NULL;
}
