#include "defs.gh"
#include "types.h"
#include "ltwd.h"
#include "strings.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "disklow.h"
#include "diskdir.h"
#include "cfgfile.h"
#include "exec.h"
#include "conio.h"
#include "ltargs.h"
#include "ltopts.h"
#include "ltbin.h"
#include "env.h"

#define CODE_DYNAMIC 0
#define CODE_STATIC 1
#define CODE_PROGRAM 2

int16 RunPlatformSpecificProgramViaPath(const char **Args,const char **Env)
{
  const char *ProgName;
  int16 Result;
#if !__b__EXEC_PATH__
  char Buffer[MAX_FILESPEC_SIZE+16];
  char *Ptr;
  const char *PathPtr;
  char Ch;
  bool Valid;
#endif

  ProgName=Args[0];
#if __b__EXEC_PATH__
  Result=ExecutePlatformSpecificProgram(Args[0],Args,Env);
  if (Result>=0) return(0);
#else
  PathPtr=GetEnv("PATH");
  if (PathPtr==NULL) {
    PrintSTDERR("Unable to execute ");
    PrintSTDERR(ProgName);
    PrintSTDERR(
      ".\n"
      "No PATH environment variable found.\n"
      "Probably you need to do 'export PATH' first to make it visible.\n"
    );
    return(fseNOENT);
  }
  Valid=true;
  Ptr=Buffer;
  Result=beUNKERR;
  do {
    Ch=*(PathPtr++);
    if (Ch=='\0' || Ch==':') {
      Valid=Valid && (Ptr>Buffer);
      if (Valid) {
        if (*(Ptr-1)!='/') *(Ptr++)='/';
        StrCpy(Ptr,ProgName);
        Args[0]=Buffer;
        Result=ExecutePlatformSpecificProgram(Buffer,Args,Env);
        Args[0]=ProgName;
        if (Result>=0) return(Result);
      }
      Valid=true;
      Ptr=Buffer;
    } else {
      Valid=Valid && (Ptr-Buffer<MAX_FILESPEC_SIZE);
      if (Valid) *(Ptr++)=Ch;
    }
  } while (Ch!='\0');
#endif
  PrintSTDERR("Utility: ");
  PrintSTDERRLn(Args[0]);
  SysERR("Unable to execute a platform specific utility");
  return(Result);
}

static void ReportError(const char *LeadMsg,const char *Msg)
{
  PrintSTDERRLn(LeadMsg);
  PrintSTDERRLn(Msg);
}

static const char *const* EnvPtr;

static const char *GetNextEnvVar(void)
{
  if (EnvPtr==NULL) {
    return(GetNextEnv());
  } else {
    return(*(EnvPtr++));
  }
}

static const char *GetFirstEnvVar(const char *const* Ptr)
{
  EnvPtr=Ptr;
  if (Ptr==NULL) {
    return(GetFirstEnv());
  } else {
    return(GetNextEnvVar());
  }
}

static TArgumentList *AddDirectoriesToVariable(
  const char *const*OrigEnv,
  const char *EnvVarName,
  const char *DirList
)
{
  TArgumentList *Environment;
  const char *EnvP;
  const char *EValue;
  bool PathConstructed;
  char *Ptr;
  char *VarName;
  char Names[(INTERNAL_DIR_SPEC_SIZE+MAX_FILESPEC_SIZE+1)*3];
  int16 CompareResult;
  const char *Msg;

  /*
    Build the enviroment for the linker. The enviroment must have the
    LIBRARY_PATH set to <work_dir>:<mem_dir>:$LIBRARY_PATH, where
    <work_dir> is the working directory, <mem_dir> is the directory for
    libraries loaded into the memory and $LIBRARY_PATH is the original
    value of the variable. All other variables should remain intact.
  */
  Environment=CreateList();
  if (Environment==NULL) {
    PlainERR("No memory for enviroment for the linker");
    return(NULL);
  }
  /*
    Populate the just created
  */
  EnvP=GetFirstEnvVar(OrigEnv);
  PathConstructed=false;
  Ptr=StrCpy(Names,EnvVarName);
  *(Ptr++)='=';
  Ptr=StrNCpy(Ptr,DirList,sizeof(Names)-(Ptr-Names+64));
  Names[sizeof(Names)-65]='\0';
  VarName=Ptr+1;
  while (EnvP!=NULL) {
    if (!PathConstructed) {
      EValue=StrChr(EnvP,'=');
      if (EValue!=NULL) {
        MemCpy(VarName,EnvP,EValue-EnvP);
        VarName[EValue-EnvP]='\0';
        CompareResult=StrCmp(VarName,EnvVarName);
        PathConstructed=CompareResult==0;
        if (PathConstructed) {
          Names[sizeof(Names)-1]='\0';
          *(Ptr++)=':';
          StrNCpy(Ptr,EValue+1,sizeof(Names)-(Ptr-Names));
          if (Names[sizeof(Names)-1]!='\0') {
            Ptr=StrCpy(Names,EnvVarName);
            StrCpy(Ptr," value too long");
            PlainERR(Names);
            return(NULL);
          }
          EnvP=Names;
        }
      }
    }
    Msg=AddArgument(Environment,EnvP);
    if (Msg!=NULL) {
      ReportError("Error building enviroment for exec",Msg);
      return(NULL);
    }
    EnvP=GetNextEnvVar();
  }
  /*
    If there is no LIBRARY_PATH, create a new one and insert it before the
    current enviroment item (the enviroment must be kept sorted).
  */
  if (!PathConstructed) {
    Msg=AddArgument(Environment,Names);
    if (Msg!=NULL) {
      ReportError("Error building enviroment for exec",Msg);
      return(NULL);
    }
  }
  /*
    Everything went OK, so we can return the result.
  */
  return(Environment);
}

static TArgumentList *AddInternalDirsToVariable(
  const char *const*OrigEnv,
  const char *EnvVarName,
  bool WorkDirToo
)
{
  char Names[(INTERNAL_DIR_SPEC_SIZE+MAX_FILESPEC_SIZE+1)*3];
  char *Ptr;

  /*
    Setup the value to be added into the variable
  */
  Ptr=Names;
  if (WorkDirToo) {
    Ptr=PutInternalWorkDirSpec(Ptr);
    Ptr--;
    *(Ptr++)=':';
  }
  Ptr=PutInternalMemDirSpec(Ptr);
  Ptr--;
  *Ptr='\0';
  /*
    Perform the asdding
  */
  return(AddDirectoriesToVariable(OrigEnv,EnvVarName,Names));
}

int16 DoCompile(const char * const *argv)
{
  char Args[8192];
  char IDir[INTERNAL_DIR_SPEC_SIZE+MAX_FILESPEC_SIZE+1];
  const char *NewArgs[1024];
  char *Ptr;
  char *Ptr1;
  int16 Result;

  /*
    Build the "-IXXX" statement for the GNU compiler to recvognize the
    OSHS specific headers.
  */
  Ptr=StrCpy(IDir,"-I");
  Ptr=StrNCpy(Ptr,CodeBaseDir,sizeof(IDir)-2);
  StrNCpy(Ptr,"include",sizeof(IDir)-(Ptr-IDir));
  IDir[sizeof(IDir)-1]='\0';
  /*
    Build the environment for the compiler to run with
  */
  NewArgs[0]="gcc";
  NewArgs[1]="-c";
  NewArgs[2]="-g";
  NewArgs[3]="-Wall";
  if (TreatWarningsAsErrors) {
    NewArgs[4]="-Werror";
  } else {
    NewArgs[4]="-Wall";
  }
  NewArgs[5]=IDir;
  Ptr=PutInternalWorkDirSpec(Args);
  Ptr=StrCpy(Ptr,"result.o")+1;
#if __i__LIB_SUPPORT__static__
  NewArgs[6]=argv[0];
  NewArgs[7]="-o";
  NewArgs[8]=Args;
  NewArgs[9]=NULL;
  Result=RunPlatformSpecificProgramViaPath(NewArgs,NULL);
#else
  Result=MkEmptyFileDirect(Args);
#endif
  RegisterFileForDeletion(Args);
  if (Result!=0) return(Result);
  Ptr1=PutInternalWorkDirSpec(Ptr);
  StrCpy(Ptr1,"result.lo");
#if __i__LIB_SUPPORT__shared__
  NewArgs[6]="-fPIC";
  NewArgs[7]="-DPIC";
  NewArgs[8]=argv[0];
  NewArgs[9]="-o";
  NewArgs[10]=Ptr;
  NewArgs[11]=NULL;
  Result=RunPlatformSpecificProgramViaPath(NewArgs,NULL);
#else
  Result=MkEmptyFileDirect(Ptr);
#endif
  RegisterFileForDeletion(Ptr);
  if (Result!=0) return(Result);
  NewArgs[0]=Ptr;
  NewArgs[1]=Args;
  NewArgs[2]=NULL;
  return(CombineCompilationResults(argv,NewArgs,"GNUL"));
}

static bool AddLibrary(TArgumentList *Args,const char *LibName)
{
  char Buffer[MAX_FILESPEC_SIZE+3];
  char *Ptr;
  char *Msg;

  Ptr=StrCpy(Buffer,"-l");
  StrToLwr(Ptr,LibName);
  Msg=AddArgument(Args,Buffer);
  if (Msg!=NULL) {
    ReportError("Unable to record a library into the arguments",Msg);
  }
  return(Msg==NULL);
}

bool AddStaticLibrary(TArgumentList *Args,const char *LibName)
{
  return(AddLibrary(Args,LibName));
}

bool AddDynamicLibrary(TArgumentList *Args,const char *LibName)
{
  return(AddLibrary(Args,LibName));
}

int32 ExtractStaticLibrary(int16 InFD,int16 OutFD,bool Verbose)
{
  return(ExtractContent(InFD,OutFD,CODE_STATIC,Verbose));
}

int32 ExtractDynamicLibrary(int16 InFD,int16 OutFD,bool Verbose)
{
  return(ExtractContent(InFD,OutFD,CODE_DYNAMIC,Verbose));
}

int32 ExtractProgramCode(int16 InFD,char *Buffer,bool Verbose)
{
  int16 OutFD;
  char *Ptr;
  int32 Result;

  Ptr=PutInternalWorkDirSpec(Buffer);
  StrCpy(Ptr,"program");
  OutFD=LowLevelOpen(Buffer,oeTRUNC+onCREAT+oWRONLY);
  if (OutFD<0) return(OutFD);
  Result=ExtractContent(InFD,OutFD,CODE_PROGRAM,Verbose);
  LowLevelClose(OutFD);
  return(Result);
}

char *BuildStaticLibName(char *Ptr,const char *Name)
{
  Ptr=StrCpy(Ptr,"lib");
  Ptr=StrToLwr(Ptr,Name);
  return(StrCpy(Ptr,".a"));
}


char *BuildDynamicLibName(char *Ptr,const char *Name)
{
  Ptr=StrCpy(Ptr,"lib");
  Ptr=StrToLwr(Ptr,Name);
  return(StrCpy(Ptr,".so"));
}

static void NoMemoryForLinkerArgumentList(void)
{
  PlainERR("No memory for argument list for the platform linker");
}

int16 DoLink(TLinkingInfo *Info)
{
  uint16 Len;
  const char *Name;
  const char *Ext;
  char Ch;
  char Names[(INTERNAL_DIR_SPEC_SIZE+MAX_FILESPEC_SIZE+1)*3];
  char ProgName[MAX_FILESPEC_SIZE];
  char LibDir[MAX_FILESPEC_SIZE+16];
  char *Ptr;
  const char *Combines[4];
  TArgumentList *ListProg;
#if __i__LIB_SUPPORT__static__
  bool HasStaticLibrary;
  TArgumentList *ListLibStatic;
#endif
#if __i__LIB_SUPPORT__shared__
  bool HasDynamicLibrary;
  TArgumentList *ListLibDynamic;
#endif
  const char *Msg;
  int16 Result;
  int16 WDSize;

  /*
    Extract the name of the output file from the specification of the result
    file. Place it at the end of the buffer for the names.
  */
  Len=StrLen(Info->Output);
  Name=Info->Output+Len;
  Ext=Name;
  do {
    Ch=*(--Name);
    if (Ch=='.') {
      Ext=Name;
    } else if (Ch=='/') {
      Name++;
    }
  } while (Ch!='/' && Name>Info->Output);
  MemCpy(ProgName,Name,Ext-Name);
  ProgName[Ext-Name]='\0';
  /*
    Create the names for the results
  */
  Combines[CODE_PROGRAM]=Names;
  Ptr=PutInternalWorkDirSpec(Names);
  WDSize=Ptr-Names;
  Ptr=StrCpy(Ptr,ProgName);
  Ptr=StrCpy(Ptr,".bin")+1;
  Combines[CODE_STATIC]=Ptr;
  Ptr=PutInternalWorkDirSpec(Ptr);
  Ptr=BuildStaticLibName(Ptr,ProgName)+1;
  Combines[CODE_DYNAMIC]=Ptr;
  Ptr=PutInternalWorkDirSpec(Ptr);
  BuildDynamicLibName(Ptr,ProgName);
  RegisterFileForDeletion(Combines[CODE_STATIC]);
  RegisterFileForDeletion(Combines[CODE_DYNAMIC]);
  RegisterFileForDeletion(Combines[CODE_PROGRAM]);
  Combines[3]=NULL;
  /*
    I need to know whether the result has a library and whether
    it also contains a static portion of it.
  */
#if __i__LIB_SUPPORT__static__
  HasStaticLibrary=Info->LibObj[0]!=NULL;
#if __i__LIB_SUPPORT__shared__
  HasDynamicLibrary=HasStaticLibrary || Info->LibLibStatic[0]!=NULL;
#endif
#elif __i__LIB_SUPPORT__shared__
  HasDynamicLibrary=Info->LibObj[0]!=NULL;
#endif
  /*
    Extract all pieces into the workspace. The object modules are named
    "00000000.o, 00000001.o etc. The libraries are named after their names
    (the ones stored in the files, not the filenames themselves) with "lib"
    prepended.
  */
  ListProg=CreateList();
  if (ListProg==NULL) {
    NoMemoryForLinkerArgumentList();
    return(1);
  }
#if __i__LIB_SUPPORT__static__
  ListLibStatic=CreateList();
  if (ListLibStatic==NULL) {
    NoMemoryForLinkerArgumentList();
    return(1);
  }
#endif
#if __i__LIB_SUPPORT__shared__
  ListLibDynamic=CreateList();
  if (ListLibDynamic==NULL) {
    NoMemoryForLinkerArgumentList();
    return(1);
  }
#endif
  AddArgument(ListProg,"gcc");
  AddArgument(ListProg,"-o");
  AddArgument(ListProg,Combines[CODE_PROGRAM]);
  LibDir[0]='-';
  LibDir[1]='L';
  PutInternalWorkDirSpec(LibDir+2);
  AddArgument(ListProg,LibDir);
  PutInternalMemDirSpec(LibDir+2);
  AddArgument(ListProg,LibDir);
  ExtractNumberedContents(
    Info->ProgObj,
#if __i__LIB_SUPPORT__static__
    CODE_STATIC,
#else
    CODE_DYNAMIC,
#endif
    "pg",
    "o",
    ListProg
  );
  DirectRemove(Combines[CODE_STATIC]);
#if __i__LIB_SUPPORT__static__
  AddArgument(ListLibStatic,"ar");
  AddArgument(ListLibStatic,"qc");
  AddArgument(ListLibStatic,Combines[CODE_STATIC]);
  ExtractNumberedContents(
    Info->LibObj,
    CODE_STATIC,
    "lib",
    "o",
    ListLibStatic
  );
#endif
  DirectRemove(Combines[CODE_DYNAMIC]);
#if __i__LIB_SUPPORT__shared__
  AddArgument(ListLibDynamic,"gcc");
  AddArgument(ListLibDynamic,"-shared");
  AddArgument(ListLibDynamic,"-o");
  AddArgument(ListLibDynamic,Combines[CODE_DYNAMIC]);
  PutInternalWorkDirSpec(LibDir+2);
  AddArgument(ListLibDynamic,LibDir);
  PutInternalMemDirSpec(LibDir+2);
  AddArgument(ListLibDynamic,LibDir);
  ExtractNumberedContents(
    Info->LibObj,
    CODE_DYNAMIC,
    "lib",
    "lo",
    ListLibDynamic
  );
#endif
  /*
    Create the library. Library is composed from a static and a dynamic
    (shared) component.
  */
#if __i__LIB_SUPPORT__static__
  if (HasStaticLibrary) {
    Result=RunPlatformSpecificProgramViaPath(ListLibStatic->ArgList,NULL);
    if (Result!=0) return(Result);
  }
#endif
#if __i__LIB_SUPPORT__shared__
  if (HasDynamicLibrary) {
    Msg=AddList(ListLibDynamic,Info->LibLibStatic);
    if (Msg==NULL) Msg=AddList(ListLibDynamic,Info->LibLibDynamic);
    if (Msg!=NULL) {
      PrintSTDERRLn("Error building args for dynamic library linking");
      PrintSTDERRLn(Msg);
      return(1);
    }
    Result=RunPlatformSpecificProgramViaPath(ListLibDynamic->ArgList,NULL);
    if (Result!=0) return(Result);
  }
#endif
  /*
    Create the program if any.
  */
  if (Info->ProgObj[0]!=NULL || Info->ProgLibStatic[0]!=NULL) {
    if (
#if __i__LIB_SUPPORT__static__
      HasStaticLibrary
#if __i__LIB_SUPPORT__shared__
      ||
#endif
#endif
#if __i__LIB_SUPPORT__shared__
      HasDynamicLibrary
#else
#if !__i__LIB_SUPPORT__static__
      0
#endif
#endif
    ) AddLibrary(ListProg,ProgName);
    Msg=AddList(ListProg,Info->ProgLibStatic);
    if (Msg==NULL) Msg=AddList(ListProg,Info->ProgLibDynamic);
    if (Msg==NULL) Msg=AddList(ListProg,Info->LibLibStatic);
    if (Msg==NULL) Msg=AddList(ListProg,Info->LibLibDynamic);
    if (Msg==NULL) Msg=AddSystemLibraries(ListProg);
    if (Msg!=NULL) {
      PrintSTDERRLn("Error building arguments for program linking");
      PrintSTDERRLn(Msg);
      return(1);
    }
    Result=RunPlatformSpecificProgramViaPath(ListProg->ArgList,NULL);
    if (Result!=0) return(Result);
  } else {
    if (MkEmptyFileDirect(Combines[CODE_PROGRAM])!=0) {
      SysERR("Error creating dummy program");
      return(1);
    }
  }
  /*
    Create dummy library files if the libraries don't exist. These dummy
    library files would obstruct linking if they were created previously.
  */
#if __i__LIB_SUPPORT__static__
  if (!HasStaticLibrary) {
#endif
    if (MkEmptyFileDirect(Combines[CODE_STATIC])!=0) {
      SysERR("Error creating dummy static library");
      return(1);
    }
#if __i__LIB_SUPPORT__static__
  }
#endif
#if __i__LIB_SUPPORT__shared__
  if (!HasDynamicLibrary) {
#endif
    if (MkEmptyFileDirect(Combines[CODE_DYNAMIC])!=0) {
      SysERR("Error creating dummy dynamic library");
      return(1);
    }
#if __i__LIB_SUPPORT__shared__
  }
#endif
  /*
    Create the resulting file. The result is composed up to three components:
    the program, the static library and the dynamic library.
  */
  return(AssembleResultingBinaryFile(Info->Output,Combines,"GNUL",ProgName));
}

TArgumentList *ConstructEnvironmentForExec(const char *const*OrigEnv)
{
  return(AddInternalDirsToVariable(OrigEnv,"LD_LIBRARY_PATH",false));
}
