#include "defs.gh"
#include "types.h"

/*
  LIBTOOL main module.

  The LIBTOOL is responsible for hiding platform differencies 

  The LIBTOOL is presented as a standalone program to the OSHS software. The
  usage is

    LIBTOOL <command> <options_and_files>

  where <command> is one of the following:

    compile     Compile a single C file
    link        Link a program or library
    load        Load a library into memory (actually this loads the library
                into space of actively used libraries)
*/

#include "conio.h"
#include "cfgfile.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "pdexec.h"
#include "diskdata.h"
#include "disklow.h"
#include "diskdir.h"
#include "ltargs.h"
#include "ltwd.h"
#include "ltbin.h"
#include "ltdeps.h"
#include "ltopts.h"
#undef __t__USE_GNU__
#if __e__COMPILER_DRV__ == __v__gnu__
#define __t__USE_GNU__ 1
#elif __e__EXEC_DRV__ == __v__mingw__
#define __t__USE_GNU__ 1
#else
#define __t__USE_GNU__ 0
#endif
#if __t__USE_GNU__
#include "ltgnu.h"
#endif
#include "exec.h"
#include "memory.h"
#include "strings.h"

#if __e__INTERNAL_DIR__ == __v__optional__
static void ReportNoInternalDir(const char *Action)
{
  PrintSTDERR("Cannot perform ");
  PrintSTDERR(Action);
  PrintSTDERR(
    " without the OSHS internal directory.\n"
    "Setup the OSHSINTERNALDIR enviroment variable and try again."
  );
}
#endif

int16 RunCompile(const char * const *argv)
{
  int16 Result;
  const char * const *NewArgV;

  OBSAreaSize=0;
#if __e__INTERNAL_DIR__ == __v__optional__
  if (InternalDir[0]=='\0') {
    ReportNoInternalDir("compilation");
    return(1);
  }
#endif
  NewArgV=ProcessCompilerOptions(argv);
  if (NewArgV==NULL) return(1);
  Result=DoCompile(NewArgV);
  DeleteAllRegisteredFiles();
  DestroyArgumentLists();
  return(Result);
}

static void ReportError(const char *Msg,const char *Report)
{
  PrintSTDERRLn(Msg);
  if (Report!=NULL) {
    PrintSTDERR("The system said: ");
    PrintSTDERRLn(Report);
  }
}

static bool PrepareLibraryForLinkCore(
  int16 InFD,
  int16 *OutFD,
  TArgumentList *ArgLibs,
  bool Static
)
{
  char Name[9];
  char FileSpec[INTERNAL_DIR_SPEC_SIZE*2];
  char *Ptr;

  if (ExtractLibraryName(InFD,Name,true)<0) return(false);
  Ptr=PutInternalWorkDirSpec(FileSpec);
  if (Static) {
    BuildStaticLibName(Ptr,Name);
  } else {
    BuildDynamicLibName(Ptr,Name);
  }
  *OutFD=LowLevelOpen(FileSpec,oeTRUNC+onCREAT+oRDWR);
  if (*OutFD<0) {
    SysERR("Error creating output file");
    return(false);
  }
  RegisterFileForDeletion(FileSpec);
  if (Static) {
    if (ExtractStaticLibrary(InFD,*OutFD,true)<0) return(false);
    if (!AddStaticLibrary(ArgLibs,Name)) return(false);
  } else {
    if (ExtractDynamicLibrary(InFD,*OutFD,true)<0) return(false);
    if (!AddDynamicLibrary(ArgLibs,Name)) return(false);
  }
  return(true);
}

static bool PrepareLibraryForLink(
  TRootStatus *Root,
  const char *File,
  TArgumentList *ArgLibs,
  bool Static
)
{
  int16 InFD;
  int16 OutFD;
  bool Result;

  InFD=ConvertAndOpenFile(Root,File,oeOPEN+oRDONLY);
  if (InFD<0) {
    SysERR("Unable to open the file");
    return(false);
  }
  OutFD=-1;
  Result=PrepareLibraryForLinkCore(InFD,&OutFD,ArgLibs,Static);
  if (LowLevelClose(InFD)<0) {
    SysERR("Error closing input file");
    Result=false;
  }
  if (LowLevelClose(OutFD)<0) {
    SysERR("Error closing output file");
    Result=false;
  }
  return(Result);
}

static bool SeparateLinkerArgs(
  TRootStatus *Root,
  const char * const *argv,
  TLinkingInfo *I
)
{
  const char *Result;
  const char *Arg;
  TArgumentList *ProgObj;
  TArgumentList *ProgLibStatic;
  TArgumentList *ProgLibDynamic;
  TArgumentList *LibObj;
  TArgumentList *LibLibStatic;
  TArgumentList *LibLibDynamic;
  TArgumentList *ArgList;
  bool ProgFiles;
  bool OnlySwitch;
  bool Invalid;
  TArgumentList *ProgList;
  TArgumentList *LibList;
  const char *ArgToAdd;
  bool Static;
  bool Library;

  /*
    The first argument to the linker is the output file name
  */
  I->Output=*(argv++);
  if (I->Output==NULL) {
    ReportError("At least output file name required",NULL);
    return(false);
  }
  /*
    Create the separated argument lists
  */
  ProgObj=CreateList();
  ProgLibStatic=CreateList();
  ProgLibDynamic=CreateList();
  LibObj=CreateList();
  LibLibStatic=CreateList();
  LibLibDynamic=CreateList();
  if (
    ProgObj==NULL || ProgLibStatic==NULL || ProgLibDynamic==NULL ||
    LibObj==NULL || LibLibStatic==NULL || LibLibDynamic==NULL
  ) {
    ReportError("No memory for argument lists for external linker",NULL);
    return(false);
  }
  /*
    Bind them with the elements of the linker info structure
  */
  I->ProgObj=ProgObj->ArgList;
  I->ProgLibStatic=ProgLibStatic->ArgList;
  I->ProgLibDynamic=ProgLibDynamic->ArgList;
  I->LibObj=LibObj->ArgList;
  I->LibLibStatic=LibLibStatic->ArgList;
  I->LibLibDynamic=LibLibDynamic->ArgList;
  /*
    Do the actual argument list separation
  */
  OBSAreaSize=0;
  ProgFiles=true;
  Invalid=false;
  Static=false;
  Library=false;
  ProgList=ProgObj;
  LibList=LibObj;
  do {
    Arg=*(argv++);
    if (Arg!=NULL) {
      OnlySwitch=false;
      ArgToAdd=Arg;
      if (Arg[0]=='-') {
        ArgToAdd=Arg+2;
        if (Arg[1]=='l') {
          ProgList=ProgLibStatic;
          LibList=LibLibStatic;
          Static=true;
          Library=true;
          OnlySwitch=true;
        } else if (Arg[1]=='d') {
          ProgList=ProgLibDynamic;
          LibList=LibLibDynamic;
          Static=false;
          Library=true;
          OnlySwitch=true;
        } else if (Arg[1]=='P' || Arg[1]=='L') {
          ProgList=ProgObj;
          LibList=LibObj;
          Library=false;
          ProgFiles=(Arg[1]=='P');
          OnlySwitch=true;
        } else if (Arg[1]=='A') {
          if (!ReadOBSAreaSize(Arg+2)) return(false);
          continue;
        } else {
          Invalid=true;
        }
      }
      if (!Invalid) {
        if (OnlySwitch) {
          Invalid=(ArgToAdd[0]!='\0');
        } else {
          ArgList=ProgFiles?ProgList:LibList;
          if (Library) {
            Result=NULL;
            if (!PrepareLibraryForLink(Root,ArgToAdd,ArgList,Static)) {
              Result="See messages above";
            }
          } else {
            Result=AddArgument(ArgList,ArgToAdd);
          }
          if (Result!=NULL) {
            PrintSTDERR("Argument: ");
            PrintSTDERRLn(Arg);
            ReportError("Failed to process argument",Result);
            return(false);
          }
        }
      }
      if (Invalid) {
        ReportError("Invalid commandline argument specified",Arg);
        return(false);
      }
    }
  } while (Arg!=NULL);
  LinkSystemLibsStatic=Static;
  return(true);
}

int16 RunLink(TRootStatus *Root, const char * const *argv)
{
  TLinkingInfo I;
  int16 Result;

#if __e__INTERNAL_DIR__ == __v__optional__
  if (InternalDir[0]=='\0') {
    ReportNoInternalDir("linking");
    return(1);
  }
#endif
  Result=1;
  if (SeparateLinkerArgs(Root,argv,&I)) {
    Result=DoLink(&I);
  }
  DeleteAllRegisteredFiles();
  DestroyArgumentLists();
  return(Result);
}

static bool CheckLibraryPresent(const upint8 *LibName)
{
  char Buffer[MAX_FILESPEC_SIZE*2];
  char *Ptr;
  int16 Handle;
  bool Result;

  Ptr=PutInternalMemDirSpec(Buffer);
  Ptr=BuildDynamicLibName(Ptr,(const char *)LibName);
  Handle=LowLevelOpen(Buffer,oeOPEN+oRDONLY);
  Result=(Handle>=0);
  if (Result) LowLevelClose(Handle);
  return(Result);
}

static void CheckLibrary(const upint8 *LibName, uint16 Len, TExecReply *R)
{
  if (CheckLibraryPresent(LibName)) {
  } else {
    if (MAX_MESSAGE_SIZE-R->ExitCode-2>=Len) {
      MemCpy(R->Info+R->ExitCode,LibName,Len);
      R->ExitCode+=Len;
    }
  }
}

static int16 AnalyzeLibraries(TDependencyInfo *I, TExecReply *R)
{
  upint8 *Ptr;
  upint8 *LibName;
  uint16 Length;
  uint8 Ch;

  Length=I->Length;
  Ptr=I->Info;
  LibName=Ptr;
  while (Length>0) {
    Ch=*(Ptr++);
    Length--;
    if (Ch==0) {
      CheckLibrary(LibName,Ptr-LibName,R);
      LibName=Ptr;
    }
  }
  if (Ch!='\0') return(execeFORMAT);
  return(0);
}

static int16 PrepareLibraries(int16 Handle, TExecReply *R)
{
  int16 Result;
  TDependencyInfo *I;

  Result=FetchDependencyInfo(Handle,&I);
  if (Result!=0) return(Result);
  if (I==NULL) return(beUNKERR);
  if (I->Length==0) {
    R->ExitCode=0x8000;
    return(0);
  }
  R->ExitCode=0;
  Result=AnalyzeLibraries(I,R);
  if (Result==0) {
    if (R->ExitCode!=0) {
      Result=R->ExitCode+2;
      R->ExitCode=execeNOLIB;
    }
  }
  KMEM__free(I);
  return(Result);
}

int16 ExtractCodeToExecute(int16 InFD,char *FileSpec)
{
  return(ExtractProgramCode(InFD,FileSpec,false));
}

static int16 PrepareProgramForExec(
  TRootStatus *Root,
  const char *Prog,
  char *PlatformSpecificSpec,
  TExecReply *R
)
{
  int16 InFD;
  int32 Result;

  InFD=ConvertAndOpenFile(Root,Prog,oeOPEN+oRDONLY);
  if (InFD<0) return(InFD);
  Result=ExtractCodeToExecute(InFD,PlatformSpecificSpec);
  if (Result>0) {
    Result=0;
  } else if (Result==0) {
    Result=execeNONPROG;
  }
  if (Result>=0) Result=PrepareLibraries(InFD,R);
  LowLevelClose(InFD);
  if (Result!=0) return(Result);
  return(Result);
}

int16 RunProgram(
  TRootStatus *Root,
  const char * const *argv,
  const char * const *env,
  TExecReply *R
)
{
  TArgumentList *Environment;
  char Buffer[MAX_FILESPEC_SIZE+16];
  int16 Output;

  Output=PrepareProgramForExec(Root,argv[0],Buffer,R);
  if (Output!=0) return(Output);
  if (R->ExitCode!=0x8000) {
    Environment=ConstructEnvironmentForExec(env);
    if (Environment==NULL) return(beNOMEM);
    env=Environment->ArgList;
  }
  Output=ExecutePlatformSpecificProgram(Buffer,argv,env);
  if (Output>=0) {
    R->ExitCode=Output;
    Output=2;
  }
  return(Output);
}

static int16 ExtractSharedLibraryPortion(int16 Handle)
{
  int16 Result;
  int32 ExtractResult;
  int16 CloseResult;
  char Buffer[MAX_FILESPEC_SIZE+16];
  char LibName[9];
  char *Ptr;
  int16 OutFD;
  TDependencyInfo *I;

  Result=ExtractLibraryName(Handle,LibName,false);
  assert(Result<=0);
  if (Result<0) return(Result);
  Ptr=PutInternalMemDirSpec(Buffer);
  Ptr=BuildDynamicLibName(Ptr,LibName);
  OutFD=LowLevelOpen(Buffer,oeTRUNC+onCREAT+oWRONLY);
  if (OutFD<0) return(OutFD);
  ExtractResult=ExtractDynamicLibrary(Handle,OutFD,false);
  if (ExtractResult<0) {
    assert(ExtractResult>=-32768);
    Result=ExtractResult;
  } else if (ExtractResult==0) {
    Result=DirectRemove(Buffer);
    assert(Result<=0);
    if (Result==0) Result=execeNONLIB;
  }
  CloseResult=LowLevelClose(OutFD);
  assert(CloseResult<=0);
  if (CloseResult<0) return(CloseResult);
  Result=FetchDependencyInfo(Handle,&I);
  if (Result<0) return(Result);
  PutDependencyInfo(LibName,I);
  SyncDependencyInfo();
  return(Result);
}

int16 LoadDynamicLibrary(const char *LibSpec)
{
  int16 Handle;
  int16 Result;
  int16 CloseResult;

  Handle=LowLevelOpen(LibSpec,oeOPEN+oRDONLY);
  if (Handle<0) return(Handle);
  Result=ExtractSharedLibraryPortion(Handle);
  CloseResult=LowLevelClose(Handle);
  if (CloseResult<0) return(CloseResult);
  return(Result);
}
