#include "defs.gh"
#include "types.h"
#include "traps.h"
#include "strings.h"
#include "cfgfile.h"
#include "kernel.h"
#include "kernmsg.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "diskdata.h"
#include "disklow.h"
#include "diskdir.h"
#include "ltargs.h"

static int16 CheckIsDir(const char *Spec)
{
  /*
    Attempt to open the specification as a file. If it is a directory, the
    call will fail with eISDIR, so "True" will be returned. Otherwise we
    return "False". Don't forget to close the file if it really was opened.
  */
  int16 fd;

  fd=LowLevelOpen(Spec,oeOPEN+oRDONLY);
  if (fd>=0) LowLevelClose(fd);
  return((fd==fseISDIR));
}

static void MakeDir(const char *Name)
{
  char Buf[128];
  char *Ptr;
  char *NumPtr;
  int16 Result;

  if (!CheckIsDir(Name)) {
    Result=DirectMkDir(Name);
    if (Result<0) {
      Result=-Result;
      NumPtr=Buf+128;
      *(--NumPtr)='\0';
      do {
        *(--NumPtr)=Result%10+'0';
        Result/=10;
      } while (Result>0);
      Ptr=StrCpy(Buf,"Unable to make directory (Error -");
      Ptr=StrCpy(Ptr,NumPtr);
      StrCpy(Ptr,")\n");
      Fatal(Buf);
    }
  }
}

void PrepareInternalDirectory(void)
{
  char DirSpec[INTERNAL_DIR_SPEC_SIZE+10];
  char *Ptr;

  Ptr=StrCpy(DirSpec,InternalDir);
  MakeDir(DirSpec);
  *(Ptr++)='/';
  StrCpy(Ptr,"work");
  MakeDir(DirSpec);
  StrCpy(Ptr,"libmem");
  MakeDir(DirSpec);
}

char *PutInternalDirSpec(const char *Dir,char *Ptr)
{
  Ptr=StrCpy(Ptr,InternalDir);
  *(Ptr++)='/';
  if (Dir!=NULL) {
    Ptr=StrCpy(Ptr,Dir);
    *(Ptr++)='/';
  }
  *Ptr='\0';
  return(Ptr);
}

char *PutInternalWorkDirSpec(char *Ptr)
{
  return(PutInternalDirSpec("work",Ptr));
}

char *PutInternalMemDirSpec(char *Ptr)
{
  return(PutInternalDirSpec("libmem",Ptr));
}

typedef char *TDirGenProc(char *Ptr);

static char *BuildInternalFileSpec(
  TDirGenProc *GenDir,
  const char *File,
  char *Buf
)
{
  assert(StrLen(File)<256);
  Buf=GenDir(Buf);
  return(StrCpy(Buf,File));
}

char *BuildInternalWorkFileSpec(const char *File, char *Buf)
{
  return(BuildInternalFileSpec(PutInternalWorkDirSpec,File,Buf));
}

char *BuildInternalMemFileSpec(const char *File, char *Buf)
{
  return(BuildInternalFileSpec(PutInternalMemDirSpec,File,Buf));
}

typedef char *TFileNameBuilder(const char *File, char *Buf);

static int16 OpenInternalFile(
  TFileNameBuilder *Builder,
  const char *File,
  int16 Mode
)
{
  char Spec[INTERNAL_DIR_SPEC_SIZE+256];

  Builder(File,Spec);
  return(LowLevelOpen(Spec,Mode));
}

int16 OpenTemporaryWorkFile(const char *File, int16 Mode)
{
  return(OpenInternalFile(BuildInternalWorkFileSpec,File,Mode));
}

int16 OpenTemporaryMemFile(const char *File, int16 Mode)
{
  return(OpenInternalFile(BuildInternalMemFileSpec,File,Mode));
}

int16 MkEmptyFileDirect(const char *File)
{
  int16 Result;

  Result=LowLevelOpen(File,oeTRUNC+onCREAT+oRDWR);
  if (Result>=0) {
    Result=LowLevelClose(Result);
    if (Result>=0) Result=0;
  }
  return(Result);
}

static TArgumentList *FilesToDelete;

void RegisterFileForDeletion(const char *File)
{
  const char *const*ListPtr;

  if (FilesToDelete==NULL) {
    FilesToDelete=CreateList();
    if (FilesToDelete==NULL) return;
  }
  ListPtr=FilesToDelete->ArgList;
  while (*ListPtr!=NULL) {
    if (StrCmp(File,*ListPtr)==0) return;
    ListPtr++;
  }
  AddArgument(FilesToDelete,File);
}

void DeleteAllRegisteredFiles(void)
{
  const char *const*ListPtr;

  if (FilesToDelete==NULL) return;
  ListPtr=FilesToDelete->ArgList;
  while (*ListPtr!=NULL) {
    DirectRemove(*ListPtr);
    ListPtr++;
  }
  FilesToDelete=NULL;
}
