#ifndef __h__SYSLIB_PDFILEIO__
#define __h__SYSLIB_PDFILEIO__

/*
  The protocol for the file I/O calls
*/

/*** High performance file I/O commands ***/
/*
  This form of file I/O commands are called "high performance", because one
  command packet can induce reading or writing huge amount of data and the
  data is transferred directly between the file at one side and the client
  buffer at the other side.

  The client sends to the server a pointer to the buffer among other things
  in the command packet. The server then maps the buffer into its own address
  space and transfers the data. When the transfer is completed or failed, the
  server sends a reply message to the client.

  If the server cannot do direct file I/O for some reason but can perform the
  file I/O using classic file I/O commands, it must return beNOTSUP error
  message. The client then may use the classic file I/O to complete the
  requested file I/O (usually the standard library takes care about this).
  The classic file I/O must always be supported.
*/

#define cmdREAD 0
#define cmdWRITE 1

typedef struct tagTReadCmd {
  upint8 Command;
  pint32 Size;
  void *Ptr;
} TReadCmd;

typedef struct tagTWriteCmd {
  upint8 Command;
  pint32 Size;
  const void *Ptr;
} TWriteCmd;

typedef struct tagTIOReply {
  pint32 Size;
} TIOReply;

/*** SEEK: The SEEK command ***/
/*
  This command moves the current file pointer to the specified location in
  the file.
*/

#define cmdSEEK 2
#define cmdTELL 3
#define cmdTRUNCATE 4

/*** The seek command request structure ***/
typedef struct tagTSeekCmd {
  /*
    The command code. It must be set to one of the cmdSEEK, cmdTELL or
    cmdTRUNCATE calls.
  */
  upint8 Command;
  /*
    The lower 32 bits of the file position.
  */
  upint32 Position;
  /*
    The higher 32 bits of the file position.
  */
  upint32 Page;
  /*
    The direction specifier. See ~Directions
  */
  upint8 Direction;
} TSeekCmd;

typedef struct tagTTellReply {
  upint32 Position;
  upint32 Page;
} TTellReply;

/*** Directions: The valid direction specifiers ***/
/*
  The direction specifier must contain one of these values. If not, the SEEK
  command will fail with fseINVAL error.
*/

/*
  Move the pointer to the specified value.
*/
#define dirABSOLUTE 0
/*
  Move the pointer the specified number of bytes forward.
*/
#define dirFORWARD 1
/*
  Move the pointer the specified number of bytes backward.
*/
#define dirBACKWARD 2

/*** FSTAT: The FSTAT command ***/
/*
  This command returns a TFileInfo structure as the reply.
*/

#define cmdFSTAT 5

/*** Classic file I/O commands ***/
/*
  These commands don't communicate the pointer to the data or the buffer for
  data. They transfer the data instead. This method is very slow, because it
  copies each byte three times (from the buffer on the client side to the
  message queue, from the message queue to the buffer on the server side and
  from the buffer on the server side to the destination place) while the
  high performance file I/O peeks the data directly from the buffer on the
  client side to the destination place. Another source of the slowness is
  that the data to be written must be splitted into relatively small chunks
  (about 1.5K each) which imposes much more overhead as with the high
  performance file I/O (which writes the whole buffer using only one
  message).

  The advantages of these commands are that they work even in networked
  enviroment and that they are much more friendly to (currently planned)
  range checking feature of the P compiler.

  The cmdREADANDSEND initiates a file read. The server responds by burst of
  cmdDATA messages terminated by an empty message when there was no error and
  by an error message when there was an error.

  A file write is even simpler. After seeking to the desired location in the
  file (using cmdSEEK) simply send the data to be written using the cmdDATA
  command. After each such command the server responds by either a TIOReply
  structure telling the size of the data written into the file or an error
  message in case of failure.
*/

#define cmdREADANDSEND 6
#define cmdDATA 7

typedef struct tagTReadAndSendCmd {
  upint8 Command;
  pint32 Size;
} TReadAndSendCmd;

typedef struct tagTDataCmd {
  upint8 Command;
  upint8 Data[MAX_MESSAGE_SIZE-1];
} TDataCmd;

#define cmdFCHMOD 8
#define cmdFCHOWN 9
#define cmdFCHGRP 10

typedef struct tagTAttrManipCmd {
  /*
    The manipulation command to be carried out.
  */
  upint8 Command;
  /*
    The new value of the attribute
  */
  upint16 Value;
} TAttrManipCmd;

#endif
