#ifndef __h__SYSLIB_PDFSYS__
#define __h__SYSLIB_PDFSYS__

/*** Index: Master filesystem server protocol ***/
/*
  This protocol imitates the content of the old DISKIO module.
*/

/*** The list of error codes ***/
#define fseINVDRV  -0x0200   /* Drive not present */
#define fseEXIST   -0x0201   /* Filesystem object already exists */
#define fseISDIR   -0x0202   /* Filesystem object is a directory */
#define fseISLNK   -0x0203   /* Filesystem object is a symlink */
#define fseISFILE  -0x0204   /* Filesystem object is a file */
#define fseISDEV   -0x0205   /* Filesystem object is a device access point */
#define fseACCESS  -0x0206   /* Filesystem access denied */
#define fseNAMELEN -0x0207   /* File specification too long */
#define fseNOENT   -0x0208   /* Filesystem object does not exist */
#define fsePLNDIR  -0x0209   /* Path leads through non-directory */
#define fseNOIO    -0x020A   /* Device or pipe not ready */
#define fseROFS    -0x020B   /* Attempt to write on a read-only filesystem */
#define fsePRGUSE  -0x020C   /* Program file is in use */
#define fseLOOP    -0x020D   /* Too many symbolic link indirections */
#define fseNOSPC   -0x020E   /* No room on storage device */
#define fseFILELEN -0x020F   /* File too large */
#define fseNOTSUP  -0x0210   /* Feature not supported by the filesystem */
#define fseNOSEEK  -0x0211   /* Descriptor is not seekable */
#define fseBADNAME -0x0212   /* Invalid filename for the filesystem */
#define fseINVDLET -0x0213   /* Invalid drive specification */
#define fseRENADRV -0x0214   /* Cannot rename across drive */
#define fseNODEV   -0x0215   /* No such device */
#define fseCDATA   -0x0216   /* Filesystem control data broken */
#define fseLONGIO  -0x0217   /* File does not support long I/O */

/*** Maximal length of a file name ***/
/*
  This constant configures the maximal length of file names accepted by the
  directory reading calls. This is the filename length valid for the protocol
  as a whole. The actual limit for filename length depends on the filesystem
  the server is handling.
*/
#define MAX_FILENAME_SIZE 255

/*** Maximal length of a file specification ***/
#define MAX_FILESPEC_SIZE 700

/*** The directory item ***/
/*
  This structure denotes the format of data read from an opened directory.
  Each directory item is sent in a single message.
*/
typedef struct tagTDirItem {
  /*
    Contains the identification number of the file. This number may be used
    by the ~GET_FILE_INFO function to obtain remaining informations about a
    file.
  */
  upint32 ID;
  /*
    Contains the name of the file (without the path).
  */
  char Name[MAX_FILENAME_SIZE+1];
} TDirItem;

/*** The file information structure ***/
/*
  This structure is used by the STAT, FSTAT and ISTAT functions of the
  filesystem server.
*/
typedef struct tagTFileInfo {
  /*
    Time when the file was created. This is in UNIX format (seconds since
    1.1.1970 00:00:00 UTC).
  */
  upint64 CreatedTime;
  /*
    Time of the last modification of the file (in UNIX format).
  */
  upint64 ModifiedTime;
  /*
    ID of the user which owns the file.
  */
  upint16 Owner;
  /*
    ID of the group which owns the file.
  */
  upint16 Group;
  /*
    Description of the file. For FAT12, FAT16 or FAT32 here will be stored
    the long file name for the file. For JNFS here will be stored the file
    description itself. These file descriptions in OSHS were designed to be
    used as a place for notes about the content of the file.
  */
  char Description[256];
  /*
    Size of the file.
  */
  upint32 Size;
  /*
    File attributes. This field is bit oriented and the
    `FileAttributes file attribute constants` can be used to examine it. Also
    constants for possible `FileTypes file types` are provided.
  */
  upint16 Attr;
  /*
    The identifier of the file.
  */
  upint32 ID;
} TFileInfo;

/*** FileAttributes: File attribute constants ***/
/*
  These constants denote the layout of the file protection flags. Also the
  mask for the file type is provided. The valid file type constants are
  listed `FileTypes here`.
*/
#define fmUSRREAD  (1<< 0)
#define fmUSRWRITE (1<< 1)
#define fmUSREXEC  (1<< 2)
#define fmGRPREAD  (1<< 3)
#define fmGRPWRITE (1<< 4)
#define fmGRPEXEC  (1<< 5)
#define fmOTHREAD  (1<< 6)
#define fmOTHWRITE (1<< 7)
#define fmOTHEXEC  (1<< 8)
#define fmSETUID   (1<< 9)
#define fmSETGID   (1<<10)
#define fmSTICKY   (1<<11)
#define fmTYPE     (7<<12)

/*** FileTypes: File type constants ***/
/*
  These constants denote the types of files. They are designed so that they
  can be used directly with a file attributes value with protection bits
  masked out using the fmTYPE constant.
*/
#define fmREGULAR   (0<<12)
#define fmDIRECTORY (1<<12)
#define fmSYMLINK   (2<<12)
#define fmFIFO      (3<<12)
#define fmPKTFIFO   (6<<12)
#define fmCHARDEV   (4<<12)
#define fmBLOCKDEV  (5<<12)

/*** The OPEN commands ***/
#define cmdOPENFILE 1
#define cmdOPENDIR 2
#define cmdOPENLINK 3

/*** The open file command request ***/
typedef struct tagTOpenCmd {
  /*
    The command code.
  */
  upint8 Command;
  /*
    The file mode flags. See `OpenFlags this` for more info.
  */
  upint8 Mode;
  /*
    The specification of the file being opened.
  */
  char FileSpec[MAX_FILESPEC_SIZE+1];
} TOpenCmd;

/*** The open file command reply ***/
typedef struct tagTOpenReply {
  /*
    The handle of the file or to the fileserver. If the handle refers to the
    file, the message covers only this field. When this handle is a
    filesystem server handle, the message covers the remaining fields of
    this structure as well. The QUERY call returns a zero-length message if
    the file is local to the filesystem server.
  */
  upint16 Handle;
} TOpenReply;
/*
  NOTES

  - The server redirection message is formatted so that all that is required
    to retry the request is to pass the content of the returned message
    starting from the <EM>Command</EM> field to the server whose handle is
    returned in <EM>Handle</EM>.
*/

/*** OpenFlags: The open file flags ***/
/*
  These flags determine the behavior of the filesystem server in various
  situations.
*/

/*
  Action taken when the file is found. The <CODE>oeOPEN</CODE> directs the
  server to simply open the file and do not touch the data. The
  <CODE>oeTRUNC</CODE> will result in the file being truncated to the size
  of zero.
*/
#define oeOPEN (1<<5)
#define oeTRUNC (2<<5)
/*
  If this flag is set, the file will be created if not found. If this flag is
  clear, an error will be returned.
*/
#define onCREAT (1<<4)
/*
  When this flag is set, any writes to the file go directly into the disk.
  This increases the safety of the data but also has great impact to the
  performance.

  By default the data is buffered in memory and written at once when the 
  buffer is idle long enough (so called staged write). This default behavior
  also may be forced for all files by a server; in that case this bit is
  silently ignored.
*/
#define oSYNC (1<<3)
/*
  The access mode. The <CODE>oTEMP</CODE> value creates a temporary file.
  Temporary files have no name (the name part of the message has to be empty)
  and their content is discarded when the file is closed. The other values
  specify access mode to be used for a normal disk file.
*/
#define oTEMP 0
#define oRDONLY 1
#define oWRONLY 2
#define oRDWR (oRDONLY+oWRONLY)
/*
  Any attempt to write data onto the file will occur at the end of the file
  just like the SEEK request atomically occured before it.
*/
#define oAPPEND (1<<7)

/*** The CWD manipulation commands ***/
#define cmdCHDIR 4
#define cmdGETCWD 5

typedef struct {
  upint8 Command;
  upint8 Cwd[MAX_MESSAGE_SIZE-1];
} TChDirCmd;

typedef struct {
  upint8 Cwd[MAX_MESSAGE_SIZE];
} TGetCwdReply;

/*** Filesystem manipulation commands ***/
#define cmdMKDIR 8
#define cmdRMDIR 9
#define cmdREMOVE 10
#define cmdRENAME 11
#define cmdSTAT 12

typedef struct tagTManipCmd {
  /*
    The manipulation command to be carried out.
  */
  upint8 Command;
  /*
    The specification of the filesystem object to be affected
  */
  char FileSpec[2*(MAX_FILESPEC_SIZE+1)];
} TManipCmd;

#define cmdISTAT 16

typedef struct tagTIStatCmd {
  /*
    The command code.
  */
  upint8 Command;
  /*
    The identifier of the attribute
  */
  upint32 ID;
} TIStatCmd;

#define cmdFIRSTDRV 16
#define cmdNEXTDRV 17
#define cmdGETDRV 18
#define cmdSETDRV 19

typedef struct tagTSetDrvCmd {
  upint8 Command;
  char Name;
} TSetDrvCmd;

typedef struct tagTDriveInfo {
  char Name;
  uint8 Type;
} TDriveInfo;

#define dtEMPTY   0
#define dtMSDOS   1
#define dtWINDOWS 2
#define dtUNIX    3

#endif
