#ifndef __h__SYSLIB_PDGFXCON__
#define __h__SYSLIB_PDGFXCON__

/*
  The screen is accessed through the device named "GFXCON". It is a packet
  pipe device. Once opened, the user can prepare command packets by filling
  the structures provided below and send them into the device using the Write
  command. Once the command is completed the Read command must be used to
  read the answer packet. If there is an answer packet pending and another
  command is issued, the pending answer packet is discarded.
*/

/*
  Error codes for the screen
*/
#define screBADSIZE	-0x0301	/* The command packet is too short */
#define screBADFLAGS	-0x0302	/* Reserved flags set */
#define screBADPXSIZE	-0x0303	/* Invalid pixel size */
#define screSCRINIT	-0x0304	/* Screen hardware initialization failure */
#define screMODESUP	-0x0305	/* Screen mode not supported */
#define screNOTSUP	-0x0306	/* Command not supported by the driver */
#define screKBDINIT	-0x0307	/* Keyboard hardware initialization failure */
#define screNOMODESET	-0x0308	/* No screen mode set */
#define screSCRHWFAIL	-0x0309	/* Screen hardware failure */
#define screREGWRT      -0x030A /* A region is already being written */
#define screBPXID       -0x030B /* Bad pixmap ID */
#define screINCINS	-0x030C /* Incomplete instruction in code */
#define screINVINS	-0x030D /* Invalid instruction in code */
#define screMRECTOUT	-0x030E /* Mount rectangle outside window */
#define screSRECTOUT	-0x030F /* Source rectangle outside boundaries */
#define screDRECTOUT	-0x0310 /* Dest rectangle outside boundaries */
#define screMOVERLAP	-0x0311 /* Mount rectangle overlaps another */
#define screNOTMNT	-0x0312 /* Attempt to unmount non-mounted pixmap */
#define screPARTPIX	-0x0313 /* Partial pixel in pixmap input data */

/*
  The mouse event definition
*/
typedef struct {
  pint32 Time;
  pint32 DeltaX;
  pint32 DeltaY;
  upint8 Buttons;
} TMouseEvent;

#define mLeftButton 1
#define mRightButton 2
#define mMiddleButton 4

/*
  Screen mode manipulation commands

  The SetMode command sets the specified screen mode. The ProbeMode
  determines if the screen is capable to set the specified mode.
*/

#define scrcmdSetMode	128
#define scrcmdProbeMode	129
/*
  Flags (put them into the "Flags" piece)
*/
#define scrflgUpdateOnDemand	128
#define scrflgGraphics		64
#define scrflgReserved		(32+16+8+4+2+1)
/*
  Pixel sizes (put them into the "PixelSize" piece)
*/
#define scrpx8bit	0
#define scrpx16bit	1
#define scrpx24bit	2
#define scrpx32bit	3
#define scrpxInvalid	4

typedef struct {
  upint8 Command;
  upint8 Flags;
  upint8 PixelSize;
  upint16 Height;
  upint16 Width;
} TSetModeCommand;

/*
  Keyboard handling commands

  The RequestKeyboard command is used to obtain a handle from which the
  keyboard keystrokes will be coming. This handle uses the PDKEYS protocol.
  A screen mode must be set before this call can be issued.
*/

#define scrcmdOpenKeyboard	130

typedef struct {
  upint8 Command;
} TOpenKeyboardCommand;

typedef struct {
  pint16 Handle;
} TOpenKeyboardReply;

/*
  Screen update command

  In update-on-demand video modes (the flag "scrflgUpdateOnDemand" in the
  video request was set), you need to call this to make the changes on the
  screen visible. Update-on-demand video modes are useful to avoid flickering
  and similar unwanted effects while instant-update video modes are useful
  during debugging to see what exactly is on the screen after a video change
  command. This command takes no parameters and returns an empty message on
  success. It returns the empty message even on instant-update video mode to
  make the two types of video modes interchangeable from the program's view.
*/
#define scrcmdUpdateScreen	131

/*
  Open screen region command

  This command is used to prepare a screen region for writing.
*/
#define scrcmdOpenScreenRegion	132

typedef struct {
  upint8 Command;
  upint16 X;
  upint16 Y;
  upint16 Width;
  upint16 Height;
} TOpenScreenRegionCommand;

/*
  Create pixmap for screen output

  This command creates a pixmap suitable for screen output and associates it
  with the specified ID. The pixmap is guaranteed to be mountable on the
  screen.
*/
#define scrcmdCreateOutPixmap	134

typedef struct {
  upint8 Command;
  upint16 ID;
  upint16 Width;
  upint16 Height;
} TCreateOutPixmapCommand;

/*
  Create pixmap

  This command creates a pixmap and associates it with the specified ID.
*/
#define scrcmdCreatePixmap	135

typedef struct {
  upint8 Command;
  upint16 ID;
  upint16 Width;
  upint16 Height;
  upint8 Flags;
} TCreatePixmapCommand;

#define bfEnableTransparent 1

/*
  Execute drawing sequence

  This command transfers a bunch of drawing commands to the driver.
*/
#define scrcmdDraw		137

typedef struct {
  upint8 Command;
  upint8 Code[MAX_MESSAGE_SIZE-1];
} TDrawCommand;

#define drwinsTransferRectangle		0
#define drwinsMountPixmap		0xF8
#define drwinsUnmountPixmap		0xF9

/*
  Destroy pixmaps

  Destroys the pixmaps associated with the specified IDs and frees them for
  reuse. Mustiple IDs can be passed to this command and their count is
  reflected in the size of the message.
*/
#define scrcmdDestroyPixmaps	138

typedef struct {
  upint8 Command;
  upint16 IDList[(MAX_MESSAGE_SIZE-1)/sizeof(upint16)];
} TDestroyPixmapsCommand;

#endif
