#include "defs.gh"

#include "types.h"
#include "traps.h"
#include "conio.h"
#include "strings.h"
#include "kernel.h"

#define __INTERFACE__
#include "thread.h"
#undef __INTERFACE__

#if __e__THREAD_DRV__ != __v__none__
#include "thrmain.h"
#endif

#define MAX_THREADS 1024
#define MAX_LOCKS 1024
#define MAX_CONDS 1024

#if __e__THREAD_DRV__ != __v__none__
typedef struct {
  void *Lock;
  bool Locked;
} TLock;

static int16 ThreadCnt;
static void *Threads[MAX_THREADS];
static void *ThreadLock;
static void *NewThrLock;
static TLock Locks[MAX_LOCKS];
static void *MasterLock;
static void *Conds[MAX_CONDS];
static void *CondLock;

static void LockMaster(void)
{
  ThrRawLock(MasterLock);
}

static void UnlockMaster(void)
{
  ThrRawUnlock(MasterLock);
}

static void LockConds(void)
{
  ThrRawLock(CondLock);
}

static void UnlockConds(void)
{
  ThrRawUnlock(CondLock);
}

static void LockThreads(void)
{
  ThrRawLock(ThreadLock);
}

static void UnlockThreads(void)
{
  ThrRawUnlock(ThreadLock);
}

typedef struct {
  int16 Handle;
  TMainFunction *Main;
  void *Data;
} TStartupData;

static void FreeHandle(int16 Handle)
{
  LockThreads();
  ThrRawDestroyHandle(Threads[Handle]);
  Threads[Handle]=NULL;
  UnlockThreads();
}

static int16 StartProc(void *Info)
{
  TStartupData *I;
  int16 Handle;
  TMainFunction *Main;
  void *Data;
  int16 Code;

  I=Info;
  Handle=I->Handle;
  Main=I->Main;
  Data=I->Data;
  ThrRawUnlock(NewThrLock);
  UnlockThreads();
  Code=Main(Data);
  FreeHandle(Handle);
  return(Code);
}

static int16 CheckLockValid(int32 LockID)
{
  bool Invalid;

  Invalid=(LockID<0) || (LockID>=MAX_LOCKS);
  if (!Invalid) Invalid=(Locks[LockID].Lock==NULL);
  if (Invalid) return(keBADHND);
  return(0);
}

static int16 CheckCondValid(int32 CondID)
{
  bool Invalid;

  Invalid=(CondID<0) || (CondID>=MAX_CONDS);
  if (!Invalid) Invalid=(Conds[CondID]==NULL);
  if (Invalid) return(keBADHND);
  return(0);
}
#endif

void SYSInitThreads(void)
{
#if __e__THREAD_DRV__ != __v__none__
  MasterLock=ThrRawCreateLock();
  if (MasterLock==NULL) ThrRawReportError("Failed to create master lock");
  CondLock=ThrRawCreateLock();
  if (CondLock==NULL) ThrRawReportError("Failed to create conditions lock");
  ThreadLock=ThrRawCreateLock();
  if (ThreadLock==NULL) ThrRawReportError("Failed to create thread table lock");
  NewThrLock=ThrRawCreateLock();
  if (NewThrLock==NULL) ThrRawReportError("Failed to create new thread lock");
  ThrRawInitThreads();
#endif
}

int16 SYSCreateNewThread(TMainFunction *MainFnc,void *Data)
{
#if __e__THREAD_DRV__ == __v__none__
  if (MainFnc==NULL);
  if (Data==NULL);
  return(keNOTSUP);
#else
  void *Thread;
  int16 ThreadID;
  TStartupData Info;

  /*
    Allocate a slot in the thread table. The thread table is employed to hide
    the details of the implementation of this module from the users. 
  */
  LockThreads();
  ThreadID=0;
  while (ThreadID<ThreadCnt) {
    if (Threads[ThreadID]==NULL) break;
    ThreadID++;
  }
  if (ThreadID==ThreadCnt) {
    if (ThreadCnt==MAX_THREADS) return(keMTHREAD);
    ThreadCnt++;
  }
  /*
    Try to create the thread
  */
  Info.Handle=ThreadID;
  Info.Main=MainFnc;
  Info.Data=Data;
  ThrRawLock(NewThrLock);
  Thread=ThrRawCreate(StartProc,&Info);
  Threads[ThreadID]=Thread;
  if (Thread==NULL) {
    ThrRawUnlock(NewThrLock);
    UnlockThreads();
    return(keNOMEM);
  }
  ThrRawLock(NewThrLock);
  ThrRawUnlock(NewThrLock);
  /*
    Put the thread number into the thread table and return its number in the
    table.
  */
  Threads[ThreadID]=Thread;
  return(ThreadID);
#endif
}

int16 SYSKillThread(int16 ThreadID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (ThreadID==0);
  return(keNOTSUP);
#else
  ThrRawKill(Threads[ThreadID]);
  FreeHandle(ThreadID);
  return(0);
#endif
}

int16 SYSWaitThread(int16 ThreadID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (ThreadID==0);
  return(keNOTSUP);
#else
  void *Thread;
  int16 Status;

  LockThreads();
  Thread=Threads[ThreadID];
  UnlockThreads();
  if (Thread==NULL) {
    return(keBADHND);
  } else {
    Status=ThrRawWait(Threads[ThreadID]);
    FreeHandle(ThreadID);
  }
  return(Status);
#endif
}

int32 SYSCreateLock(void)
{
#if __e__THREAD_DRV__ == __v__none__
  return(keNOTSUP);
#else
  void *Lock;
  int32 LockNum;
  TLock *LockH;

  /*
    Lock the lost of locks
  */
  LockMaster();
  /*
    Determine the number for the new lock. If the lock table is full, abort.
    Remember to unlock the list in that case.
  */
  LockNum=0;
  while ((LockNum<MAX_LOCKS) && (Locks[LockNum].Lock)) LockNum++;
  if (LockNum==MAX_LOCKS) {
    UnlockMaster();
    return(keNOHND);
  }
  /*
    Create the lock. Also remember to unlock the list in the case of any
    failure.
  */
  Lock=ThrRawCreateLock();
  if (Lock==NULL) {
    UnlockMaster();
    return(keNOMEM);
  }
  /*
    Fill out the table item
  */
  LockH=Locks+LockNum;
  LockH->Lock=Lock;
  LockH->Locked=false;
  /*
    Unlock the list of locks
  */
  UnlockMaster();
  /*
    Return the lock ID
  */
  return(LockNum);
#endif
}

int16 SYSLock(int32 LockID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (LockID==0) return(keNOTSUP);
  return(keNOTSUP);
#else
  TLock *LockH;
  int16 Error;

  /*
    Lock the list of locks
  */
  LockMaster();
  /*
    Check if the lock number is valid. If not, abort the program
  */
  Error=CheckLockValid(LockID);
  if (Error!=0) {
    UnlockMaster();
    return(Error);
  }
  LockH=Locks+LockID;
  /*
    Set the lock locking flag to locked
  */
  LockH->Locked=true;
  /*
    Unlock the master lock. It must be done here, because the next command
    may block the thread and if the master lock would be locked, the program
    would get deadlock (also Unlock() locks the master lock).
  */
  UnlockMaster();
  /*
    Lock the semaphore
  */
  ThrRawLock(LockH->Lock);
  /*
    Return success
  */
  return(0);
#endif
}

int16 SYSUnlock(int32 LockID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (LockID==0) return(keNOTSUP);
  return(keNOTSUP);
#else
  TLock *LockH;
  int16 Error;

  /*
    Lock the master lock
  */
  LockMaster();
  /*
    Check if the lock number is valid. If not, abort the program
  */
  Error=CheckLockValid(LockID);
  if (Error!=0) {
    UnlockMaster();
    return(Error);
  }
  LockH=Locks+LockID;
  /*
    There is no fear from blocked thread here; this call cannot block the
    thread.
  */
  ThrRawUnlock(LockH->Lock);
  /*
    Set the lock flag to unlocked
  */
  LockH->Locked=false;
  /*
    Unlock the master lock
  */
  UnlockMaster();
  /*
    Return success
  */
  return(0);
#endif
}

int16 SYSDestroyLock(int32 LockID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (LockID==0) return(keNOTSUP);
  return(keNOTSUP);
#else
  TLock *LockH;
  int16 Error;

  LockMaster();
  Error=CheckLockValid(LockID);
  if (Error!=0) {
    UnlockMaster();
    return(Error);
  }
  LockH=Locks+LockID;
  if (LockH->Locked) ReportBug("Attempt to destroy locked lock");
  ThrRawDestroyLock(LockH->Lock);
  LockH->Lock=NULL;
  UnlockMaster();
  return(0);
#endif
}

int32 SYSCreateCond(void)
{
#if __e__THREAD_DRV__ == __v__none__
  return(keNOTSUP);
#else
  void *Cond;
  int32 CondNum;

  /*
    Lock the list of events
  */
  LockConds();
  /*
    Determine the number for the new event. If the event table is full,
    abort. Remember to unlock the list in that case.
  */
  CondNum=0;
  while ((CondNum<MAX_LOCKS) && (Conds[CondNum])) CondNum++;
  if (CondNum==MAX_LOCKS) {
    UnlockMaster();
    return(keNOHND);
  }
  /*
    Create the event. Also remember to unlock the list in the case of any
    failure.
  */
  Cond=ThrRawCreateCond();
  if (Cond==NULL) {
    UnlockConds();
    return(keNOMEM);
  }
  /*
    Fill out the table item
  */
  Conds[CondNum]=Cond;
  /*
    Unlock the list of events
  */
  UnlockConds();
  /*
    Return the event ID
  */
  return(CondNum);
#endif
}

int16 SYSSetCond(int32 CondID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (CondID==0) return(keNOTSUP);
  return(keNOTSUP);
#else
  int16 Error;

  /*
    Lock the list of events to avoid other threads interfering with our
    events, set the event and unlock it. If the event handle passed to this
    call is invalid, abort. This call never blocks.
  */
  LockConds();
  Error=CheckCondValid(CondID);
  if (Error==0) ThrRawSetCond(Conds[CondID]);
  UnlockConds();
  return(Error);
#endif
}

int16 SYSResetCond(int32 CondID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (CondID==0);
  return(keNOTSUP);
#else
  int16 Error;

  /*
    The same as the above but the event is reset this time.
  */
  LockConds();
  Error=CheckCondValid(CondID);
  if (Error==0) ThrRawResetCond(Conds[CondID]);
  UnlockConds();
  return(Error);
#endif
}

int16 SYSWaitCond(int32 CondID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (CondID==0);
  return(keNOTSUP);
#else
  void *CPtr;
  int16 Error;

  /*
    This call may sleep so we need to take proper precautions for it. The
    list of event is unlocked before the sleepy call is made.
  */
  LockConds();
  Error=CheckCondValid(CondID);
  if (Error==0) CPtr=Conds[CondID];
  UnlockConds();
  if (Error==0) ThrRawWaitCond(CPtr);
  return(Error);
#endif
}

int16 SYSDestroyCond(int32 CondID)
{
#if __e__THREAD_DRV__ == __v__none__
  if (CondID==0);
  return(keNOTSUP);
#else
  void *CPtr;
  int16 Error;

  /*
    Wake up any threads waiting on that event and destroy it.
  */
  LockConds();
  Error=CheckCondValid(CondID);
  if (Error==0) {
    CPtr=Conds[CondID];
    ThrRawSetCond(CPtr);
    ThrRawDestroyCond(CPtr);
  }
  UnlockConds();
  return(Error);
#endif
}
