/*
  thread low level module: using pthread library directly

  pthread library should be available in most GNU/Linux platforms and
  eventually other POSIX platforms.
*/

#include "defs.gh"

#include <stdio.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <string.h>

#include "types.h"
#include "traps.h"
#include "kernel.h"
#include "pdbase.h"
#include "strings.h"
#include "thread.h"
#include "memory.h"

#define __INTERFACE__
#include "thrpthr.h"
#undef __INTERFACE__

#include "typesnd.h"

void ThrReportError(const char *Msg)
{
  char Buffer[1024];
  char *Ptr;

  Ptr=StrCpy(Buffer,Msg);
  Ptr=StrCpy(Ptr,"\nPTHREADS said: ");
  if (strerror_r(errno,Ptr,1023-(Ptr-Buffer))) {
    StrCpy(Ptr,"Unknown error");
  }
  Buffer[1023]='\0';
  Fatal(Buffer);
}

/**********
** Locks **
**********/

static void InitializeMutex(pthread_mutex_t *Mutex)
{
  if (pthread_mutex_init(Mutex,NULL)!=0) {
    ThrReportError(
      "pthread_mutex_init() failed but POSIX does not allow that"
    );
  }
}

static void LockMutex(pthread_mutex_t *Mutex)
{
  if (pthread_mutex_lock(Mutex)!=0) {
    ThrReportError("pthread_mutex_lock() failed");
  }
}

static void UnlockMutex(pthread_mutex_t *Mutex)
{
  if (pthread_mutex_unlock(Mutex)!=0) {
    ThrReportError("pthread_mutex_unlock() failed");
  }
}

static void DestroyMutex(pthread_mutex_t *Mutex)
{
  if (pthread_mutex_destroy(Mutex)!=0) {
    ThrReportError("pthread_mutex_destroy() failed");
  }
}

/*
  There are two implementations of OSHS locks using pthread library. One
  uses traditional mutexes and the other one uses rwlocks.

  Mutexes are probably faster than using rwlocks but manuals states that it
  is nonportable. More specifically, OSHS allows a lock to be unlocked by
  thread other than the one that owns the lock and this is stated as a
  nonportable behavior of the mutexes (when the OSHS locks are implemented
  as mutexes).
*/

typedef struct {
#if __b__USE_RWLOCK__
  pthread_rwlock_t Lock;
#else
  pthread_mutex_t Mutex;
#endif
} TLock;

void *ThrCreateLock(void)
{
  TLock *L;

  L=KMEM__malloc(sizeof(*L));
  if (L!=NULL) {
#if __b__USE_RWLOCK__
    if (pthread_rwlock_init(&L->Lock,NULL)!=0) {
      KMEM__free(L);
      L=NULL;
    }
#else
    InitializeMutex(&L->Mutex);
#endif
  }
  return(L);
}

void ThrDestroyLock(void *Lock)
{
  TLock *L;

  L=Lock;
#if __b__USE_RWLOCK__
  if (pthread_rwlock_destroy(&L->Mutex)!=0) {
    ThrReportError("pthread_rwlock_destroy() failed");
  }
#else
  DestroyMutex(&L->Mutex);
#endif
  KMEM__free(Lock);
}

void ThrLock(void *Lock)
{
  TLock *L;

  L=Lock;
#if __b__USE_RWLOCK__
  if (pthread_rwlock_wrlock(&L->Lock)!=0) {
    ThrReportError("pthread_rwlock_wrlock() failed");
  }
#else
  LockMutex(&L->Mutex);
#endif
}

void ThrUnlock(void *Lock)
{
  TLock *L;

  L=Lock;
#if __b__USE_RWLOCK__
  if (pthread_rwlock_unlock(&L->Lock)!=0) {
    ThrReportError("pthread_rwlock_unlock() failed");
  }
#else
  UnlockMutex(&L->Mutex);
#endif
}

/**************************
** Conditional variables **
**************************/

typedef struct {
  /*
    The value of the conditional. If it is "false", ThrWaitCond() will
    suspend the calling thread until someone sets it to "true". If it is
    "true", ThrWaitCond() will do nothing and returns immediately.
  */
  bool Value;
  /*
    Implementation specific workspace. This is for the pthread
    implementation. The Cond is the pthread conditional and Mutex is a
    pthread lock. They both act together, because pthread semantics of a
    conditional variable is that it has no value and the waiting call always
    waits until a signal from other thread is executed.
  */
  pthread_cond_t Cond;
  pthread_mutex_t Mutex;
} TCond;

static bool InitializeCond(TCond *C)
{
  C->Value=false;
  InitializeMutex(&C->Mutex);
  return(pthread_cond_init(&C->Cond,NULL)==0);
}

void *ThrCreateCond(void)
{
  TCond *C;

  C=KMEM__malloc(sizeof(*C));
  if (C!=NULL) {
    if (!InitializeCond(C)) {
      KMEM__free(C);
      C=NULL;
    }
  }
  return(C);
}

void ThrSetCond(void *VV)
{
  TCond *C;

  C=VV;
  LockMutex(&C->Mutex);
  if (!C->Value) {
    C->Value=true;
    if (pthread_cond_broadcast(&C->Cond)!=0) {
      ThrReportError("pthread_cond_broadcast() failed");
    }
  }
  UnlockMutex(&C->Mutex);
}

void ThrResetCond(void *VV)
{
  TCond *C;

  C=VV;
  LockMutex(&C->Mutex);
  C->Value=false;
  UnlockMutex(&C->Mutex);
}

void ThrWaitCond(void *VV)
{
  TCond *C;

  C=VV;
  LockMutex(&C->Mutex);
  if (!C->Value) {
    if (pthread_cond_wait(&C->Cond,&C->Mutex)!=0) {
      ThrReportError("pthread_cond_wait() failed");
    }
  }
  UnlockMutex(&C->Mutex);
}

static void DestroyCondContent(TCond *C)
{
  DestroyMutex(&C->Mutex);
  if (pthread_cond_destroy(&C->Cond)!=0) {
    ThrReportError("pthread_cond_destroy() failed");
  }
}

void ThrDestroyCond(void *VV)
{
  DestroyCondContent(VV);
  KMEM__free(VV);
}

/************
** Threads **
************/

typedef struct {
  pthread_t ID;
} TThreadHandle;

typedef struct {
  TCond Cond;
  TMainFunction *Main;
  void *Arg;
} TStartupInfo;

void ThrInitThreads(void)
{
}

void *RunThread(void *Data)
{
  TStartupInfo *I;
  TMainFunction *Main;
  void *Arg;
  int16 Result;

#if __b__THR_USE_CANCEL__
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
#endif
  I=Data;
  Main=I->Main;
  Arg=I->Arg;
  ThrSetCond(&I->Cond);
  Result=Main(Arg);
  return((void *)Result);
}

void *ThrCreate(TMainFunction *MainFnc,void *Data)
{
  TThreadHandle *H;
  TStartupInfo I;

  if ( sizeof(pthread_t)!=sizeof(void *) ) {
    Fatal("Threads not supported on this platform");
  }
  if (!InitializeCond(&I.Cond)) return(NULL);
  H=KMEM__malloc(sizeof(*H));
  if (H!=NULL) {
    I.Main=MainFnc;
    I.Arg=Data;
    if ( pthread_create(&H->ID, NULL, RunThread, &I) != 0 ) {
      ThrDestroyHandle(H);
      H=NULL;
    }
    ThrWaitCond(&I.Cond);
    DestroyCondContent(&I.Cond);
  }
  return (H);
}

int16 ThrWait(void *Thread)
{
  void *Result;
  TThreadHandle *H;

  H=Thread;
  if (pthread_join(H->ID,&Result)) return(keNOTHR);
  return((int16)Result);
}

bool ThrKill(void *Thread)
{
  TThreadHandle *H;

  H=Thread;
#if __b__THR_USE_CANCEL__
  return(!pthread_cancel(H->ID));
#else
  return(!pthread_kill(H->ID,SIGKILL));
#endif
}

void ThrDestroyHandle(void *Handle)
{
  KMEM__free(Handle);
}
