#include "defs.gh"

#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>

#include "types.h"
#include "traps.h"
#include "strings.h"
#include "kernel.h"
#include "thread.h"
#include "memory.h"

#include "typesnd.h"

void ThrReportError(const char *Msg)
{
  char Buffer[1024];
  char *Ptr;

  Ptr=StrCpy(Buffer,Msg);
  Ptr=StrCpy(Ptr,"\nSDL said: ");
  StrCpy(Ptr,SDL_GetError());
  Fatal(Buffer);
}

#ifdef __DEBUG__
static void ThrTrapUndefined(const char *Msg)
{
  char Buffer[1024];
  char *Ptr;

  Ptr=StrCpy(Buffer,Msg);
  Ptr=StrCpy(Ptr,"\nSDL said: ");
  StrCpy(Ptr,SDL_GetError());
  TrapUndefined(Buffer);
}
#endif

void *ThrCreateLock(void)
{
  return(SDL_CreateSemaphore(1));
}

void ThrDestroyLock(void *Lock)
{
  SDL_DestroySemaphore(Lock);
}

void ThrLock(void *Lock)
{
  if (SDL_SemWait(Lock)) {
#ifdef __DEBUG__
    ThrTrapUndefined("ThrLock() failed");
#endif
  }
}

void ThrUnlock(void *Lock)
{
  if (SDL_SemPost(Lock)) {
#ifdef __DEBUG__
    ThrTrapUndefined("ThrUnlock() failed");
#endif
  }
}

void *ThrCreate(TMainFunction *MainFnc,void *Data)
{
  return(SDL_CreateThread(MainFnc,Data));
}

int16 ThrWait(void *Thread)
{
  int16 Status;

  SDL_WaitThread(Thread,&Status);
  return(Status);
}

bool ThrKill(void *Thread)
{
  SDL_KillThread(Thread);
  return(true);
}

void ThrDestroyHandle(void *Handle)
{
  if (Handle==NULL);
}

typedef struct {
  SDL_cond *Cond;
  SDL_mutex *Mutex;
  bool State;
  uint16 RefCount;
} TCondVar;

void *CondLock;

static void LockCondVars(TCondVar *V)
{
  ThrLock(CondLock);
  V->RefCount++;
}

static void UnlockCondVars(TCondVar *V)
{
  V->RefCount--;
  if (V->RefCount==0) {
    SDL_DestroyCond(V->Cond);
    SDL_DestroyMutex(V->Mutex);
    KMEM__free(V);
  }
  ThrUnlock(CondLock);
}

void *ThrCreateCond(void)
{
  TCondVar *V;

  V=KMEM__malloc(sizeof(*V));
  if (V!=NULL) {
    V->Cond=SDL_CreateCond();
    V->Mutex=SDL_CreateMutex();
    if (V->Cond==NULL || V->Mutex==NULL) {
      if (V->Cond!=NULL) SDL_DestroyCond(V->Cond);
      if (V->Mutex!=NULL) SDL_DestroyMutex(V->Mutex);
      KMEM__free(V);
      V=NULL;
    } else {
      V->State=false;
      V->RefCount=1;
    }
  }
  return(V);
}

void ThrSetCond(void *VV)
{
  TCondVar *V;

  V=VV;
  LockCondVars(V);
  V->State=true;
  if (SDL_CondBroadcast(V->Cond)==-1) {
#ifdef __DEBUG__
    ThrTrapUndefined("ThrSetCond() failed");
#endif
  }
  UnlockCondVars(V);
}

void ThrResetCond(void *VV)
{
  TCondVar *V;

  V=VV;
  LockCondVars(V);
  V->State=false;
  UnlockCondVars(V);
}

void ThrWaitCond(void *VV)
{
  TCondVar *V;
  bool Wait;

  V=VV;
  LockCondVars(V);
  Wait=!V->State;
  if (Wait) {;
    ThrUnlock(CondLock);
    SDL_mutexP(V->Mutex);
    if (SDL_CondWait(V->Cond,V->Mutex)==-1) {
#ifdef __DEBUG__
      ThrTrapUndefined("ThrWaitCond() failed");
#endif
    }
    SDL_mutexV(V->Mutex);
    ThrLock(CondLock);
  }
  UnlockCondVars(V);
}

void ThrDestroyCond(void *VV)
{
  TCondVar *V;

  V=VV;
  LockCondVars(V);
  V->RefCount--;
  UnlockCondVars(V);
}

void ThrInitThreads(void)
{
  CondLock=ThrCreateLock();
  if (CondLock==NULL) Fatal("Unable to create lock for cond variables");
}
