#include "defs.gh"

#include <windows.h>

#include "types.h"
#include "traps.h"
#include "memory.h"
#include "strings.h"
#include "kernel.h"
#include "thread.h"

#define __INTERFACE__
#include "thrwin.h"
#undef __INTERFACE__

#include "typesnd.h"

#define ERROR_BUFFER_SIZE 1024

static void FormatError(const char *Msg,char *Buffer)
{
  DWORD Error;
  DWORD WrittenChars;
  char *Ptr;

  Error=GetLastError();
  Ptr=StrCpy(Buffer,Msg);
  Ptr=StrCpy(Ptr,"\nWindows said: ");
  WrittenChars=FormatMessage(
    FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,Error,0,Ptr,ERROR_BUFFER_SIZE-(Ptr-Buffer),NULL
  );
}

void ThrReportError(const char *Msg)
{
  char Buffer[ERROR_BUFFER_SIZE];

  FormatError(Msg,Buffer);
  Fatal(Buffer);
}

#ifdef __DEBUG__
static void ThrTrapUndefined(const char *Msg)
{
  char Buffer[ERROR_BUFFER_SIZE];

  FormatError(Msg,Buffer);
  TrapUndefined(Buffer);
}
#endif

void *ThrCreateLock(void)
{
  return(CreateSemaphore(NULL,1,1,NULL));
}

void ThrDestroyLock(void *Lock)
{
  CloseHandle(Lock);
}

void ThrLock(void *Lock)
{
  if (WaitForSingleObject(Lock,INFINITE)!=WAIT_OBJECT_0) {
#ifdef __DEBUG__
    ThrTrapUndefined("ThrLock() failed");
#endif
  }
}

void ThrUnlock(void *Lock)
{
  if (!ReleaseSemaphore(Lock,1,NULL)!=WAIT_OBJECT_0) {
#ifdef __DEBUG__
    ThrTrapUndefined("ThrUnlock() failed");
#endif
  }
}

void *ThrCreateThreadCore(void *MainFnc,void *Data)
{
  DWORD ThreadID;

  return(CreateThread(
    NULL,
    0,
    MainFnc,
    Data,
    0,
    &ThreadID
  ));
}

static HANDLE NewThreadLock;

typedef struct {
  TMainFunction *Main;
  void *Data;
} TStartupInfo;

static DWORD WINAPI RunThread(void *Pointer)
{
  TStartupInfo *Info;
  TMainFunction *Main;
  void *Data;

  Info=Pointer;
  Main=Info->Main;
  Data=Info->Data;
  ThrUnlock(NewThreadLock);
  return(Main(Data));
}

void *ThrCreate(TMainFunction *MainFnc,void *Data)
{
  TStartupInfo Info;
  void *Result;

  if (NewThreadLock==NULL) {
    NewThreadLock=ThrCreateLock();
    if (NewThreadLock==NULL) return(NULL);
  }
  ThrLock(NewThreadLock);
  Info.Main=MainFnc;
  Info.Data=Data;
  Result=ThrCreateThreadCore(&RunThread,&Info);
  ThrLock(NewThreadLock);
  ThrUnlock(NewThreadLock);
  return(Result);
}

int16 ThrWait(void *Thread)
{
  DWORD ExitCode;

  WaitForSingleObject(Thread,INFINITE);
  if (!GetExitCodeThread(Thread,&ExitCode)) ExitCode=-1;
  return(ExitCode);
}

bool ThrKill(void *Thread)
{
  /*
    KillThread() not implemented under Windows
  */
  return(false);
}

void ThrDestroyHandle(void *Handle)
{
  CloseHandle(Handle);
}

void *ThrCreateCond(void)
{
  return(CreateEvent(NULL,TRUE,FALSE,NULL));
}

void ThrSetCond(void *VV)
{
  if (SetEvent(VV)==0) {
    ThrReportError("SetEvent() failed");
  }
}

void ThrResetCond(void *VV)
{
  if (ResetEvent(VV)==0) {
    ThrReportError("ResetEvent() failed");
  }
}

void ThrWaitCond(void *VV)
{
  if (WaitForSingleObject(VV,INFINITE)==WAIT_FAILED) {
    ThrReportError("WaitForSingleObject() failed");
  }
}

void ThrDestroyCond(void *VV)
{
  if (CloseHandle(VV)==0) {
    ThrReportError("CloseHandle() failed");
  }
}

void ThrInitThreads(void)
{
}
