/*
  Test: Conditional variables test

  A conditional variable is a boolean variable capable of suspending threads
  until it becomes true. This test checks whether its behavior obeys the
  specification.
*/

#include "types.h"
#include "kernel.h"
#include "thread.h"
#include "conio.h"
#include "process.h"
#include "time.h"

int16 TestCond;

int16 CounterLock;

int16 Counter;

void Abort(const char *Msg)
{
  PrintSTDERR(Msg);
  PrintSTDERR("\n");
  PerformProgramHalting(255);
}

/*
  This is used as an brake. When the main test deadlocks, this thread awakens
  and performs shutdown. This avoids infinite execution of the test.
*/
int16 EmergencyBrake(void *Data)
{
  if (Data==NULL);
  DelayThread(2000000);
  Abort("Emergency brake engaged");
  return(0);
}

/*
  This thread waits on the conditional variable and then increments the
  counter.
*/
int16 Waiter(void *Data)
{
  if (Data==NULL);
  SYSWaitCond(TestCond);
  SYSLock(CounterLock);
  Counter++;
  SYSUnlock(CounterLock);
  return(0);
}

#define TESTING_THREADS 10

void CreateWaiters(void)
{
  int16 i;

  SYSLock(CounterLock);
  Counter=0;
  SYSUnlock(CounterLock);
  for (i=0;i<TESTING_THREADS;i++) {
    if (SYSCreateNewThread(Waiter,NULL)<0) {
      Abort("Unable to create testing thread");
    }
  }
}

void TestThreadsOut(void)
{
  SYSLock(CounterLock);
  if (Counter!=TESTING_THREADS) {
    Abort("Some threads blocked on true conditional");
  }
  SYSUnlock(CounterLock);
}

void TestConditionalFalse(void)
{
  CreateWaiters();
  DelayThread(100000);
  SYSLock(CounterLock);
  if (Counter!=0) Abort("Some threads sneaked past false conditional");
  SYSUnlock(CounterLock);
  SYSSetCond(TestCond);
  DelayThread(100000);
  TestThreadsOut();
}

void TestConditionalTrue(void)
{
  CreateWaiters();
  DelayThread(100000);
  TestThreadsOut();
}

int16 Main(void)
{
  /*
    Initialize the testing enviroment
  */
  if (SYSCreateNewThread(EmergencyBrake,NULL)<0) {
    Abort("Unable to create emergency brake");
  }
  TestCond=SYSCreateCond();
  if (TestCond<0) Abort("Unable to create test condition variable");
  CounterLock=SYSCreateLock();
  if (CounterLock<0) Abort("Unable to create counter lock");
  /*
    A conditional is initially set to false. Test if it is really is.
  */
  TestConditionalFalse();
  /*
    Now change it to true and test its behavior.
  */
  SYSSetCond(TestCond);
  TestConditionalTrue();
  /*
    Change it back to false and test its behavior.
  */
  SYSResetCond(TestCond);
  TestConditionalFalse();
  /*
    Now the conditional is true (it was been set by the false conditional
    test executed at previous line). Try setting it to true again. This
    action should do nothing.
  */
  SYSSetCond(TestCond);
  TestConditionalTrue();
  /*
    Now set the conditional to false but two times. The first attempt should
    flip it and the second attempt should be ignored.
  */
  SYSResetCond(TestCond);
  SYSResetCond(TestCond);
  TestConditionalFalse();
  /*
    If got here, everything went ok.
  */
  return(0);
}
