#include "kernel.h"
#include "pdbase.h"
#include "pdfsys.h"
#include "pdfileio.h"
#include "strings.h"

typedef struct tagTServerInfo {
  int16 Server;
  void *Cmd;
  int16 CmdSize;
  void *Reply;
  int16 RepSize;
} TServerInfo;

static void SetErrNo(int16 Code)
{
  ErrNo=Code;
}

static bool SendCmd(TServerInfo *I)
{
  int16 ErrNo;

  do {
    ErrNo=KCALL__SEND(I->Server,I->Cmd,I->CmdSize);
    if (ErrNo<0) break;
    ErrNo=KCALL__RECV(I->Server,I->Reply,I->RepSize,NULL);
  } while(false);
  SetErrNo(ErrNo);
  return(ErrNo>=0);
}

static int16 OpenCmd(uint8 Command,const char *FileSpec,uint8 Mode)
{
  TOpenCmd C;
  TOpenReply R;
  TServerInfo I;
  char *Ptr;

  C.Command=Command;
  C.Mode=Mode;
  Ptr=StrNCpy(C.FileSpec,FileSpec,MAX_FILESPEC_SIZE+1);
  if (Ptr-C.FileSpec>MAX_FILESPEC_SIZE) {
    SetErrNo(fseNAMELEN);
    return(false);
  }
  I.Server=hndROOTFS;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=&R;
  I.RepSize=sizeof(R);
  if (!SendCmd(&I)) return(-1);
  return(R.Handle);
}

int16 NewOpen(const char *FileSpec,uint8 Mode)
{
  return(OpenCmd(cmdOPENFILE,FileSpec,Mode));
}

static uint16 ReadWrite(uint8 Cmd,int16 fd,void *Buf,uint16 Size)
{
  TReadCmd C;
  TIOReply R;
  TServerInfo I;

  C.Command=Cmd;
  C.Size=Size;
  C.Ptr=Buf;
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=&R;
  I.RepSize=sizeof(R);
  if (!SendCmd(&I)) return(0);
  return(R.Size);
}

uint16 NewRead(int16 fd,void *Buf,uint16 Size)
{
  return(ReadWrite(cmdREAD,fd,Buf,Size));
}

int32 NewRead32(int16 fd,void *Buf,int32 Size)
{
  return(ReadWrite(cmdREAD,fd,Buf,Size));
}

uint16 NewWrite(int16 fd,const void *Buf,uint16 Size)
{
  return(ReadWrite(cmdWRITE,fd,(void *)Buf,Size));
}

int32 NewWrite32(int16 fd,const void *Buf,int32 Size)
{
  return(ReadWrite(cmdWRITE,fd,(void *)Buf,Size));
}

bool NewSeek(int16 fd,uint32 Where)
{
  TSeekCmd C;
  TServerInfo I;

  C.Command=cmdSEEK;
  C.Position=Where;
  C.Page=0;
  C.Direction=dirABSOLUTE;
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.RepSize=0;
  return(SendCmd(&I));
}

bool NewSeekRel(int16 fd,int32 Where)
{
  TSeekCmd C;
  TServerInfo I;

  C.Command=cmdSEEK;
  C.Page=0;
  if (Where>0) {
    C.Position=Where;
    C.Direction=dirFORWARD;
  } else {
    C.Position=-Where;
    C.Direction=dirBACKWARD;
  }
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.RepSize=0;
  return(SendCmd(&I));
}

uint32 NewTell(int16 fd)
{
  TNoParamCmd C;
  TTellReply R;
  TServerInfo I;

  C.Command=cmdTELL;
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=&R;
  I.RepSize=sizeof(R);
  if (!SendCmd(&I)) return(0xFFFFFFFF);
  if (R.Page!=0) return(0xFFFFFFFF);
  return(R.Position);
}

bool NewFStat(int16 fd,TFileInfo *Info)
{
  TNoParamCmd C;
  TServerInfo I;

  C.Command=cmdFSTAT;
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=Info;
  I.RepSize=sizeof(*Info);
  return(SendCmd(&I));
}

bool NewClose(int16 fd)
{
  int16 ErrNo;

  ErrNo=KCALL__CLOSE(fd);
  SetErrNo(ErrNo);
  return(ErrNo==0);
}

int16 NewOpenDir(const char *DirSpec)
{
  return(OpenCmd(cmdOPENDIR,DirSpec,oeOPEN+oRDONLY));
}

bool NewReadItem(int16 dd,TDirItem *Buf)
{
  int32 Size;

  Size=NewRead(dd,Buf,sizeof(*Buf));
  if (Size==0) {
    Size=sizeof(*Buf);
    Buf->ID=0xFFFFFFFF;
  }
  return(Size==sizeof(*Buf));
}

bool NewCloseDir(int16 dd)
{
  return(NewClose(dd));
}

bool NewReadFileInfo(uint32 Number,TFileInfo *Info)
{
  TIStatCmd C;
  TServerInfo I;

  C.Command=cmdISTAT;
  C.ID=Number;
  I.Server=hndROOTFS;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=Info;
  I.RepSize=sizeof(*Info);
  return(SendCmd(&I));
}

static bool ManipAttr(uint8 Command,int16 fd,uint16 NewAttr)
{
  TAttrManipCmd C;
  TServerInfo I;

  C.Command=Command;
  C.Value=NewAttr;
  I.Server=fd;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.RepSize=0;
  return(SendCmd(&I));
}

bool NewChMod(int16 fd,uint16 NewAttr)
{
  return(ManipAttr(cmdFCHMOD,fd,NewAttr));
}

bool NewChOwn(int16 fd,uint16 NewOwner)
{
  return(ManipAttr(cmdFCHOWN,fd,NewOwner));
}

bool NewChGrp(int16 fd,uint16 NewGroup)
{
  return(ManipAttr(cmdFCHGRP,fd,NewGroup));
}

static bool ManipFS(uint8 Command,const char *Spec,const char *Another)
{
  TManipCmd C;
  TServerInfo I;
  char *Ptr;
  char *APtr;

  C.Command=Command;
  Ptr=StrNCpy(C.FileSpec,Spec,MAX_FILESPEC_SIZE+1);
  APtr=Ptr+1;
  if (Another!=NULL) APtr=StrNCpy(APtr,Another,MAX_FILESPEC_SIZE+1);
  if (Ptr-C.FileSpec>MAX_FILESPEC_SIZE || APtr-(Ptr+1)>MAX_FILESPEC_SIZE) {
    SetErrNo(fseNAMELEN);
    return(false);
  }
  I.Server=hndROOTFS;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.RepSize=0;
  return(SendCmd(&I));
}

bool NewMkDir(const char *DirSpec)
{
  return(ManipFS(cmdMKDIR,DirSpec,NULL));
}

bool NewRmDir(const char *DirSpec)
{
  return(ManipFS(cmdRMDIR,DirSpec,NULL));
}

bool NewRemove(const char *FileSpec)
{
  return(ManipFS(cmdREMOVE,FileSpec,NULL));
}

bool NewRename(const char *OldName,const char *NewName)
{
  return(ManipFS(cmdRENAME,OldName,NewName));
}

static bool ManipDisk(uint8 Command,TDriveInfo *Info)
{
  TNoParamCmd C;
  TServerInfo I;

  C.Command=Command;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.Reply=Info;
  I.RepSize=sizeof(*Info);
  return(SendCmd(&I));
}

bool NewGetFirstDisk(TDriveInfo *Info)
{
  return(ManipDisk(cmdFIRSTDRV,Info));
}

bool NewGetNextDisk(TDriveInfo *Info)
{
  return(ManipDisk(cmdNEXTDRV,Info));
}

char NewGetCurrentDrive(void)
{
  TDriveInfo Info;

  return(ManipDisk(cmdGETDRV,&Info)?Info.Name:'\0');
}

bool NewSetCurrentDrive(char DriveLetter)
{
  TSetDrvCmd C;
  TServerInfo I;

  C.Command=cmdSETDRV;
  C.Name=DriveLetter;
  I.Server=hndROOTFS;
  I.Cmd=&C;
  I.CmdSize=sizeof(C);
  I.RepSize=0;
  return(SendCmd(&I));
}
